#======================================================================================
# returns either "0-360" or "-180-180"
#
determine_lon_branch = function( xvals ) {

	if( max( xvals ) > 181 ) 
		return( "0-360" )

	if( min( xvals ) < 0 ) 
		return( "-180-180" )

	#------------------------------------------------------------------------
	# If we get here it could be a location, say, in Europe. Don't know which
	# branch. Choose one
	#------------------------------------------------------------------------
	return( "-180-180" )
}

#=============================================================================
# Usage: set 'fill' to color to fill with, otherwise just does outlines.
#
coastlines <- function( fill=NA, border='grey', ... ) {
	ns <- 385
	basedir <- "/data/obs/topo/etopo_5/world_coastline_segments_hires"

	if( (! is.na(fill)) && is.na(border) )
		border <- fill

	print(paste("drawing",ns,"segments"))
	for( i in 1:ns ) {
		fname <- paste( "seg.", formatC(i,flag="0",width=4), sep="" )
		fname <- paste(basedir,"/",fname,sep="")
		data <- read.table(fname)
		polygon(data[,1],data[,2],col=fill,border=border,...)

		# do other branches
		branchx <- data[,1]-360.0
		polygon(branchx,data[,2],col=fill,border=border,...)

		branchx <- data[,1]+360.0
		polygon(branchx,data[,2],col=fill,border=border,...)
		}
}

#=============================================================================
# if plot is FALSE, then simply returns the segments as a list of length
# nseg, where each is a list with $x[1..npts_in_seg] and $y[1..npts_in_seg]; for
# an example of useage, see /data/obs/boundary_gis/hydro1k/plot_closeup_bil_v2.R
# 
myusa <- function( plot=FALSE ) {

	home = Sys.getenv('HOME')
	ncid   <- open.ncdf(paste0(home,"/lib/coastlines_and_geography/OUTLUSAM.nc"))
	segdir <- get.var.ncdf( ncid, 'segment_directory')
	nseg   <- ncid$dim[["segment"]]$len
	verts  <- get.var.ncdf( ncid, 'verts')

	retval = list()
	for( i in 1:nseg ) {
		s0 <- segdir[5,i]+1
		nv <- segdir[6,i]
		s1 <- s0+nv-1
		latlons <- verts[s0:s1]
		index   <- 1:(nv/2)
		lats    <- latlons[index*2-1]
		lons    <- latlons[index*2] - 360.
		if( plot ) 
			lines(lons,lats)
		else
			retval[[i]] = list( x=lons, y=lats )
		}
	close.ncdf(ncid)

	if( ! plot )
		return( retval )
}

#=============================================================================
# Given a point (ptX,ptY) and a polygon specified by the vertices
# given in polyX and polyY, this returns TRUE if the point is INSIDE
# the polygon and FALSE otherwise. NOTE NOTE NOTE that you should NOT
# specify the final point of the polygon as being a repeat of the
# first point.  It is assumed to be closed in this way. 
# NOTE: ptX and ptY can be arrays, in which case an array of logicals 
# is returned.  In places where skip_mask is TRUE, a FALSE result is
# returned without actually checking the point.
#
# **NOTE** supplanted by pointInPolygonC, the C version ***
#
pointInPolygonR <- function( ptX, ptY, polyX, polyY, skip_mask=NA ) {

	nqpts <- length(ptX)
	if( length(ptY) != nqpts )
		stop(paste("length of ptX and ptY must be the same!  NX=",length(ptX),"  NY=", length(ptY)))
	
	np <- length(polyX)
	if( length(polyY) != np ) {
		stop( 'eror in pointInPolygon, #x points != #y points!')
		}

	if( (length(skip_mask)==1) && (is.na(skip_mask)) )
		retval <- array(TRUE,dim=dim(ptX))   # TRUE=in polygon, FALSE=outside polygon
	else
		retval <- (!skip_mask)

	#-----------------------------------------------------------
	# quick sanity check ... if pt is outside bounds of polygon,
	# no chance it could be inside the polygon
	#-----------------------------------------------------------
	max_polyX <- max(polyX)
	min_polyX <- min(polyX)
	max_polyY <- max(polyY)
	min_polyY <- min(polyY)
	retval <- ifelse( (ptY>max_polyY) | (ptY < min_polyY) | (ptX > max_polyX) | (ptX < min_polyX), FALSE, retval )
	if( sum(retval) == 0 ) {  # done if all are FALSE from this simple check
		return(retval)
		}
	
	yMax <- -1.e30
	xofyMax <- yMax
	for( i in 1:np ) {
		if( polyY[i] >= yMax ) {
			if( polyY[i] > yMax ) {
				yMax <- polyY[i]
				xofyMax <- polyX[i]
				maxIndex <- i
				}
			else
				{
				# Must be equal to a prev yMax we've seen 
				if( polyX[i] > xofyMax ) {
					yMax <- polyY[i]
					xofyMax <- polyX[i]
					maxIndex <- i
					}
				}
			}
		}

	#-----------------------------------------------------------
	# OK, we now have a point that is the easternmost of the
	# set of northermost points.  Drop SLIGHTLY down towards the
	# midpoint of the segment connecting the previous and next
	# points in the polygon.  This method fails for extremely
	# concave polygons that almost cross themselves at this
	# particular point.  Hopefully that is pathologically rare
	#-----------------------------------------------------------
	iprev <- maxIndex - 1
	inext <- maxIndex + 1
	if( iprev == 0 ) 
		iprev <- np
	if( inext == (np+1) ) 
		inext <- 1
	p0x <- polyX[iprev]
	p0y <- polyY[iprev]
	p1x <- polyX[inext]
	p1y <- polyY[inext]

	midX <- 0.5*(p0x+p1x)
	midY <- 0.5*(p0y+p1y)

	dx <- polyX[maxIndex] - midX
	dy <- polyY[maxIndex] - midY

	scalefact <- 1.e-4
	pointInPolyX <- polyX[maxIndex] - dx*scalefact
	pointInPolyY <- polyY[maxIndex] - dy*scalefact

	polyX <- c(polyX, polyX[1])  # because we access [np+1], below
	polyY <- c(polyY, polyY[1])

	iqdebug <- 270297
	for( iq in 1:nqpts ) {
		if( iq%%10000 == 0 )
			print(paste('iq=',iq,'/',nqpts))
		if( retval[iq] ) {   # skip pts we've already found to be outside of the polygon
			if( iq == iqdebug )
				print(paste("pt X,Y=",ptX[iq],ptY[iq],"   point in poly=",pointInPolyX,pointInPolyY))
			lseg <- c(ptX[iq],ptY[iq],pointInPolyX,pointInPolyY)
			ncrossings <- 0
			for( i in 1:np ) {
				lsegPoly <- c(polyX[i],polyY[i],polyX[i+1],polyY[i+1])
				res <- intersectLineSegs( lseg, lsegPoly )
				if( res[1] == 1 ) {
					ncrossings <- ncrossings + 1
				if( iq == iqdebug )
					print(paste("found intersection with seg",i, "(",polyX[i],polyY[i],")  (",polyX[i+1],polyY[i+1],")"))
					}
				else if( res[1] == 2 ) {
					print('error, polygon alg fails, segments are colinear')
					stop('colinear segments')
					}
				}
			if( (ncrossings %% 2) != 0 )
				retval[iq] <- FALSE
			if( iq == iqdebug )
				print(paste("final # of corssings, retval=",ncrossings, retval[iq]))
			}
		}

	return(retval)
}

#=============================================================================
# Given the specification of 2 line segments, this returns:
#	(0,NA,NA) if the segments do not intersect or overlap
#	(1,xret,yret) if the segments intersect at one point, possibly an endpoint,
#		which is then returned in xret, yret
#	(2,x1,y1,x2,y2) if the segments are colinear and overlap, in which event
#		the segment of overlap is returned in (x1,y1) to (x2,y2)
#
# segments are specified as:
#	lseg <- c(x0,y0,x1,y1)
#
intersectLineSegs <- function( lseg0, lseg1 ) {

	if( lseg0[1] == lseg0[3] ) 
		lseg0Vert = TRUE
	else
		lseg0Vert = FALSE
	if( lseg1[1] == lseg1[3] ) 
		lseg1Vert = TRUE
	else
		lseg1Vert = FALSE

	if( lseg0Vert && lseg1Vert ) {
		if( lseg0[1] != lseg1[1] ) # Lines are vertical and parallel, don't intersect
			return(c(0,NA,NA))
		#--------------------------------------------------
		# lines are vertical and colinear, do they overlap?
		#--------------------------------------------------
		ls0ymax = lseg0[2]
		ls0ymin = lseg0[4]
		if( lseg0[4] > ls0ymax ) {
			ls0ymax = lseg0[4]
			ls0ymin = lseg0[2]
			}
		ls1ymax = lseg1[2]
		ls1ymin = lseg1[4]
		if( lseg1[4] > ls1ymax ) {
			ls1ymax = lseg1[4]
			ls1ymin = lseg1[2]
			}

		if( (ls0ymax < ls1ymin) || (ls0ymin > ls1ymax)) 
			return(c(0,NA,NA))

		#------------------------------------------------
		# See if are end-to-end, sharing only an endpoint
		#------------------------------------------------
		if( ls0ymax == ls1ymin ) 
			return( c(1,lseg0[1],ls0ymax))
		if( ls1ymax == ls0ymin ) 
			return( c(1,lseg0[1],ls1ymax))

		#-------------------------------------------------------
		# Is one segment contained exclusively within the other?
		#-------------------------------------------------------
		if( (ls0ymin >= ls1ymin) && (ls0ymax <= ls1ymax ))
			return( c(2,lseg0) )
		if( (ls1ymin >= ls0ymin) && (ls1ymax <= ls0ymax ))
			return( c(2,lseg1) )

		#--------------------------------------------------
		# If we get here they must be partially overlapping
		#--------------------------------------------------
		if( (ls0ymin >= ls1ymin) && (ls0ymin <= ls1ymax))	# bottom pt of lseg1 lies in lseg2, top goes up beyond lseg2
			return( c(2,lseg0[1],ls0ymin,lseg0[1],ls1ymax))

		if( (ls0ymax >= ls1ymin) && (ls0ymax <= ls1ymax))	# top pt of lseg1 lies in lseg2, bottom starts below lseg2
			return( c(2,lseg0[1],ls1ymin,lseg0[1],ls0ymax))

		stop('unhandled case, you forgot one!')
		}

	if( lseg0Vert || lseg1Vert ) {	# ONE segment is vertical, but not both. They cannot be collinear
		if( lseg0Vert ) {
			lsVert = lseg0
			lsNVert = lseg1
			}
		else
			{
			lsVert = lseg1
			lsNVert = lseg0
			}
		xval <- lsVert[1]
		mNVert <- (lsNVert[4]-lsNVert[2])/(lsNVert[3]-lsNVert[1])
		xInt <- xval
		yInt <- lsNVert[2] + mNVert*(xval - lsNVert[1])
		if( pointBetween( c(xInt,yInt), lsVert ) &&
		    pointBetween( c(xInt,yInt), lsNVert ))
		    	return(c(1,xInt,yInt))
		else
			return(c(0,NA,NA))
		}

	#-----------------------------------------------
	# OK, no vertical lines -- can safely use slopes
	#-----------------------------------------------
	m0 <- (lseg0[4]-lseg0[2])/(lseg0[3]-lseg0[1])
	m1 <- (lseg1[4]-lseg1[2])/(lseg1[3]-lseg1[1])
	b0 <- lseg0[2] - m0*lseg0[1]
	b1 <- lseg1[2] - m1*lseg1[1]

	if( m0 == m1 ) {
		#-------------------------------------------
		# Parallel ... are they possibly coincident?
		#-------------------------------------------
		if( b0 != b1 ) 	# Parallel, cannot overlap
			return(c(0,NA,NA))
		#---------------------------------------------------
		# Segments are from the same line.  Do they overlap?
		#---------------------------------------------------
		if( pointBetween( c(lseg0[1],lseg0[2]), lseg1 ) ||
		    pointBetween( c(lseg0[3],lseg0[4]), lseg1 ) ||
		    pointBetween( c(lseg1[1],lseg1[2]), lseg0 ) ||
		    pointBetween( c(lseg1[3],lseg1[4]), lseg0 )) {
			#-------------------------------------------------
		    	# If we get here they must overlap. Now figure out
			# the line segment of overlap. We can safely do
			# this by just using the X values since they are
			# known not to be vertical.
			#-------------------------------------------------
			# See if one seg is entirely within the other
			#--------------------------------------------
			if( lseg0[1] < lseg0[3] ) {
				ls0xmin = lseg0[1]
				ls0ymin = lseg0[2]	# NOTE: Y corresponding to xmin, NOT the min y value!
				ls0xmax = lseg0[3]
				ls0ymax = lseg0[4]
				}
			else
				{
				ls0xmin = lseg0[3]
				ls0ymin = lseg0[4]	# NOTE: Y corresponding to xmin, NOT the min y value!
				ls0xmax = lseg0[1]
				ls0ymax = lseg0[2]
				}
			if( lseg1[1] < lseg1[3] ) {
				ls1xmin = lseg1[1]
				ls1ymin = lseg1[2]	# NOTE: Y corresponding to xmin, NOT the min y value!
				ls1xmax = lseg1[3]
				ls1ymax = lseg1[4]
				}
			else
				{
				ls1xmin = lseg1[3]
				ls1ymin = lseg1[4]	# NOTE: Y corresponding to xmin, NOT the min y value!
				ls1xmax = lseg1[1]
				ls1ymax = lseg1[2]
				}
			#--------------------------------------------
			# Check if one is completely within the other
			#--------------------------------------------
			if( (ls0xmin >= ls1xmin) && (ls0xmax <= ls1xmax))
				return( c(2,lseg0))
			if( (ls1xmin >= ls0xmin) && (ls1xmax <= ls0xmax))
				return( c(2,lseg1))
			#--------------------------------------------
			# If we get here, they must partially overlap
			#--------------------------------------------
			if( ls0xmin >= ls1xmin)
				return( c(2,ls0xmin,ls0ymin,ls1xmin,ls1ymin))
			if( ls1xmin >= ls0xmin)
				return( c(2,ls1xmin,ls1ymin,ls0xmin,ls0ymin))
			stop("error uncaught case")
		    	}
		return(c(0,NA,NA))
		}

	
	#-------------------------------------
	# See if intersect at their end points
	#-------------------------------------
	if( (lseg0[1] == lseg1[1]) && (lseg0[2] == lseg1[2])) 
		return( c(1, lseg0[1], lseg0[2] ))
	if( (lseg0[1] == lseg1[3]) && (lseg0[2] == lseg1[4])) 
		return( c(1, lseg0[1], lseg0[2] ))
	if( (lseg0[3] == lseg1[3]) && (lseg0[4] == lseg1[4])) 
		return( c(1, lseg0[3], lseg0[4] ))

	#---------------------------------------------------------
	# All special cases taken care of ... neither are vertical
	# lines, and they are neither colinear nor parallel.  
	# Get intersection.
	#---------------------------------------------------------
	xInt <- (b1-b0)/(m0-m1)
	yInt <- m1*xInt + b1
	if( pointBetween( c(xInt,yInt), lseg0 ) &&
	    pointBetween( c(xInt,yInt), lseg1 ))
	    	return(c(1,xInt,yInt))
	else
		return(c(0,NA,NA))
}

#======================================================================================
# Given the slope and intersecpt of two lines, returns status:
#	0 = do not intersect
#	1 = intersect at a point
#	2 = colinear
#
# and point of intersection, if applicable.
# Return value is ALWAYS a 3-element array, with c( status, x0, y0 )
#
intersectLines <- function( m1, b1, m2, b2 ) {

	if( m1 == m2 ) {
		if( b1 == b2 ) 
			return( c( 2, 0, b1 ) )	# slope and int same :: co-linear
		else
			return( c( 0, 0, 0 ) )	# slope same but int diff :: parallel
		}

	y = (m2*b1 - m1*b2) / (m2 - m1)
	x = (y - b1)/m1

	return( c( 1, x, y ))
}

#=============================================================================
# Given a point that is KNOWN TO BE ON A LINE, and the endpoints
# of a SEGMENT OF THAT LINE, returns true if the point is inside
# the segment and false otherwise.
#
pointBetween <- function( pt, lseg ) {
	ptX <- pt[1]
	ptY <- pt[2]

	lsXmin <- lseg[1]
	lsXmax <- lseg[3]
	if( lseg[3] < lsXmin ) {
		lsXmin <- lseg[3]
		lsXmax <- lseg[1]
		}
	lsYmin <- lseg[2]
	lsYmax <- lseg[4]
	if( lseg[4] < lsYmin ) {
		lsYmin <- lseg[4]
		lsYmax <- lseg[2]
		}

	if( (lsXmin <= ptX) && (ptX <= lsXmax) &&
	    (lsYmin <= ptY) && (ptY <= lsYmax))
	    	return(TRUE)
	else
		return(FALSE)
}

#=============================================================================
inOregon <- function( origlon, lat ) {

	# Lon is assumed to be in DEGREES WEST
	lon <- origlon
	if( lon < 0. ) 
		lon <- -lon
	if( lon > 180. )
		lon <- 360 - lon

	pts <- c(46.26,124.05,
		46.115, 122.943,
		45.648, 122.749,
		45.575, 122.344,
		45.711, 121.830,
		45.664, 121.209,
		45.607, 121.177,
		45.658, 121.054,
		45.744, 120.568,
		45.687, 120.475,
		45.934, 119.622,
		45.943, 119.091,
		46.0, 119.,
		46., 116.9,
		45.765, 116.577,
		45.587, 116.487,
		44.469, 117.219,
		44.257, 117.140,
		44.177, 116.892,
		43.849, 117.025,
		41.999, 117.027,
		41.999, 124.217,
		42.8, 124.6)
	np <- length(pts)/2
	yvals <- pts[(1:np)*2-1]
	xvals <- pts[(1:np)*2]
	#mplot(xvals,yvals,pch='+')

	retval <- pointInPolygon( lon, lat, xvals, yvals )
	return(retval)
}

#=============================================================================
cliptoUSA <- function( col=rgb(.8,.8,.8), otherbranch=FALSE ) {

	home = Sys.getenv('HOME')
	d1 <- read.table(paste0(home,'/src/mine/R/mylib/clipusa_seg1.txt'))
	d2 <- read.table(paste0(home,'/src/mine/R/mylib/clipusa_seg2.txt'))

	x1 <- d1$V1
	y1 <- d1$V2

	x2 <- d2$V1
	y2 <- d2$V2

	if( otherbranch ) {
		x1 <- ifelse( x1<0, 360+x1, x1 )
		x2 <- ifelse( x2<0, 360+x2, x2 )
		}

	#mplot(c(x1,x2), c(y1,y2))
	#lines(x1,y1,col='green')
	#lines(x2,y2,col='red')

	poly1x <- c(-180,-180,x1,360,360)
	poly1y <- c(-90,90,y1,90,-90)

	poly2x <- c(360,360,x2,-180,-180)
	poly2y <- c(90,-90,y2,-90,90)

	polygon(poly1x,poly1y,col=col, border=col)
	polygon(poly2x,poly2y,col=col, border=col)
}

#=============================================================================
# if dat2d is supplied, then any NA value in dat2d will yield a NA in 
# the final result
#
inUSA_reg_grid = function( xvals, yvals, dat2d=NULL, thin=NA, desample_km=NA ) {

	nx = length( xvals )
	ny = length( yvals )

	lon = array(0., nx*ny )
	lat = array(0., nx*ny )

	iout = 0
	for( jj in 1:ny )
	for( ii in 1:nx ) {
		if( is.null( dat2d )) {
			iout = iout + 1
			lon[iout] = xvals[ii]
			lat[iout] = yvals[jj]
			}
		else
			{
			if( ! is.na( dat2d[ii,jj] )) {
				iout = iout + 1
				lon[iout] = xvals[ii]
				lat[iout] = yvals[jj]
				}
			}
		}

	lon = lon[ 1:iout ]
	lat = lat[ 1:iout ]

	rv = inUSA( lon, lat, thin=thin, desample_km=desample_km )

	if( is.null( dat2d )) {
		dim( rv ) = c( nx, ny )
		}
	else
		{
		retval = array( NA, dim=c(nx,ny))
		iout = 0
		for( jj in 1:ny )
		for( ii in 1:nx ) {
			if( ! is.na( dat2d[ii,jj] )) {
				iout = iout + 1
				retval[ii,jj] = rv[iout]
				}
			}
		rv = retval
		}

	return( rv )
}

#=============================================================================
inUSA <- function( origlon, lat, thin=1., desample_km=NA ) {

	npin <- length(origlon)
	if( length(lat) != npin ) 
		stop(paste("error in inUSA: # of lat and lon points must be the same!  Nlon=",length(origlon),
				"  Nlat=",length(lat)))

	print(paste('inUSA: entering with npts=', npin ))

	retval <- array(FALSE,dim=c(npin))

	if( max(origlon) > 185 ) 
		lon <- origlon - 360
	else
		lon <- origlon

	#-----------------------------------------------------------
	# First do really coarse estiamtes. Only have to do the full
	# evaluation for points near the border
	#-----------------------------------------------------------
	point_done = array( FALSE, npin )

	print(paste('inUSA: part 1, coarse Canada...'))
	inCAc = inCA_super_coarse_util( lon, lat )

	print(paste('inUSA: part 2, coarse USA...'))
	inUSc = inUS_super_coarse_util( lon, lat )

	print(paste('inUSA: part 3, coarse MX...'))
	inMXc = inMX_super_coarse_util( lon, lat )

	for( kk in 1:npin ) {
		if( inCAc[kk] ) {
			retval[kk] = FALSE
			point_done[kk] = TRUE
			}
		else if( inUSc[kk] ) {
			retval[kk] = TRUE
			point_done[kk] = TRUE
			}
		else if( inMXc[kk] ) {
			retval[kk] = FALSE
			point_done[kk] = TRUE
			}
		}

	ndone = sum( point_done )
	print(paste('inUSA: after parts 1-3, number of pts done:', ndone, '=', ndone/npin*100, '%' ))

	q <- map("usa",plot=FALSE)
	npmap <- length(q$x)
	#--------------------------------------------------------------
	# NA's denote multiple line segments -- take the first one only
	#--------------------------------------------------------------
	xvals <- array(0,npmap)
	yvals <- array(0,npmap)
	seenna <- FALSE
	j <- 0
	for( i in 1:length(q$x) ) {
		if( is.na(q$x[i]) || is.na(q$y[i]))
			seenna <- TRUE
		if( ! seenna ) {
			j <- j + 1
			xvals[j] <- q$x[i]
			yvals[j] <- q$y[i]
			}
		}
	xvals <- xvals[1:j]
	yvals <- yvals[1:j]
	npmap <- j

	if( ! is.na( desample_km )) {
		print(paste("Desampling polygon to", desample_km, 'km; orig npts=', length(xvals) ))
		rvds = desample_path( xvals, yvals, desample_km )
		xvals = rvds$lon
		yvals = rvds$lat
		print(paste("...done desampling polygon to", desample_km, 'km; new npts=', length(xvals) ))
		}

	#------------------
	# Thin if requested
	#------------------
	if( ! is.na(thin) ) {
		n2 <- 0
		lonold <- -999.
		latold <- -999.
		xvals2 <- array(0.,npmap)
		yvals2 <- array(0.,npmap)
		thinsqr <- thin*thin
		for( i in 1:npmap ) {
			distsqr <- (lonold-xvals[i])*(lonold-xvals[i]) + (latold-yvals[i])*(latold-yvals[i])
			if( is.na(distsqr))
				distsqr <- thinsqr*10
			if( distsqr > thinsqr ) {
				n2 <- n2 + 1
				xvals2[n2] <- xvals[i]
				yvals2[n2] <- yvals[i]
				lonold <- xvals[i]
				latold <- yvals[i]
				}
			}
		xvals <- xvals2[1:n2]
		yvals <- yvals2[1:n2]
		print(paste("thinning to dist=",thin,"; old # pts=",npmap," thinned # pts=",n2))
		npmap <- n2
		}

	n2do = sum( ! point_done )
	print(paste("Processing the remaining", n2do, 'pts' ))

	idx2do = (1:npin)[! point_done ]

	for( kk in 1:n2do ) {

		i = idx2do[kk]

		if( (kk %% 100) == 0 )
			print(paste(kk,"/",n2do))

		retval[i] <- pointInPolygonC( lon[i], lat[i], xvals, yvals )
		}

	return(retval)
}

#=============================================================================
# See also 'instate', below
#
inCalifornia <- function( origlon, lat ) {

	return( instate( 'ca', origlon, lat ))

###--- unused

	if( (lat > 42.) || (lat < 32.5)) 
		return(FALSE);

	# Lon is assumed to be in DEGREES WEST
	lon <- origlon
	if( lon < 0. ) 
		lon <- -lon
	if( lon > 180. )
		lon <- 360 - lon

	if( (lon > 124.4) || (lon < 114.1))
		return(FALSE);

	pts <- c(42.001,120, 39,120, 35,114.6, 34.76,114.54,
		34.69,114.45, 34.4, 114.26,
		34.32, 114.145, 34.25, 114.149,
		34.102, 114.420, 33.943,114.525,
		33.576,114.529, 33.356,114.697,
		33.057,114.669, 33.016,114.507,
		32.857,114.462, 32.724,114.624,
		32.713, 114.709, 32.632, 117.130,
		34.432,120.541, 40.364,124.522,
		42.001,124.177)

	np <- length(pts)/2
	yvals <- pts[(1:np)*2-1]
	xvals <- pts[(1:np)*2]
	#mplot(xvals,yvals,pch='+')

	retval <- pointInPolygonC( lon, lat, xvals, yvals )
	return(retval)
}

#=============================================================================
# In this vesion lon is dim [nx] and lat is dim [ny], and they are assumed to
# describe a regular lon/lat grid. See version 'instate', below, for a more
# flexible interface.
#
instate_reg_grid <- function( state, origlon, lat, verbose=FALSE ) {

	lon = origlon
	if( max(lon) > 181 )
		lon = lon - 360

	nx = length( lon )
	ny = length( lat )

	lon_flat = array(0., nx*ny )
	lat_flat = array(0., nx*ny )
	src_ii = array(0., nx*ny )
	src_jj = array(0., nx*ny )

	for( jj in 1:ny )
	for( ii in 1:nx ) {
		idx = ii + (jj-1)*nx  
		lon_flat[idx] = lon[ii]
		lat_flat[idx] = lat[jj]

		src_ii[idx] = ii
		src_jj[idx] = jj 
		}

	rr = instate( state, lon_flat, lat_flat, verbose=verbose )

	#----------------------------
	# Unpack back to regular grid
	#----------------------------
	retval = array( FALSE, dim=c(nx,ny))
	for( idx in 1:(nx*ny) ) {
		ii = src_ii[idx]
		jj = src_jj[idx]
		retval[ii,jj] = rr[idx]
		}

	return( retval )	# [nx,ny] logical array, TRUE if in specified state
}

#=============================================================================
# county should be a string, for example "california,san diego"
#
make_county_polygon = function( county ) {

stop('dont do this use the data in data obs boundaries us_counties instead')

	globalframenumber <<- 0

	rv = map( 'county', 'california,san diego', plot=FALSE )
	print(rv)
	npts = length( rv$x )

	xlim = range( rv$x, na.rm=TRUE )
	ylim = range( rv$y, na.rm=TRUE )

	plotit = TRUE

	if( plotit ) canvas( xlim=xlim, ylim=ylim )

	dg = 1/npts

	xout = array(0., npts)
	yout = array(0., npts)

	np_out = 0
	for( it in 1:npts ) {
		
		if( plotit ) {
			print(paste(rv$x[it], rv$y[it]))
			points( rv$x[it], rv$y[it], col=grey(it*dg), pch=19 )
			text  ( rv$x[it], rv$y[it], it, adj=c(-0.3,0.5) )
			if( it > 1 ) 
				lines( c(rv$x[it-1],rv$x[it]), c(rv$y[it-1],rv$y[it]) ) 
			}

		if( !is.na(rv$x[it])) {
			np_out = np_out + 1
			xout[np_out] = rv$x[it]
			yout[np_out] = rv$y[it]
			}
		}

	xout = xout[1:np_out]
	yout = yout[1:np_out]

	return( list( x=xout, y=yout ))
}

#=============================================================================
# 'state' is a two-letter abbreviation for the state, in upper or lower case, 
# such as 'ca', 'wa', or 'ky'
#
# 'lon' and 'lat' are NOT the ordinary 1-D arrays that give the edges of a
# matrix ... rather then are 1..nn, where nn is the total number of points.
# For a version using two "ordinary" 1-D lon/lat arrays, use instate_reg_grid
#
instate <- function( state, origlon, lat, verbose=TRUE ) {

	if( verbose ) print(paste('instate: entering for', state))

	nn = length( lat )
	if( length(origlon) != nn ) 
		stop(paste("Error, length of lon and lat must be identical!"))
	#print(paste("Entering routine instate for state ", state, "npts=", nn ))

	retval <- array(FALSE,nn)

	lon = origlon
	mask_change_branch = (lon > 180)
	lon[ mask_change_branch ] = origlon[ mask_change_branch ] - 360

	#-------------------------------------------------------------------------
	# Get polygons for the state. Lons here are degrees E, i.e., are generally 
	# negative (-122 to -60 or so) for the U.S.
	#-------------------------------------------------------------------------
	state_uc = toupper(state)

	fname = paste('/data/obs/boundary_gis/us_states_DWP/DWP_state_outline.', state_uc, '.R_dat', sep='' )
	load( file=fname )
	nsegs = length( seg_out )

	if( verbose ) print(paste('instate: nsegs=', nsegs))

	#---------------------------------------------------------------------------------
	# now have seg_out[[1...nsegs]], each seg_out[[]] is itself a list with $x and $y;
	# i.e., this is the format I use for my GIS routines
	#---------------------------------------------------------------------------------
	for( iseg in 1:nsegs ) {	# These are, for example, offshore islands. I.e., Colorado only has 1 segment, while Florida has many

		lon_st = seg_out[[iseg]]$x
		lat_st = seg_out[[iseg]]$y

		mx_lon = max( lon_st, na.rm=TRUE )
		mn_lon = min( lon_st, na.rm=TRUE )
		mx_lat = max( lat_st, na.rm=TRUE )
		mn_lat = min( lat_st, na.rm=TRUE )
		#print(paste("state:", state, "segment:", iseg, "lon range:", mn_lon, mx_lon, "lat range:", mn_lat, mx_lat))

		for( ipt in 1:nn ) {

			if( verbose && (ipt%%10000 == 0)) print(paste('instate: ipt=', ipt, '/', nn, ipt/nn*100, '%'))

			if( (lon[ipt] >= mn_lon) && (lon[ipt] <= mx_lon) &&
			    (lat[ipt] >= mn_lat) && (lat[ipt] <= mx_lat)) {
				#-----------------------------------------------------------
				# Note: don't need to check that the state polygon is valid,
				# I ensured that they all were when I created them
				#-----------------------------------------------------------
				if( pointInPolygonC( lon[ipt], lat[ipt], lon_st, lat_st, check_valid=FALSE )) {
					retval[ipt] = TRUE
					}
				}
			}
		}

	return( retval )
}

#=============================================================================
# 'state' is a two-letter abbreviation for the state, in lower case, such
# as 'ca', 'wa', or 'ky'
#
# 'lon' and 'lat' are NOT the ordinary 1-D arrays that give the edges of a
# matrix ... rather then are 1..nn, where nn is the total number of points.
#
instate_old <- function( state, origlon, lat ) {

	nn <- length(origlon)

	retval <- array(FALSE,nn)

	lon <- origlon

	#----------------------------
	# Get the state boundary info
	#----------------------------
	pts <- drawstate( state, returnpts=TRUE )
	pts$lon <- -pts$lon	# switch to our lon convention

	for( ipt in 1:nn ) {
		if( (lat[ipt] > 55.) || (lat[ipt] < 20)) 
			retval[ipt] <- FALSE
		else
			{
			# Lon is assumed to be in DEGREES WEST
			if( lon[ipt] < 0. ) 
				lon[ipt] <- -lon[ipt]
			if( lon[ipt] > 180. )
				lon[ipt] <- 360 - lon[ipt]

			if( (lon[ipt] > 180) || (lon[ipt] < 30))
				retval[ipt] <- FALSE
			else 	
				retval[ipt] <- pointInPolygonC( lon[ipt], lat[ipt], pts$lon, pts$lat )
			}
		}

	return(retval)
}

#=============================================================================
maskCalifornia <- function( colleft='white', colright='white') {

	dleft <- read.table('/data/obs/usdiv/old_drd964x_style/boundaries/gis/ca_polygon_left.txt')
	xleft <- dleft$V1
	yleft <- dleft$V2

	dright <- read.table('/data/obs/usdiv/old_drd964x_style/boundaries/gis/ca_polygon_right.txt')
	xright <- dright$V1
	yright <- dright$V2

	xright <- c(xright, -117.5, -113, -113, -124 )
	yright <- c(yright, 0,       0,     90,   90 )

	xleft <- c(xleft, -122, -180, -180, -117 )
	yleft <- c(yleft,  90,  90,    0,   0 )

	#mplot(c(-124,-114),c(32,43),type='p')

	polygon(xleft,yleft,col=colleft,border=colleft)
	polygon(xright,yright,col=colright,border=colright)
}

#=============================================================================
# 'state' is the two-letter state name in lower case.  For example,
# "ca", "or", or "ky".
#
# If you get weird geometry errors, might as well try with version=2
# which accesses a different database
#
drawstate <- function( state, verbose=FALSE, returnpts=FALSE, version=1 ) {

	if( version==2 ) {
		fname = paste0( '/data/obs/boundary_gis/us_states_DWP/DWP_state_outline.', toupper(state), '.R_dat' )
		load( file=fname )
		iseg = 1
		lon = seg_out[[iseg]]$x
		lat = seg_out[[iseg]]$y
		return( list(lon=lon, lat=lat ))
		}

	fname <- paste('/data/obs/usdiv/old_drd964x_style/boundaries/',state,'sta.asc', sep='')
	print(paste("mlib_map_utils: drawstate: reading file", fname))

	f <- file(fname,"r")
	eof <- FALSE
	l <- readLines(f,n=1)
	if( verbose )
		print(paste("# points:",l))
	npairs <- as.integer(l)
	lon <- array(0.,npairs)
	lat <- array(0.,npairs)
	for( i in 1:npairs ) {
		l   <- readLines(f,n=1)
		if( verbose )
			print(paste("point number",i,":",l))
		lat[i] <- as.double(substr(l,1,7))
		lon[i] <- as.double(substr(l,8,15))
		}

	close(f)

	if( returnpts ) 
		return( list(lon=lon, lat=lat ))

	lines( lon, lat )
}

#=============================================================================
inIndia <- function( origlon, lat ) {

	if( (lat > 36.) || (lat < 7.8)) 
		return(FALSE);

	lon <- origlon
	if( (lon > 90.0) || (lon < 68.0))
		return(FALSE);

	pts <- c(
		77.5,8.0, 79.8,10.4,
		80.1,15.3, 86.9,20.4,
		87.0,21.5, 89.0,21.7,
		88.5,26.4, 83.3,27.5,
		80.1,29.0, 80.8,30.2,
		78.9,31.4, 78.5,32.5,
		79.4,32.7, 78.0,35.6,
		76.9,34.8, 73.9,34.6,
		74.0,33.2, 75.3,32.3,
		74.5,31.7, 74.6,31.1,
		71.6,27.8, 70.6,28.1,
		69.5,27.0, 71.0,24.4,
		68.8,24.4, 68.2,23.7,
		69.0,22.9, 70.4,22.8,
		68.8,22.4, 70.9,20.6,
		72.5,21.9, 73.5,16.0,
		76.6,8.6 )
	
	np <- length(pts)/2
	xvals <- pts[(1:np)*2-1]
	yvals <- pts[(1:np)*2]
	#mplot(xvals,yvals,pch='+')

	retval <- pointInPolygonC( lon, lat, xvals, yvals )
	return(retval)
}

#===============================================================================
# To use my fonts, specify fontinfo:
# fontinfo <- dphershey_fontinfo(1)
# height_inch <- array(0.1, level )
#
# Input 'state' is a 2-letter state abbreviation, such as "CA", "NV", or "AZ"
#
# If force_label is set, then level is ignored, and the city list must be the version with underscores instead of spaces
# (for example, "La_Jolla")
#
label_state_cities <- function(state, cex_text=1.0, adj=c(-0.1,0.5), level=2, pch=19, cex_pch=0.7, fontinfo=dphershey_fontinfo(1), 
			airport_codes=FALSE,
			lwd=array(0.5,20),
			force_label=NULL,	# LIST OF CITIES TO LABEL
			height_inch=array(0.1,20), dx=0.1, 
			col_label='black', 
			col_outline='white',
			lwd_outline=lwd*8,
			col_point='black',
			add=TRUE,		# if add is false, then DON'T actually draw/emit anything, just return city info 
			lwd_white=5*lwd) {

	cex0 = par()$cex
	height_inch = height_inch * cex0
	
	source('/home/pierce/src/mine/label_points/label_points.R')

	state = tolower( state )

	if( ! is.null( force_label )) {
		nfl = length( force_label )
		did_force_label = array( FALSE, nfl )
		level = -1
		}

	home = Sys.getenv('HOME')
	fname_db = paste0( home, "/src/mine/R/mylib/", state, "_city_location_info.txt")
	if( filesize( fname_db ) < 10 ) 
		stop(paste("Sorry, not yet done for state", state, ": need to make file ", fname_db, ' run routine make_state_city_db(state) to help' ))

	print(paste("### label_state_cities: reading file", fname_db ))
	d <- read.table(fname_db)
	dlevel<- as.integer(d$V1)
	lat   <- as.double(d$V2)
	lon   <- as.double(d$V3)
	name  <- as.character(d$V4)

	nsta = length( lon )

	has_airport_codes = FALSE
	air3 = NULL
	if( ! is.null(d$V5)) {
		has_airport_codes = TRUE
		air3 = as.character(d$V5)
		}
print(paste('maps QQQQQ, here is air3:'))
print(air3)
print('---')

	#------------------------------------------------------
	# Collect point x, y, label, and label size into arrays
	#------------------------------------------------------
	nflat = 0
	px_flat = array( 0., nsta )
	py_flat = array( 0., nsta )
	label_flat = array( ' ', nsta )
	height_in_flat = array(0.08, nsta )
	lwd_flat = array(1, nsta )

	ret_x	= numeric()
	ret_y	= numeric()
	first = TRUE

print(paste("maps label_pt_cities AAA: airport_codes=", airport_codes, " force_label="))
print(force_label)
print('---')

	n <- length(lat)
	ret_strs = ''
	for( i in 1:n ) {

		forceit = FALSE
		if( ! is.null(force_label )) {
			if( airport_codes ) {
				forceit = (air3[i] %in% force_label)
print(paste('maps before d, here is air3:'))
print(air3)
print('---')
print(paste("maps label_pt_cities DDD: i=", i, "air3=", air3[i], "force_label=", pv(force_label), " in:", (air3[i] %in% force_label) ))
				}
			else 
				{
				forceit = (name[i] %in% force_label)
print(paste("maps label_pt_cities CCC: name=", name[i], "force_label=", pv(force_label), " in:", (name[i] %in% force_label) ))
				}
			}

print(paste("maps label_pt_cities BBB: name=", name[i], "forceit=>", forceit, "< i=", i, " dlevel=", dlevel[i] ))

		if( forceit || (dlevel[i] <= level)) {

			ret_x = c( ret_x, lon[i] )
			ret_y = c( ret_y, lat[i] )
			if( first ) {
				ret_strs  = name[i]
				first = FALSE
				}
			else
				{
				ret_strs  = c( ret_strs, name[i] )
				}

			if( (length(fontinfo)==1) && is.na(fontinfo)) {
				#----------------------------------
				# THIS IS THE NON-DPHERSHEY VERSION
				#----------------------------------
				if( add  )
					text(lon[i],lat[i],name[i],adj=adj,cex=cex_text[dlevel[i]], col=col)
				}
			else
				{
				if( airport_codes ) {
					if( (is.null(air3)) || (air3[i] == 'xxx') ) 
						nam = name[i]
					else
						nam = air3[i]
					}
				else
					nam = name[i]

				for( kk in 1:nchar(nam) )
					if( substr(nam,kk,kk) == '_' )
						substr(nam,kk,kk) = ' ' 
				dy = 0.

				nflat = nflat + 1
				px_flat[nflat] = lon[i]
				py_flat[nflat] = lat[i]
				label_flat[nflat] = nam
				height_in_flat[nflat] = height_inch[dlevel[i]]
				lwd_flat[nflat] = lwd[dlevel[i]]
				}
			}

			if( forceit ) {
				izz = -1
				for( kk in 1:nfl ) {
					if( airport_codes ) 
						nnn = air3[i]
					else
						nnn = name[i]
					if( nnn == force_label[kk] )
						izz = kk
					}
				if( izz == -1 ) stop(paste("coding error, izz==-1??? Did not find index"))
				did_force_label[izz] = TRUE
				}
		}

	px_flat = px_flat[1:nflat]
	py_flat = py_flat[1:nflat]
	label_flat = label_flat[1:nflat]
	height_in_flat = height_in_flat[1:nflat]

	print('label_state_cities: calling label_points_dphershey')
	if( storage.mode(fontinfo) != 'list' )
		stop(paste("Error, called label_state_cities with invalid fontinfo!"))

	if( add ) 
		label_points_dphershey( px_flat, py_flat, label_flat, height_in_flat, fontinfo,
			col_outline=col_outline, off_dx=0.5, off_dy=0.5, xpd=FALSE, lwd=lwd, 
			lwd_outline=lwd_outline, col_label=col_label )

	print('label_state_cities: back from label_points_dphershey')
	if( add ) 
		points( px_flat, py_flat, pch=array(pch,nflat), cex=array(cex_pch,nflat), col=col_point )

	if( ! is.null( force_label )) {
		for( kk in 1:nfl ) {
			if( ! did_force_label[kk] )
				print(paste("NOTE, user specified force_label with city name", force_label[kk], "but that city was not found in the database file named", fname_db))
			}
		}

	retval = list(
		xvals	= ret_x,
		yvals	= ret_y,
		strs	= ret_strs )

	return( retval )
}

#===============================================================================
# To use my fonts, specify fontinfo:
# fontinfo <- dphershey_fontinfo(1)
# height_inch <- array(0.1, level )
#
# Input 'state' is a 2-letter state abbreviation, such as "CA", "NV", or "AZ"
#
# old version does not use 'label_points'
#
label_state_cities_old <- function(state, cex_text=1.0, adj=c(-0.1,0.5), level=2, pch=21, cex_pch=0.7, fontinfo=NA, 
			airport_codes=FALSE,
			height_inch=0.1, dx=0.1, col='black', lwd_white=5) {
	
	state = tolower( state )

	home = Sys.getenv('HOME')
	fname_db = paste0( home, "/src/mine/R/mylib/", state, "_city_location_info.txt")
	if( filesize( fname_db ) < 10 ) 
		stop(paste("Sorry, not yet done for state", state, ": need to make file ", fname_db, ' run routine make_state_city_db(state) to help' ))

	d <- read.table(fname_db)
	dlevel<- as.integer(d$V1)
	lat   <- as.double(d$V2)
	lon   <- as.double(d$V3)
	name  <- as.character(d$V4)

	has_airport_codes = FALSE
	if( ! is.null(d$V5)) {
		has_airport_codes = TRUE
		air3 = as.character(d$V5)
		}

	n <- length(lat)
	for( i in 1:n ) {
		if( dlevel[i] <= level ) {
			if( (length(fontinfo)==1) && is.na(fontinfo))
				text(lon[i],lat[i],name[i],adj=adj,cex=cex_text[dlevel[i]], col=col)
			else
				{
				if( airport_codes ) {
					if( air3[i] == 'xxx' ) 
						nam = name[i]
					else
						nam = air3[i]
					}
				else
					nam = name[i]

				for( kk in 1:nchar(nam) )
					if( substr(nam,kk,kk) == '_' )
						substr(nam,kk,kk) = ' ' 
				dy = 0.
				if( nam == 'Tijuana' ) dy = -dx
				dphershey_emit_string( lon[i]+dx, lat[i]+dy, nam, height_inch[dlevel[i]], fontinfo, col='white', lwd=lwd_white )
				dphershey_emit_string( lon[i]+dx, lat[i]+dy, nam, height_inch[dlevel[i]], fontinfo, col=col, lwd=1 )
				}
			if( pch != -1 ) {
				points(lon[i],lat[i],pch=pch,cex=cex_pch, col=grey(0.9), bg='black')
				}
			}
		}
}

#=====================================================================
make_state_city_db = function( state, nkeep=50 ) {

	state = tolower( state )

	home = Sys.getenv('HOME')
	fname_db = paste0( home, "/src/mine/R/mylib/", state, "_city_location_info.txt")
	if( filesize( fname_db ) > 10 ) 
		stop(paste("Error, file ", fname_db, ' already exists' ))

	cdb = read_city_dbase()
		#lon	= lon[1:lnum],
		#lat	= lat[1:lnum],
		#country	= country[1:lnum],
		#state	= state[1:lnum],
		#pop	= pop[1:lnum],
		#elev	= elev[1:lnum],
		#tz	= tz[1:lnum],
		#name	= name[1:lnum] )

	nc = length( cdb$lon )

	#-------------------------------------------------------------------
	# To avoid so many levels of indirection, simply copy city info over
	#-------------------------------------------------------------------
	lon = array(0, nc )
	lat = array(0, nc )
	pop = array(0, nc )
	name = array(' ', nc )

	nin = 0		# number of cities in the state
	for( ic in 1:nc ) {
		dbstate = tolower( cdb$state[ic] )
		if( dbstate == state ) {
			#print(paste( cdb$name[ic], cdb$pop[ic] ))
			nin = nin + 1
			lon[nin] = cdb$lon[ ic ]
			lat[nin] = cdb$lat[ ic ]
			pop[nin] = cdb$pop[ ic ]
			name[nin] = cdb$name[ ic ]
			}
		}

	lon 	= lon[1:nin]
	lat 	= lat[1:nin]
	pop 	= pop[1:nin]
	name 	= name[1:nin]

	#------------
	# Sort by pop
	#------------
	six = sort( pop, index.return=TRUE, decreasing=TRUE )

	lon = lon[ six$ix ]
	lat = lat[ six$ix ]
	pop = pop[ six$ix ]
	name = name[ six$ix ]

	for( ic in 1:nin ) {
		print(paste(name[ic], pop[ic] ))
		}

	#--------------------------------------
	# Keep only top 'nkeep' populous cities
	#--------------------------------------
	if( nkeep > nin ) 
		nkeep = nin

	lon = lon[ 1:nkeep ]
	lat = lat[ 1:nkeep ]
	pop = pop[ 1:nkeep ]
	name = name[ 1:nkeep ]

	dist_matrix = array(0, dim=c(nkeep,nkeep))
	for( j in 1:nkeep ) 
	for( i in j:(nkeep-1)) {
		p1 = c( lon[j], lat[j] )
		p2 = c( lon[i], lat[i] )
		if( i == j ) 
			dd = 1.e30
		else
			dd = great_circle_dist( p1, p2 )
		dist_matrix[i,j] = dd
		dist_matrix[j,i] = dd
		}
	print(dist_matrix)

	#-------------------------------------------------------------
	# Generate 'levels' based on pop and how far from anything else
	# level 1 = major city, level 4 = minor town
	#-------------------------------------------------------------
	level = array(0, nkeep)
	n4 = nkeep/4
	for( i in 1:nkeep ) {
		if( i < n4 ) 
			level[i] = 1
		else if( i < n4*2 ) 
			level[i] = 2
		else if( i < n4*3 ) 
			level[i] = 3
		else
			level[i] = 4
		}

	#---------------------------------------------------
	# Don't allow any level 1 cities too near each other
	#---------------------------------------------------
	mindist_crit = 50	# km
	for( i in 1:(nkeep-1) ) {
		for( j in (i+1):nkeep ) {
			if( (i != j) && (level[i] == 1) && (level[j] == 1)) {
				dd = dist_matrix[i,j]
				if( dd < mindist_crit ) {
					print(paste("demoting", name[j], "from level", level[j], " to that +1 based on distance to ", 
						name[i], ' being only ', dd ))
					level[j] = level[j] + 1	
					}
				}
			}
		}

	#------------------------------------------------------------------------
	# Any city that is very near a higher rated city is bumped down 2 notches
	#------------------------------------------------------------------------
	maxlevel = 4
	for( tclevel in 1:2 ) 	# "too close to" level
	for( i in 1:nkeep ) {
		if( level[i] > tclevel ) {
			# Get min dist from this city to city of level 'tclevel' 
			mindist = 1.e30
			ixmin = -1
			for( j in 1:nkeep ) {
				if( i != j ) {
					if( level[j] == tclevel ) {
						dd = dist_matrix[i,j]
						if( dd < mindist ) {
							mindist = dd
							ixmin = j
							}
						}
					}
				}
			print(paste(name[i], 'is level', level[i], ' and closest city of level ', tclevel, 
				'is', mindist, ' city ', name[ixmin], 'level', level[ixmin] ))
			if( mindist < mindist_crit ) { 	# kilometers
				level[i] = level[i] + 2
				if( level[i] > maxlevel )
					level[i] = maxlevel
				}
			}
		}

	#-------------
	# Write output
	#-------------
	fout = file( fname_db, 'w' )
	for( i in 1:nkeep ) {
		nn = name[i]
		for( kk in 1:nchar(nn))
			if( substr(nn,kk,kk) == ' ' )
				substr(nn,kk,kk) = "_"
		cat( file=fout, level[i], ' ', mformatC( '%10.7f', lat[i]), ' ', 
			mformatC('%12.7f', lon[i]), ' ', nn, '\n', sep='' )
		}
	close( fout )
}

#=====================================================================
# Changes the branch that the longtitude is on from 0,360 to
# -180,180, or vice versa (selected automatically).  Input data is 2 
# or 3-d; orig longitudes must also be given.
#
change_lon_branch <- function( data, origlon ) {
	
	s <- dim(data)
	ndims <- length(s)

	if( (ndims < 2) || (ndims > 3))
		stop(paste("change_lon_branch: only works on 2 or 3-d data. Given: ndims=",ndims))

	nx <- s[1]
	ny <- s[2]
	newlon <- array(0.,nx)

	if( nx != length(origlon))
		stop(paste("error, first dim of 2-d array is",nx,"but was passed",length(origlon),"X values!"))

	if( ndims == 2 )
		newdat <- array(0.,dim=c(nx,ny))
	else
		{
		nz <- s[3]
		newdat <- array(0.,dim=c(nx,ny,nz))
		}

	#----------------------------------------
	# Figure out what branch we are on alredy
	#----------------------------------------
	branch_m180_180 <- 1
	branch_0_360    <- 2

	if( min(origlon) < -100 )
		branch_orig <- branch_m180_180
	else
		branch_orig <- branch_0_360

	if( branch_orig == branch_0_360 ) 
		tt <- ifelse( origlon > 180, origlon-360, origlon )
	else
		tt <- ifelse( origlon < 0, 360+origlon, origlon )

	srt <- sort( tt, index.return=TRUE)
	for( i in 1:nx ) {
		newlon[i]  <- tt[srt$ix[i]]
		if( ndims == 2 ) 
			newdat[i,] <- data[srt$ix[i],]
		else
			{
			for( iz in 1:nz )
				newdat[i,,iz] <- data[srt$ix[i],,iz]
			}
		}

	rv <- list()

	#------------------
	# RETURNED VALS:
	#   rv$lon
	#   rv$data
	#------------------
	rv$lon  <- newlon
	rv$data <- newdat
	return(rv)	 #   RETURNED components: $lon  $data
}

#=======================================================================
# Returns:
#	1 - Atlantic
#	2 - Pacific
#	3 - Indian
#	4 - Arctic
#	5 - Med
#	0 - otherwise
#
# NOTE: there is NO land-sea mask applied, so it may return 1,2, or 3 for
# a land point!
#
ocean_sector <- function( lon, lat ) {

	#------------------------------------------------
	# Get various branches of the lon for convenience
	#------------------------------------------------
	lon_m180_180 <- ifelse( lon>180, lon-360, lon )
	lon_0_360    <- ifelse( lon<0,   360+lon, lon )

	#------------------------------------------------
	# Runs from N to S down the spine of the Americas
	#------------------------------------------------
	seg_americas <- c(258,65.1118,   257,22,  275,13,  276,10, 279.1,7.7,  281.3,9.5,   285,7,  
			297,-18,  288,-52,  290.1118,-54, 290.1118,-89 ) + 0.01118

	#--------------------------------------------------------------
	# To get the indices necessary to reverse a segment, we need
	# (for example, when npts=6) "11 12  9 10  7 8  5 6  3 4  1 2".
	# This can be generated by: (n:1) + array((-1,1),n) where n
	# is the # of points*2
	#--------------------------------------------------------------
	n <- length(seg_americas)
	rev_seg_americas <- (n:1) + array(c(-1,1),n)

	#-----------------------------------------------------------------------------
	# Runs from N to S down from Norway, through Spain, then Africa, to Antarctica
	#-----------------------------------------------------------------------------
	seg_africa <- c(42,65.1118,  42,58, 355,39, 355,32, 25.1118,25, 25.1118,-89 )
	rev_seg_africa <- (12:1) + array(c(-1,1),12)
	seg_africa_m180_180 <- ifelse( seg_africa>180, seg_africa-360, seg_africa)

	#-------------------------------------------------------------------
	# Runs from N to S starting in Siberia, then down through Indonesia,
	# down Australia, and to Antarctica
	#-------------------------------------------------------------------
	seg_indonesia <- c(93.1118,71.1118,  99.5,14,  98.9,8.5,  103.5,2.1118,   109.1118,0.1118,
			131.1118,0.1118,   140.1118,-6.1118,   141.1118,-9.1118,   142.1118,-13.1118,   
			146.1118,-43.1118,   146.1118,-89.1118)
	rev_seg_indonesia <- (22:1) + array(c(-1,1),22)

	#---------------------------------------------
	# We go around the polygons in a CCW direction
	#---------------------------------------------
	poly_atl <- c( seg_africa[rev_seg_africa], seg_americas )

	np_atl <- length(poly_atl)/2
	polyx_atl <- poly_atl[(0:(np_atl-1))*2+1]
	polyy_atl <- poly_atl[(0:(np_atl-1))*2+2]

	#--------------------------------
	# put atl lons on -180,180 branch
	#--------------------------------
	polyx_atl <- ifelse( polyx_atl>180, polyx_atl-360, polyx_atl )

	poly_pac <- c( seg_indonesia, seg_americas[rev_seg_americas] )
	np_pac <- length(poly_pac)/2
	polyx_pac <- poly_pac[(0:(np_pac-1))*2+1]
	polyy_pac <- poly_pac[(0:(np_pac-1))*2+2]

	poly_ind <- c(seg_africa_m180_180, seg_indonesia[rev_seg_indonesia])
	np_ind <- length(poly_ind)/2
	polyx_ind <- poly_ind[(0:(np_ind-1))*2+1]
	polyy_ind <- poly_ind[(0:(np_ind-1))*2+2]

	poly_med <- c(-6,38,  -5.5,34,  19.5,27.5,   32.2,30.1,  39,30.,  39,39,
			11,47,  2.5,44) + 0.01118
	np_med <- length(poly_med)/2
	polyx_med <- poly_med[(0:(np_med-1))*2+1]
	polyy_med <- poly_med[(0:(np_med-1))*2+2]

#mplot(c(-180,-90),c(180,90),col='white')
#print(polyx_med)
#print(polyy_med)
#mplot(polyx_med,polyy_med)
#for( i in 1:length(polyx_med))
#text(polyx_med[i],polyy_med[i],i)
#map(add=TRUE)

	np <- length(lon)
	if( length(lat) != np ) 
		stop(paste("must have same number of lon and lat points!  #lon=",np,
			"#lat=",length(lat)))

	in_something <- array(FALSE,dim=dim(lon))

	print("test med")
	in_med <- pointInPolygonC( lon_m180_180, lat, polyx_med, polyy_med )

	print("test atl")
	in_something <- in_med
	in_atl <- pointInPolygonC( lon_m180_180, lat, polyx_atl, polyy_atl, skip_mask=in_something )

	print("test pac")
	in_something <- (in_something | in_atl)
	in_pac <- pointInPolygonC( lon_0_360,    lat, polyx_pac, polyy_pac, skip_mask=in_something )

	print("test ind")
	in_something <- (in_something | in_pac)
	in_ind <- pointInPolygonC( lon_m180_180, lat, polyx_ind, polyy_ind, skip_mask=in_something )

	print("test arctic")
	retval <- ifelse( lat>65., 4, 0 )

	retval <- ifelse( (retval==0) & in_med, 5, retval )

	retval <- ifelse( (retval==0) & in_atl, 1, retval )

	retval <- ifelse( (retval==0) & in_pac, 2, retval )

	retval <- ifelse( (retval==0) & in_ind, 3, retval )

	return(retval)
}

#=======================================================================
test_ocean_sector <- function() {

	#xvals <- seq(260,310,by=.5)
	#yvals <- seq(-10,20,by=.5)
	xvals <- seq(0,360,by=2)
	yvals <- seq(-90,90,by=2)
	#xvals <- seq(-10,42,by=2)
	#yvals <- seq(20,50,by=2)

	nx <- length(xvals)
	ny <- length(yvals)

	lonvals <- array(0.,dim=c(nx,ny))
	latvals <- array(0.,dim=c(nx,ny))

	lonvals[,1] <- xvals
	latvals[1,] <- yvals

	for( iy in 2:ny )
		lonvals[,iy] <- lonvals[,1]
	for( ix in 2:nx )
		latvals[ix,] <- latvals[1,]

	res <- ocean_sector( lonvals, latvals )


	res_br2_rv <- change_lon_branch( res, lonvals[,1] )
	res <- res_br2_rv$data
	lon <- res_br2_rv$lon

	ccontour(lon, latvals[1,], res )
	map(add=TRUE)

}

#=======================================================================================
region_mask_ocean_name = function( ibasin ) {

	retval = array(' ', 16)
	retval[1] = 'South Pacific'
	retval[2] = 'Tropical South Pacific'
	retval[3] = 'Tropical North Pacific'
	retval[4] = 'North Pacific'
	retval[5] = 'North Indian'
	retval[6] = 'South Atlantic'
	retval[7] = 'North Atlantic'
	retval[8] = 'Arctic'
	retval[9] = 'Weddell'
	retval[10] = 'Mediterranean'
	retval[11] = 'South Indian'
	retval[12] = 'Enderby'
	retval[13] = 'Ross'
	retval[14] = 'Tropical North Atlantic'
	retval[15] = 'Tropical South Atlantic'
	retval[16] = 'Tropical South Indian'

	return( retval[ibasin] )
}

#=======================================================================================
# See comments under land_sea_mask
#
region_mask <- function( lons, lats, critfrac=0.5 ) {

	LAND 		<- 0

	#--------------------------
	# Following are sea entries
	#--------------------------
	S_PAC 		<- 1
	TROP_S_PAC 	<- 2
	TROP_N_PAC	<- 3
	N_PAC		<- 4
	N_IND		<- 5
	S_ATL		<- 6
	N_ATL 		<- 7
	ARCTIC		<- 8
	WEDDELL		<- 9
	MED		<- 10
	S_IND		<- 11
	ENDERBY		<- 12
	ROSS		<- 13
	TROP_N_ATL	<- 14
	TROP_S_ATL	<- 15
	TROP_S_IND	<- 16

	#---------------------------
	# Following are land entries
	#---------------------------
	TROP_N_AMER	<- 100
	TROP_S_AMER	<- 101
	N_AMER		<- 102
	S_AMER		<- 103
	AUSTRALIA	<- 104
	TROP_N_AFRICA	<- 105
	TROP_S_AFRICA	<- 106
	N_AFRICA	<- 107
	S_AFRICA	<- 108
	TROP_N_ASIA	<- 109
	TROP_S_ASIA	<- 110
	N_ASIA		<- 111
	S_ASIA		<- 112
	EUROPE		<- 113
	ANTARCTICA	<- 114

	ret_labs = list()
	ret_labs[['South_Pacific']] = S_PAC
	ret_labs[['Tropical_South_Pacific']] = TROP_S_PAC
	ret_labs[['Tropical_North_Pacific']] = TROP_N_PAC
	ret_labs[['North_Pacific']] = N_PAC
	ret_labs[['North_Indian_Ocean']] = N_IND
	ret_labs[['South_Atlantic_Ocean']] = S_ATL
	ret_labs[['North_Atlantic_Ocean']] = N_ATL
	ret_labs[['Arctic_Ocean']] = ARCTIC
	ret_labs[['Weddell_Sea']] = WEDDELL
	ret_labs[['Mediterranean_Sea']] = MED
	ret_labs[['South_Indian_Ocean']] = S_IND
	ret_labs[['Enderby_Sea']] = ENDERBY
	ret_labs[['Ross_Sea']] = ROSS
	ret_labs[['Tropical_North_Atlantic']] = TROP_N_ATL
	ret_labs[['Tropical_South_Atlantic']] = TROP_S_ATL
	ret_labs[['Tropical_South_Indian']] = TROP_S_IND

	ret_labs[['Tropical_North_America']] = TROP_N_AMER
	ret_labs[['Tropical_South_America']] = TROP_S_AMER
	ret_labs[['North_America']] = N_AMER
	ret_labs[['South_America']] = S_AMER
	ret_labs[['Australia']] = AUSTRALIA
	ret_labs[['Tropical_North_Africa']] = TROP_N_AFRICA
	ret_labs[['Tropical_South_Africa']] = TROP_S_AFRICA
	ret_labs[['North_Africa']] = N_AFRICA
	ret_labs[['South_Africa']] = S_AFRICA
	ret_labs[['Tropical_North_Asia']] = TROP_N_ASIA
	ret_labs[['Tropical_South_Asia']] = TROP_S_ASIA
	ret_labs[['North_Asia']] = N_ASIA
	ret_labs[['South_Asia']] = S_ASIA
	ret_labs[['Europe']] = EUROPE
	ret_labs[['Antarctica']] = ANTARCTICA

	lsm <- land_sea_mask( lons, lats, critfrac=critfrac )

	nx <- length(lons)
	ny <- length(lats)

	basinno <- array(as.integer(0), dim=c(nx,ny))
	landno  <- array(as.integer(0), dim=c(nx,ny))

	for( j in 1:ny )
	for( i in 1:nx ) {
		if( lsm[i,j] > 0.5 ) {

			lon <- lons[i]
			lat <- lats[j]

			if( lat < -58.0 )
				landno[i,j] <- ANTARCTICA

			else if( (lon > 180) && (lon < 330)) {
				# Americas
				if( lat > 0. ) {
					# North America
					if( lat > 23.5 )
						landno[i,j] <- N_AMER
					else
						landno[i,j] <- TROP_N_AMER
					}
				else
					{
					# South America
					if( lat < -23.5 )
						landno[i,j] <- S_AMER
					else
						landno[i,j] <- TROP_S_AMER
					}
				}
			else if( (lon < 180) || (lon > 340) ) {
				# Austroasia
				if( (lon>100) && (lat>-50.) && (lat< -5))
					landno[i,j] <- AUSTRALIA
				else
					{
					af_x0 <- 30
					af_y0 <- 35
					af_x1 <- 50
					af_y1 <- 10
					af_m  <- (af_y1-af_y0)/(af_x1-af_x0)
					if( ((lon<50)||(lon>340)) && (lat<40) &&
						(lat < af_m*(lon-af_x1) + af_y1) ) {
						# Africa
						if( lat > 23.5 )
							landno[i,j] <- N_AFRICA
						else if( lat > 0. )
							landno[i,j] <- TROP_N_AFRICA
						else if( lat < -23.5 )
							landno[i,j] <- S_AFRICA
						else 
							landno[i,j] <- TROP_S_AFRICA
						}
					else	
						{
						# Asia/Europe
						if( lat > 23.5 ) {
							if( lon < 40 )
								landno[i,j] <- EUROPE
							else
								landno[i,j] <- N_ASIA
							}
						else if( lat > 0. )
							landno[i,j] <- TROP_N_ASIA
						else if( lat < -23.5 )
							landno[i,j] <- S_ASIA
						else
							landno[i,j] <- TROP_S_ASIA
						}
					}
				}
			}
		}

	southern_cutoff_lat <- -50.0

	for( j in 1:ny )
	for( i in 1:nx ) {

		lon <- lons[i]
		lat <- lats[j]

		if( lat > 60.0 ) 
			basinno[i,j] <- ARCTIC
		
		if( lat < 0.0 ) {
			# Southern hemisphere
			if( (lon >= 130) && (lon <= 290)) {
				# Pacific
				if( lat < southern_cutoff_lat ) 
					basinno[i,j] <- ROSS
				else if( lat < -23.5 ) 
					basinno[i,j] <- S_PAC
				else 
					basinno[i,j] <- TROP_S_PAC
				}
			else if( (lon > 290) || (lon <= 23)) {
				if( lat < southern_cutoff_lat ) 
					basinno[i,j] <- WEDDELL
				else if( lat > -23.5 )
					basinno[i,j] <- TROP_S_ATL
				else
					basinno[i,j] <- S_ATL
				}

			else
				{
				if( lat < southern_cutoff_lat ) 
					basinno[i,j] <- ENDERBY
				else if( lat > -23.5 )
					basinno[i,j] <- TROP_S_IND
				else
					basinno[i,j] <- S_IND
				}
			}

		else
			{
			# Northern Hemisphere
			if( (lon >= 110) && (lon <= 252) && (lat >= 23.5)) {
				if( lat < 65.5 )
					basinno[i,j] <- N_PAC
				}
			
			else if( (lon >= 110) && (lon <= 290)) {
				if( lat > (26.0-(lon-250.)/40.*23.5)) {
					if( lat < 67.0) {
						if( lat < 23.5 )	
							basinno[i,j] <- TROP_N_ATL
						else
							basinno[i,j] <- N_ATL
						}
					}
				else
					basinno[i,j] <- TROP_N_PAC
				}

			else if( (lon > 30) && (lon < 180) && (lat < 30) ) {
				basinno[i,j] <- N_IND
				}

			else if( ((lon >=355) || (lon <= 34)) && (lat > 30) && (lat < 43))
				basinno[i,j] <- MED

			else if( (lat < 67.0) && ((lon>240.)||(lon<28.))) {
				if( lat < 23.5 )
					basinno[i,j] <- TROP_N_ATL
				else
					basinno[i,j] <- N_ATL
				}
				
			}
		if( lsm[i,j] > 0.5 )
			basinno[i,j] <- landno[i,j]
		}

	retval = list( basinno=basinno, ret_labs=ret_labs )

	return(retval)	# $basinno[nx,ny], $ret_labs[[1..nregions]] with value of region flag
}

#=======================================================================================
# Returns a '1' if the passed lat/lon is land, and 0 if it is ocean.
#
# Parameter critfrac: if there is THIS MUCH land in the subsampled area OR MORE, then
#  the return value for that square is set to land.  So, smaller values of critfrac
#  lead to 'fatter' continents.
#
# NOTE: it is assumed lon and lat describe a regular grid, so they can be different
# sizes, and the returnd value is a grid with nx*ny elements
#
land_sea_mask <- function( lon, lat, critfrac=0.2, hires_usa=FALSE ) {

	if( hires_usa ) 
		topo_db <- '/data/obs/topo/us_topo.nc'
	else
		topo_db <- '/data/obs/topo/etopo.nc'
	ncid <- nc_open( topo_db )

	topo <- ncvar_get( ncid )
	topo[ is.na(topo) ] = -999.
	lon_topo <- ncid$dim[['X']]$vals
	lat_topo <- ncid$dim[['Y']]$vals

	nx_topo <- length(lon_topo)
	ny_topo <- length(lat_topo)
	dx_topo <- abs(lon_topo[2] - lon_topo[1])
	dy_topo <- abs(lat_topo[2] - lat_topo[1])

	nc_close( ncid )

	nx <- length(lon)
	ny <- length(lat)

	#--------------------------------------------
	# make sure lons are monotonically increasing
	#--------------------------------------------
	for( i in 1:(nx-1))
		if( lon[i+1] < lon[i] )
			stop("lons must be monotonically increasing")

	#--------------------------------------------------------------
	# Get a version of lons on the same branch as the topo (0->360)
	#--------------------------------------------------------------
	lon_0_360 <- lon
	lon_0_360[ lon_0_360 < 0 ] <- 360 + lon_0_360
		
	idx_x <- array(0,nx)
	idx_y <- array(0,ny)

	for( i in 1:nx )
		idx_x[i] <- wherenearest( lon_0_360[i], lon_topo )
	for( j in 1:ny )
		idx_y[j] <- wherenearest( lat[j], lat_topo )

	retval <- array(0,dim=c(nx,ny))

	#---------------------------------
	# Loop over the OUTPUT lat/lons...
	#---------------------------------
	for( j in 1:ny )
	for( i in 1:nx ) {
		
		if( (i==1) || (i==nx) || (j==1) || (j==ny)) {
			if( topo[idx_x[i],idx_y[j]] >= 0.1 )
				retval[i,j] <- 1
			}
		else
			{
			xl <- 0.5*(lon[i-1]+lon[i])
			xr <- 0.5*(lon[i+1]+lon[i])
			dx <- abs(xl-xr)

			ya <- 0.5*(lat[j+1]+lat[j])
			yb <- 0.5*(lat[j-1]+lat[j])

			dy <- abs(ya-yb)

			subsampx <- floor( dx/dx_topo ) + 1
			subsampy <- floor( dy/dy_topo ) + 1

			ssxh <- floor( subsampx/2 ) + 1
			ssyh <- floor( subsampy/2 ) + 1

			ninsamp <- 0
			accum   <- 0
			for( jj in -ssyh:ssyh )
			for( ii in -ssxh:ssxh ) {
				ip <- idx_x[i] + ii
				jp <- idx_y[j] + jj
				if( (ip>=1) && (ip<=nx_topo) && (jp>=1) && (jp<=ny_topo)) {
					ninsamp <- ninsamp + 1
					if( topo[ip,jp] >= 0. )
						locmask <- 1
					else
						locmask <- 0
					accum <- accum + locmask
					}
				}
			if( (ninsamp>0) && (accum/ninsamp > critfrac) )
				retval[i,j] <- 1
			}
		}

	return(retval)
}

#============================================================================================
# Given 3-dimensional data in array data3d, which is at locations xvals, yvals,
# this generates a vertical section along a line from (x0,y0) to (x1,y1); the
# final output will be of dimension ( nseg , nz )
#
get_sect_along_track <- function( xvals, yvals, data3d, x0, y0, x1, y1, nsegs ) {

	library('fields')

	nx <- length(xvals)
	ny <- length(yvals)
	nz <- dim(data3d)[3]

	dx0 <- abs(xvals[2] - xvals[1])
	dy0 <- abs(yvals[2] - yvals[1])

	if( yvals[2] < yvals[1] ) {
		for( iz in 1:nz ) 
			data3d[,,iz] <- flip(data3d[,,iz])
		yvals <- yvals[ny:1]	
		}

	if( (x0 < xvals[1]) || (x1 < xvals[1]) || (x0 > xvals[nx]) || (x1 > xvals[nx]))
		stop("specified x0 or x1 out of range of xvals")
	if( (y0 < yvals[1]) || (y1 < yvals[1]) || (y0 > yvals[ny]) || (y1 > yvals[ny]))
		stop("specified y0 or y1 out of range of yvals")

	#-----------------------------------------------------------
	# Make list of X and Y locations where we will be generating
	# values from the 3-D array
	#-----------------------------------------------------------
	dx <- x1 - x0
	dy <- y1 - y0
	delx <- dx/(nsegs-1)
	dely <- dy/(nsegs-1)

	newx <- (0:(nsegs-1))*delx + x0
	newy <- (0:(nsegs-1))*dely + y0

	locs <- array(0,dim=c(nsegs,2))
	locs[,1] <- newx
	locs[,2] <- newy

	slice <- array(0.,dim=c(nsegs,nz))
	#------------------
	# Loop over heights
	#------------------
	for( iz in 1:nz ) {
		#---------------------------
		# Make our interpolation obj
		#---------------------------
		interpobj <- list( x=xvals, y=yvals, z=data3d[,,iz] )
		slice[,iz] <- interp.surface( interpobj, locs )
		}

	retval <- list( x=newx, y=newy, data=slice )
	return(retval)
}

#=======================================================================================
# Given a boolean array, this returns the points of the polygon outline
# of the TRUE points in the array.
#
raster_to_outline_polygon <- function( xvals, yvals, dd ) {
	
stop('sorry, this does not work. Use fortran code in /data/obs/topo/etopo5 instead')
	nx <- length(xvals)
	ny <- length(yvals)

	#-------------------------------------------------------------------
	# Begin by getting rid of any islands or lakes in the interior of a 
	# filled region.  They are identifiable by not being connected to
	# the other exterior locations.  1=exterior, 2=interior
	#-------------------------------------------------------------------
	regno <- ifelse( dd, 2, 0 )	# begin by just setting interior pts
	if( dd[1,1] != FALSE )
		stop('alg fails')
	regno[1,1] <- 1
	made_a_change <- TRUE
	iter <- 1
	while( made_a_change ) {
		print(paste("island filling, iter=",iter))
		made_a_change <- FALSE
		for( jj in 1:ny )
		for( ii in 1:nx )
			if( (regno[ii,jj] == 0) && (dd[ii,jj] == FALSE)) {
				im1 <- ii - 1
				if( im1 < 1 ) im1 <- 1
				ip1 <- ii + 1
				if( ip1 > nx ) ip1 <- nx
				jm1 <- jj - 1
				if( jm1 < 1 ) jm1 <- 1
				jp1 <- jj + 1
				if( jp1 > ny ) jp1 <- ny
				if( (regno[im1,jj]==1) || (regno[ip1,jj]==1) || (regno[ii,jp1]==1) || (regno[ii,jm1]==1)) {
					regno[ii,jj] <- 1
					made_a_change <- TRUE
					}
				}
		iter <- iter + 1
		if( iter > nx*ny )
			stop("coding error, too many iterations!")
		}
	n_island_pts <- sum( regno == 0 )
	if( n_island_pts > 0 ) {
		print(paste("Eliminating",n_island_pts,"island points"))
		dd[regno==0] <- TRUE
		}

	midx <- array(nx-1)
	for( ii in 1:nx )
		midx[ii] <- (xvals[ii+1]+xvals[ii])*.5

	midy <- array(ny-1)
	for( jj in 1:ny )
		midy[jj] <- (yvals[jj+1]+yvals[jj])*.5

	nmax <- nx*ny
	ptx <- array(0,nmax)
	pty <- array(0,nmax)
	nn  <- 0

	for( jj in 1:ny )
		for( ii in 1:(nx-1) )
			if( dd[ii,jj] != dd[ii+1,jj] ) {
				nn <- nn + 1
				ptx[nn] <- midx[ii]
				pty[nn] <- yvals[jj]
				}

	for( ii in 1:nx )
		for( jj in 1:(ny-1))
			if( dd[ii,jj] != dd[ii,jj+1]) {
				nn <- nn + 1
				ptx[nn] <- xvals[ii]
				pty[nn] <- midy[jj]
				}

	ptx <- ptx[1:nn]
	pty <- pty[1:nn]
	print(paste("There are",nn,"points on the perimeter"))

	# Pick first point
	ptx_in_order <- array(0,nn)
	pty_in_order <- array(0,nn)

	ptx_in_order[1] <- ptx[1]
	pty_in_order[1] <- pty[1]
print(paste(ptx_in_order[1],pty_in_order[1]))
	ptx_remain <- ptx[2:nn]
	pty_remain <- pty[2:nn]
	for( ipt in 1:(nn-1) ) {
		srcx <- ptx_in_order[ipt]
		srcy <- pty_in_order[ipt]
		dist <- (srcx-ptx_remain)**2 + (srcy-pty_remain)**2
		sv <- sort( dist, index.return=TRUE )
		ptx_in_order[ipt+1] <- ptx_remain[ sv$ix[1] ]
		pty_in_order[ipt+1] <- pty_remain[ sv$ix[1] ]
print(paste(ipt,ptx_in_order[ipt+1],pty_in_order[ipt+1],ptx_in_order[ipt+1]-ptx_in_order[ipt],pty_in_order[ipt+1]-pty_in_order[ipt]))
		selarr <- array(TRUE,length(ptx_remain))
		selarr[sv$ix[1]] <- FALSE
		ptx_remain <- ptx_remain[selarr]
		pty_remain <- pty_remain[selarr]
		}

	testn <- 250
	ptx_in_order <- ptx_in_order[1:testn]
	pty_in_order <- pty_in_order[1:testn]

	retval <- list( x=ptx_in_order, y=pty_in_order )

	return(retval)
}

#==================================================================================
earth_color <- function( xvals, yvals ) {

	ncid <- open.ncdf( "/home/pierce/images/earth/land_shallow_topo_21600.ppm.v2.nc" )

	lon <- ncid$dim$lon$vals
	lat <- ncid$dim$lat$vals

	nx_want <- length(xvals)
	ny_want <- length(yvals)

	idxx <- array(0, nx_want)
	idxy <- array(0, ny_want)
	for( i in 1:nx_want )
		idxx[i] <- wherenearest( xvals[i], lon )
	for( i in 1:ny_want )
		idxy[i] <- wherenearest( yvals[i], lat )

	start <- c( min(idxx), min(idxy) )
	end   <- c( max(idxx), max(idxy) )
	count <- end - start + 1

	lon_subset <- lon[start[1]:end[1]]
	lat_subset <- lat[start[2]:end[2]]

	for( i in 1:nx_want )
		idxx[i] <- wherenearest( xvals[i], lon_subset )
	for( i in 1:ny_want )
		idxy[i] <- wherenearest( yvals[i], lat_subset )

	varname <- c('Red', 'Green', 'Blue' )
	nvars   <- length(varname)
	retval  <- array(0., dim=c(nx_want,ny_want,nvars))
	for( ivar in 1:nvars ) {
		print(paste("earth_color: loading", varname[ivar]))
		tt <- get.var.ncdf( ncid, varname[ivar], start=start, count=count )
		tt[ is.na(tt) ] <- 0
		for( j in 1:ny_want )
		for( i in 1:nx_want ) {
			retval[i,j,ivar] <- tt[ idxx[i], idxy[j] ]/255.0
			}
		}

	close.ncdf( ncid )

	return( retval )	# retval[nx,ny,3] where last index is R, G, B
				#
}

#========================================================================
# cx is in degreesE
# cy is in degreesN
# dx is in degrees
#
draw_geographic_circle <- function( cx, cy, dx, col='yellow', lwd=3 ) {

	#-----------------------
	# get length in KM of dx
	#-----------------------
	cy_rad <- cy/180.*pi
	dx_km <- 110.0*dx*cos(cy_rad)

	#----------------------------------
	# Get dy in degrees that is KM long
	#----------------------------------
	dy_deg <- dx_km/110.0

	lines( c(cx-dx, cx+dx), c(cy, cy), col='black', lwd=lwd*2 )
	lines( c(cx-dx, cx+dx), c(cy, cy), col=col, lwd=lwd )
	lines( c(cx, cx), c(cy+dy_deg, cy-dy_deg), col='black', lwd=lwd*2 )
	lines( c(cx, cx), c(cy+dy_deg, cy-dy_deg), col=col, lwd=lwd )

#return()

	#----------------
	# Draw the circle
	#----------------
	nseg <- 100
	px <- array(cx-dx, nseg+1)
	py <- array(cy, nseg+1)
	for( i in 1:nseg ) {
		theta_deg <- i/nseg*360.
		theta     <- theta_deg/180.*pi
		r <- dx
		if( theta_deg < 90 )
			r <- dx*cos(theta) + dy_deg*sin(theta)
		px[i+1] <- cx - r*cos(theta)
		py[i+1] <- cy + r*sin(theta)
		}

	lines( px, py, col=col, lwd=lwd )
}

#=================================================================================
# A fast approximation to the distance between two points along a great circle
# path, from Wikipedia
#
#	p1[2]: 		starting point, where p1[1] is lon and p1[2] is lat (both in degrees)
#	p2[2]: 		ending point, where p2[1] is lon and p2[2] is lat (both in degrees)
#
great_circle_dist_haversine = function( p1, p2 ) {

	r = 6371	# in km

	#---------------------------
	# Convert degrees to radians
	#---------------------------
	lon1 = p1[1]/180*pi
	lon2 = p2[1]/180*pi
	lat1 = p1[2]/180*pi
	lat2 = p2[2]/180*pi

	t1 = ( sin( (lat2 - lat1)/2 ) )**2
	t2 = cos(lat1) * cos(lat2) * ( sin( (lon2 - lon1)/2 ) )**2

	dist_km = 2 * r * asin( sqrt( t1 + t2 ))

	return( dist_km )
}

#=================================================================================
# Returns the distance (in km) bewteen two points along a great circle path
#
# Inputs:
#	p1[2]: 		starting point, where p1[1] is lon and p1[2] is lat (both in degrees)
#	p2[2]: 		ending point, where p2[1] is lon and p2[2] is lat (both in degrees)
#	return_bearing: if TRUE, then a 2-element vector is returned, with rv[1] being distance (km)
#			and rv[2] being bearing from the origin point to the dest point along
#			the great circle path degrees, clockwise from 0=North
#	bearing_res:	if at high latitudes, set to 0.1 instead of default of 1.0 (this slows things down)
#
# See also find_closest_points, below
#
great_circle_dist <- function( p1, p2, plot=FALSE, return_bearing=FALSE, bearing_res=1, fast=FALSE ) {

	if( fast ) 
		return( great_circle_dist_haversine( p1, p2 ))

	if( (p1[1] == p2[1]) && (p1[2] == p2[2]))
		return( 0 ) 

	if( ! return_bearing ) {
		dd <- great_circle_path( p1, p2, 10000000 )
		return( dd$dist[2] )
		}
	else
		{
		dd <- great_circle_path( p1, p2, bearing_res )
		x = dd$x[2]
		y = dd$y[2]
		#----------------------------------------------------------------
		# Find angle from p1 to (x,y). Since the distance is only 0.1 km,
		# I just use local cartesian coords
		#----------------------------------------------------------------
		dx_deg = x - p1[1]
		dy_deg = y - p1[2]
		cy_deg = p1[2] + dy_deg/2

		dx_km = dx_deg*cos(cy_deg/180*pi)
		dy_km = dy_deg

		theta_rad = atan2( dx_km, dy_km )
		theta = theta_rad/pi*180
		np = length( dd$dist )
		return( c(dd$dist[np], theta) )
		}
}

#=================================================================================
# Given a center point and 1d array of points (pxx, pyy), retrun sorted INDICES 
# and DISTANCES of closest points using either the slow (1) or fast (2) alg
#
find_closest_points_inner = function( cx, cy, pxx, pyy, type_1_slow_2_fast, debug=TRUE ) {

	nn = length( pxx )
	dist = array(0., nn)

	if( type_1_slow_2_fast == 1 ) {
		# SLOW
		p1 = c( cx, cy )
		for( kk in 1:nn ) {
			p2 = c( pxx[kk], pyy[kk] )	
			dist[kk] = great_circle_dist( p1, p2 )
			if( debug )
				print(paste("SLOW: cx, cy:", cx, cy, "p2:", pv(p2), 'dist:', dist[kk] ))
			}
		}
	else
		{
		# FAST
		dx = cx - pxx
		dy = cy - pyy
		for( kk in 1:nn ) {
			midy = 0.5*( cy + pyy[kk] )	# in degrees
			midy_rad = midy / 180. * pi		# in radians

			tdx = dx[kk] * cos( midy_rad )
			tdy = dy[kk]

			dist[kk] = tdx*tdx + tdy*tdy

			if( debug )
				print(paste("FAST: cx, cy:", cx, cy, "p2:", pxx[kk], pyy[kk], 'dist:', dist[kk] ))
			}
		}

	six = sort( dist, index.return=TRUE )

	ret_ix = six$ix
	ret_dist = array(0., nn)
	for( kk in 1:nn ) {
		ret_dist[kk] = dist[ six$ix[kk] ]

		if( debug )
			print(paste("1=SLOW_2=FAST:", type_1_slow_2_fast, "final nearest", kk, '/', nn, ':', ret_ix[kk], ret_dist[kk] ))
		}

	return( list(
		ix	= ret_ix,	# index into pxx and pyy arrays such that entry 1 is closest to cx, cy
		dist	= ret_dist ))	# dist array such that entry 1 is closest to cx, cy
}

#=================================================================================
# This takes as input a center lon (cx) and lat (cy), and 2-d arrays of lon and
# lat. It returns the X and Y indices of the 2d arrays sorted from closest to
# cx, cy to farthest.
#
# cx, cy, lon2d, lat2d must be in degrees
#
# npts : number of closest points to find. The larger this is, the slower this
#	 routine will run
#
find_closest_points = function( cx, cy, lon2d, lat2d, npts_out, verbose=TRUE, ndebug=5 ) {

	nx = dim( lon2d )[1]
	ny = dim( lon2d )[2]

	npts = nx*ny

	lon2d_flat = array(0., npts)
	lat2d_flat = array(0., npts)
	ixx_2_ii = array(0, npts)
	ixx_2_jj = array(0, npts)
	for( jj in 1:ny )
	for( ii in 1:nx ) {
		ixx = ii + (jj-1)*nx
		lon2d_flat[ixx] = lon2d[ii,jj]
		lat2d_flat[ixx] = lat2d[ii,jj]

		ixx_2_ii[ ixx ] = ii
		ixx_2_jj[ ixx ] = jj
		}

	#---------------------------------------------------------------------
	# Do an initial pass NOT using great circle dist to get the candidates
	# for the npts closest points
	#---------------------------------------------------------------------
	type_slow = 1
	type_fast = 2
	rv1 = find_closest_points_inner( cx, cy, lon2d_flat, lat2d_flat, type_fast )	# in range 1..npts
	ix_sort_fast   = rv1$ix
	dist_sort_fast = rv1$dist

	#-------------------------------------------------------------------
	# Second pass: get the great circle distance for just the candidates
	#-------------------------------------------------------------------
	npts_out_pad = floor( npts_out * 1.2 + 0.5 )
	if( npts_out_pad <   10 ) npts_out_pad = 10
	if( npts_out_pad > npts ) npts_out_pad = npts

	cand_ixx = ix_sort_fast[1:npts_out_pad]	# these are ii + (jj-1)*nx values for the closest pts
	dist = array(0., npts_out_pad )
	lon2d_flat_cand = array(0., npts_out_pad )
	lat2d_flat_cand = array(0., npts_out_pad )
	ixx_of_cand = array(0, npts_out_pad )
	for( kk in 1:npts_out_pad ) {
		ixx = ix_sort_fast[ kk ]
		lon2d_flat_cand[kk] = lon2d_flat[ixx]
		lat2d_flat_cand[kk] = lat2d_flat[ixx]

		ixx_of_cand[kk] = ixx
		}

	rv2 = find_closest_points_inner( cx, cy, lon2d_flat_cand, lat2d_flat_cand, type_slow )	# in range 1:npts_out_pad
	ix_sort_slow   = rv2$ix
	dist_sort_slow = rv2$dist

	#---------------------------------------
	# Returned ii, jj indices sorted by dist
	#---------------------------------------
	dsort_ii = array(0., npts )
	dsort_jj = array(0., npts )
	dsort_dist = array(0., npts )
	for( kk in 1:npts ) {
		ixx_for_cand = ix_sort_slow[ kk ]
		ixx = ixx_of_cand[ ixx_for_cand ]

		dsort_dist[kk] = dist_sort_slow[ ixx_for_cand ]

		dsort_ii[kk] = ixx_2_ii[ ixx ]
		dsort_jj[kk] = ixx_2_jj[ ixx ]


		if( (ndebug > 0) && (kk <= ndebug )) {

			 o_ii = dsort_ii[kk]
			 o_jj = dsort_jj[kk]

			print(paste('SLOW alg closest point', kk, ': dist=', dsort_dist[ kk ], 
				'orig ii, jj:', o_ii, o_jj, 
				'orig loc:', lon2d[ o_ii, o_jj ], lat2d[ o_ii, o_jj ] ))
			}
		}

	return( list( 
		dist		= dsort_dist,	# [npts], great circle dist to specified center point
		dsort_ii	= dsort_ii,	# [npts], ii index of entry in lon2d, lat2d sorted by closest to cx, cy
		dsort_jj	= dsort_jj 	# [npts], jj index of entry in lon2d, lat2d sorted by closest to cx, cy
		))
}

#=================================================================================
# Given a starting and ending point in degrees lon and lat, as well as an along-track
# distance in km, this returns the lon,lat locations along a great circle joining
# the start and end points, separated by a constant distance of 'dist' km.
#
# Inputs:
#	p1[2]: 		starting point, where p1[1] is lon and p1[2] is lat (both in degrees)
#	p2[2]: 		ending point, where p2[1] is lon and p2[2] is lat (both in degrees)
# 	dist:   	along-track length to use, in km
#
# See also 'great_circle_project'
#
great_circle_path <- function( p1, p2, dist, plot=FALSE ) {

	if( plot ) {
		canvas( xlim=c(-180,180), ylim=c(-90,90), xaxs='i', yaxs='i')
		g <- .7
		col <- rgb(g,g,g)
		map('world', add=TRUE, col=col)

		points( p1[1], p1[2], pch=19, col='red' )
		points( p2[1], p2[2], pch=19, col='red' )
		}

	#----------------
	# project command
	#----------------
	cmd <- paste( 'gmt project -C',p1[1],'/',p1[2],' -E',p2[1],'/',p2[2], ' -Q -G', dist, 
			' -Lw ', sep='' )
	res <- system( cmd, intern=TRUE )

	#---------------------------------------------------------------------
	# Parse output from project command, which is a series of points along
	# the great circle separated by a distance 'dist' (in km)
	#---------------------------------------------------------------------
	npts <- length( res )
	xx <- array(0., npts)
	yy <- array(0., npts)
	ll <- array(0., npts)
	for( ipt in 1:npts ) {
		tsplit <- strsplit( res[ipt], '\t' )
		xx[ipt] <- as.double(tsplit[[1]][1])
		yy[ipt] <- as.double(tsplit[[1]][2])
		ll[ipt] <- as.double(tsplit[[1]][3])
		}

	if( plot )
		points( xx, yy, pch='.', col='blue' )

	return( list( x=xx, y=yy, dist=ll ))	# note: dist is distance along circle in km
}


#=================================================================================
# Interpolates data along a great circle path.  Input array 'data' has shape [npts,3] 
# where [,1] is longitude (degrees), [,2] is latitude (degrees), and [,3] is some
# data value measured at that lon and lat.  The returned value will also have the
# data sampled at constant 'dist' intervals.
#
# Inputs:
#	data[N,3]:	Data values, where [,1] is lon (deg), [,2] is lat (deg), and	
#			[,3] is the measured data value at that location.
# 	dist:   	along-track length to use, in km
#
# Output:
#	an array of length Z, which is the data interpolated to points along the great circle.
#
great_circle_project <- function( data, dist, debug=TRUE ) {

	if( debug ) print(paste("great_circle_project: entering with dist (km)=", dist ))

	#--------------------------------
	# Convert dist from km to degrees
	#--------------------------------
	dist <- dist/110.0

	#-------------------------------------------------------------------
	# Put the point coords into a temporary file that fitcircle can read
	#-------------------------------------------------------------------
	tfile <- '/tmp/tmp.Rfitcircle.txt'
	ftmp  <- file(tfile,'w')
	npts <- dim(data)[1]
	for( ipt in 1:npts ) 
		cat( file=ftmp, data[ipt,1], ' ', data[ipt,2], ' ', data[ipt,3], '\n', sep='' )
	close( ftmp )

	#-------------------------------------------------------------------------
	# Construct and run the fitcircle command to get the center point and pole
	#-------------------------------------------------------------------------
	cmd <- paste( 'gmt fitcircle -L1 ', tfile )
	res <- system( cmd, intern=TRUE )

	if( debug ) {
		print(paste("return value from gmt fitcircle:"))
		print(res)
		print('---')
		}

	#-----------------------------------------------------------------
	# Parse result of fitcircle to get center (cx,cy) and pole (px,py)
	#-----------------------------------------------------------------
	nlines <- length(res)
	for( iline in 1:nlines ) {
		tsplit <- strsplit( res[iline], '\t' )
		if( iline == 2 ) {
			cx <- tsplit[[1]][1]
			cy <- tsplit[[1]][2]
			}
		if( iline == 3 ) {
			px <- tsplit[[1]][1]
			py <- tsplit[[1]][2]
			}
		}

	#----------------
	# project command
	#----------------
	cmd <- paste( 'gmt project ', tfile, ' -C',cx,'/',cy,' -T',px,'/',py, ' -S -Fprsz > tmp.project_out', sep='')
	msystem( cmd )
	cmd <- paste( 'uniq tmp.project_out > tmp.project_out_u', sep='' )
	msystem( cmd )
	cmd <- paste( 'gmt sample1d -I', dist, ' tmp.project_out_u', sep='' )
	print(cmd)
	res <- system( cmd, intern=TRUE )

	#-------------------------------------------------------------
	# Parse results -- 2 cols, along-track distance and data value
	#-------------------------------------------------------------
	npts <- length(res)
	ret_dist <- array(0., npts)	# along-track distance (km)
	ret_x <- array(0., npts)
	ret_y <- array(0., npts)
	ret_z <- array(0., npts)
	for( ipt in 1:npts ) {
		tt <- strsplit( res[ipt], '\t' )
		ret_dist[ipt] <- as.double( tt[[1]][1] )*110.0	# convert from deg to km
		ret_x[ipt] <- as.double( tt[[1]][2] )
		ret_y[ipt] <- as.double( tt[[1]][3] )
		ret_z[ipt] <- as.double( tt[[1]][4] )
		}

	return( list( x=ret_x, y=ret_y, z=ret_z, dist=ret_dist )) 
}

#=====================================================================================
# Ordinarily, map('state') only plots states on the -180 to +180 longitude
# branch.  This call plots the states on the 0 to 360 longitude branch.
#
# REMEMBER you generally want to call this with ( add=TRUE )
#
plot_state_other_lon_branch = function( add=TRUE, ... ) {

	p = par()
	porig = p$usr
	user_new = c(porig[1]-360, porig[2]-360, porig[3], porig[4])
	par(usr=user_new)
	map('state', add=add, ... )
	par(usr=porig)
}

#======================================================================================
# This returns 1 if the polygon is clockwise, 2 if anti-clockwise, and 0 if invalid
#
get_polygon_dir = function( px, py, plot=FALSE ) {

	RES_CLOCKWISE = 1
	RES_ANTICLOCKWISE = 2

	if( !validPolygon(px,py)) return(0)	# from the pointInPolygon package

	#---------------------------------------------------------------------------------
	# Method: consider the line segment FROM point 1 TO point 2. Extend a line segment
	# 90 degrees out to the right of that line segment. If the extended line segment
	# intersects another part of the polygon, then the polygon is drawn clockwise.
	#---------------------------------------------------------------------------------
	xmin = min(px)
	xmax = max(px)
	ymin = min(py)
	ymax = max(py)

	#----------------------------------------------------------------
	# Get points known to lie completely outside the polygon's region
	#----------------------------------------------------------------
	xleft  = xmin - (xmax-xmin)*0.1
	xright = xmax + (xmax-xmin)*0.1
	ybelow = ymin - (ymax-ymin)*0.1
	yabove = ymax + (ymax-ymin)*0.1

	#--------------------------------------------
	# Get center point of seg between pts 1 and 2
	#--------------------------------------------
	cx = 0.5*(px[1] + px[2])
	cy = 0.5*(py[1] + py[2])

	#---------------------------------------------------------------------
	# Get angle of this extended line segment, which must be perpindicular
	# to the original line segment
	#---------------------------------------------------------------------
	if( px[1] == px[2] ) {
		# Orig line seg is vertical, so extended seg must be horiz
		if( py[1] < py[2] ) 	# orig line seg is going UP, so line extends to RIGHT
			tseg = c(cx,cy,xright,cy)
		else
			tseg = c(cx,cy,xleft,cy)
		}
	else if( py[1] == py[2] ) {
		# Orig seg is horiz, so extended seg must be vert
		if( px[1] < px[2] )	# orig line seg is going RIGHT, so line extends DOWN
			tseg = c(cx,cy,cx,ybelow)
		else
			tseg = c(cx,cy,cx,yabove)
		}
	else
		{
		#-----------------------------------------------------------
		# Figure out what sector  line seg is pointing
		# into; 1=up, 2=right, 3=down, 4=left:
		#        \1/
		#       4 X 2
		#        /3\
		#-----------------------------------------------------------
		m0 = (py[2]-py[1])/(px[2]-px[1])
		b0 = py[1] - m0*px[1]
		if( (m0 >= 1) || (m0 <= -1)) {
			if( py[1] < py[2] )
				sector = 1
			else
				sector = 3
			}
		else
			{
			if( px[1] < px[2] )
				sector = 2
			else
				sector = 4
			}

		#-----------------------
		# Make full test segment
		#-----------------------
		m1 = -1/m0
		b1 = cy - m1*cx
		if( sector == 1 ) {
			yright = m1*xright + b1
			tseg = c(cx,cy,xright,yright)
			}
		else if( sector == 2 ) {
			xbelow = (ybelow - b1)/m1
			tseg = c(cx,cy,xbelow,ybelow)
			}
		else if( sector == 3 ) {
			yleft = m1*xleft + b1
			tseg = c(cx,cy,xleft,yleft)
			}
		else
			{
			xabove = (yabove - b1)/m1
			tseg = c(cx,cy,xabove,yabove)
			}
		}

	#----------------------------------------------------
	# Now see if our test line seg intersects the polygon
	# (other than seg1, which it originates from)
	#----------------------------------------------------
	nseg = length(px)
	res = RES_ANTICLOCKWISE
	nisect = 0
	xisect = array(nseg,0)
	yisect = array(nseg,0)
	for( iseg in 2:nseg ) {
		if( iseg == nseg )
			seg2 = c(px[iseg], py[iseg], px[1], py[1] )
		else
			seg2 = c(px[iseg], py[iseg], px[iseg+1], py[iseg+1] )
		isect = intersectLineSegs( tseg, seg2 )
		if( isect[1] != 0 ) {
			if( isect[1] == 1 ) {
				#-------------------------------------
				# yes, test seg DOES intersect polygon
				#-------------------------------------
				nisect = nisect + 1
				xisect[nisect] = isect[2]
				yisect[nisect] = isect[3]
				}
			else
				{
				# isect[1] ==2, are collinear
				nisect = nisect + 1
				xisect[nisect] = isect[2]
				yisect[nisect] = isect[3]
				nisect = nisect + 1
				xisect[nisect] = isect[4]
				yisect[nisect] = isect[5]
				}
			}
		}

	if( nisect%%2 == 0 ) 
		res = RES_ANTICLOCKWISE
	else
		res = RES_CLOCKWISE

	if( plot ) {
		xlim = range( c(px,tseg[1],tseg[3]))
		ylim = range( c(py,tseg[2],tseg[4]))
		if( res == RES_CLOCKWISE )
			main = 'Clockwise'
		else
			main = 'AntiClockwise'
		main = paste( main, 'nisect=', nisect)
		#canvas( xlim=xlim, ylim=ylim, main=main )
	canvas( xlim=c(0,100), ylim=c(0,100), grid=FALSE, main=main)
		polygon( px, py, border='black' )
		ll = 0.08
		arrows( px[1], py[1], px[2], py[2], col='blue', length=ll )
		arrows( tseg[1], tseg[2], tseg[3], tseg[4], col='red', length=ll )
		if( nisect > 0 ) {
			xisect = xisect[1:nisect]
			yisect = yisect[1:nisect]
			pch = 18+(1:nisect)
			points( xisect, yisect, pch=pch, cex=2, col='darkgreen' )
			}
		}

	return( res )
}

#======================================================================================
# Given a VALID (non self-intersecting) polygon, this arranges it to be
# in a given direction. IN GENERAL, polygons should be written clockwise,
# and holes should be anti-clockwise.
#
set_polygon_dir = function( px, py, clockwise=TRUE ) {

	RES_CLOCKWISE = 1
	RES_ANTICLOCKWISE = 2

	dir = get_polygon_dir( px, py )
	if( dir == RES_CLOCKWISE ) {
		if( clockwise )
			return( list(x=px, y=py))
		else
			return( list(x=reverse(px), y=reverse(py)))
		}
	else
		{
		# dir == RES_ANTICLOCKWISE
		if( clockwise )
			return( list(x=reverse(px), y=reverse(py)))
		else
			return( list(x=px, y=py))
		}
}

#======================================================================================
# Given two polygons that do not self-intsersect, this gives the polygon of their overlap
#
# Given the specification of 2 line segments, this returns:
#	(0,NA,NA) if the segments do not intersect or overlap
#	(1,xret,yret) if the segments intersect at one point, which is then returned
#		in xret, yret
#	(2,NA,NA) if the segments are colinear and overlap
#
# segments are specified as:
#	lseg <- c(x0,y0,x1,y1)
#
polygon_overlap = function( p1x, p1y, p2x, p2y, plot=FALSE ) {

	n1 = length( p1x )
	if( length(p1y) != n1 ) stop(paste("Error, polygon 1 has different number of X and Y points!"))
	n2 = length( p2x )
	if( length(p2y) != n2 ) stop(paste("Error, polygon 2 has different number of X and Y points!"))

	#---------------------------------------------------
	# Make sure polygons follow our clockwise convention
	#---------------------------------------------------
	rr = set_polygon_dir( p1x, p1y )
	p1x = rr$x
	p1y = rr$y
	rr = set_polygon_dir( p2x, p2y )
	p2x = rr$x
	p2y = rr$y

	#--------------------------------------------------------------
	# Get list of all points where the polygon borders intersect.
	#--------------------------------------------------------------
	nisectmax = n1*n2
	isectx = array(0., nisectmax)
	isecty = array(0., nisectmax)
	isectsegno1 = array( as.integer(0), nisectmax) # which segment number, starting with 1, in polygon 1
	isectsegno2 = array( as.integer(0), nisectmax) # which segment number, starting with 1, in polygon 2
	nisect = 0
	for( i2 in 1:n2 )
	for( i1 in 1:n1 ) {
		if( i1 == n1 ) 
			seg1 = c( p1x[i1], p1y[i1], p1x[1], p1y[1] )
		else
			seg1 = c( p1x[i1], p1y[i1], p1x[i1+1], p1y[i1+1] )
		if( i2 == n2 ) 
			seg2 = c( p2x[i2], p2y[i2], p2x[1], p2y[1] )
		else
			seg2 = c( p2x[i2], p2y[i2], p2x[i2+1], p2y[i2+1] )
		isect = intersectLineSegs( seg1, seg2 )
		if( isect[1] != 0 ) {
			#-------------------
			# Segments intersect
			#-------------------
			nisect = nisect + 1
			isectx[nisect] = isect[2]
			isecty[nisect] = isect[3]
			isectsegno1[nisect] = i1
			isectsegno2[nisect] = i2
			}
		}

	if( plot) {
		canvas( xlim=range(c(p1x,p2x)), ylim=range(c(p1y,p2y)))
		polygon( p1x, p1y, border='red' )
		polygon( p2x, p2y, border='blue' )
		for( i in 1:nisect )
			points( isectx[i], isecty[i], pch=19, col='green' )
		}

}

#===================================================================================
# Minimum distance between a point and a line
#
# Line: specfied by m and b where y = m*x + b
# Point: specified by x1, y1
#
# IF THE LINE IS VERTICAL, set vertical_x_val to the X value of the vertical line! In
# this case "m" and "b" are not referenced, so can be anything
#
# If return_point, then returns a list with the minimum dist and the point on the specified line
# at which the distance to point (x1,y1) is minimized
#
perpendicular_dist_point_line = function( m, b, x1, y1, return_point=FALSE, plot=FALSE, vertical_x_val=NA ) {

	if( !is.na(vertical_x_val) ) {
		dist = abs( x1 - vertical_x_val )
		if( return_point ) {
			return( list( dist=dist, x=vertical_x_val, y=y1) )
			}
		else
			return( dist )
		}

	#---------------------------------------------------------------
	# Note that if y = mx + b, then we could also write this in the
	# form ax + (bb)y + c = 0 where a == -m, bb == 1, c == -b
	# The standard wikipedia formulas can then be used
	#---------------------------------------------------------------
	num = abs(y1 - m*x1 - b)
	denom = sqrt( m**2 + 1 )
	dist = num / denom

	a  = -m
	bb = 1
	c  = -b
	x0 = x1
	y0 = y1
	num = bb*( bb*x0 - a*y0 ) - a*c
	denom = a*a + bb*bb
	x_mindist = num / denom

	num = a*( -bb*x0 + a*y0 ) - bb*c
	y_mindist = num / denom

	if( plot ) {
		#----------
		# Draw line
		#----------
		xvals = seq(-100,100,by=1)
		yvals = m*xvals + b
		lines( xvals, yvals, col='orange' )

		points( x1, y1, pch=1, col='orange' )
		points( x_mindist, y_mindist, pch=4, col='orange' )
		lines( c(x1,x_mindist), c(y1,y_mindist), col='orange' )
print(paste(">>> perpendicular_dist_point_line: line from", x1, y1, "to", x_mindist, y_mindist ))
		}

	if( return_point ) 	
		return( list( dist=dist, x=x_mindist, y=y_mindist) )
	else
		return( dist )
}

#===================================================================================
# Minimum distance, and point of minimum distance, between a point and a line seg
#
# Line seg: specfied by segx1, segy1, segx2, segy2
# Point: specified by px, py
#
min_dist_point_lineseg = function( segx1, segy1, segx2, segy2, px, py, plot_userunits=FALSE, plot_color=NA ) {

	if( is.na(px) || is.na(py)) stop(paste("Error, entered with px or py as NA: ", px, py ))
	if( is.na(segx1) || is.na(segy1)) stop(paste("Error, entered with seg1 x or y as NA: ", segx1, segy1 ))
	if( is.na(segx2) || is.na(segy2)) stop(paste("Error, entered with seg2 x or y as NA: ", segx2, segy2 ))

	if( plot_userunits ) 
		print(paste("min_dist_point_lineseg: line seg: ", segx1, segy1, 'to', segx2, segy2, ' to pt ', px, py ))

	#----------------------------
	# Disallow collapsed segments
	#----------------------------
	if( (segx1 == segx2) && (segy1 == segy2))
		stop(paste("Error, degenerate line segment, x1==x2 and y1==y2"))

	#-------------------------------------------------------
	# Handle case of a vertical line segment by special code
	#-------------------------------------------------------
	if( segx1 == segx2 ) {
		m = 0.0		# dummy
		b = 0.0		# dummy
		rv = perpendicular_dist_point_line( m, b, px, py, return_point=TRUE, vertical_x_val=segx1 )
		}
	else
		{
		dx = segx2 - segx1
		dy = segy2 - segy1

		m = dy/dx
		b = segy2 - m*segx2

		rv = perpendicular_dist_point_line( m, b, px, py, return_point=TRUE )
		}

	#--------------------------------------------------------------
	# If the min dist point is ON the line seg, that's our solution
	#--------------------------------------------------------------
	xi = rv$x
	yi = rv$y
	x_min = min(c(segx1,segx2))
	x_max = max(c(segx1,segx2))
	y_min = min(c(segy1,segy2))
	y_max = max(c(segy1,segy2))
	if( (xi >= x_min) && (xi <= x_max) && (yi >= y_min) && (yi <= y_max)) {
		#---------------------------------------
		# Min dist point must be on our line seg
		#---------------------------------------
		rv_min_dist = rv$dist
		rv_min_x    = rv$x
		rv_min_y    = rv$y

		if( plot_userunits ) {
			xyuu = inches_to_user_units( rv_min_x, rv_min_y )
			points( xyuu[1], xyuu[2], pch=2, col=plot_color )
			seg1uu = inches_to_user_units( segx1, segy1 )
			seg2uu = inches_to_user_units( segx2, segy2 )
			lines( c(seg1uu[1],seg2uu[1]), c(seg1uu[2],seg2uu[2]), col=plot_color ) 
			print(paste("min_dist_point_lineseg: ON LINE SEG, min dist=", rv_min_dist, "pt=", rv_min_x, rv_min_y ))
			}

		return( list(
			dist	= rv_min_dist,
			x	= rv_min_x,
			y	= rv_min_y ))
		}

	#--------------------------------------------------------------------------------------
	# If we get here, the min dist is NOT on our line seg, so must be one of the end points
	#--------------------------------------------------------------------------------------
	dx1 = abs( segx1 - px )
	dy1 = abs( segy1 - py )
	dist1 = sqrt( dx1*dx1 + dy1*dy1 )

	dx2 = abs( segx2 - px )
	dy2 = abs( segy2 - py )
	dist2 = sqrt( dx2*dx2 + dy2*dy2 )

	if( dist1 < dist2 ) {

		if( plot_userunits ) {
			xyuu = inches_to_user_units( segx1, segy1 )
			points( xyuu[1], xyuu[2], pch=2, col=plot_color )
			seg1uu = inches_to_user_units( segx1, segy1 )
			seg2uu = inches_to_user_units( segx2, segy2 )
			lines( c(seg1uu[1],seg2uu[1]), c(seg1uu[2],seg2uu[2]), col=plot_color ) 
			print(paste("min_dist_point_lineseg: endpoint 1, min dist=", dist1 ))
			}

		return( list(
			dist	= dist1,
			x	= segx1,
			y	= segy1 ))
		}

	if( plot_userunits ) {
		xyuu = inches_to_user_units( segx2, segy2 )
		points( xyuu[1], xyuu[2], pch=2, col=plot_color )
		seg1uu = inches_to_user_units( segx1, segy1 )
		seg2uu = inches_to_user_units( segx2, segy2 )
		lines( c(seg1uu[1],seg2uu[1]), c(seg1uu[2],seg2uu[2]), col=plot_color ) 
		print(paste("min_dist_point_lineseg: endpoint 2, min dist=", dist2 ))
		}

	return( list(
		dist	= dist2,
		x	= segx2,
		y	= segy2 ))
}

#===================================================================================
# Given: specification of a polygon, and X and Y values for a regular grid,
# this returns a logical 2-D mask of TRUE if you are inside (or on the edges of)
# the polygon, and FALSE otherwise.
#
polygon_to_mask = function( polyx, polyy, maskx, masky ) {

	if( ! validPolygon( polyx, polyy )) stop(paste("Error, supplied polygon is not valid (self-intersects)"))

	nx = length( maskx )
	ny = length( masky )
	retval = array(FALSE, dim=c(nx,ny))
	np = length( polyx )
	if( length(polyy) != np ) stop(paste("Error, polygon X and Y point lengths must be identical"))

	for( j in 1:ny )
	for( i in 1:nx ) {
		if( pointInPolygonC( maskx[i], masky[j], polyx, polyy, check_valid=FALSE ))
			retval[i,j] = TRUE
		}

	return( retval )
}

#============================================================================================
# This adds an edge to a map when it's drawn using ccontour with a projection
#
# For an example see: /home/pierce/projects/NSF_aerosols/2014-01-23_second_dec_var_anal_prescribed_vs_interactive/plot_enso_spec_v2.R
#
map_add_edge = function( x0, x1, ... ) {

	yvals = seq(-90,90,by=1)
	nn = length(yvals)
	xvals = array(x0,nn)
	rv = mapproject( xvals, yvals, '' )
	lines( rv$x, rv$y, ... )

	yvals = seq(-90,90,by=1)
	nn = length(yvals)
	xvals = array(x1,nn)
	rv = mapproject( xvals, yvals, '' )
	lines( rv$x, rv$y, ... )
}

#===================================================================================================
map_proj4 = function( dbase, proj4_str, ... ) {

	mm = map(dbase, plot=FALSE, ... )
	npts = length(mm$x)
	xy = array(0., dim=c(npts,2))
	xy[,1] = mm$x
	xy[,2] = mm$y

	pp = project( xy, proj4_str, degrees=TRUE )

	dx = pp[2:npts,1] - pp[1:(npts-1),1]
	dy = pp[2:npts,2] - pp[1:(npts-1),2]
	len = sqrt( dx*dx + dy*dy )

	too_long = 1.e6
	bad = (len > too_long)
	pp[bad,1] = NA
	pp[bad,2] = NA

	lines( pp[,1], pp[,2], ... )
}

#=========================================================================================================
read_city_dbase = function() {

	fname = '/data/obs/boundary_gis/us_highways_and_cities/geonames_cities_database/US.txt.parsed'

	fin = file( fname, 'r' )

	tokbnd = c( 11, 21, 24, 27, 36, 44, 75, 150 )

	nlines_max = 10000
	lon 	= array(0.,  nlines_max)
	lat 	= array(0.,  nlines_max)
	state  	= array(' ', nlines_max)
	country = array(' ', nlines_max)
	pop 	= array(0,   nlines_max)
	elev	= array(0.,  nlines_max)
	tz	= array(' ', nlines_max)
	name	= array(' ', nlines_max)

	l <- readLines( fin, 1 )
	n <- nchar(l)

	lnum = 0
	while( length(n) > 0 ) {

		lnum = lnum + 1
		
		toks = tokenize( l, tokbnd )

		lon[lnum] 	= as.double( toks[[1]] )
		lat[lnum] 	= as.double( toks[[2]] )
		country[lnum] 	= str_trim( toks[[3]] )
		state[lnum] 	= str_trim( toks[[4]] )
		pop[lnum] 	= as.double( toks[[5]] )
		elev[lnum] 	= as.double( toks[[6]] )
		tz[lnum] 	= str_trim( toks[[7]] )
		name[lnum] 	= str_trim( toks[[8]] )

		l <- readLines( fin, 1 )
		n <- nchar(l)
		}

	db = list(
		lon	= lon[1:lnum],
		lat	= lat[1:lnum],
		country	= country[1:lnum],
		state	= state[1:lnum],
		pop	= pop[1:lnum],
		elev	= elev[1:lnum],
		tz	= tz[1:lnum],
		name	= name[1:lnum] )

	close( fin )

	return( db )
}

#=========================================================================================================
get_city_info = function( city_name, state, city_db=NA, debug=FALSE ) {

	if( debug ) print(paste("mlib_map_utils.R::get_city_info: entering with name, state=", city_name, state ))

	if( length(city_db) < 2 ) {
		city_db = read_city_dbase()
		}

	nc = length( city_db$name )
	if( debug ) print(paste("mlib_map_utils.R::get_city_info: ncities known=", nc ))

	for( ic in 1:nc ) {
		if( (city_name == city_db$name[ic]) && (state == city_db$state[ic])) {

			if( debug ) print(paste('get_city_info: found match with index', ic))

			retval = list( 
				lon=city_db$lon[ic], 
				lat=city_db$lat[ic],
				country=city_db$country[ic],
				state=city_db$state[ic],
				pop=city_db$pop[ic],
				elev=city_db$elev[ic],	# IN METERS
				tz=city_db$tz[ic],
				name=city_db$name[ic],
				city_db=city_db )

			return( retval )
			}
		}

	stop(paste("Error: did not find requested city", city_name, "in state", state, "; try looking in file", 
		'/data/obs/boundary_gis/us_highways_and_cities/geonames_cities_database/US.txt.parsed' ))

	return( list( city_db=city_db ) ) 
}

#============================================================================================================
# Get the points of intersection between a circle and a line
#
# The line is described by its slope and interscept
# The circle is described by its center and radius
#
# Returns a list with 0, 1, or 2 two-element arrays:
#	0 : line and circle do not intersect
#	1 : line and circle intersect at a point
#	2 : line and circle intersect at two points
#
intersect_line_circle = function( slope, int, cx, cy, rad ) {

	#-------------------------------------------------------------
	# Get closest distance between line and circle's center point.
	# This is defined by another line perpindicular to the main
	# line, and which passes through the circle's center
	#-------------------------------------------------------------
	mindist = perpendicular_dist_point_line( slope, int, cx, cy ) 
	if( mindist > rad ) 
		return( list() )

	if( mindist == rad ) {
		#------------------------------------------------------------
		# Line is known to pass through center and have slope 1/slope
		#------------------------------------------------------------
		slope_perp = 1/slope
		b_perp = cy - slope_perp*cx
		rv = intersectLines( slope, int, slope_perp, b_perp )
		if( rv[1] != 1 )
			stop(paste("Error, the lines should intersect but dont???"))
		return( list( c( rv[2], rv[32] )))
		}

	#---------------------------------
	# Code taken from web cuz I'm lazy
	#---------------------------------
	x1 = 0
	y1 = int
	x2 = 1
	y2 = slope*x2 + int
	cr = rad

	dx = x2 - x1
	dy = y2 - y1
	a = dx*dx + dy*dy
	b = 2*(dx*(x1-cx)+dy*(y1-cy))
	c = cx*cx + cy*cy +x1*x1 + y1*y1 - 2*(cx*x1 + cy*y1) - cr*cr
	bb4ac = b*b - 4*a*c

	mmu = (-b + sqrt( b*b - 4*a*c )) / (2*a)
	ix1 = x1 + mmu*dx
	iy1 = y1 + mmu*dy
	mmu = (-b - sqrt( b*b - 4*a*c )) / (2*a)
	ix2 = x1 + mmu*dx
	iy2 = y1 + mmu*dy

	return( list(
		c(ix1, iy1),
		c(ix2, iy2)) )
}

#==================================================================================
add_equator_line = function( dx, dx_gap, dy ) {

	dxtot = dx + dx_gap

	x0 = seq( -180, 360, by=dxtot )
	nx = length( x0 )

	for( i in 1:nx ) {
		llx = x0[i]
		lly = -dy/2
		urx = llx + dx
		ury = dy/2

		#---------------------
		# Draw outer white box
		#---------------------
		ddx = dx / 10
		ddy = dy / 4
		llxw = llx - ddx
		llyw = lly - ddy
		urxw = urx + ddx
		uryw = ury + ddy

		polygon( c(llxw,llxw,urxw,urxw), c(llyw,uryw,uryw,llyw), 
			col='white', border=NA )

		polygon( c(llx,llx,urx,urx), c(lly,ury,ury,lly), 
			col='black', border=NA )
		}
}

#====================================================================================================================
# This is true or false whether boxes overlap. See also rect_overlap, which returns 
# the coordinates of the intersection box
#
boxes_overlap = function( llx0, lly0, urx0, ury0, llx1, lly1, urx1, ury1 ) {

	tots_left_or_right_of = (urx0 < llx1) || (llx0 > urx1)
	if( tots_left_or_right_of ) return( FALSE )

	tots_above_or_below = (ury0 < lly1) || (lly0 > ury1)
	if( tots_above_or_below ) return( FALSE )

	enclosing_x_01 = ((llx0 >= llx1) && (llx0 <= urx1)) || ((urx0 >= llx1) && (urx0 <= urx1))
	enclosing_x_10 = ((llx1 >= llx0) && (llx1 <= urx0)) || ((urx1 >= llx0) && (urx1 <= urx0))
	if( (! enclosing_x_01) && (! enclosing_x_10)) return( FALSE )

	enclosing_y_01 = ((lly0 >= lly1) && (lly0 <= ury1)) || ((ury0 >= lly1) && (ury0 <= ury1))
	enclosing_y_10 = ((lly1 >= lly0) && (lly1 <= ury0)) || ((ury1 >= lly0) && (ury1 <= ury0))
	if( (! enclosing_y_01) && (! enclosing_y_10)) return( FALSE )

	return( TRUE )
}

#====================================================================
# Following taken from stackoverflow.com:
# http://stackoverflow.com/questions/9324339/how-much-do-two-rectangles-overlap
#
# Returns an area of 0.0 if the rectangles do not overlap, or the area of the overlap if they do along
# with the coordinates of the overlap rectangle
#
# if area == 0.0, then  llxi, llyi, urxi, uryi do not have any information in them
#
rect_overlap = function( cllx0, clly0, curx0, cury0, cllx1, clly1, curx1, cury1 ) {

	#---------------------------------------
	# Check for collapsed rectangles (lines) 
	#---------------------------------------
	if( (cllx0 == curx0) || (cury0 == clly0) || (cllx1 == curx1) || (cury1 == clly1)) {
		area = 0.0
		return( list( area=0.0 ))
		}

	#--------------------------------------------------------------------------
	# If gave bad specifications of upper and lower or left and right, fix that
	#--------------------------------------------------------------------------
	if( cllx0 < curx0 ) {
		llx0 = cllx0
		urx0 = curx0
		}
	else
		{
		llx0 = curx0
		urx0 = cllx0
		}
	if( clly0 < cury0 ) {
		lly0 = clly0
		ury0 = cury0
		}
	else
		{
		lly0 = cury0
		ury0 = clly0
		}

	if( cllx1 < curx1 ) {
		llx1 = cllx1
		urx1 = curx1
		}
	else
		{
		llx1 = curx1
		urx1 = cllx1
		}

	if( clly1 < cury1 ) {
		lly1 = clly1
		ury1 = cury1
		}
	else
		{
		lly1 = cury1
		ury1 = clly1
		}


	area = 0.0
	llxi = 1.e30
	llyi = 1.e30
	urxi = 1.e30
	uryi = 1.e30

	if( (urx0 <= llx1) || (urx1 <= llx0) || (ury0 <= lly1) || (lly0 >= ury1)) {
		area = 0.0
		return( list( area=area ))
		}

	aRight = urx0
	bRight = urx1
	aLeft  = llx0
	bLeft  = llx1

	aTop   = ury0
	bTop   = ury1
	aBottom = lly0
	bBottom = lly1

    if( (aRight >= bRight) && ( aLeft <= bLeft )) {
        # From x axis point of view, b is wholly contained in a
	llxi = bLeft
	urxi = bRight
        width = bRight - bLeft
	}

    else if( ( bRight >= aRight) && ( bLeft <= aLeft )) {
        # From x axis point of view, a is wholly contained in b
	llxi = aLeft
	urxi = aRight
        width = aRight - aLeft
	}

    else if( aRight >= bRight ) {
    	llxi = aLeft
	urxi = bRight
        width = bRight - aLeft
	}

    else
	{
    	llxi = bLeft
	urxi = aRight
        width = aRight - bLeft
	}

    if( (aTop >= bTop) && ( aBottom <= bBottom )) {
        # From y axis point of view, b is wholly contained in a
	llyi = bBottom
	uryi = bTop
        height = bTop - bBottom
	}

    else if ( (bTop >= aTop) && ( bBottom <= aBottom )) {
        # From y axis point of view, a is wholly contained in b
	llyi = aBottom
	uryi = aTop
        height = aTop - aBottom
	}

    else if( aTop >= bTop ) {
    	llyi = aBottom
	uryi = bTop
        height = bTop - aBottom
	}

    else
    	{
    	llyi = bBottom
	uryi = aTop
        height = aTop - bBottom
	}

	area = width * height

	return( list(
		area	= area,
		llx	= llxi,
		lly	= llyi,
		urx	= urxi,
		ury	= uryi ))
}

#==================================================================================================
draw_LCRB = function( outline=FALSE, ... ) {

	t = reads( '/data/obs/boundary_gis/colorado_river_basin/entire_basin/from_shapefile/lcrb.coords.txt' )
	lon = t[,1]
	lat = t[,2]

	usr = par()$usr
	x0 = usr[1]
	x1 = usr[2]
	if( x0 > 180 ) 
		lon = lon + 360.

	if( outline ) 
		lines( lon, lat, lwd=5, col='white' )

	lines( lon, lat, ... )
}

#==================================================================================================
draw_UCRB = function( outline=FALSE, ... ) {

	t = reads( '/data/obs/boundary_gis/colorado_river_basin/upper/Upper_Colorado_River_Basin_Boundary/ucrb.coords.txt' )
	lon = t[,1]
	lat = t[,2]

	usr = par()$usr
	x0 = usr[1]
	x1 = usr[2]
	if( x0 > 180 ) 
		lon = lon + 360.

	if( outline ) 
		lines( lon, lat, lwd=5, col='white' )

	lines( lon, lat, ... )
}

#================================================================================================
lon_covers_whole_earth = function( xvals ) {

	nx = length( xvals )

	#------------------------------------
	# Put everything on the 0->360 branch
	#------------------------------------
	branch = determine_lon_branch( xvals ) 

	if( branch != '0-360' ) {
		ny = 30
		dat = array( 0., dim=c(nx,ny))
		rv = change_lon_branch( dat, xvals )
		xvals = rv$lon
		}

	bin_10deg = array(FALSE, 36 )
	for( ii in 1:nx ) {
		ibin = floor( xvals[ii] / 10. + 0.5 ) + 1
		if( ibin > 36 ) ibin = 36
		bin_10deg[ibin] = TRUE
		}

	ntrue = sum( bin_10deg )
	return( ntrue == length(bin_10deg) )
}

#================================================================================================
same_grid = function( fname1, x1, y1, fname2, x2, y2 ) {

	nx1 = length(x1)
	nx2 = length(x2)
	ny1 = length(y1)
	ny2 = length(y2)

	err = FALSE

	if( nx1 != nx2 ) err = TRUE
	if( ny1 != ny2 ) err = TRUE

	if( x1[1] != x2[1] ) err = TRUE
	if( y1[1] != y2[1] ) err = TRUE

	if( x1[nx1] != x2[nx2] ) err = TRUE
	if( y1[ny1] != y2[ny2] ) err = TRUE

	if( ! err ) return( TRUE )

	print(paste("ERROR: grids differ!"))
	print(paste("FILE 1:", fname1, "is shape:", nx1, 'x', ny1 ))
	print(paste("FILE 1: xrange=", x1[1], x1[nx1] ))
	print(paste("FILE 1: yrange=", y1[1], y1[ny1] ))

	print(paste("FILE 2:", fname2, "is shape:", nx2, 'x', ny2 ))
	print(paste("FILE 2: xrange=", x2[1], x2[nx2] ))
	print(paste("FILE 2: yrange=", y2[1], y2[ny2] ))

	return( FALSE )
}

#========================================================================
# Given:
#	* the name of an input netcdf file
#	* the name of a binary mask file (0 to NOT include, 1 to include)
#	* the name of the output file
#
# this makes a regional average of the grid cells included in the mask file
#
make_reg_avg_from_mask = function( fname_in, fname_mask, fname_out, verbose=FALSE ) {

	if( verbose ) print(paste("make_reg_avg_from_mask: entering with fname_in=", fname_in, "fname_mask=", 
		fname_mask ))

	#--------------
	# Get mask data
	#--------------
	if( verbose ) print(paste("make_reg_avg_from_mask: getting mask data"))
	rvmask = get_dat( fname_mask, fname_mask )

	nreg = max( rvmask$dat, na.rm=TRUE )
print(paste("***************"))
print(paste("NREG=", nreg))
print(paste("***************"))

	#-------------------
	# Get ALL input data
	#-------------------
	rvdat = get_dat( fname_in, fname_in )
	ncid = nc_open( fname_in )
	var_in = nc_main_var( ncid )
	tdim_in = ncid$dim[[ ncid$unlimdimid ]]

	#----------------------
	# Make sure grids match
	#----------------------
	if( verbose ) print(paste("make_reg_avg_from_mask: check grids match"))
	nx = length( rvmask$xvals )
	ny = length( rvmask$yvals )

	nxd = length( rvdat$xvals )
	nyd = length( rvdat$yvals )
	if( (nx != nxd) || (ny != nyd)) {
		print(paste("Error, shape of mask does not match input file"))
		print(paste("Shape of mask:", nx, ny ))
		print(paste("Shape of data:", nxd, nyd ))
		stop(paste("This is not allowed"))
		}

	if( verbose ) print(paste("make_reg_avg_from_mask: check grids match 2"))
	dx = rvmask$xvals[2] - rvmask$xvals[1]
	dy = rvmask$yvals[2] - rvmask$yvals[1]
	dxd = rvdat$xvals[2] - rvdat$xvals[1]
	dyd = rvdat$yvals[2] - rvdat$yvals[1]
	eps = abs( dx * 1.e-6 )
	if( (abs( dx - dxd ) > eps ) || (abs( dy - dyd ) > eps )) {
		print(paste("Error, dx and/or dy of mask does not match data!"))
		print(paste("dx, dy of mask:", dx, dy ))
		print(paste("dx, dy of data:", dxd, dyd ))
		stop(paste("This is not allowed"))
		}

	nt = dim( rvdat$dat )[3]

	if( nreg == 1 ) 
		retval = array(0, nt)
	else
		retval = array(0, c(nreg,nt))

	for( ireg in 1:nreg ) {

		keepmask = ( rvmask$dat == ireg )

		for( it in 1:nt ) {
			tt = rvdat$dat[,,it]
			if( nreg == 1 ) 
				retval[it] = mean( tt[keepmask], na.rm=TRUE )
			else
				retval[ireg,it] = mean( tt[keepmask], na.rm=TRUE )
			}
		}

	#-----------------
	# Make output file
	#-----------------
	if( verbose ) print(paste("make_reg_avg_from_mask: making output files"))

	if( nreg > 1 ) 
		dim_reg_out = ncdim_def( 'region', '1', 1:nreg )

	if( is.null(tdim_in$calendar ))
		dim_t_out = ncdim_def( tdim_in$name, tdim_in$units, tdim_in$vals, unlim=TRUE )
	else
		dim_t_out = ncdim_def( tdim_in$name, tdim_in$units, tdim_in$vals, unlim=TRUE, calendar=tdim_in$calendar )

	if( nreg > 1 ) 
		var_out = ncvar_def( var_in$name, var_in$units, list(dim_reg_out,dim_t_out), var_in$missval )
	else
		var_out = ncvar_def( var_in$name, var_in$units, list(dim_t_out), var_in$missval )

	print(paste("make_reg_avg_from_mask: creating output file", fname_out ))
	ncid_out = nc_create( fname_out, var_out )
	ncvar_put( ncid_out, var_out, retval )

	nc_close( ncid_out )
	nc_close( ncid )

	return( retval )
}

#====================================================================================
# For some reason, map('county') has egregious errors, such as putting Oceansice
# offshore. This is my version that draws the boundaries from a different data base
#
dp_map_county = function( ... ) {

	fname_shp_base = '/data/obs/boundary_gis/us_counties/census.gov/raw/cb_2017_us_county_500k'
	rv = dpread_shapefile( fname_shp_base )
	nseg = length( rv )
	for( iseg in 1:nseg ) {
		lines( rv[[iseg]]$x, rv[[iseg]]$y, ... )
		}
}

#=====================================================================================
# sabbrev is a 2-letter state abbreviation, such as "CA"
# requries:
#	library('USAboundaries')
#
draw_state_hires = function( sabbrev, ... ) {

	z = us_states( resolution='high', states=sabbrev )

	gg = z$geometry

	nseg = length( gg[[1]] )
	for( iseg in 1:nseg ) {
		tt = gg[[1]][[iseg]][[1]] 
		lon = tt[,1]
		lat = tt[,2]

		lines( lon, lat, ... )
		}
}

#=====================================================================================
state_name_to_abbrev = function( state_name ) {

	return( state.abb[ match( state_name, state.name )] )

}

#=====================================================================================
clim_div_name_init = function() {

	fname = '/home/pierce/src/mine/R/mylib/clim_div_names.txt' 
	fin = file( fname, 'r' )

	l <- readLines( fin, 1 )
	n <- nchar(l)

	map_state_num_2_div_name = list()

	max_ndivs_in_a_state = 12

	statename = NULL
	while( length(n) > 0 ) {

		l = str_trim(l)
		toks = tokenize( l, sepchar=' ' )
		ntoks = length( toks )

		c1 = substr( l, 1, 1 )
		if( dp_isDigit( c1 )) {
			cdivnum = as.integer( toks[1] )
			map_state_num_2_div_name[[ statename ]][  cdivnum ] = pv( toks, skip=1 )
			}
		else
			{
			statename = pv( toks[1:(ntoks-1)] )
			map_state_num_2_div_name[[ statename ]] = array(' ', max_ndivs_in_a_state )
			}

		l <- readLines( fin, 1 )
		n <- nchar(l)
		}

	#print(map_state_num_2_div_name)

	#-------------------
	# Save in global var
	#-------------------
	map_state_num_2_div_name_glob <<- map_state_num_2_div_name

	close( fin )
}

#=====================================================================================
# Given a two letter state name or abbrev, such as "Idaho" or "ID", and a number, returns
# the full climate division name
#
clim_div_name = function( state, num ) {

	if( nchar(state) == 2 ) {
		state = state_abbrev_to_name( abbrev=state, camelcase=TRUE )
		}

	varname_glob = 'map_state_num_2_div_name_glob'
	if( ! exists( varname_glob ))
		clim_div_name_init()

	known_states = names( map_state_num_2_div_name_glob )
	if( ! (state %in% known_states ))
		stop(paste0("Error, state name '", state, "' not known" ))

	retval =  map_state_num_2_div_name_glob[[ state ]][num] 
	if( nchar(retval) < 2 ) 
		stop(paste("Error, invalid clim div state/num combo:", state, num ))
		
	return( retval )
}

#=====================================================================================
# a VERY COARSE vesion of in usa. Do not use this by itself; it is only intended
# to be used as part of an overall larger process!
#
inUS_super_coarse_util = function( lon, lat, plot=FALSE, return_poly=FALSE ) {
	
	npin = length( lon )

	pts = c(
		48.4,  -126.67,
		48.4, -124.7,
		48.3,  -123.1,
		48.97, -122.67,
		48.91,  -95.286,
		47.6,   -90.97,
		47.54,  -85.89,
		46.77, -84.7,
		45.456, -82.539,
		43.39,  -82.36,
		42.02, -83.56,
		41.81, -83.93,
		41.27, -83.70,
		41.698, -81.618,
		44.72, -74.69,
		44.92, -71.50,
		47.06, -69.2,
		47.10,  -68.12,
		44.76,  -67.39,
		24.82, -80.63,
		26.12,  -97.39,
		30.04, -101.598,
		32.00, -106.55,
		31.94, -112.143,
		32.92, -114.68,
		32.0, -125.485
		)

	np = length( pts ) / 2

	ilon = (1:np)*2 
	clon = pts[ ilon ]

	ilat = ilon - 1
	clat = pts[ ilat ]

	if( plot ) {
		map( 'usa' )
		polygon( clon, clat, col=NA, lwd=2, border='red' )
		}

	if( return_poly ) 
		return( list(
			lon	= clon,
			lat	= clat  
			))

	retval = array( FALSE, npin )
	for( ip in 1:npin )
		retval[ip] <- pointInPolygonC( lon[ip], lat[ip], clon, clat )

	return( retval )
}

#=====================================================================================
# a VERY COARSE vesion of in Canada. Do not use this by itself; it is only intended
# to be used as part of an overall larger process!
#
inCA_super_coarse_util = function( lon, lat, plot=FALSE, return_poly=FALSE ) {

	npin = length( lon )

	pts = c(
		49.18,  -137.27,
		49.37, -88.74,
		46, -82, 
		42.5, -80.8 ,
		48.24, -69.6 ,
		48.24, -68.6 ,
		42.24, -64.6 ,
		48.53, -48.28,
		72.94, -61.01,
		71.03, -139.767
		)

	np = length( pts ) / 2

	ilon = (1:np)*2 
	clon = pts[ ilon ]

	ilat = ilon - 1
	clat = pts[ ilat ]

	if( return_poly ) 
		return( list(
			lon	= clon,
			lat	= clat  
			))

	if( plot ) {
		map( 'world' )
		polygon( lon, lat, col=NA, lwd=2, border='red' )
		}

	retval = array( FALSE, npin )
	for( ip in 1:npin )
		retval[ip] <- pointInPolygonC( lon[ip], lat[ip], clon, clat )

	return( retval )
}

#=====================================================================================
# a VERY COARSE vesion of in Canada. Do not use this by itself; it is only intended
# to be used as part of an overall larger process!
#
inMX_super_coarse_util = function( lon, lat, plot=FALSE, return_poly=FALSE ) {

	pts = c(
		32.34, -117.84,
		31.1, -110.96,
		28.908, -103.37,
		25.59, -96.95,
		21.578, -86.20,
		16.14, -94.28,
		14.78, -119.116
		)

	np = length( pts ) / 2

	ilon = (1:np)*2 
	clon = pts[ ilon ]

	ilat = ilon - 1
	clat = pts[ ilat ]

	if( return_poly ) 
		return( list(
			lon	= clon,
			lat	= clat  
			))

	if( plot ) {
		map( 'world' )
		polygon( clon, clat, col=NA, lwd=2, border='red' )
		}

	npin = length( lon )

	retval = array( FALSE, npin )
	for( ip in 1:npin )
		retval[ip] <- pointInPolygonC( lon[ip], lat[ip], clon, clat )

	return( retval )
}

