#======================================================================================
# Add (or concatenate onto) the history attribute for the var or file a message
# indicating what R script generated the file.
# By default adds to global history, if ncvar is set then adds to that var as well
#
# ncid must be open for writing
#
nc_add_history = function( ncid, myname, ncvar=NULL ) {

	rv = ncatt_get( ncid, 0, 'history' )
	if( rv$hasatt ) 
		hist_old = paste( rv$value, '|' )
	else
		hist_old = ''

	dstr = Sys.time()	# "2015-10-05 10:48:39 PDT"

	whoami = system( 'whoami', intern=TRUE )
	mydir  = system( 'pwd', intern=TRUE )

	new_hist = paste0( hist_old, ' ', whoami, ' ', dstr, ' ', mydir, '/', myname )

	ncatt_put( ncid, 0, 'history', new_hist )

	#---------

	if( ! is.null( ncvar )) {
		rv = ncatt_get( ncid, ncvar, 'history' )
		if( rv$hasatt ) 
			hist_old = paste( rv$value, '|' )
		else
			hist_old = ''

		new_hist = paste0( hist_old, ' ', whoami, ' ', dstr, ' ', mydir, '/', myname )
		ncatt_put( ncid, ncvar, 'history', new_hist )
		}
}

#======================================================================================
# A little utility to make it faster to access a single entry from a netcdf file.
# Give it the time units, first and second time values, and number of time entries,
# along with the desired y/m/d, and this gives the index in the file. The idea
# is that you extract this from the file only once, hardcoded if necessary, then
# open the file without reading the entire time axis, and go directly to the entry
# of interest.
#
# See also a more complete but slower version, nc_calc_time_entry_index_full()
#
nc_calc_time_entry_index_fast = function( tunits, tval0, tval1, nt, year, month, day, hour=12, minute=0., calendar='standard' ) {

	dt = tval1 - tval0
	tvals_in_file = (0:(nt-1)) * dt + tval0

	#--------------------------------------------------------
	# Get tval for passed y/m/d in the specified file's units
	#--------------------------------------------------------
	dd = list( year=year, month=month, day=day, hour=hour, minute=minute, seconds=0.0 )
	tval = utInvCalendar( dd, tunits, calendar=calendar )

	ix = wherenearest( tval, tvals_in_file, mindist=dt )

	print(paste("nc_calc_time_entry_index: date", year, '/', month, '/', day, 'is entry', ix ))

	return( ix )
}

#==========================================================================================
# Given a set of (possibly time discontinuous) files, this reads in the dates of all
# of them, and calculates the index for the passed date. It returns the data read-in,
# so on subsequent calls this can be supplied to make the process much faster.
#
# The return list is "as if" it is returning information from the one file that holds
# the requested date, but also has extra info so that the second and subsequent times
# this routine is called, it is much faster.
#
# Note this OPENS and LEAVES OPEN all the files as netcdf files
#
# Useage:
#	flist = getfilelist( "data.*.nc" )
#	date_want = list( list( year=, month=, etc), list( year=year2, )) etc for N dates
#	ndates = length( date_want )
#	eps_days  = 0.5
#
#	for( idate in 1:ndates ) {
#		if( idate == 1 ) {
#			# First call, specify flist NOT rvprev
#			rv = nc_calc_time_entry_index_full( date_want, eps_days, flist=flist )	# takes a long time
#			rv_first = rv
#			}
#		else
#			rv = nc_calc_time_entry_index_full( date_want, eps_days, rvprev=rv_first )
#
#		print(paste("Reading from file", flist[ rv$findex ], "date", fmt_utdates( rv$dates, index=rv$tindex )))
#		dat = ncvar_get( rv$ncid, start=c(1,1,rv$tindex), count=c(-1,-1,1) )
#		}
#
nc_calc_time_entry_index_full = function( date_want, eps_days, flist=NULL, rvprev=NULL, trace=TRUE ) {

	if( trace ) print(paste("nc_calc_time_entry_index_full: entering with date_want=", fmt_utdate( date_want )))

	if( ! is.null( rvprev )) {
		if( trace ) print(paste("nc_calc_time_entry_index_full: previouly called branch"))

		if( ! is.null(flist))
			stop(paste("Error, if you pass rvprev you must NOT pass flist"))

		nf 	= length( rvprev$flist )
		dates_all = rvprev$dates_all
		ncid_all  = rvprev$ncid_all
		flist 	= rvprev$flist

		xvals 	= rvprev$xvals
		yvals 	= rvprev$yvals
		}
	else
		{
		if( trace ) print(paste("nc_calc_time_entry_index_full: first time called branch"))
		if( is.null(flist))
			stop(paste("Error, if do NOT pass rvprev you MUST pass flist"))

		nf = length( flist )
		dates_all = list()
		ncid_all  = list()
		for( ifile in 1:nf ) {
			ncid_all[[ifile]] = nc_open( flist[ifile] )
			tdim = ncid_all[[ifile]]$dim[[ ncid_all[[ifile]]$unlimdimid ]]
			dates_all[[ifile]] = utCalendar( tdim$vals, tdim$units )

			if( ifile == 1 ) {
				xvals = ncid_all[[ifile]]$var[[1]]$dim[[1]]$vals
				yvals = ncid_all[[ifile]]$var[[1]]$dim[[2]]$vals
				}
			}
		}

	ix    = array( as.integer(0), nf)
	ret_tix = -1	# time index in the matching file
	ret_fix = -1	# matching file number
	if( trace ) print(paste("nc_calc_time_entry_index_full: ==e" ))
	for( ifile in 1:nf ) {
		ix[ifile] = wherenearest_date( date_want, dates_all[[ifile]], mindist_hours=eps_days*24 )
		if( ix[ifile] != -1 ) {
			ret_tix = ix[ifile]
			ret_fix = ifile
			print(paste("nc_calc_time_entry_index_full: found matching time index=",
				ret_tix, 'date=', fmt_utdate( dates_all[[ifile]], index=ret_tix ),
				'in file', ifile, ':', flist[ifile] ))
			}
		}

	if( trace ) print(paste("nc_calc_time_entry_index_full: ==f" ))

	#--------------------------------------
	# Only allow ONE return file/index pair
	#--------------------------------------
	if( sum( ix > 0 ) > 1 ) 
		ret_ix = -1

	if( trace ) print(paste("nc_calc_time_entry_index_full: returning" ))

	retval = list(
		xvals		= xvals,
		yvals		= yvals,
		flist		= flist,
		dates		= dates_all[[ret_fix]],	 
		ncid		= ncid_all[[ret_fix]],		# the OPENED netcdf files, used to get the data from the right file/time index
		date_want	= date_want,
		eps_days	= eps_days,
		findex		= ret_fix,	# index of file name in flist that matches the desired date, or -1 if none match
		tindex		= ret_tix,	# time index in flist[findex] that matches date_want, or -1 if no file entry found that is near date_want
		dates_all	= dates_all,
		ncid_all	= ncid_all
		)

	return( retval )
}

#======================================================================================
# used to be get_indices_for_vals
#
nc_get_indices_for_geobox = function( ncvar, llx, lly, urx, ury, x_offset=0 ) {

	xvals = ncvar$dim[[1]]$vals
	yvals = ncvar$dim[[2]]$vals
	if( (length(xvals) < 2) || (length(yvals) < 2))
		stop(paste("Error, unreasonable length of x and/or y arrays:", length(xvals), length(yvals), "varname:", ncvar$name )) 

	return( get_indices_for_geobox( xvals, yvals, llx, lly, urx, ury, x_offset, varname=ncvar$name ))
}

#=======================================================================================================
get_indices_for_geobox = function( xvals, yvals, llx, lly, urx, ury, x_offset=0, varname='' ) {

	if( (length(xvals) < 2) || (length(yvals) < 2))
		stop(paste("Error, unreasonable length of x and/or y arrays:", length(xvals), length(yvals) ))

	nnax = sum( is.na( xvals ))
	nnay = sum( is.na( yvals ))
	if( nnax > 0 ) stop(paste("error, xvals must have no na's, but has:", nnax, "varname:", varname ))
	if( nnay > 0 ) stop(paste("error, yvals must have no na's, but has:", nnay, "varname:", varname ))

	xvals = xvals + x_offset

	dx = max(xvals) - min(xvals)
	eps = abs(dx/20.)

	illx = wherenearest( llx, xvals, mindist=eps )
	iurx = wherenearest( urx, xvals, mindist=eps )
	if( (illx < 0) || (iurx < 0)) {
		quickstats( xvals, 'xvals' )
		illx = wherenearest( llx, xvals )
		iurx = wherenearest( urx, xvals )
		print(paste("closest file X value to user-specifed llx value of", llx, ':', xvals[illx] ))
		print(paste("closest file X value to user-specifed urx value of", urx, ':', xvals[iurx] ))
		stop(paste("did not find specified X value:", llx, urx, 'eps=', eps ))
		}

	illy = wherenearest( lly, yvals, mindist=eps )
	iury = wherenearest( ury, yvals, mindist=eps )
	if( (illy < 0) || (iury < 0)) {
		quickstats( yvals, 'yvals' )
		illy = wherenearest( lly, yvals )
		iury = wherenearest( ury, yvals )
		print(paste("closest file Y value to user-specifed lly value of", lly, ':', yvals[illy] ))
		print(paste("closest file Y value to user-specifed ury value of", ury, ':', yvals[iury] ))
		stop(paste("did not find specified Y value:", lly, ury, 'eps=', eps ))
		}

	countx = iurx - illx + 1
	county = iury - illy + 1

	return( c( illx, illy, iurx, iury, countx, county ))
}

#======================================================================================
nc_var_exists = function( ncid, varname ) {

	if( class(ncid) != 'ncdf4' )
		stop(paste("Error, first argument must be the return value from a nc_open command!"))

	if( storage.mode(varname) != 'character' )
		stop("Error, second argument must be a character string, with the name of the variable to be checked")

	return( ! is.null( ncid$var[[varname]] ))
}

#======================================================================================
# This is the multi-file version of get_dat.
#
# if geobox is supplied, it must be in degrees lon/lat with order: c(llx, lly, urx, ury)
#
# if tz_offset_hours is specified, it is ASSUMED that the date units in the file are hours,
# and the passed value is added to each time value. This effectively changes the time zone
# in which the data values are valid.
#
get_dat = function( fname, tag=fname, varname='', geobox=NULL, change_lon_branch=TRUE, start=NA, count=NA,
	tzoffset_hours=NA, collapse_degen=TRUE, ignore_dates=FALSE, ... ) {

	nfiles = length( fname )

	if( nfiles == 1 ) 
		return( get_dat_onefile( fname, tag=tag, varname=varname, geobox=geobox,
			change_lon_branch=change_lon_branch, 
			start=start, count=count,
			tzoffset_hours=tzoffset_hours,
			ignore_dates=ignore_dates,
			collapse_degen=collapse_degen, ... ))

	for( ifile in 1:nfiles ) {
		trv = get_dat_onefile( fname[ifile], tag=tag, varname=varname, geobox=geobox,
			change_lon_branch=change_lon_branch, 
			start=start, count=count,
			tzoffset_hours=tzoffset_hours,
			ignore_dates=ignore_dates,
			collapse_degen=collapse_degen, ... )

		if( ifile == 1 ) {
			dat	= trv$dat
			xvals	= trv$xvals
			yvals	= trv$yvals
			dates	= trv$dates
			tdim	= trv$tdim
			tvals	= trv$tvals
			tunits	= trv$tunits
			fp_years = trv$fp_years
			varname = trv$varname
			varunits = trv$varunits
			missval  = trv$var$missval
			var	= trv$var
			tag	= trv$tag 
			}
		else
			{
			dat 	= c_arrays( dat, trv$dat )
			dates	= cat_dates( dates, trv$dates )
			tvals	= c( tvals, trv$tvals )
			fp_years = c( fp_years, trv$fp_years )
			}
		}

	return( list(
		dat	= dat,
		xvals	= xvals,
		yvals	= yvals,
		dates	= dates,
		tdim	= tdim,
		tvals	= tvals,	# It is important to note that tvals MAY NOT BE VALID for multiple files, if the time units are different in each!
		tunits	= tunits,
		fp_years = fp_years,
		fname	= fname,
		varname = varname,
		varunits = varunits,
		missval  = missval,
		var	= var,
		tag	= tag ))
}

#======================================================================================
# if geobox is supplied, it must be in degrees lon/lat with order: c(llx, lly, urx, ury)
#
# if tz_offset_hours is specified, it is ASSUMED that the date units in the file are hours,
# and the passed value is added to each time value. This effectively changes the time zone
# in which the data values are valid.
#
get_dat_onefile = function( fname, tag=fname, varname='', geobox=NULL, change_lon_branch=TRUE, start=NA, count=NA,
	tzoffset_hours=NA, collapse_degen=TRUE, ignore_dates=FALSE, quiet=FALSE, force_dim3_time=FALSE ) {

	if( ! quiet ) print(paste0('get_dat: opening file "', fname, '"' ))
	ncid = nc_open( fname )
	if( nchar( varname ) > 1 ) {
		var = ncid$var[[ varname ]]
		if( is.null(var)) stop(paste("error, specified varname", varname, "not found in file", fname ))
		}
	else
		{
		var = nc_main_var( ncid )
		}
	if( ! quiet ) print(paste("get_dat: using var", var$name, "from file", fname ))

	varname = var$name
	varunits = var$units
	calendar = NULL

	ndims = var$ndims

	if( !is.null(geobox)) {

		if( !is.na(start) || !is.na(count))
			stop(paste("Error, cannot specify BOTH geobox AND start/count"))

		gbox = nc_get_indices_for_geobox( var, geobox[1], geobox[2], geobox[3], geobox[4] )
		illx = gbox[1]
		illy = gbox[2]
		iurx = gbox[3]
		iury = gbox[4]

		start = c(illx,illy)
		if( ndims > 2 )
			start = c( start, 1 )
		count = c( iurx-illx+1, iury-illy+1 )
		if( ndims > 2 )
			count = c( count, -1 )
		}

	sl1 = (length(start) == 1)
	cl1 = (length(count) == 1)
	specified_start = ( (! sl1) || (sl1 && (!is.na(start))))
	specified_count = ( (! cl1) || (cl1 && (!is.na(count))))

	if( specified_start && (! specified_count))
		stop(paste("sorry must specify BOTH start and count if either"))

	if( specified_count && (! specified_start))
		stop(paste("sorry must specify BOTH start and count if either"))

	xvals = var$dim[[1]]$vals
	nx = length( xvals )

	cwe = lon_covers_whole_earth( xvals ) 
	if( change_lon_branch && (! cwe) ) {
		if( max( xvals ) > 181 ) xvals = xvals - 360
		}

	yvals = NA
	if( ndims > 1 ) {
		yvals = var$dim[[2]]$vals
		ny = length( yvals )
		}

	tdim_force = NULL
	is_climo_file = (ncid$unlimdimid == -1)
	if( force_dim3_time ) {
		is_climo_file = FALSE
		tdim_force = var$dim[[3]]
		}
	
	if( (ndims == 3) && is_climo_file ) {

		#--------------------
		# A climatology file?
		#--------------------
		dates = NULL
		fp_years = NULL

		tdim = var$dim[[3]]
		tunits = tdim$units
		nt = tdim$len
		tvals = 1:nt
		if( specified_start ) {
			if( count[3] == -1 )
				count_t = nt
			else
				count_t = count[3]
			ee = start + count - 1
			ee = start[3] + count_t - 1
			tvals = tvals[start[3]:ee]
			}
		}

	else if( (ndims >= 3) && (! is_climo_file)) {

		if( is.null( tdim_force ))
			tdim = ncid$dim[[ ncid$unlimdimid ]]
		else
			tdim = tdim_force

		tunits = tdim$units
		nt = tdim$len
		count_t = nt
		rvcal = ncatt_get( ncid, tdim$name, 'calendar' )
		if( rvcal$hasatt ) {
			calendar = rvcal$value
			print(paste('    .... using calendar >', calendar, '<' ))
			}
		else
			{
			calendar = 'standard'
			print(paste('    .... NO CALENDAR SPECIFIED, using default calendar >', calendar, '<' ))
			}

		tvals = tdim$vals
		if( specified_start ) {
			if( count[3] == -1 )
				count_t = nt
			else
				count_t = count[3]
			ee = start + count - 1
			ee = start[3] + count_t - 1
			tvals = tvals[start[3]:ee]
			}
			
		if( ! is.na( tzoffset_hours ))
			tvals = tvals + tzoffset_hours

		if( is.null( calendar ))
			ignore_dates = TRUE

		if( ignore_dates ) 
			dates = NULL
		else
			{
			dates = utCalendar( tvals, tdim$units, calendar=calendar )
			}

		if( ! is.na( tzoffset_hours ))
			dates$tzoffset_hours = array( tzoffset_hours, length( dates$tzoffset_hours ))

		fp_years = tvals_to_fp_year( dates=dates )

		if( ! ignore_dates ) {
			print(paste("...dates in that file:", fmt_utdate( dates, index=1), 'to', fmt_utdate( dates, index=count_t)))
			}
		}
	
	else if( (ndims <= 2) && (ncid$unlimdimid != -1) ) {
		tdim = ncid$dim[[ ncid$unlimdimid ]]
		tunits = tdim$units
		rvcal  = ncatt_get( ncid, tdim$name, 'calendar' )
		if( rvcal$hasatt )
			calendar = rvcal$value

		tvals = tdim$vals
		if( ! is.na( tzoffset_hours ))
			tvals = tvals + tzoffset_hours

		if( ignore_dates ) 
			dates = NULL
		else
			dates = utCalendar( tvals, tunits )

		if( ! is.na( tzoffset_hours ))
			dates$tzoffset_hours = array( tzoffset_hours, length( dates$tzoffset_hours ))

		fp_years = tvals_to_fp_year( dates=dates )
		}

	else
		{
		dates = NA
		fp_years = NA
		tvals = NA
		tunits = NA
		tdim  = NULL
		}

	if( ! specified_start )
		start = array(1, ndims)
	if( ! specified_count ) {
		if( ndims == 1 ) 
			count = nx
		else if( ndims == 2 ) 
			count = c(nx,ny)
		else if( ndims == 3 ) 
			count = c(nx,ny,nt)
		else
			count = var$size
		}

	dat = ncvar_get( ncid, var, start=start, count=count, collapse_degen=collapse_degen )

	act_count = dim(dat)	# use instead of "count" since count might have -1's

	if( specified_start ) {
		ee = start + act_count - 1
		xvals = xvals[start[1]:ee[1] ]
		if( ndims >= 2 ) 
			yvals = yvals[start[2]:ee[2] ]
		}

	nc_close( ncid )

	return( list(
		dat	= dat,
		xvals	= xvals,
		yvals	= yvals,
		dates	= dates,
		tdim	= tdim,
		tvals	= tvals,
		tunits	= tunits,
		calendar = calendar,
		fp_years = fp_years,
		fname	= fname,
		varname = varname,
		varunits = varunits,
		missval  = var$missval,
		var	= var,
		tag	= tag ))
}

#===========================================================================================
# Given a regex type pattern, this returns the associated variable. Or, NULL if no match
# is found
#
# Example:
#
#	var = nc_grep_var( ncid, '^Weighted_EOF-*' )
#
nc_grep_var = function( ncid, pattern, verbose=FALSE ) {

	if( verbose ) print(paste("nc_grep_var: entering with file=", ncid$filename, 'pattern=', pattern ))

	nv = ncid$nvars
	vname = array(' ', nv)

	for( ivar in 1:nv ) {
		vname[ivar] = ncid$var[[ivar]]$name
		}

	if( verbose ) print(paste('nc_grep_var: vars in file: ', pv(vname)))

	z = grep( pattern, vname )
	
	if( verbose ) {
		print(paste("z result:"))
		print(z)
		print('---')
		}

	if( length(z) == 1 ) {
		return( ncid$var[[z]] )
		}

	if( length(z) == 0 ) stop(paste("error, no variable name in file", ncid$filename, 'matches pattern', pattern ))

	stop(paste("error, multiple variables in file", ncid$fname, 'match pattern', pattern ))
}

#===========================================================================================
# Returns a ncvar structure (list). If there are NO vars in the file, then returns NULL.
# This can happen b/c dimvars are excluded!
#
nc_main_var = function( ncid, debug=FALSE ) {

	if( debug ) print(paste('nc_main_var: entering for file', ncid$filename ))

	nv = ncid$nvars
	if( nv == 0 ) return( NULL )

	if( nv == 1 ) {
		if( debug ) print(paste('nc_main_var: just one var in file, returning:', ncid$var[[1]]$name ))
		return( ncid$var[[1]] )
		}

	vsize = array(0., nv)
	for( ivar in 1:nv ) {
		nd = ncid$var[[ivar]]$ndims
		if( nd > 0 ) {
			size = 1
			for( idim in 1:nd ) 
				size = size * ncid$var[[ivar]]$dim[[idim]]$len
			vsize[ivar] = size
			}
		else
			vsize[ivar] = 1
		if( debug ) print(paste("nc_main_var: var", ivar, '/', nv, "=", ncid$var[[ivar]]$name, 'size:', vsize[ivar] ))
		}

	im = which.max(vsize)

	if( debug ) print(paste('nc_main_var: max size is index', im, "size", vsize[im], 'var', ncid$var[[im]]$name ))
	return( ncid$var[[ im ]] )
}

#===========================================================================================
# Different ways to call this:
#
# Can call with a file name or an already opened ncid
#
# Can supply all of (year, month, day) OR a date (a list with $year, $month, $day)
# If you supply a date, you can optionally supply an index for an array of dates
#
ncinfo_date2index = function( fname='', ncid=NA, year=NA, month=NA, day=NA, date=NA, index=NA ) {

	if( nchar(fname) == 0 ) 
		fname = ncid$filename

	if( !is.na(year)) {
		yy = year
		mm = month
		dd = day
		}
	else
		{
		if( !is.na(index)) {
			yy = date$year [index]
			mm = date$month[index]
			dd = date$day  [index]
			}
		else
			{
			if( length( date$year ) > 1 ) 
				stop(paste("Error, array of dates not supported; use index= arg to pass one date from array. Passed index=", index ))
			yy = date$year 
			mm = date$month
			dd = date$day  
			}
		}

	cmd = paste0( 'ncinfo -date2index ', yy, '/', mformatC('%02d',mm), '/', mformatC('%02d',dd), ' ', pv(fname) )
	print(cmd)
	rv = system( cmd, intern=TRUE )

	return( as.integer( rv ))
}

#===========================================================================================
# Note: 'varname_src' can be integer 0 for global atts
#
# NOTE: new versions of the netcdf library stop with an error if you try to define the
# attribute _FillValue when there is already data in the file. In this event, set flag
# skip_fill_value and that att will not be defined.
#
nc4_copy_atts = function( ncid_src, ncid_dest, varname_src, varname_dest, 
		clobber_existing=FALSE, verbose=FALSE,
		stoplist=NULL,
		skip_fill_value=FALSE)  {

	if( verbose ) print(paste("nc4_copy_atts: entering with src, dest=",
		varname_src, varname_dest ))

	#--------------------------------
	# Get attributes from source file
	#--------------------------------
	att_src = ncatt_get( ncid_src, varname_src )
	names_src = names( att_src )
	natts = length(att_src)

	if( verbose ) {
		print('names_src:')
		print(names_src)
		print(paste("natts src:", natts))
		}

	if( natts == 0 ) 
		return()

	for( iatt in 1:natts ) {
		#-----------------------------------
		# See if already exists in dest file
		#-----------------------------------
		att_dest = ncatt_get( ncid_dest, varname_dest, names_src[iatt] )

		skipit = ((names_src[iatt] == "_FillValue") && skip_fill_value)
		if( ! is.null( stoplist ))
			if( names_src[iatt] %in% stoplist )
				skipit = TRUE

		if( att_dest$hasatt ) {
			if( clobber_existing && (! skipit) ) {
				ncatt_put( ncid_dest, varname_dest, names_src[iatt], att_src[[iatt]] )
				}
			}
		else
			if( ! skipit ) 
				ncatt_put( ncid_dest, varname_dest, names_src[iatt], att_src[[iatt]] )
		}
}

#==================================================================================================
# Given a file with a big domain and a file with a small domain, this returns the start and
# count needed to make the big domain file be on the small domain
#
nc_determine_overlap = function( fname_big, fname_small, xoffset_big=0 ) {

	ncid_big = nc_open( fname_big )
	ncid_small = nc_open( fname_small )

	var_big = nc_main_var( ncid_big )
	var_small = nc_main_var( ncid_small )

	xvals_big = var_big$dim[[1]]$vals + xoffset_big
	yvals_big = var_big$dim[[2]]$vals
	xvals_small = var_small$dim[[1]]$vals
	yvals_small = var_small$dim[[2]]$vals

	quickstats( xvals_big, 'xvals_big' )
	quickstats( yvals_big, 'yvals_big' )
	quickstats( xvals_small, 'xvals_small' )
	quickstats( yvals_small, 'yvals_small' )

	eps = (xvals_small[2] - xvals_small[1]) * 0.01

	ix0 = wherenearest( xvals_small[1], xvals_big, mindist=eps )
	if( ix0 < 0 ) stop(paste("Error, did not find LLX value in overlap!"))

	iy0 = wherenearest( yvals_small[1], yvals_big, mindist=eps )
	if( ix0 < 0 ) stop(paste("Error, did not find LLY value in overlap!"))

	nxs = length( xvals_small )
	nys = length( yvals_small )

	ix1 = wherenearest( xvals_small[nxs], xvals_big, mindist=eps )
	if( ix1 < 0 ) stop(paste("Error, did not find URX value in overlap!"))

	iy1 = wherenearest( yvals_small[nys], yvals_big, mindist=eps )
	if( iy1 < 0 ) stop(paste("Error, did not find URY value in overlap!"))

	print(paste0("ncunary -out t.nc -start ", ix0, ',', iy0, ',1 -end ', ix1, ',', iy1, ',: ', fname_big ))
}

#==================================================================================================
# Given 2 variables in 2 netcdf files, this determines the 3-d start and count for the overlap
# of the two arrays. See also nc_determine_overlap
# 
# Useage:
#
#	rv = nc4_start_count_overlap( ncid1, ncid2, varname1, varname2 )
#
#	dat1 = ncvar_get( ncid1, var1, start=c(rv$start1,it0), count=c(rv$count,nt))
#	dat2 = ncvar_get( ncid2, var2, start=c(rv$start2,it0), count=c(rv$count,nt))
#
#	xcommon = xvals1[rv$start1[1]:rv$end1[1]]
#	ycommon = yvals1[rv$start1[2]:rv$end1[2]]
#
# This version does the X, Y, and T overlap. For only the 2-d overlap, see:
#	nc4_start_count_overlap_2d
#
# If you want to pass FILENAMES instead of NCIDS, pass NULL as ncid1 and ncid2 and fill out fname1 and fname2
#
nc4_start_count_overlap_3d = function( ncid1, ncid2, varname1='', varname2='', retain_2_lons=FALSE, fname1='', fname2='', verbose=FALSE ) {

	I_NCIDS = 1
	I_FNAMES = 2

	if( nchar(fname1) == 0 )
		imode = I_NCIDS
	else
		imode = I_FNAMES

	if( imode == I_FNAMES ) {
		ncid1 = nc_open( fname1 )
		ncid2 = nc_open( fname2 )
		}
	else
		{
		fname1 = ncid1$fname
		fname2 = ncid2$fname
		}

	X = 1
	Y = 2

	if( nchar(varname1) < 1 )
		var1 = nc_main_var( ncid1 )
	else
		{
		var1 = ncid1$var[[varname1]]
		if( is.null(var1)) stop(paste("Error, specified var name", varname1, 'for file 1, but that file holds no such var' ))
		}
	
	if( nchar(varname2) < 1 )
		var2 = nc_main_var( ncid2 )
	else
		{
		var2 = ncid2$var[[varname2]]
		if( is.null(var2)) stop(paste("Error, specified var name", varname2, 'for file 2, but that file holds no such var' ))
		}

	if( verbose ) {
		print(paste("nc4_start_count_overlap: from file 1 using var", var1$name ))
		print(paste("nc4_start_count_overlap: from file 2 using var", var2$name ))
		}

	x1 = var1$dim[[X]]$vals
	x2 = var2$dim[[X]]$vals
	y1 = var1$dim[[Y]]$vals
	y2 = var2$dim[[Y]]$vals

	rvo = nc4_get_overlap( x1, y1, x2, y2, retain_2_lons=retain_2_lons, verbose=verbose )	# $start1, $start2, $end1, $end2, $count

	#-----------------
	# Get time overlap
	#-----------------
	tol = 1/24
	ta = list()
	ta[[1]] = ncid1
	ta[[2]] = ncid2
	rvto = nc_overlapping_period( NULL, tol, ncids=ta, verbose=verbose )		# $start, $count, $sc_string 

	ss1 = paste0( 
		'-start ', rvo$start1[1], ',', rvo$start1[2], ',', rvto$start[1],  ' ',
		'-count ', rvo$count[1], ',', rvo$count[2], ',', rvto$count[1], ' ', 
		fname1 )

	ss2 = paste0( 
		'-start ', rvo$start2[1], ',', rvo$start2[2], ',', rvto$start[2],  ' ',
		'-count ', rvo$count[1], ',', rvo$count[2], ',', rvto$count[2], ' ', 
		fname2 ) 

	if( verbose ) {
		cat( 'ncbinop ', ss1, ' -- ', ss2, '\n', sep='' )
		}

	if( imode == I_FNAMES ) {
		nc_close( ncid1 )
		nc_close( ncid2 )
		}

	return( c( ss1, ss2 ))
}

#==================================================================================================
# Given 2 variables in 2 netcdf files, this determines the 2-d start and count for the overlap
# of the two arrays. See also nc_determine_overlap
# 
# Useage:
#
#	rv = nc4_start_count_overlap( ncid1, ncid2, varname1, varname2 )
#
#	dat1 = ncvar_get( ncid1, var1, start=c(rv$start1,it0), count=c(rv$count,nt))
#	dat2 = ncvar_get( ncid2, var2, start=c(rv$start2,it0), count=c(rv$count,nt))
#
#	xcommon = xvals1[rv$start1[1]:rv$end1[1]]
#	ycommon = yvals1[rv$start1[2]:rv$end1[2]]
#
# This version does only the X and Y overlap. See also:
#	nc4_start_count_overlap_3d
#
nc4_start_count_overlap_2d = function( ncid1, ncid2, varname1='', varname2='', retain_2_lons=FALSE, verbose=TRUE ) {

	X = 1
	Y = 2

	if( nchar(varname1) < 1 )
		var1 = nc_main_var( ncid1 )
	else
		{
		var1 = ncid1$var[[varname1]]
		if( is.null(var1)) stop(paste("Error, specified var name", varname1, 'for file 1, but that file holds no such var' ))
		}
	
	if( nchar(varname2) < 1 )
		var2 = nc_main_var( ncid2 )
	else
		{
		var2 = ncid2$var[[varname2]]
		if( is.null(var2)) stop(paste("Error, specified var name", varname2, 'for file 2, but that file holds no such var' ))
		}

	print(paste("nc4_start_count_overlap: from file 1 using var", var1$name ))
	print(paste("nc4_start_count_overlap: from file 2 using var", var2$name ))

	x1 = var1$dim[[X]]$vals
	x2 = var2$dim[[X]]$vals
	y1 = var1$dim[[Y]]$vals
	y2 = var2$dim[[Y]]$vals

	return( nc4_get_overlap( x1, y1, x2, y2, retain_2_lons=retain_2_lons, verbose=verbose ))
}

#======================================================================================================
# This is the lowest level overlap function, it is just given the x and y arrays for the 2 data sets
#
#	dat1 = ncvar_get( ncid1, var1, start=c(rv$start1,it0), count=c(rv$count,nt))
#	dat2 = ncvar_get( ncid2, var2, start=c(rv$start2,it0), count=c(rv$count,nt))
#
#	xcommon = xvals1[rv$start1[1]:rv$end1[1]]
#	ycommon = yvals1[rv$start1[2]:rv$end1[2]]
#
nc4_get_overlap = function( x1, y1, x2, y2, retain_2_lons=FALSE, verbose=TRUE ) {

	X = 1
	Y = 2

	nx1 = length( x1 )
	ny1 = length( y1 )
	nx2 = length( x2 )
	ny2 = length( y2 )

	#-----------------------
	# On different branches?
	#-----------------------
	if( max(x1) > 181 ) x1 = x1 - 360
	if( (! retain_2_lons) && (max(x2) > 181)) x2 = x2 - 360

	if( verbose ) {
		print(paste( "X range: array 1:", pv(range(x1)), " array 2:", pv(range(x2)) ))
		print(paste( "Y range: array 1:", pv(range(y1)), " array 2:", pv(range(y2)) ))
		}

	#-------------------------
	# Different reversed in y?
	#-------------------------
	rev1 = (y1[1] > y1[2])
	rev2 = (y2[1] > y2[2])
	if( rev1 != rev2 ) 
		stop(paste("Error, one array is flipped in Y and the other is not"))

	#-------------------------------
	# Completely outside each other?
	#-------------------------------
	if( x1[nx1] < x2[1] ) stop(paste("Error, array 1 is completely to left of array 2"))
	if( x2[nx2] < x1[1] ) stop(paste("Error, array 2 is completely to left of array 1"))
	if( y1[ny1] < y2[1] ) stop(paste("Error, array 1 is completely below array 2"))
	if( y2[ny2] < y1[1] ) stop(paste("Error, array 2 is completely below array 1"))

	xleft  = max( c( x1[1], x2[1] ))
	xright = min( c( x1[nx1], x2[nx2] ))
	ybot   = max( c( y1[1], y2[1] ))
	ytop   = min( c( y1[ny1], y2[ny2] ))

	start1 = array(0, 2)
	start2 = array(0, 2)
	end1   = array(0, 2)
	end2   = array(0, 2)
	count1 = array(0, 2)
	count2 = array(0, 2)

	start1[X] = wherenearest( xleft, x1 )
	start2[X] = wherenearest( xleft, x2 )
	end1[X]   = wherenearest( xright, x1 )
	end2[X]   = wherenearest( xright, x2 )

	start1[Y] = wherenearest( ybot, y1 )
	start2[Y] = wherenearest( ybot, y2 )
	end1[Y]   = wherenearest( ytop, y1 )
	end2[Y]   = wherenearest( ytop, y2 )

	#----------------------
	# Check if flipped in Y
	#----------------------
	if( start1[Y] > end1[Y] ) {
		ii = start1[Y]
		start1[Y] = end1[Y]
		end1[Y] = ii

		ii = start2[Y]
		start2[Y] = end2[Y]
		end2[Y] = ii
		}

	count1[X] = end1[X] - start1[X]	+ 1 # x component
	count1[Y] = end1[Y] - start1[Y]	+ 1 # y component

	count2[X] = end2[X] - start2[X]	+ 1 # x component
	count2[Y] = end2[Y] - start2[Y]	+ 1 # y component

	#---------------------------------------
	# Make sure count is same in both arrays
	#---------------------------------------
	if( count1[X] != count2[X] ) stop(paste("Error, in array 1 X count is ", count1[X], " but in array 2 it is", count2[X] ))
	if( count1[Y] != count2[Y] ) stop(paste("Error, in array 1 Y count is ", count1[Y], " but in array 2 it is", count2[Y] ))

	#-----------------------------------------------
	# Sanity check -- make sure actual vals are same
	#-----------------------------------------------
	epsx = abs( x1[2] - x1[1] )
	epsy = abs( y1[2] - y1[1] )
	eps = min( c( epsx, epsy ))

	for( ii in 1:count1[X] ) {
		ii1 = start1[X] + ii - 1
		ii2 = start2[X] + ii - 1
		diff = abs( x1[ii1] - x2[ii2] )
		if( diff > eps ) stop(paste("Error in sanity check; X entry", ii1, "in array 1 should equal entry", 
			ii2, "in array 2, but vals are:", x1[ii1], x2[ii2] ))
		}
		
	for( jj in 1:count1[Y] ) {
		jj1 = start1[Y] + jj - 1
		jj2 = start2[Y] + jj - 1
		diff = abs( y1[jj1] - y2[jj2] )
		if( diff > eps ) stop(paste("Error in sanity check; Y entry", jj1, "in array 1 should equal entry", 
			jj2, "in array 2, but vals are:", y1[jj1], y2[jj2] ))
		}

	if( verbose ) {
		print(paste("Orig    shape 1:", nx1, 'x', ny1, ' 2:', nx2, 'x', ny2 ))
		print(paste("Overlap shape  :", count1[1], 'x', count1[2] ))

		print(paste("File 1 overlap range X:", x1[ start1[1] ], x1[ end1[1] ] ))
		print(paste("File 1 overlap range Y:", y1[ start1[2] ], y1[ end1[2] ] ))

		print(paste("File 2 overlap range X:", x2[ start2[1] ], x2[ end2[1] ] ))
		print(paste("File 2 overlap range Y:", y2[ start2[2] ], y2[ end2[2] ] ))

		print('')
		print(paste0("These will give the same region:"))
		print(paste0("ncunary -start ", start1[1], ',', start1[2], ' -count ', count1[1], ',', count1[2], ' infile1.nc' ))
		print(paste0("ncunary -start ", start2[1], ',', start2[2], ' -count ', count1[1], ',', count1[2], ' infile2.nc' ))
		}

	rv = list(
		start1	= start1,	# [2], x and y
		start2	= start2,	# [2], x and y
		end1	= end1,		# [2], x and y
		end2	= end2,		# [2], x and y
		count	= count1 )	# [2], x and y

	return( rv )
}

#=====================================================================================================
copy_missval_mask_file1_file2 = function( fname_mvmask_src, fname_dest ) {

	print(paste("copy_missval_mask_file1_file2: copying mask from"))
	print(paste(fname_mvmask_src))
	print(paste("to"))
	print(paste(fname_dest))

	ncid = nc_open( fname_mvmask_src, readunlim=FALSE )
	var = ncid$var[[1]]
	ndims = var$ndims
	if( ndims == 2 ) {
		start = c(1,1)
		count = c(-1,-1)
		}
	else
		{
		start = c(1,1,1)
		count = c(-1,-1,1)
		}
		
	dat_mask = ncvar_get( ncid, var, start=start, count=count )

	nxm = dim( dat_mask )[1]
	nym = dim( dat_mask )[2]

	print(paste("copy_missval_mask_file1_file2: mask shape:", nxm, nym ))

	nc_close( ncid )

	#--------
	ncid = nc_open( fname_dest, readunlim=FALSE, write=TRUE )
	var = ncid$var[[1]]
	ndims = var$ndims
	nx = var$dim[[1]]$len
	ny = var$dim[[2]]$len

	if( (nx != nxm) || (ny != nym))
		stop(paste('Error, data file to impose mask upon does not match shape of mask. Mask shape:', 
			nxm, nym, 
			'Data file data shape:', nx, ny ))

	if( ndims == 2 ) {
		dat = ncvar_get( ncid, var )
		dat[ is.na( dat_mask ) ] = NA
		ncvar_put( ncid, var, dat, start=c(1,1), count=c(nx,ny))
		nc_close( ncid )
		return()
		}

	blocksize = 1000
	nt = var$dim[[3]]$len
	nb = floor( (nt-1)/blocksize ) + 1

	for( ib in 1:nb ) {
		it0 = (ib-1)*blocksize + 1
		it1 = it0 + blocksize - 1
		if( it1 > nt ) it1 = nt

		ntt = it1 - it0 + 1

		dat = ncvar_get( ncid, var, start=c(1,1,it0), count=c(-1,-1,ntt))
		for( iti in 1:ntt ) {
			tt = dat[,,iti] 
			tt[ is.na( dat_mask ) ] = NA
			dat[,,iti] = tt
			}

		ncvar_put( ncid, var, dat, start=c(1,1,it0), count=c(-1,-1,ntt))
		}

	nc_close( ncid )
}

#==========================================================================================
nc_change_lon_branch = function( fname_in, fname_out ) {
	
	mydo( paste( 'cp', fname_in, fname_out ))

	ncid = nc_open( fname_out, write=TRUE )
	var  = nc_main_var( ncid )
	xvals = var$dim[[1]]$vals

	print(paste("Range of orig x vals:", pv(range(xvals)) ))
	
	if( max(xvals) > 181 ) {
		print('Changing from the 0 -> 360 branch to -180 -> 180')
		xvals_out = xvals - 360.
		}
	else
		{
		print('Changing from the  -180 -> 180 branch to 0 -> 360' )
		xvals_out = xvals + 360.
		}
		
	ncvar_put( ncid, var$dim[[1]]$name, xvals_out )

	print(paste("Range of new x vals:", pv(range(xvals_out)) ))
	nc_close( ncid )
}

#==========================================================================================
# This takes a list of input files (which may be just one), and performs a timestep-by-timestep
# regional average of the input file(s). The output file is created, and data written to it.
# 
# 'mask' is a list with the following elements: 
#	$tag	: string name of the mask. For example, "UCRB" or "Yosemite"
#	$nx, $ny: shape of the mask data
#	$mask	: an array of shape [nx,ny] with values >= 1 where the data is to be averaged, 0 elsewhere
#
# The regions can be identified either by just a '1' in the mask file, or there can be multiple regions
# identified by integers from 1 to N. If that is the case, then this routine ALSO outputs a "pseudo region",
# which is the area-averaged value of ALL the regions 
#
# For an example of how to use, see:
#	/net/loca/CA_NV_2016-04-02/region_Yosemite/make_reg_avg_files_v2_daily_Livneh_PRISM.R
#
make_reg_avg_inner = function( fnames_in, fname_out, mask, reg_names, varname=NULL, do_area_weight=TRUE, force=FALSE, trace=TRUE ) {

	if( trace ) print(paste("make_reg_avg_inner: entering for fname_out=", fname_out ))

	if( storage.mode(mask) != 'list' )
		stop('Error on input mask argument. It must be a list with the following elements: $tag    : string name of the mask. For example, "UCRB" or "Yosemite". $nx, $ny: shape of the mask data. $mask   : an array of shape [nx,ny] with values >= 1 where the data is to be averaged, 0 elsewhere' )

	if( length( fnames_in ) < 1 ) 
		stop(paste("error, called with ZERO input files!"))

	if( (! force) && (filesize( fname_out ) > 10)) {
		print(paste("make_reg_avg_inner: skipping, output file exists:", fname_out ))
		return()
		}

	nreg = max( mask$mask, na.rm=TRUE )
	if( trace ) print(paste("make_reg_avg_inner: nreg=", nreg ))

	#-------------------------------------------------------------------------------
	# Open first (possibly only) input file as a template. Check for errors using it
	#-------------------------------------------------------------------------------
	if( trace ) print(paste("make_reg_avg_inner: opening template file", fnames_in[1] ))
	ncid1 = nc_open( fnames_in[1], readunlim=FALSE )
	if( is.null( varname )) {
		var = nc_main_var( ncid1 )
		varname = var$name
		}
	else
		var = ncid1$var[[ varname ]]

	if( is.null(var))
		stop(paste("error, did not find specified var in file:", varname ))

	#--------------------------------
	# Get key info from template file
	#--------------------------------
	xvals = var$dim[[1]]$vals
	yvals = var$dim[[2]]$vals
	dat1 = ncvar_get( ncid1, varname, count=c(-1,-1,1))	# just get one timeslice
	nx = dim( dat1 )[1]
	ny = dim( dat1 )[2]
	if( (nx != mask$nx) || (ny != mask$ny)) 
		stop(paste("Error, mask has shape", mask$nx, mask$ny, "but data has shape", nx, ny ))

	tdim1 = ncid1$dim[[ ncid1$unlimdimid ]]
	tunits1 = tdim1$units
	if( trace ) print(paste("make_reg_avg_inner: nx=", nx, "ny=", ny, "time units:", tunits1 ))

	#------------------------------------------
	# Get area of each cell using template file
	#------------------------------------------
	area = array(0., dim=c(nx,ny))
	for( jj in 1:ny )
	for( ii in 1:nx ) {
		if( ii < nx ) 
			dx = abs(xvals[ii+1] - xvals[ii])
		else
			dx = xvals[nx] - xvals[nx-1]
		if( jj < ny ) 
			dy = abs( yvals[jj+1] - yvals[jj] )
		else
			dy = abs( yvals[ny] - yvals[ny-1] )

		lat_rad = yvals[jj]/180. * pi
		dx = dx * cos( lat_rad )

		km_per_deg = 110.
		area[ii,jj] = km_per_deg * dx * km_per_deg * dy
		}

	nc_close( ncid1 )

	nf = length( fnames_in )
	if( trace ) print(paste("make_reg_avg_inner: nfiles_in=", nf ))

	tser_out  = NULL
	tvals_out = NULL

	#----------------------------------------------
	# Pre-compute the keep masks for a bit of speed
	#----------------------------------------------
	keep_mask = array( TRUE, dim=c(nx,ny,nreg))
	sum_weights = array(0., nreg)
	keep_w = list()
	for( ireg in 1:nreg ) {
		tt = (mask$mask == ireg)
		nkeep = sum( tt )
print(paste('ireg, nkeep:', ireg, nkeep ))
		keep_mask[,,ireg] = tt

		keep_w[[ireg]] = area[tt]
		sum_weights[ireg] = sum( keep_w[[ireg]] )
		}

	for( ifile in 1:nf ) {

		print(paste("...reading full 3-d data from input file", ifile, '/', nf, ':', fnames_in[ifile] ))
		rv = get_dat( fnames_in[ifile] )

		#--------------------------
		# Make sure same time units
		#--------------------------
		if( rv$tunits != tunits1 ) 
			stop(paste("Error, template file has time units", tunits1, 
				"but input file", 
				fnames_in[ifile],
				"has time units", rv$tunits,
				"I am not coded to handle that case!"))

		#-------------------------
		# Perform regional average
		#-------------------------
		nt_this_file = dim( rv$dat )[3]
		tser_this_file = array(0., dim=c(nreg,nt_this_file) )
		for( it in 1:nt_this_file ) {
			tdat = rv$dat[,,it]	# 2-d field for this time step
			for( ireg in 1:nreg ) {
				keep = keep_mask[,,ireg]
				tt = tdat[keep]
				if( do_area_weight )
					tser_this_file[ireg,it] = sum( tt * keep_w[[ireg]], na.rm=TRUE ) / sum_weights[ireg]
				else
					tser_this_file[ireg,it] = mean( tt, na.rm=TRUE )
				}
			}

		#-----------------------------------------
		# Add values from this file to output tser
		#-----------------------------------------
		if( is.null( tser_out ))
			tser_out = tser_this_file
		else
			tser_out  = c_arrays( tser_out, tser_this_file )
		tvals_out = c( tvals_out, rv$tvals )
		}

	nt_out = length( tvals_out )

	#-----------------
	# Make output file
	#-----------------
	dim_nchar = ncdim_def( 'nchar', '', 1:50, create_dimvar=FALSE )
	dim_t     = ncdim_def( tdim1$name, tunits1, tvals_out, unlim=TRUE )
	dim_reg   = ncdim_def( 'region', '', 1:nreg, create_dimvar=FALSE )

	if( nreg == 1 ) 
		var_out  = ncvar_def( var$name, var$units, list(dim_t), var$mv )
	else
		var_out  = ncvar_def( var$name, var$units, list(dim_reg,dim_t), var$mv )

	var_regname = ncvar_def( 'region_name', '', list(dim_nchar,dim_reg), ' ', prec='char' )

	print(paste("Creating output file", fname_out ))
	ncid_out = nc_create( fname_out, list(var_regname,var_out) )

	#------------------
	# Write output data
	#------------------
	if( nreg == 1 ) 
		ncvar_put( ncid_out, var_out, tser_out, start=1, count=nt_out )
	else
		ncvar_put( ncid_out, var_out, tser_out, start=c(1,1), count=c(nreg,nt_out) )
		
	ncvar_put( ncid_out, var_regname, reg_names )

	nc_close( ncid_out )
}

#==============================================================================================================
nc_util_getdlist = function( ncid, ivar, dim_out ) {

	vv = ncid$var[[ivar]]
	if( is.null(vv))
		stop(paste("Internal error, got null for var with ivar=", ivar ))

	ndims = vv$ndims
	dlist = list()
	for( idim in 1:ndims ) {
		dname2find = vv$dim[[idim]]$name
		ixd = -1
		for( kk in 1:length(dim_out) ) {
			if( dname2find == dim_out[[kk]]$name ) {
				ixd = kk
				break
				}
			}
		if( ixd == -1 ) 
			stop(paste("Error, did not find this dim in dim_out:", dname2find ))

		dlist[[idim]] = dim_out[[ixd]]
		}

	return( dlist )
}

#==============================================================================================================
# This keeps only the requested time steps from an input file. The way this differs from ncunary and nccat
# is that it keeps ALL the vars in the source file. Vars along the time axis are all segmented to keep
# only time entries it0 to it1 (1-based counting). Vars that do not have a time axis are copied unchanged.
#
nc_time_segment_file = function( fname_in, it0, it1, fname_out ) {

	print(paste("Executing nc_time_segment_file: fname_in=", fname_in, 'it0, 1:', it0, it1, "fname_out:", fname_out ))

	ncid = nc_open( fname_in )
	tdim = ncid$dim[[ ncid$unlimdimid ]]

	#---------------------
	# Make all output dims
	#---------------------
	dim_out = list()
	for( idim in 1:ncid$ndims ) {
		dd = ncid$dim[[idim]]
		if( dd$name == tdim$name ) {
			dim_out[[idim]] = ncdim_def( dd$name, dd$units, dd$vals[it0:it1], unlim=TRUE )
			}
		else
			dim_out[[idim]] = ncdim_def( dd$name, dd$units, dd$vals )
		}

	#---------------------
	# Make all output vars
	#---------------------
	var_out = list()
	for( ivar in 1:ncid$nvars ) {
		vv = ncid$var[[ivar]]
		#-----------------------------
		# Get dims to use for this var
		#-----------------------------
		dlist = nc_util_getdlist( ncid, ivar, dim_out )
		var_out[[ivar]] = ncvar_def( vv$name, vv$units, dlist, vv$missval, prec=vv$prec )
		}

	ncid_out = nc_create( fname_out, var_out )

	#----------------
	# Copy var values
	#----------------
	t_count = it1 - it0 + 1
	for( ivar in 1:ncid$nvars ) {
		ndims = ncid$var[[ivar]]$ndims
		var_is_unlim = FALSE
		for( kk in 1:ndims ) {
			if( var_out[[ivar]]$dim[[kk]]$name == tdim$name ) {
				var_is_unlim = TRUE
				break
				}
			}

		#----------------------------
		# Accumulate size of this var
		#----------------------------
		start_r = array(0, ndims)
		count_r = array(0, ndims)
		start_w = array(0, ndims)
		count_w = array(0, ndims)
		for( idim in 1:ndims ) {
			start_r[idim] = 1
			count_r[idim] = var_out[[ivar]]$dim[[idim]]$len
			start_w[idim] = 1
			count_w[idim] = var_out[[ivar]]$dim[[idim]]$len
			}

		if( var_is_unlim ) {
			start_r[idim] = it0
			count_r[idim] = t_count
			start_w[idim] = 1
			count_w[idim] = t_count
			}

		dat = ncvar_get( ncid, ncid$var[[ivar]]$name, start=start_r, count=count_r )
		ncvar_put( ncid_out, ncid$var[[ivar]]$name, dat, start=start_w, count=count_w )
		}

	nc_close( ncid_out )
}

#==============================================================================================================
# You often want to get or extract multiple time series at individual points from a large set of 3d data files.
# For example, getting data at multiple stations from the livneh yearly files. This is a slow operation
# via the typical methods. This makes it much more efficient.
#
# If fill_gaps_with_NA is TRUE, then any temporal gaps in the input data are filled with NA's
#
# Inputs:
#	fnames[nfiles]		: array of filenames, ideaelly temporally sequential, MUST all be on the same grid
#	ix[nsta], iy[nsta]	: arrays of INDICES (not lat/lon values) of the locations to extract time series for
#	fnames_out[nsta]	: optional, names of netcdf files to be created, one per output station
#	meta_out[[ nsta ]]	: optional, a list of nsta LISTS that hold metadata to be written to the output netcdf files
#
# Example meta_out: 
#	meta_out = list()
#	for( ista in 1:nsta ) 
#		meta_out[[ista]] = list( list( 'lon', lon[ista] ), list( 'lat', lat[ista] ), list( 'name', name[ista] ))
#
# Output:
#	rv$tser[nsta,nt]	: time series at all the stations
#	rv$dates		: output dates
#
nc_get_multi_tser_from_files = function( fnames, ix, iy, varname=NULL, fnames_out=NULL, meta_out=NULL, fill_gaps_with_NA=TRUE ) {

	if( is.null( fnames )) stop(paste("error, passed NULL fnames!"))
	if( length( fnames ) < 1 ) stop(paste("error, passed zero-length input fnames!"))

	nsta = length( ix )
	if( length(iy) != nsta ) stop(paste("Error, length ix not equal to length iy"))

	if( ! is.null(meta_out)) {
		if( storage.mode(meta_out) != 'list' )
			stop(paste("Error, if passed, meta_out must be a LIST OF LISTS holding netcdf metadata to be written to output files"))
		if( length( meta_out ) != nsta ) {
			stop(paste("Error, if passed, meta_out must be a LIST OF LISTS holding netcdf metadata for each station to be written to output files"))
			}
		}

	nf = length( fnames ) 

	dates = NULL
	tser  = NULL

	WRF_HACK = FALSE

	for( ifile in 1:nf ) {
		progress_meter( 'reading file', ifile, nf )

		#------------------------------------
		# Open file, read in entire 3-d field
		#------------------------------------
		ncid = nc_open( fnames[ifile] )
		if( is.null( varname ))
			var  = nc_main_var( ncid )
		else
			{
			var  = ncid$var[[ varname ]]
			if( is.null(var))
				stop(paste("Error, user specified varname not found in file:", varname, 'file:', fnames[ifile] ))
			}

		dat  = ncvar_get( ncid, var )

		#--------------------------------------------------------
		# Dat may be returned as 2d if there is only 1 time entry
		#--------------------------------------------------------
		ndims = length( dim(dat) )
		if( ndims == 2 ) 
			dim(dat) = c( dim(dat), 1 )

		#---------------------------------------------------
		# Get dates for this file, and append to total dates
		#---------------------------------------------------
		tdim = ncid$dim[[ ncid$unlimdimid ]]
		if( ifile == 1 ) 
			rv_calatt = ncatt_get( ncid, tdim$name, 'calendar' )

		#---------
		# WRF HACK 
		#---------
		if( (tdim$dimvarid$id == -1 ) && (! is.null(ncid$var$Times) )) {
			WRF_HACK = TRUE
			#print('nc_get_multi_tser_from_filesj: WRF TIME/DATE HACK ENGAGED')
			wrfdates = ncvar_get( ncid, 'Times' )
			ndates = tdim$len
			wyear  = array(0, ndates )
			wmonth = array(0, ndates )
			wday   = array(0, ndates )
			whour  = array(0, ndates )
			wmin   = array(0, ndates )
			wsec   = array(0, ndates )

			for( kk in 1:ndates ) {
				dd = wrfdates[kk]
				wyear [kk] = as.integer( substr( dd,  1,  4 ))
				wmonth[kk] = as.integer( substr( dd,  6,  7 ))
				wday  [kk] = as.integer( substr( dd,  9, 10 ))
				whour [kk] = as.integer( substr( dd, 12, 13 ))
				wmin  [kk] = as.integer( substr( dd, 15, 16 ))
				wsec  [kk] = as.integer( substr( dd, 18, 19 ))
				}
			tdates = list( year=wyear, month=wmonth, day=wday, hour=whour, minute=wmin, second=wsec )
			}
		else
			{
			tdates = utCalendar( tdim$vals, tdim$units )
			}

		#--------------------------------------
		# concat dates. Check for missing dates
		#--------------------------------------
		dates = cat_dates( dates, tdates )

		if( ifile == 1 ) {
			tname  = tdim$name
			if( WRF_HACK ) 
				tunits = 'hours since 1980-01-01 00:00'
			else
				tunits = tdim$units
			}

		nt_this_file = tdim$len
		ttser = array(0., dim=c(nsta,nt_this_file))

		#------------------------------------------
		# Get data from this file, for each station
		#------------------------------------------
		for( ista in 1:nsta ) 
			ttser[ista,] = dat[ ix[ista], iy[ista], ]
		
		tser = c_arrays( tser, ttser )	# [nsta,nt]

		nc_close( ncid )
		}

	#--------------------------------------
	# Check for gaps in data, fix any found
	#--------------------------------------
	tvals = utInvCalendar( dates, tunits )
	rvfill = fill_discontin_time_gaps( tvals, tser )
	tvals = rvfill$tvals
	tser  = rvfill$data
	dates = utCalendar( tvals, tunits )

	#----------------------------------------------
	# Write output netcdf format files if reqeusted
	#----------------------------------------------
	if( ! is.null( fnames_out )) {
		nt_out = length( dates$year )
		if( length( fnames_out ) != nsta ) stop(paste("Error, length of fnames_out must match nsta!"))

		for( ista in 1:nsta ) {

			#---------------------------------------------
			# M<ake tdim, preserving the original calendar
			#---------------------------------------------
			tvals = utInvCalendar( dates, tunits )

			if( rv_calatt$hasatt ) 
				cal2use = rv_calatt$value
			else
				cal2use = NA

			tdim = ncdim_def( tname, tunits, tvals, unlim=TRUE, calendar=cal2use )

			varout = ncvar_def( var$name, var$units, tdim, var$missval )

			print(paste("Creating output file", fnames_out[ista] ))
			ncid_out = nc_create( fnames_out[ista], varout )

			ncvar_put( ncid_out, varout, tser[ista,], start=1, count=nt_out )

			#----------------------
			# Write metadata output
			#----------------------
			if( ! is.null( meta_out )) {
				nitems = length( meta_out[[ista]] )
				for( kk in 1:nitems ) 
					ncatt_put( ncid_out, 0, meta_out[[ista]][[kk]][[1]], meta_out[[ista]][[kk]][[2]] )
				}

			nc_close( ncid_out )
			}
		}

	return( list(
		tser	= tser,		# [nsta,nt]
		dates	= dates 
		))
}

