
#=============================================================================
dotplot = function( lon, lat, val, cmap='bluered_cblind', discretevals=NA,
	xlim=range(lon,na.rm=TRUE), ylim=range(lat,na.rm=TRUE), 
	col_map=grey(0.3), pch=19, cex=1.0, map='', zrange=range(val,na.rm=TRUE), 
	... ) {

	canvas( xlim=xlim, ylim=ylim, grid=FALSE, ... )

	if( nchar(map) > 0 ) 
		map( map, col=col_map, add=TRUE )

	npts = length( lon )
	if( length(lat) != npts ) stop(paste("Error, got", npts, "lon points but",
		length(lat), 'lat points'))
	if( length(val) != npts ) stop(paste("Error, got", npts, "lon points but",
		length(val), 'vals to plot'))

	col = mcolormap( cmap, val, zrange=zrange )

	points( lon, lat, pch=pch, cex=cex, col=col )
}

#=============================================================================
# Mean of a histogram
#
# can also consider hx as the values, and hy as the weights, in which case
# hist_mean is a weighted mean, wmean, weightmean, weight_mean
#
hist_mean = function( hx, hy ) {

	nn = length( hx )
	if( length( hy ) != nn )
		stop(paste("error, #x (vals) is", nn, "but #y (weights) is", length(hy) ))

	tsum = 0
	for( kk in 1:nn ) 
		tsum = tsum + hx[kk]*hy[kk]

	return( tsum / sum(hy) )
}

#=============================================================================
# 'classnum' can be used to add coloration to the histogram bars indicating
# which class of events contributed to the count.  It is an integer array
# with one entry for each data point showing the class number of that data
# point.  The colors for the bars are given in 'classcolor', which is an array
# of colors, with length equal to the maximum value of the classnum array.
#
# Argument 'norm' indicates the normalization to apply to the curve.
#	If: 'norm' == 'count' (default), then simple counts are plotted
#	    'norm' == 'pdf', then the total area under the plotted curve will be 1.
#	    'norm' == 'max1', then the maximum value plotted will be 1.0
#	    'norm' == 'frac', then the fraction of total accounted for by each bin is plotted
#
# Argument 'type' is:
#	'bars'   == draw bars
#	'lines'  == draw lines
#	'points' == plot lines
#
mhist <- function( data_orig, lty=1, breaks=NA, col=NA, main='', ..., norm='count', forceFDR=TRUE, type='bars', bar_width=0.8,
	scalefact=1.0, classnum=NA, classcolor=NA, add=FALSE, plot=TRUE, offsetfact=0.03, voffsetfact=0.01,
	col_pos=NA, col_neg=NA, hist=NA,
	log='', zero_log_value=1.e-10, verbose=FALSE, xlim=NA, ylim=NA ) {

	data = data_orig
	dim(data) = length(data)
	data = data[ ! is.na(data) ]
	data = data[ is.finite(data) ]

	#-------------------------------------
	# Make sure range of breaks spans data
	#-------------------------------------
	if( length(breaks) > 1) {
		rd = range( data )
		rb = range( breaks )
		if( (rd[1] < rb[1]) || (rd[2] > rb[2])) 
			stop(paste("Error, need to make range of breaks wider; current range:", pv(range(breaks)), "needs to be:", pv(range(data))  ))
		}

	#-------------------------------------------------------------------------------
	# Consists of two parts -- first part, make object to hist. Second part, plot it
	#-------------------------------------------------------------------------------

	if( (type != 'bars') && (length(col) == 1) && is.na(col))
		col = 'black'

	NORM_COUNT <- 1
	NORM_PDF   <- 2
	NORM_MAX1  <- 3
	NORM_FRAC  <- 4
	if( substr(norm,1,4) == 'coun' )
		norm2use <- NORM_COUNT
	else if( substr(norm,1,3) == 'pdf' )
		norm2use <- NORM_PDF
	else if( substr(norm,1,3) == 'max' )
		norm2use <- NORM_MAX1
	else if( substr(norm,1,4) == 'frac' )
		norm2use <- NORM_FRAC
	else
		stop(paste("error, passed arg 'norm' must be 'count', 'pdf', or 'max1'; I got: ", norm))

#	if( ! exists(globalframenumber))
#		globalframenumber <<- 0
#
#	if( !add )
#		globalframenumber <<- globalframenumber + 1

	#------------------------------------------------------
	# Estimate ideal # of cells from Freeman-Diaconis method
	# Freedman, D. and Diaconis, P. (1981) On this histogram as 
	# a density estimator: L2 theory. Zeit. Wahr. ver. Geb., 
	# 57, 453476.
	#------------------------------------------------------
	nobs     <- length(data)
	if( nobs == 0 ) {
		print(paste(">>>>>>>>>>>>>>> mhist: no data, returning! <<<<<<<<<<<<<<<"))
		canvas( main=paste('NO DATA',main ))
		return()
		}
	if( verbose ) print(paste("mhist: nobs=", nobs))

	iqr      <- IQR(data,na.rm=TRUE)
	binwidth <- 2.*iqr*(nobs^-.333333)
	range    <- max(data,na.rm=TRUE)-min(data,na.rm=TRUE)
	ncells   <- as.integer(range/binwidth + 0.5)
	if( (length(breaks) == 1) && (is.na(ncells)) ) {

		dnz = c(0, data[ data != 0 ])
		iqr      <- IQR(dnz,na.rm=TRUE)
		binwidth <- 2.*iqr*(nobs^-.333333)
		range    <- max(dnz,na.rm=TRUE)-min(dnz,na.rm=TRUE)
		ncells   <- as.integer(range/binwidth + 0.5)

		if( (length(breaks) == 1) && (is.na(ncells)) ) {
			print(paste(">>>>>>>>>>>>>>> mhist: iqr=", iqr, "binwidth=", binwidth, "range=", range, "ncells=", ncells, "<<<<<<<<<<<<<<<<<<<<<<<"))
			print(paste(">>>>>>>>>>>>>>> mhist: nbins==0, returning! <<<<<<<<<<<<<<<"))
			print(data)
			#canvas( main=paste('NO BINS',main ))

			#range = 3*iqr
			#ncells   <- as.integer(range/binwidth + 0.5)

			return()
			}
		}

	if( verbose )
		print(paste("Nobs:",nobs,"Range:",range,"IQR:",iqr,"IQR-based binwidth:",binwidth,"Est # of cells by Freeman-Diaconis rule:",ncells))

	if( (length(breaks)==1) && is.na(breaks) ) {
		if( forceFDR ) {
			breaks <- (0:(ncells-1))/(ncells-1.0)*range + min(data,na.rm=TRUE)
			}
		else
			breaks <- ncells
		}

	if( (is.nan(breaks)) || ((length(breaks)==1) && (breaks==2)) || (length(breaks)==2))
		breaks <- 3

	if( (length(hist)==1) && (is.na(hist))) 
		h <- hist( data, plot=FALSE, breaks=breaks, xlim=xlim, ylim=ylim, ... )
	else
		{
		print(paste("Using user-specified histogram object, nmids=", pv(range(hist$mids))))
		h <- hist
		}

	mids <- h$mids
	# Get binwidth
	bw <- mids[2] - mids[1]

	#----------------------------
	# Apply desired normalization
	#----------------------------
	if( norm2use == NORM_PDF ) {
		dat <- h$counts / sum(h$counts*bw)
		min_valid = 0.5 / sum( h$counts*bw)
		}

	else if( norm2use == NORM_COUNT ) {
		dat <- h$counts
		min_valid = 0.5
		}

	else if( norm2use == NORM_MAX1 ) {
		dat <- h$counts / max(h$counts)
		min_valid = 0.5 / max( h$counts )
		}

	else if( norm2use == NORM_FRAC ) {
		dat <- h$counts / sum(h$counts)
		min_valid = 0.5 / sum( h$counts)
		}

	else
		stop('bad value of norm2use')

	#---------------------------------
	# Apply any scale factor requested
	#---------------------------------
	dat <- dat * scalefact

	#-----------------------------------------------
	# Set ylim if doing Y log and not specified ylim
	#-----------------------------------------------
	log_y_set = (log=='y') || (log=='xy') || (log=='yx') 
	ylim_set  = ( length(ylim) > 1 )
	if( log_y_set && (! ylim_set)) {
		#stop(paste("Error, if you specify log(y) you must also specify ylim.  Log=",log,"ylim=", ylim))
		ymin_valid = min_valid 
		ymax_valid = max( dat )
		ylim = c( ymin_valid, ymax_valid )
		print(paste("Log-Y invoked but no ylim set: setting ylim range to", ylim[1], ylim[2] ))
		}

	#---------------------------
	# Pad edges with zero counts 
	#---------------------------
	dat <- c(0,dat,0)
	mids <- c(mids[1]-bw,mids,mids[length(mids)]+bw)

	if( log != '' ) {
		if( (log == 'y')||(log == 'xy')) {
			dat <- ifelse(dat==0,zero_log_value,dat)
			}
		}

	if( plot ) {
		offset <- bw*offsetfact
		if( add ) {
			xvals  <- mids-bw*.5+offset
			yoff <- max(dat)*offsetfact*voffsetfact
			if( type=='bars' ) {
				nm <- length(mids)
				for( im in 1:nm ) {
					x0 <- mids[im]-bw*bar_width/2 + offset
					x1 <- x0 + bw*bar_width
					y0 <- 0
					y1 <- dat[im]
					if( length(col) > 1 )
						col2use <- col[im]
					else
						col2use <- col

					if( (! is.na(col_pos)) && (mids[im] > 0))
						col2use = col_pos
					else if( (! is.na(col_neg)) && (mids[im] < 0))
						col2use = col_neg

					polygon( c(x0,x1,x1,x0), c(y0,y0,y1,y1), col=col2use, ... )
					}
				}
			else if( type == 'lines' )
				lines(xvals,dat+yoff,log=log,...,type='s',lty=lty,col=col)
			else if( type == 'points' )
				points(xvals,dat+yoff,log=log,...,lty=lty,col=col)
			else
				stop(paste("Error, type must be one of: lines points bars"))
			}
		else	
			{
			if(type=='bars') {
				
				if( length(xlim==1) && is.na(xlim))
					xlim <- range(mids-bw*.5,na.rm=TRUE)
				if( length(ylim==1) && is.na(ylim))
					ylim <- range(dat,na.rm=TRUE)
				canvas( xlim=xlim, ylim=ylim,log=log,main=main,...,type="s",lty=lty)
				nm <- length(mids)
				for( im in 1:nm ) {
					x0 <- mids[im]-bw*bar_width/2 + offset
					x1 <- x0 + bw*bar_width
					if( log_y_set )	# remember ylim must be set if log_y is set
						y0 = ylim[1]
					else
						y0 <- 0
					y1 <- dat[im]
					if( length(col) > 1 )
						col2use <- col[im]
					else
						col2use <- col

					xc = 0.5*(x0+x1)
					if( (! is.na(col_pos)) && (xc > 0))
						col2use = col_pos
					else if( (! is.na(col_neg)) && (xc < 0))
						col2use = col_neg

					polygon( c(x0,x1,x1,x0), c(y0,y0,y1,y1), col=col2use, ... )
					}
				}

			else if( type == 'lines' ) {
				if( length(xlim==1) && is.na(xlim))
					xlim <- range(mids-bw*.5,na.rm=TRUE)
				if( length(ylim==1) && is.na(ylim))
					ylim <- range(dat,na.rm=TRUE)
				plot(mids-bw*.5,dat,log=log,xlim=xlim,ylim=ylim,main=main,...,type='s',lty=lty, col=col)
				}

			else if( type == 'points' ) {
				if( length(xlim==1) && is.na(xlim))
					xlim <- range(mids-bw*.5,na.rm=TRUE)
				if( length(ylim==1) && is.na(ylim))
					ylim <- range(dat,na.rm=TRUE)
				plot(mids-bw*.5,dat,log=log,xlim=xlim,ylim=ylim,main=main,...,type='p',pch=19, col=col)
				}

			else
				stop(paste("Error, type must be one of: lines points bars"))
			}
		}
		
	if( verbose )
		print(paste("Actual binwidth used:",bw,"  Actual # of cells used:",length(mids)))

	if( ! is.na(classnum) ) {
		nclasses <- max(classnum)
		if( length(classcolor) != nclasses ) {
			stop("error, there must be one entry in the classcolor array for each class!")
			}
		nmids <- length(mids)
		for( i in 1:nmids ) {
			countbyclass <- array(0,nclasses)
			binstart <- mids[i]-bw/2.
			binend   <- binstart + bw
			for( j in 1:nobs ) {
				if( (!is.na(data[j])) && (data[j] >= binstart) && (data[j] < binend)) 
					countbyclass[classnum[j]] <- countbyclass[classnum[j]] + 1
				}
			cc <- 0		# currentcount
			for( k in 1:nclasses ) {
				if( countbyclass[k] > 0 ) {
					polygon(c(binstart,binend,binend,binstart),
				       	        c(cc,cc,cc+countbyclass[k],cc+countbyclass[k]),
						col=classcolor[k] )
					}
				cc <- cc + countbyclass[k]
				}
			}
		}

	return(h)	# $breaks, $counts, $density, $mids
}

#=====================================================================================
# 'h' is the return from hist or mhist.  It has the following fields:
# $breaks
# $counts
# $intensities
# $density
# $mids
# $xname
#
plot_hist_retval <- function( h, add=FALSE, bars=TRUE, log='', xlim=NA, ylim=NA, 
		bar_width=0.8, lty=1, offsetfact=0.03, ... ) {

	mids <- h$mids
	bw   <- mids[2] - mids[1]
	dat  <- h$counts

	offset <- bw*offsetfact
	if( add ) {
		xvals  <- mids-bw*.5+offset
		yoff <- 0.0
		if( bars ) {
			nm <- length(mids)
			for( im in 1:nm ) {
				x0 <- mids[im]-bw*bar_width/2 + offset
				x1 <- x0 + bw*bar_width
				y0 <- 0
				y1 <- dat[im]
				polygon( c(x0,x1,x1,x0), c(y0,y0,y1,y1), ... )
				}
			}
		else
			lines(xvals,dat+yoff,log=log,...,type='s',lty=lty)
		}
	else	{
		if(bars) {
			if( length(xlim==1) && is.na(xlim))
				xlim <- range(mids-bw*.5,na.rm=TRUE)
			if( length(ylim==1) && is.na(ylim))
				ylim <- range(dat,na.rm=TRUE)
			canvas( xlim=xlim, ylim=ylim,log=log,...,type="s",lty=lty)
			nm <- length(mids)
			for( im in 1:nm ) {
				x0 <- mids[im]-bw*bar_width/2 + offset
				x1 <- x0 + bw*bar_width
				y0 <- 0
				y1 <- dat[im]
				polygon( c(x0,x1,x1,x0), c(y0,y0,y1,y1), ... )
				}
			}
		else
			{
			if( length(xlim==1) && is.na(xlim))
				xlim <- range(mids-bw*.5,na.rm=TRUE)
			if( length(ylim==1) && is.na(ylim))
				ylim <- range(dat,na.rm=TRUE)
			plot(mids-bw*.5,dat,log=log,xlim=xlim,ylim=ylim,...,type='s',lty=lty)
			}
		}
}
	
#=============================================================================
hist2poly <- function( h ) {
	
	val <- h$counts

	nb <- length(h$breaks)
	x <- array(0,nb*2)
	y <- array(0,nb*2)
	x[1] <- h$breaks[1]
	y[1] <- 0.
	x[2] <- x[1]
	y[2] <- val[1]
	for( i in 2:nb ) {
		x[i*2-1]   <- h$breaks[i]
		x[i*2] <- h$breaks[i]
		y[i*2-1]   <- val[i-1]
		y[i*2] <- val[i]
		}
	y[nb*2] <- 0

	rv <- list()
	rv$x <- x
	rv$y <- y
	return(rv)
}

#=============================================================================
textbox <- function( x, y, text, padfact=NA, cex=NA, boxfill=NA, 
					boxborder=NA, textcolor=NA ) {

	if( is.na(cex))
		cex <- 1.0

	if( is.na(boxfill))
		boxfill <- "White"

	if( is.na(boxborder))
		boxborder <- "Black"

	if( is.na(textcolor))
		textcolor <- "Black"

	if( is.na(padfact))
		padfact <- .3

	sw <- strwidth(text,cex=cex)
	sh <- strheight(text,cex=cex)
	xpad <- sw/nchar(text)*padfact
	ypad <- sh*padfact

	llx  <- x-sw/2-xpad
	urx  <- x+sw/2+xpad
	lly  <- y-sh/2-ypad
	ury  <- y+sh/2+ypad

	polygon(c(llx, urx, urx, llx),
	        c(lly, lly, ury, ury),
		border=boxborder, col=boxfill )
	text( x, y, text, col=textcolor, cex=cex )
}

#=============================================================================
# NOTE: ylab MUST be set to an array of 2 strings: ylab <- c('Y label 1', 'Y label 2')
# 'line' is which margin line to write in, starting at 1.  Increase to
# move Y axis labels away from the plot, decrease to move closer to plot.
#
# To add a line to either the left (red) or right (blue) axes, do this:
#
# to left (red)
# rv = dualplot()
# par( usr=rv$usr1 )
# lines( ..., col='red' )
#
# Right (blue)
# same, but par( usr=rv$usr2 )
#
dualplot <- function( x, y1, y2, col=NA, ylab=NA, xlab='', line=2.5, docolor=TRUE, col.axes=NA,
		lwd=c(1,1), axis_lwd=2, type=c('l','l'), pch=c(NA,NA), main='', 
		xlim=range(x), ylim1=NA, ylim2=NA, xaxs='r', yaxs='r', cex.ylab=1.0, 
		add=FALSE,
		lty=c(1,1),
		oldstyle=FALSE, ... ) {

	if( !add ) 
		globalframenumber <<- globalframenumber + 1

	if( is.na(ylab)) {
		print("dualplot REQUIRES the ylab parameter to be set!")
		stop("error in dualplot")
		}

	#----------------------------------------------------------------
	# This is needed to make the Y axis labels come out perpindicular
	# to the axis, as per normal
	#----------------------------------------------------------------
	par(las=1)

	#-------------------------------------------------
	# Adjust right margin to make room for axis labels
	#-------------------------------------------------
	origpar <- par()
	origmar <- origpar$mar
	marnew <- c(origmar[1], origmar[2], origmar[3], origmar[4]+4)
	if( oldstyle )
		par(mar=marnew)

	if( is.na(col)) {
		if( docolor ) 
			col <- c( 'red', 'blue' )
		else
			col <- c('black','black')
		}

	if( (! docolor) && (lty[2] == lty[1]))
		lty <- c(1,6)

	if( is.na(col.axes))
		col.axes <- col

	#----------------
	# Make FIRST plot
	#----------------
	if( (length(ylim1)==1) && (is.na(ylim1)))
		ylim1 <- range(y1)
	plot( x, y1, axes=FALSE, col=col[1], ylab='',  lwd=lwd[1], ylim=ylim1,
		xaxs=xaxs, yaxs=yaxs, xlim=xlim, xlab=xlab, lty=lty[1], type=type[1], pch=pch[1])
	xticks <- pretty(xlim)
	axis( 1, xticks, lwd=axis_lwd, xaxs=xaxs )

	parorig <- par()
	mfgorig <- parorig$mfg

	#--------------------
	# First plot's Y axis
	#--------------------
	y1ticks <- pretty(ylim1)
	axis( 2, y1ticks, line=ifelse(yaxs=='r',1,0.2), col=col.axes[1], lty=lty[1], lwd=axis_lwd, yaxs=yaxs )
	mtext( ylab[1], side=2, las=3, line=line, cex=cex.ylab )

	par(mfg=mfgorig)

	#-----------------
	# Make SECOND plot
	#-----------------
	if( (length(ylim2)==1) && (is.na(ylim2)))
		ylim2 <- range(y2)
	plot( x, y2, col=col[2], lwd=lwd[2], xlab='', ylab='', axes=FALSE, ylim=ylim2,
		xlim=xlim, lty=lty[2], type=type[2], pch=pch[2], xaxs=xaxs, yaxs=yaxs )

	#---------------------
	# Second plot's Y axis
	#---------------------
	y2ticks <- pretty(ylim2)
	axis( 4, y2ticks, line=ifelse(yaxs=='r',1,0.2), col=col.axes[2], lty=lty[2], lwd=axis_lwd, yaxs=yaxs )
	mtext( ylab[2], side=4, las=3, line=line, cex=cex.ylab )

	if( oldstyle )
		par(mar=origmar)

	if( main != '' )
		title( main )

	usr1 = parorig$usr
	tpar = par()
	usr2 = tpar$usr

	par( usr=parorig$usr )

	return( list( usr1=usr1, usr2=usr2 ))
}

#=============================================================================
mplot <- function(x,y,...,grid=TRUE,xvals=NA,
			gglook=FALSE,
			posnegbars=NA, barbase=0., posnegbar_pos='red', posnegbar_neg='blue',
			xlab='',
			ylab='',
			lty=1,
			xaxs='r',
			yaxs='r',
			xlabline=1.3,
			ylabline=2,
			cex_xlabline=0.9,
			cex_ylabline=0.9,
			type='l',
			col.axis='red',
			lty.grid=3,
			subgridx=NA,
			subgridy=NA,
			col.grid=rgb(0.5,0.5,0.5),
			axes=NA,
			box=TRUE,
			yxform_func=NA,	  # a function to transform the Y cordinate for plotting purposes
			xlim=NA, ylim=NA,
			gmtborder=NA,  # set to fractional width of border (try 0.01)
			saveplot=FALSE,
			mgp_x_axis=0.,
			loadplot='',
			incr_global_frameno=TRUE,
			is_geographical=FALSE,
			yvals=NA,ylabels=NA) {

	trace <- FALSE

	if( trace ) {
		print("mplot: entering" )
		print(paste("mplot: ylim=", pv(ylim) ))
		}

	if( sum(is.na(y)) == length(y)) {
		print(paste("mplot: error, called with all NA Y values, n=", length(y) ))
		canvas( xlim=range(x, na.rm=TRUE), ylim=c(0,1), ... )
		s = paste( 'All Y values are NA (n=', length(y), ')' )
		text( min(x, na.rm=TRUE), 0.5, s, adj=c(0.0, 0.5))
		return()
		}

	if( length( ylim ) == 2 ) {
		if( is.na(ylim[1]) || (is.na(ylim[2]))) 
			stop(paste("Error, called mplot with bad ylim:", pv(ylim) ))
		if( (! is.finite(ylim[1])) || (! is.finite(ylim[2]))) 
			stop(paste("Error, called mplot with bad (not finite) ylim:", pv(ylim) ))
		}

	if( ! exists('globalframenumber'))
		globalframenumber <<- 0

	if( incr_global_frameno ) 
		globalframenumber <<- globalframenumber + 1

	if( class(x) == 'SpatialPolygons' ) {
		np1 = length( x@polygons )
		print(paste('npoly level 1:', np1))
		for( ipoly in 1:np1 ) {
			np2 = length( x@polygons[[ipoly]]@Polygons )
			print(paste('For polygong', ipoly, ' # lev 2 polygons:', np2 ))
			for( ip2 in 1:np2 ) {
				txx = x@polygons[[ipoly]]@Polygons[[ip2]]@coords[,1]
				tyy = x@polygons[[ipoly]]@Polygons[[ip2]]@coords[,2]
				print(paste("Calling mplot with iouter, inner polygons:", ipoly, ip2 ))
				mplot( txx, tyy, main='SP mplot', ... )
				}
			}

		return()
		}

	if( saveplot ) {
		saveplot_fname = paste( myname, ".frame_", mformatC('%02d',globalframenumber), '.Rdat', sep='' )
		sav   		= list()
		sav$x 		= x
		sav$y 		= y
		sav$grid 	= grid
		sav$xlab 	= xlab
		sav$ylab 	= ylab
		sav$ylabline	= ylabline
		sav$type	= type
		sav$col		= col
		sav$lty		= lty
		sav$axes	= axes
		sav$box		= box
		sav$xlim	= xlim
		sav$ylim	= ylim
		sav$xaxs	= xaxs
		sav$yaxs	= yaxs
		save( file=saveplot_fname, sav )
		}
	else if( nchar(loadplot) > 0 ) {
		load( loadplot )
		x 		= sav$x
		y 		= sav$y
		grid 		= sav$grid
		xlab 		= sav$xlab
		ylab 		= sav$ylab
		ylabline	= sav$ylabline
		type		= sav$type
		col		= sav$col
		lty		= sav$lty
		axes		= sav$axes
		box		= sav$box
		xlim		= sav$xlim
		ylim		= sav$ylim
		xaxs		= sav$xaxs
		yaxs		= sav$yaxs
		}

	if( (length(y) == 1) && (is.na(y)) )
		stop(paste('mplot: error, must supply y values!'))

	if( length(x) != length(y)) {
		print("Error in mplot, have different number of X and Y values!")
		print(paste("# X values:",length(x)))
		print(paste("# Y values:",length(y)))
		stop('X and Y must be same length')
		}

	if( trace )
		print("mplot: ==a" )

	if( (length(xlim)==1) && (is.na(xlim)) )
		xlim <- range(x, na.rm=TRUE )
	if( (length(ylim)==1) && (is.na(ylim)) )
		ylim <- range(y, na.rm=TRUE )

	if( is.na(ylim[1]) || is.na(ylim[2])) {
		print(paste("Error, called mplot with NA in the ylim! This is not allowed"))
		print(paste("ylim:", pv(ylim) ))
		quickstats( x, 'xvals' )
		quickstats( y, 'yvals' )
		stop('exiting now')
		}

	#---------------------------------------------------------
	# Check the geometry of the plot and make sure it's square
	# We assume X and Y are in degrees
	# Note: aspect ratio is width/height
	#---------------------------------------------------------
	if( is_geographical ) check_plot_aspect_ratio( 'ccontour', xlim, ylim )

	if( trace )
		print("mplot: ==b" )

	if( !is.na(yxform_func)) {
		if( is.na(yvals) ) 
			stop("if specify yxform_func, must specify yvals as well (values to label)")
		y <- yxform_func(y)
		axes <- FALSE
		ylim <- yxform_func(ylim)
		}

	if( trace )
		print("mplot: ==c" )

	par(tck=0.02,las=1,xaxs=xaxs,yaxs=yaxs)

	if( ! is.na(axes)) 
		myaxes <- axes
	else
		myaxes <- TRUE
	if( (!is.na(xvals)) || (!is.na(yvals)) ) {
		myaxes <- FALSE
		}

	if( trace )
		print("mplot: ==d" )

	if( ! is.na(posnegbars) ) {
		if( trace )
			print("mplot: ==d2" )
		nx <- length(x)
		if( length(posnegbars) == nx )
			barcolors = posnegbars
		else
			barcolors = ifelse( y >= 0, posnegbar_pos, posnegbar_neg )
		xtics <- pretty(x)
		ytics <- pretty(y)
		mnx <- min(x)
		mxx <- max(x)
		mny <- min(y)
		mxy <- max(y)
		plot(c(mnx,mxx),c(mny,mxy),type='p',pch='.',col='white',...,xlim=xlim,ylim=ylim)
		dx <- x[2]-x[1]
		for( i in 1:nx ) {
			xleft  <- x[i]-dx/2.
			xright <- xleft + dx
			ybot   <- barbase
			ytop   <- y[i]
			polygon(c(xleft,xright,xright,xleft),
				c(ybot,ybot,ytop,ytop), 
				border=barcolors[i], col=barcolors[i] )
			}
		}
	else
		{
		if( trace )
			print("mplot: ==d5" )

		#--------------------------------------------------------
		# HERE IS WHERE WE MAKE THE MAIN CANVAS AND PLOT THE LINE
		#--------------------------------------------------------
		plot(x,y,xlab='',ylab='', lty=lty, ...,axes=FALSE,type=type,xlim=xlim,ylim=ylim,bty='n')

		if( gglook ) {
			uu = par()$usr	# llx, urx, lly, ury
			llx = uu[1]
			urx = uu[2]
			lly = uu[3]
			ury = uu[4]
			polygon( c(llx,llx,urx,urx), c(lly,ury,ury,lly), col=grey(0.9), border=NA )

			xaxp = par()$xaxp
			print(paste('xaxp:', pv(xaxp) ))

			#-------------------
			# White X grid lines
			#-------------------
			txrange = xaxp[2] - xaxp[1]
			tdx = txrange / xaxp[3]
			txtics = seq( xaxp[1], xaxp[2], by=tdx )
			tnxt = length( txtics )
			for( ii in 1:tnxt ) 
				abline( v=txtics[ii], col='white', lwd=2 )

			# Do half tics in X
			for( ii in 1:tnxt ) 
				abline( v=txtics[ii]+tdx/2, col=grey(0.95), lwd=1 )

			yaxp = par()$yaxp
			print(paste('yaxp:', pv(yaxp) ))

			#-------------------
			# White Y grid lines
			#-------------------
			tyrange = yaxp[2] - yaxp[1]
			tdy = tyrange / yaxp[3]
			tytics = seq( yaxp[1], yaxp[2], by=tdy )
			tnyt = length( tytics )
			for( ii in 1:tnyt ) 
				abline( h=tytics[ii], col='white', lwd=2 )

			# Do half tics in X
			for( ii in 1:tnyt ) 
				abline( h=tytics[ii]+tdy/2, col=grey(0.95), lwd=1 )

			lines( x, y, lty=lty, ..., type=type )
			}

		if( myaxes ) {
			#---------------------------------------------------------------------
			# Adjust the offset of the X axis TIC MARKS (not label) to be less
			# than the default of mgp[2] = 0.2. This is awkward because that value
			# works fine for the Y axis, but is too large for the X axis. So, have
			# do do the axes separately
			#---------------------------------------------------------------------
			mgp_orig = par()$mgp
			mgp = mgp_orig
			mgp[2] = mgp_x_axis
			par(mgp=mgp)
			axis( 1 )

			par(mgp=mgp_orig)
			par(las=1)
			axis(2)
			}
		}

	#---------------------------------------------
	# Put on the X axes LABEL (not the tick marks)
	#---------------------------------------------
	mtext(xlab,1,line=xlabline,cex=cex_xlabline)

	if( trace )
		print("mplot: ==e" )

	par(las=0)
	mtext(ylab,2,line=ylabline,cex=cex_ylabline)

	if( trace )
		print("mplot: ==f" )

	if( grid ) {
		if( trace )
			print("mplot: ==g" )

		log <- par("xlog")
		ax <- par("xaxp")
		at <- axTicks(1, axp = ax, log=log)
		nxticks <- length(at)
		for( i in 1:nxticks )
			abline( v=at[i], lty=lty.grid, col=col.grid )
		if( !is.na(subgridx)) {
			delat <- at[2]-at[1]
			at <- c(at[1]-delat,at,at[nxticks]+delat)
			nxticks <- length(at)
			for( i in 1:(nxticks-1) ) {
				dx <- (at[i+1]-at[i])/subgridx
				for( st in 1:(subgridx-1) ) 
					abline(v=at[i]+dx*st, lty=lty.grid, col=col.grid )
				}
			}

		if( is.na(yxform_func) ) {
			log <- par("ylog")
			ax <- par("yaxp")
			at <- axTicks(2, axp = ax, log=log)
			nyticks <- length(at)
			for( i in 1:nyticks )
				abline( h=at[i], lty=lty.grid, col=col.grid )
			if( !is.na(subgridy)) {
				delat <- at[2]-at[1]
				at <- c(at[1]-delat,at,at[nyticks]+delat)
				nyticks <- length(at)
				for( i in 1:(nyticks-1) ) {
					dy <- (at[i+1]-at[i])/subgridy
					for( st in 1:(subgridy-1) ) {
						abline(h=at[i]+dy*st, lty=lty.grid, col=col.grid )
						}
					}
				}
			}

		}
	else
		{
		if( trace )
			print("mplot: ==h" )

		log <- par("ylog")
		ax <- par("yaxp")
		at <- axTicks(2, axp = ax, log=log)
		nyticks <- length(at)
		if( !is.na(subgridy)) {
			delat <- at[2]-at[1]
			at <- c(at[1]-delat,at,at[nyticks]+delat)
			nyticks <- length(at)
			dxtic <- (xlim[2]-xlim[1])*.02
			for( i in 1:(nyticks-1) ) {
				dy <- (at[i+1]-at[i])/subgridy
				for( st in 1:(subgridy-1) ) {
					abline(h=at[i]+dy*st, lty=lty.grid, col=col.grid )
					}
				}
			}
		}

	if( trace )
		print("mplot: ==i" )

	if( ! is.na(xvals) ) {
		if( is.na(xlabels) ) 
			labelToUse <- T
		else
			labelToUse <- xlabels
		axis(1,xvals,labels=labelToUse)
		axis(3,xvals,labels=F)

		if( is.na(yvals) ) {
			axis(2)
			axis(4,labels=F)
			}

		if( grid )
			abline(v=xvals,lty=3)
		}

	if( trace )
		print("mplot: ==j" )

	if( ! is.na(yvals) ) {
		if( is.na(yxform_func)) {
			if( is.na(ylabels) ) 
				labelToUse <- T
			else
				labelToUse <- ylabels
			axis(2,yvals,labels=labelToUse,col.axis=col.axis)

			if( is.na(xvals) ) {
				axis(1)
				axis(3,labels=F)
				}
			if( grid )
				abline(h=yvals,lty=3)
			}
		else
			{
			ny <- length(yvals)
			width <- xlim[2]-xlim[1]
			xoffset <- xlim[1] - width*.04
			for( ii in 1:ny ) {
				yoffset <- yxform_func( yvals[ii] )
				text(xoffset,yoffset,yvals[ii],xpd=NA, adj=c(1.2,0.5))
				if( grid )
					abline(h=yoffset,lty=lty.grid, col=col.grid )
				# tic
				lines(c(xoffset,xoffset+width*.01), c(yoffset,yoffset))
				}
			}
		}

	if( trace )
		print("mplot: ==k" )

	if( !is.na(yxform_func))
		axis(1)

	if( box )
		box( col=par()$col.axis )

	if( trace )
		print("mplot: ==l : exiting" )

	if( !is.na(gmtborder) ) {
		addgmtborder( gmtborder )
		}
}

#====================================================================================
addgmtborder <- function( width, xvals=NA, yvals=NA ) {

	usercoords <- par('usr')
	if( (length(xvals)==1) && (is.na(xvals)))
		xaxis <- par('xaxp')
	else
		xaxis <- c(min(xvals), max(xvals), length(xvals)-1)

	if( (length(yvals)==1) && (is.na(yvals)))
		yaxis <- par('yaxp')
	else
		yaxis <- c(min(yvals), max(yvals), length(yvals)-1)

	plotinch <- par('pin')
	aspect <- plotinch[2]/plotinch[1]	# width/height

	border <- 'black'

	# left 
	x0 <- usercoords[1]
	x1 <- x0 + (usercoords[2]-x0)*width*aspect
	dy <- (yaxis[2]-yaxis[1])/yaxis[3]
	flipflop <- 1
	for( i in -1:yaxis[3]) {
		col <- ifelse( flipflop==0, 'black', 'white' )
		y0 <- yaxis[1] + dy*i
		y1 <- y0 + dy
		polygon(c(x0,x1,x1,x0),c(y0,y0,y1,y1),col=col,border=border)
		flipflop <- 1 - flipflop
		}
	abline(v=x1)
	
	# right
	x1 <- usercoords[2]
	x0 <- x1 - (usercoords[2]-x0)*width*aspect
	dy <- (yaxis[2]-yaxis[1])/yaxis[3]
	flipflop <- 1
	for( i in -1:yaxis[3]) {
		col <- ifelse( flipflop==0, 'black', 'white' )
		y0 <- yaxis[1] + dy*i
		y1 <- y0 + dy
		polygon(c(x0,x1,x1,x0),c(y0,y0,y1,y1),col=col,border=border)
		flipflop <- 1 - flipflop
		}
	abline(v=x0)
	
	# bottom
	y0 <- usercoords[3]
	y1 <- y0 + (usercoords[4]-usercoords[3])*width
	dx <- (xaxis[2]-xaxis[1])/xaxis[3]
	flipflop <- 0
	for( i in -1:xaxis[3]) {
		col <- ifelse( flipflop==0, 'black', 'white' )
		x0 <- xaxis[1] + dx*i
		x1 <- x0 + dx
		polygon(c(x0,x1,x1,x0),c(y0,y0,y1,y1),col=col,border=border)
		flipflop <- 1 - flipflop
		}
	abline(h=y1)

	# top
	y1 <- usercoords[4]
	y0 <- y1 - (usercoords[4]-usercoords[3])*width
	dx <- (xaxis[2]-xaxis[1])/xaxis[3]
	flipflop <- 0
	for( i in -1:xaxis[3]) {
		col <- ifelse( flipflop==0, 'black', 'white' )
		x0 <- xaxis[1] + dx*i
		x1 <- x0 + dx
		polygon(c(x0,x1,x1,x0),c(y0,y0,y1,y1),col=col,border=border)
		flipflop <- 1 - flipflop
		}
	abline(h=y0)

	box()
}

#=============================================================================
mgrid <- function( col='grey', xval=NA, yval=NA, lty=3 ) {
	
	if( ! is.na(xval) ) {
		for( i in 1:length(xval) )
			abline(v=xval[i],col=col,lty=lty)
		for( i in 1:length(yval) )
			abline(h=yval[i],col=col,lty=lty)
		}
	else
		grid(col=col)
}

#=============================================================================
cbimage <- function( xvals, yvals, data, satmask=NA, min=NA, max=NA, nc=10,
		desat = .7, cbmain=NA, ... ) {
	
	if( is.na(min) )
		min <- min(data)
	else
		data <- ifelse( data<min, min, data )
	if( is.na(max) ) 
		max <- max(data)
	else
		data <- ifelse( data>max, max, data )
	colLevs <- pretty(c(min,max),nc)
	nc <- length(colLevs)
	col <- mrainbow(nc,satmask=satmask,desat=desat)

	nfigs <- 2
	figs <- array(0.,c(nfigs,4))
	figs[1,] <- c(0,1,0,.146)
	figs[2,] <- c(0,1,.1,1.0)
	split.screen(figs)

	screen(1,new=F)
	colorbar(c(min,max),col,colLevs,main=cbmain)

	screen(2,new=F)
	mgp <- c(1.5,0.6,0)
	par(mgp=mgp,tck=-0.02,las=1)
	image(xvals,yvals,data,col=col, ...)
}

#===============================================================================
ccontour_2dlonlat = function( lon2d, lat2d, zz, ... ) {

	xr = range( lon2d )
	yr = range( lat2d )

	nx = dim( lon2d )[1]
	ny = dim( lon2d )[2]

	inx2 = floor( nx/2 )
	iny2 = floor( ny/2 )

	dx = lon2d[inx2+1, iny2] - lon2d[inx2,iny2]
	dy = lon2d[inx2, iny2+1] - lon2d[inx2,iny2]

	dx = abs(dx)
	dy = abs(dy)

	xvals = seq( xr[1], xr[2], by=dx )
	yvals = seq( yr[1], yr[2], by=dy )

	print('running gmt_surface...')
	rvsurf = gmt_surface( lon2d, lat2d, zz, 
		west=xr[1], east=xr[2], 
		south=yr[1], north=yr[2],
		xinc=dx, yinc=dy )
	print('...done running gmt_surface')

	ccontour( rvsurf$x, rvsurf$y, rvsurf$z, ... )
}

#===============================================================================
# See discussion under 'mcolorbar' for useage of 'discretevals'.  This 
# routine ASSUMES edges==TRUE, since if it were integer values to be colorcontoured,
# there would be no need to use discretevals to begin with.
#
# To use truecolor: it must be a list with $r, $g, and $b, AND you MUST
# explicitly set colormap=NA.  All the $r, $g, $b values must be in the range
# of [0,1], NOT 0 to 255
#
# To plot on a map projection using mapproject:
#  1. Run mapproject with the full X and Y limits that you want to include (this
#		sets the projection paramters):
#	rv = mapproject( c(-180,180), c(-90,90), proj='azequalarea')
#
#  2. Call ccontour with "use_proj=TRUE". 
#
#  3. You may or may not want to set proj_allow_backwards=FALSE; for exmaple, 
#	set proj_allow_backwards=FALSE for mollweide when you are centering
#	the plot on the Pacific ocean
#
# For an example see: /home/pierce/projects/NSF_aerosols/2014-01-23_second_dec_var_anal_prescribed_vs_interactive/plot_enso_spec_v2.R
#
# To plot on a map projection using proj4, which has many more projections than mapproject:
#  1. Call ccontour with proj4="proj4 string". Note that generally you will want to set
#	xlim-=c(-1.7e7,1.7e7), ylim=c(-9e6,9e6)
#
# NOTE: if you are using a discrete color map, then nacolor is not set here, it's
#	set via make_discretevals( col_na=... ), which sets dv$col_na
#
ccontour <- function( x, y, z, nlevels=10, add=FALSE, xlab='', ylab='', levels=NA, 
		axes=TRUE, docolor=TRUE, irregular=FALSE, showpointspch=NA, showpointscex=1.0, 
		zrange=NA, EW_lons=FALSE, collapserects=TRUE,
		cmapround=100, cmapexp=1, satmask=NA, specialcolors=NA, box=TRUE, regridnx=NA,
		regrid_xfact=NA, regrid_yfact=NA, brightness=1.0, discretevals=NA, 
		use_proj=FALSE, proj_allow_backwards=TRUE, proj4='',
		regridny=regridnx, showvalscex=NA, lambda=NA, xlim=NA, ylim=NA, 
		cmap='bluered_cblind', cmap_params=NA, nacolor='grey', truecolor=NA, 
		do_not_plot_NA=FALSE, trace=FALSE, main='', split=NA, 
		cbar_show=TRUE, cbar_title='', cbar_titlecex=0.9, cbar_title_yadj=2.5, cbar_whole_width=NA,
		cbar_labelcex=1.0, cbar_horiz=TRUE, cbar_fudge_width=1.0,
		cbar_fudge_height=1.2, cbar_fudge_yoffset=1.0, cbar_fudge_xoffset=1.0, contour_labcex=0.8,
		cbar_labelvals=NA, cbar_boxticvals=NA, cbar_label_srt=NA, apply_land_sea_mask='',
		contour_show=NULL, contour_negposcolors=c(rgb(0,0,.2),rgb(.3,0,0)),
		contour_levels=NULL, contour_nozero=FALSE, clip=TRUE, clip_poly=NULL,
		use_TPS=FALSE, use_gmt_surface=FALSE, use_gmt_xyz2grd=FALSE,
		gmt_t=0.25, gmt_xinc=NA, gmt_yinc=NA, saveplot=FALSE, use_saveplot=NA,
		xlabline=10.0, ylabline=2.0, cex_xlabline=0.9, cex_ylabline=0.9, 
		raster=FALSE, cmap_verbose=FALSE, is_geographical=TRUE, save_nc_list=NULL,
		allow_single_value=FALSE, quiet=FALSE,
		... ) {

	if( is.null( contour_show )) {	# if user does not specify contour_show, figure it out
		if( is.null( contour_levels )) 
			contour_show = FALSE
		else
			contour_show = TRUE
		}

	if( trace ) print(paste("ccontour: entering, irregular=", irregular))

	if( ! exists('globalframenumber'))
		globalframenumber <<- 0

	if( !add ) 
		globalframenumber <<- globalframenumber + 1

	if( (length(zrange) == 1) && (is.na(zrange[1])))
		zrange = c(min(z,na.rm=TRUE),max(z,na.rm=TRUE))

	if( storage.mode( save_nc_list ) == 'list' ) {
		#------------------------------------------------
		# save_nc_list must have following components:
		#	$xunits
		#	$yunits
		#	$zunits
		#	$fname_out
		#------------------------------------------------
		xdim = ncdim_def( 'x', save_nc_list$xunits, x, longname=xlab )
		ydim = ncdim_def( 'y', save_nc_list$yunits, y, longname=ylab )
		var_out = ncvar_def( 'z', save_nc_list$zunits, list(xdim,ydim), 1.e30 )

		print(paste("ccontour: saving data to netcdf file", save_nc_list$fname_out ))
		ncid_out = nc_create( save_nc_list$fname_out, var_out )

		if( ! is.na( xlim )) ncatt_put( ncid_out, var_out$name, 'xlim', xlim )
		if( ! is.na( ylim )) ncatt_put( ncid_out, var_out$name, 'ylim', ylim )

		z2 = z  + 0.0
		ncvar_put( ncid_out, var_out, z2 )
		
		nc_close( ncid_out )
		}

	if( saveplot ) {
		saveplot_fname = paste( myname, ".frame_", mformatC('%02d',globalframenumber), '.Rdat', sep='' )
		sav = list()
		sav$x		= x
		sav$y		= y
		sav$z		= z
		sav$xlim	= xlim
		sav$ylim	= ylim
		sav$zrange	= zrange
		sav$xlab	= xlab
		sav$ylab	= ylab
		sav$main	= main
		sav$contour_show = contour_show
		sav$cmap	= cmap
		sav$cmapexp	= cmapexp
		sav$contour_levels 	= contour_levels
		sav$discretevals 	= discretevals
		sav$cbar_labelvals 	= cbar_labelvals
		sav$cbar_boxticvals 	= cbar_boxticvals
		save( file=saveplot_fname, sav )
		}

	if( !is.na(use_saveplot)) {
		x		= use_saveplot$x
		y		= use_saveplot$y
		z		= use_saveplot$z
		xlim		= use_saveplot$xlim
		ylim		= use_saveplot$ylim
		zrange		= use_saveplot$zrange
		xlab		= use_saveplot$xlab
		ylab		= use_saveplot$ylab
		main		= use_saveplot$main
		contour_show	= use_saveplot$contour_show
		cmap		= use_saveplot$cmap
		cmapexp		= use_saveplot$cmapexp
		if( is.null(cmapexp)) cmapexp = 1
		contour_levels	= use_saveplot$contour_levels
		discretevals 	= use_saveplot$discretevals
		cbar_labelvals	= use_saveplot$cbar_labelvals
		if( is.null(cbar_labelvals)) cbar_labelvals = NA
		cbar_boxticvals = use_saveplot$cbar_boxticvals
		if( is.null(cbar_boxticvals)) cbar_boxticvals = NA
		}

	if( ! irregular ) {
		if( length(x) != dim(z)[1] ) 
			stop(paste("Error, entered with", length(x), 'x values, but dim(z) = ', pv(dim(z)) ))
		if( length(y) != dim(z)[2] ) 
			stop(paste("Error, entered with", length(y), 'y values, but dim(z) = ', pv(dim(z)) ))
		}

	#-------------------
	# Check for all NA's
	#-------------------
	nna = sum( is.na(z) )
	if( nna == length(z)) {
		canvas( xlim=range(x,na.rm=TRUE), ylim=range(y,na.rm=TRUE), grid=FALSE, xaxs='i', yaxs='i', main=main )
		text( mean(xlim,na.rm=TRUE), mean(ylim,na.rm=TRUE), paste("All data is NAs"), cex=1.6)
		return()
		}

	#------------------------------
	# Check x, y for reasonableness
	#------------------------------
	if( is.null(x) || (length(x) < 2)) {
		print(paste("Error, bad X array passed to ccontour:"))
		print(x)
		stop("Needs to be an array of length at least 2!")
		}
	if( is.null(y) || (length(y) < 2)) {
		print(paste("Error, bad Y array passed to ccontour:"))
		print(y)
		stop("Needs to be an array of length at least 2!")
		}

	if( length(zrange) != 2 ) 
		stop(paste("Error, zrange must be a two element array with (zmin, zmax)"))

	if( is.na(zrange[1]))
		zrange[1] = min( z, na.rm=TRUE ) 
	if( is.na(zrange[2]))
		zrange[2] = max( z, na.rm=TRUE ) 

	#-------------------------
	# Refine grid if requested
	#-------------------------
	if( (!is.na(regrid_xfact)) && (regrid_xfact != 1)) {
		if( trace ) print("ccontour: Putting data on a finer grid")
		rv <- mrefinegrid( x, y, z, regrid_xfact, regrid_yfact ) 
		x <- rv$x
		y <- rv$y
		z <- rv$z
		if( trace ) print("ccontour: done putting data on a finer grid")
		}

	#-----------------------------------
	# If data is irregular, then grid it
	#-----------------------------------
	if( irregular ) {
		if( trace ) print("regridding irregular data")
		rr <- collapse3missing( x, y, z )
		x <- rr$a
		y <- rr$b
		z <- rr$c
		xirreg <- x
		yirreg <- y
		zirreg <- z
		if(trace) print("contouring irregular data -- B")
		if( use_TPS ) {
			library("fields")
			np <- length(x)
			if(trace) print(paste("np=",np))
			if( (length(y) != np) || (length(z) != np)) 
				stop("length of x, y, and z must all be same if irregular data!")
			bigx <- array(0.,dim=c(np,2))
			bigx[,1] <- x[1:np]
			bigx[,2] <- y[1:np]
			if(trace) print("about to call Tps")
			Tps_result <- Tps(bigx,z,lambda=lambda) #,m=Tps_m)
	##Tps_result <- Krig(bigx,z,lambda=lambda)
			print(paste("lambda used:",Tps_result$lambda))

			if( trace ) {
				fmt = '%9.3f'
				for( i in 1:np ) 
					print(paste('x:', mformatC(fmt,x[i]),
						    'y:', mformatC(fmt,y[i]),
						    'z:', mformatC(fmt,z[i]),
						    'fitted value:', mformatC(fmt,Tps_result$fitted.values[i]),
						    'resid:',mformatC(fmt,Tps_result$residuals[i])))
				}
			rmse = sqrt(mean( Tps_result$residuals**2 ))
			print(paste("RMS of residuals:", rmse ))
			p <- predict.surface(Tps_result,nx=regridnx,ny=regridny)
			x <- p$x
			y <- p$y
			z <- p$z
			}

		else 
			{
			if(trace) print("gridding irregular data with gmt_surface")
			rv = gmt_surface( x, y, z, xinc=gmt_xinc, yinc=gmt_yinc, tension=gmt_t, use_xyz2grd=use_gmt_xyz2grd )
			if(trace) print("done gridding irregular data with gmt_surface")
			x = rv$x
			y = rv$y
			z = rv$z
			}


		retval <- list()
		retval$x <- x
		retval$y <- y
		retval$z <- z
		if(trace) {
			print("done manip irreg data")
			}
		}

	ndims_z = length( dim( z ))
	if( ndims_z != 2 ) 
		stop(paste("Error, Z field to ccontour should be 2 dimensional, but got ndims_z=", ndims_z ))

	#------------------
	# Handle regridding
	#------------------
	if(trace) print("Handle regridding")
	if( ! is.na(regridnx)) {

		if( trace ) print("ccontour: doing regridnx")
		#-------------------------------------
		# Turn our gridded x, y, z into a list
		#-------------------------------------
		npts <- 0
		nx <- length(x)
		ny <- length(y)
		newz <- array(0.,nx*ny)
		newx <- newz
		newy <- newz
		for( j in 1:ny ) 
		for( i in 1:nx ) {
			xok <- TRUE
			if( length(xlim)>1 ) 
				xok <- ((x[i]>=xlim[1]) && (x[i]<=xlim[2]))
			yok <- TRUE
			if( length(ylim)>1 ) 
				yok <- ((y[j]>=ylim[1]) && (y[j]<=ylim[2]))
			if( xok && yok )  {
				npts <- npts + 1
				newx[npts] <- x[i]
				newy[npts] <- y[j]
				newz[npts] <- z[i,j]
				}
			}
		x <- newx[1:npts]
		y <- newy[1:npts]
		z <- newz[1:npts]

		if( trace ) print("ccontour: making a surface that fits x,y,z")
		#-------------------------------------------------------------
		# Using our new list-form x,y,z, make a surface that fits them
		#-------------------------------------------------------------
		rr <- collapse3missing( x, y, z )
		x <- rr$a
		y <- rr$b
		z <- rr$c
		library("fields")
		np <- length(x)
		if( (length(y) != np) || (length(z) != np)) 
			stop("length of x, y, and z must all be same if irregular data!")
		bigx <- array(0.,dim=c(np,2))
		bigx[,1] <- x[1:np]
		bigx[,2] <- y[1:np]
		if( trace ) print("ccontour: about to call Tps")
		result <- Tps(bigx,z,lambda=lambda)
		print(paste("lambda used:",result$lambda))
		if( trace ) print("ccontour: about to call predict.surface")
		p <- predict.surface(result,nx=regridnx,ny=regridny)
		xirreg <- x
		yirreg <- y
		zirreg <- z
		x <- p$x
		y <- p$y
		z <- p$z
		retval <- list()
		retval$x <- x
		retval$y <- y
		retval$z <- z
		}

	#-----------------------
	# Apply a land-sea mask? 
	#-----------------------
	if( apply_land_sea_mask != '' ) {
		lsm = land_sea_mask( x, y, critfrac=0.4 )
		if( apply_land_sea_mask == 'land' ) 
			z[ lsm == 0 ] = NA
		else if( apply_land_sea_mask == 'sea' ) 
			z[ lsm == 1 ] = NA
		else
			stop(paste("error, passed parameter apply_land_sea_mask must be either the string land or sea"))
		}

	#----------------------
	# Apply a clip polygon?
	#----------------------
	if( !is.null( clip_poly )) {
		nx = length(x)
		ny = length(y)
		inpoly = array(FALSE, dim=c(nx,ny))

		polyx = clip_poly$x
		polyy = clip_poly$y
		npp = length( polyx )
		print(paste("# clip_poly polygon pts:", npp ))
		#----------------------------------------------------
		# If the polygon repeats the last point, eliminate it
		#----------------------------------------------------
		is_rep = ( (polyx[1] == polyx[npp]) && (polyy[1] == polyy[npp]))
		if( is_rep ) {
			print(paste("removing repeated polgon point!"))
			polyx = polyx[1:(npp-1)]
			polyy = polyy[1:(npp-1)]
			}

		for( jj in 1:ny )
		for( ii in 1:nx )
			inpoly[ii,jj] = pointInPolygonC( x[ii], y[jj], polyx, polyy )

		z[ ! inpoly ] = NA
		}

	#--------------------------------------------
	# Set xlim and ylim if the user did not do so
	#--------------------------------------------
	if( trace ) print("ccontour: setting xlim and ylim")
	if( (length(xlim)==1) && is.na(xlim)) {
		nx = length(x)
		dx = x[2] - x[1]
		x0 = x[1] - dx/2
		dx = x[nx] - x[nx-1]
		x1 = x[nx] + dx/2
		xlim <- c(x0,x1)
		}
	if( (length(ylim)==1) && is.na(ylim)) {
		ny = length(y)
		dy = y[2] - y[1]
		y0 = y[1] - dy/2
		dy = y[ny] - y[ny-1]
		y1 = y[ny] + dy/2
		#ylim <- c(y0,y1)
		if( y0 < y1 ) 
			ylim = c(y0,y1)
		else
			ylim = c(y1,y0)
		}
	if( trace ) print("ccontour: done with x and y lim")

	#------------------------
	# Check for all one value
	#------------------------
	if( trace ) print("ccontour: check for all one value")
	zr = range( z, na.rm=TRUE )
	if( ! allow_single_value ) {
		if( zr[1] == zr[2] ) {
			canvas( xlim=xlim, ylim=ylim, grid=FALSE, xaxs='i', yaxs='i', main=main )
			text( mean(xlim), mean(ylim), paste("All data is ", zr[1] ), cex=1.6)
			return()
			}
		}

	#----------------------------------
	# Clip to only include active areas
	#----------------------------------
	if( clip ) {
		if( trace ) print(paste("ccontour: clip to active areas, ndims_z=", ndims_z ))
		rvclip = clipout( xlim, ylim, x, y, z )
		x = rvclip$x
		y = rvclip$y
		z = rvclip$z
		}

	zorig_not_bin = z

	#---------------------------------------------------------
	# Check the geometry of the plot and make sure it's square
	# We assume X and Y are in degrees
	# Note: aspect ratio is width/height
	#---------------------------------------------------------
	if( trace ) print(paste("ccontour: checking is_geographical=", is_geographical ))
	if( is_geographical ) check_plot_aspect_ratio( 'ccontour', xlim, ylim )

	#-----------------------
	# Handle raster plotting
	#-----------------------
	if( raster ) {
		if( (length(discretevals) == 1) && (is.na(discretevals)))
			stop("Must specify discretevals for raster color contouring")

		zz = t(flip(z))
		czz = array_to_color_array( zz, discretevals )
		}

	#--------------------------------------------------------------------
	# Handle discrete vals by binning.  We assume this is an 'edges' case
	# of discrete bins since we don't need to specify discrete vals if
	# they are actually integer bins already.
	#--------------------------------------------------------------------
	if( (!raster) && (!is.na(discretevals))) {
		if( trace ) print("ccontour: handle discrete vals ")

		if( is.null( discretevals$arrow_left ))
			discretevals$arrow_left = FALSE
		if( is.null( discretevals$arrow_right ))
			discretevals$arrow_right = FALSE

		#----------------------------------------------------------
		# Are we using values off the edges of the discrete values?
		#----------------------------------------------------------
		ndcol = length( discretevals$colors )
		ndval = length( discretevals$values )
		col_idx0 = 0
		if( discretevals$arrow_left ) 
			col_idx0 = 1

		if( discretevals$edges ) {
			if( trace ) {
				print("ccontour: ==discretevals with following values:")
				print(discretevals$values)
				}
			databin <- array(NA,dim=dim(z))
			nbins   <- ndcol

			if( discretevals$arrow_left ) 
				databin[ z < discretevals$values[1] ] = 1
			else
				databin[ z < discretevals$values[1] ] = NA

			if( discretevals$arrow_right ) 
				databin[ z > discretevals$values[ndval] ] = ndcol
			else
				databin[ z > discretevals$values[ndval] ] = NA

			for( ib in 1:(ndval-1) )
				databin[ (z>=discretevals$values[ib]) & (z<=discretevals$values[ib+1]) ] <- ib + col_idx0

			databin[ is.na(z) ] <- NA
			z <- databin
			zrange <- c(1,nbins)
			}
		else
			{
			#--------
			# Centers
			#--------
			databin <- array(NA,dim=dim(z))
			nbins   <- ndcol
			tol = mean(abs(discretevals$values)) * 1.e-5

			if( discretevals$arrow_left ) 
				databin[ z < (discretevals$values[1]-tol) ] = 1
			else
				databin[ z < (discretevals$values[1]-tol) ] = NA

			if( discretevals$arrow_right ) 
				databin[ z > (discretevals$values[ndval]+tol) ] = ndcol
			else
				databin[ z > (discretevals$values[ndval]+tol) ] = NA

			for( ib in 1:nbins ) {
				databin[ abs(z - discretevals$values[ib]) < tol ] <- ib + col_idx0
				}

			databin[ is.na(z) ] <- NA
			z <- databin
			zrange <- c(1,nbins)
			}
		}

	#=================================
	# HERE IS WHERE WE MAKE THE CANVAS
	#=================================
	if( trace ) print("ccontour: ==about to call mplot")
	if( ! add ) {

		if( axes )
			m_axes = (!EW_lons)
		else
			m_axes = FALSE

		mplot(0,0,xlim=xlim,ylim=ylim,type='p',pch='.',col='white',	# Note: internal axes
			incr_global_frameno=FALSE,
			xlabline=xlabline, ylabline=ylabline,
			box=box,
			cex_xlabline=cex_xlabline, cex_ylabline=cex_ylabline,
			axes=m_axes, xlab=xlab, ylab=ylab,
			xaxs='i',yaxs='i', main=main, grid=FALSE, ...)

		if( EW_lons ) {

			par(las=1)
			axis(2)	# Put on Y (latitude) axis

			#------------------------------------------------------------------
			# Label X axis as if longitudes with "East" and "West" designations
			#------------------------------------------------------------------
			xaxp = par()$xaxp
			lonlabs = seq( xaxp[1], xaxp[2], length.out=xaxp[3]+1 )
			nlabs = length( lonlabs )
			for( ilab in 1:nlabs ) {
				lonval = lonlabs[ilab]
				if( lonval < 180 ) {
					val2use = lonval
					hemi = 'E'
					}
				else if( lonval > 180 ) {
					val2use = 360 - lonval
					hemi = 'W'
					}
				ymin = par()$usr[3]
				lonstr = paste( val2use, hemi, sep='' )
				text( lonval, ymin, lonstr, adj=c(0.5,1.5), xpd=NA)
				}
			}
		}

	if( raster )
		rasterImage( czz, min(x), min(y), max(x), max(y), interpolate=FALSE )
	else
		{
		#-----------
		# NOT raster 
		#-----------
		nx <- length(x)
		ny <- length(y)

		if( trace ) print("ccontour: ==about to get Z range")
		zmin   = zrange[1]
		zmax   = zrange[2]
		if( is.na(zmin)) stop(paste("ccontour: problem, zmin is NA!"))
		if( is.na(zmax)) stop(paste("ccontour: problem, zmax is NA!"))
		deltaz = zmax - zmin
		if( deltaz == 0 ) {
			print("zero range of data!")
			zmin = zmin - abs(0.01*zmin)
			zmax = zmax + abs(0.01*zmin)
			zrange = c(zmin, zmax)
			deltaz = zmax - zmin
			}

		if( trace ) print("ccontour: ==about to calc X positions")
		xmids  <- (x[1:(nx-1)] + x[2:nx])*0.5	# length: nx-1
		xleft  <- c(x[1] - 0.5*(x[2]-x[1]), xmids )
		xright <- c(xmids, x[nx] + 0.5*(x[nx] - x[nx-1]))

		xpoly <- array(NA,nx*5)
		for( i in 1:nx ) {
			xpoly[(i-1)*5+1] <- xleft [i]
			xpoly[(i-1)*5+2] <- xright[i]
			xpoly[(i-1)*5+3] <- xright[i]
			xpoly[(i-1)*5+4] <- xleft [i]
			}

		if( is.na(cmap)) {
			#------------
			# IS.NA(CMAP)
			#------------
			if( trace ) print("ccontour: == no cmap specified")
			if( ! is.na(truecolor)) {
				print("using truecolor image")
				if( is.null( truecolor$rgb)) {
					col <- rgb( truecolor$r, truecolor$g, truecolor$b )
					dim(col) <- dim(truecolor$r)
					}
				else
					col <- truecolor$rgb
				}
			else
				{
				if( trace ) print("using default colormap")
				fz <- (z-zmin)/deltaz
				fz[fz<0] <- 0
				fz[fz>1] <- 1

				r <- fz*fz
				g <- 1. - 2*abs(fz-0.5)
				b <- 1.-(fz*fz)

				r[is.na(r)] <- 0.
				g[is.na(g)] <- 0.
				b[is.na(b)] <- 0.
				col <- ifelse(is.na(fz),nacolor,mrgb(r,g,b))
				dim(col) <- dim(r)
				n_special <- length(specialcolors$vals)
				if( n_special > 0 ) {
					for( spc in 1:n_special )
						col[ abs(z - specialcolors$vals[spc]) < specialcolors$tol ] <- specialcolors$col[spc]
					}
				}
			}
		else
			{
			#----------------
			# NOT IS.NA(CMAP)
			#----------------
			if( trace ) print('ccontour: ==using specified colormap')
			if( !is.na(discretevals)) {
				if( trace ) print('ccontour: ==using discrete vals')
				col <- array('black',dim=c(nx,ny))
				ndv <- length( discretevals$colors )
				for( j in 1:ny )
				for( i in 1:nx ) {
					zz <- z[i,j]
					if( is.na(zz)) {
						col[i,j] <- discretevals$col_na
						}
					else
						{
						if( (zz < 1) || (zz > ndv))
							stop(paste("bad binnded z value: ", zz))
						col[i,j] <- discretevals$colors[ zz ]
						}
					}
				if( trace ) print('ccontour: ==done using discrete vals')
				}
			else
				{
				col <- mcolormap( cmap, z, exp=cmapexp, satmask=satmask, 
						nacolor=nacolor, params=cmap_params,
						zrange=zrange, specialcolors=specialcolors,
						round=cmapround, brightness=brightness,
						verbose=cmap_verbose,
						split=split )
				}
			}

		if( trace ) print("ccontour: == e")
		if( do_not_plot_NA )
			collapserects = FALSE
		if( use_proj || (nchar(proj4) > 1)) {
			collapserects = FALSE
			if( contour_show )
				stop(paste("sorry, cannot currently do ccontour with projection AND contours"))
			}

		xstart <- array(0,nx)
		xend   <- array(0,nx)
		xcol   <- array('white',nx)

		#=================================================
		# Here is where we actually draw the colored boxes
		#=================================================
		if( trace ) print("ccontour: == about to enter j loop")
		for( j in 1:ny ) {

			if( j == 1 ) 
				y0 <- y[1] - 0.5*(y[2]-y[1])
			else
				y0 <- 0.5*(y[j-1]+y[j])
			if( j == ny ) 
				y1 <- y[ny] + 0.5*(y[ny] - y[ny-1])
			else
				y1 <- 0.5*(y[j] + y[j+1])

			if( trace ) print(paste("y0,y1=",y0,y1))
			pattern <- c(y0,y0,y1,y1,NA)
			ypoly <- array(pattern,nx*5)

			if( collapserects ) {
				n <- 0
				colprev <- -1
				for( i in 1:nx ) {
					if( col[i,j] != colprev ) {
						n <- n + 1
						xstart[n] <- xleft[i]
						xend[n] <- xright[i]
						xcol[n] <- col[i,j]
						}
					else
						xend[n] <- xright[i]
					colprev <- col[i,j]
					}
				rect( xstart[1:n], y0, xend[1:n], y1, col=xcol[1:n], border=NA )
				}
			else
				{
				if( do_not_plot_NA ) {
					for( i in 1:nx ) {
						if( ! is.na(z[i,j]) ) {
							rect( xleft[i], y0, xright[i], y1, col=col[i,j], border=NA )
							}
						}
					}
				else
					{
					if( use_proj || (nchar(proj4)>1)) {
						for( i in 1:nx ) {
							if( ! is.na(z[i,j]) ) {
								txx = c(xleft[i],xright[i],xright[i],xleft[i])
								tyy = c(y0,y0,y1,y1)
								if( nchar(proj4) > 1 ) {
									nn = length(txx)
									txx[ txx < -180 ] = -180
									txx[ txx >  180 ] = 180
									tyy[ tyy < -90 ] = -90
									tyy[ tyy >  90 ] =  90
									xy = array(0, dim=c(nn,2))
									xy[,1] = txx
									xy[,2] = tyy
									trv = project( xy, proj4, degrees=TRUE )
									rvp = list( x=trv[,1], y=trv[,2] )
									}
								else
									rvp = mapproject( txx, tyy, '' )  # must NOT specify proj here!! Use existing!
								ismissing = ((is.na(rvp$x[1])) || is.na(rvp$x[2]))
								if( ! ismissing ) {
									doit = TRUE
									if( ! proj_allow_backwards ) {
										backwards = rvp$x[1] > rvp$x[2]
										doit = (! backwards)
										}
									if( doit )
										polygon(rvp$x,rvp$y,col=col[i,j],border=NA)
									}
								}
							}
						}
					else
						rect( xleft, y0, xright, y1, col=col[,j], border=NA )
					}
				}
			}
		}

	if(box)
		box()

	#-----------------------------
	# Add a color bar if requested
	#-----------------------------
	if( cbar_show && (! use_proj)) {

		if( cbar_horiz ) {
			y_reversed = (ylim[1] > ylim[2])
			x0 = xlim[1]
			x1 = xlim[2]
			yr = abs(ylim[2] - ylim[1])

			mr = par()$mfrow
			mc = par()$mfcol
			if( (mr[1] == 1) && (mr[2] == 1) && (mc[1] == 1) && (mc[2] == 1)) {
				cfnew = cbar_fudge_yoffset * 0.5
				print(paste("ccontour: only 1 plot detected: changing cbar_fudge_yoffset from",
					cbar_fudge_yoffset, "to", cfnew ))
				cbar_fudge_yoffset = cfnew
				}

			yoffset = yr*0.12*cbar_fudge_yoffset
			if( y_reversed ) 
				y1      = ylim[1] + abs(yoffset)
			else
				y1      = ylim[1] - abs(yoffset)
			y0	= y1 - abs(yr)*0.03*cbar_fudge_height

			if( y_reversed ) {
				t = y0
				y0 = y1
				y1 = t
				}
			}
		else
			{
			y0 = min(ylim)
			y1 = max(ylim)

			xr = abs(xlim[2] - xlim[1])
			xoffset = xr*0.03*cbar_fudge_xoffset
			x0 = max(xlim[2]) + xoffset
			x1 = x0 + xr*0.03*cbar_fudge_width
			}

		if( ! is.na(cmap)) {
			mcolorbar( zrange, cmap, x0, y0, x1, y1,
				discretevals=discretevals,
				cmapexp=cmapexp,
				satmask=satmask,
				labelvals=cbar_labelvals,
				boxticvals=cbar_boxticvals,
				whole_width=cbar_whole_width,
				labelcex=cbar_labelcex,
				horiz=cbar_horiz,
				title=cbar_title, 
				cbar_title_yadj=cbar_title_yadj,
				label_srt=cbar_label_srt,
				titlecex=cbar_titlecex ) 
			}
		}

	#-------------------------------------
	# Add our contours on top if requested
	#-------------------------------------
	if( contour_show ) {
		z2use = z
		if( !is.na(discretevals)) 
			z2use = zorig_not_bin
		mcontour( x, y, z2use, add=TRUE, negposcolors=contour_negposcolors, levels=contour_levels, 
			use_proj=use_proj,
			labcex=contour_labcex, 
			nozero=contour_nozero )
		}
}

#===============================================================================
# Note: Add arg labcex=1.0 to change size of text on contour labels.
# Ex: mcontour( x, y, z, labcex=1.0 )
# lty is a 2-element vector, with lty[1] for negative, lty[2] for positive vals
#
mcontour <- function( x, y, z, nlevels=10, add=F, levels=NA, nozero=FALSE, lty=NA,
		lwd=1, docolor=TRUE, irregular=FALSE, showpointspch=NA, showpointscex=1.0, 
		xlim=NA, ylim=NA, col='black', zerothick=3, showplot=TRUE,
		regridnx=30, regridny=30, negposcolors=c(rgb(0,0,0), rgb(0,0,0), rgb(0,0,0)),
		forceregridirregular=FALSE, showvalscex=NA, showvalscol='black', lambda=NA, irreg_land_sea_mask=FALSE,
		verbose=FALSE, is_geographical=TRUE,
		trace=FALSE, use_TPS=FALSE, return_Tps=FALSE, Tps_m=NULL,
		use_proj=FALSE, proj4='',
		gmt_t=0.25, gmt_xinc=NA, gmt_yinc=NA, ... ) {

	if( trace ) print('mcontour: entering')

	if( ! add ) 
		globalframenumber <<- globalframenumber + 1

	if( length(lty) == 1 )
		lty = c(lty,lty)

	if( forceregridirregular ) {
		if( trace ) print("forcing regrid irregular")
		npts <- 0
		nx <- length(x)
		ny <- length(y)
		newz <- array(0.,nx*ny)
		newx <- newz
		newy <- newz
		if( trace ) print("regrid irregular: XY loop")
		for( j in 1:ny ) 
		for( i in 1:nx ) {
			xok <- TRUE
			if( length(xlim)>1 ) 
				xok <- ((x[i]>=xlim[1]) && (x[i]<=xlim[2]))
			yok <- TRUE
			if( length(ylim)>1 ) 
				yok <- ((y[j]>=ylim[1]) && (y[j]<=ylim[2]))
			if( xok && yok )  {
				npts <- npts + 1
				newx[npts] <- x[i]
				newy[npts] <- y[j]
				newz[npts] <- z[i,j]
				}
			}
		irregular <- TRUE
		x <- newx[1:npts]
		y <- newy[1:npts]
		z <- newz[1:npts]
		#print(x)
		#print(y)
		#print(z)
		print(paste("new # of irregular points to contour:",npts))
		}

	if( irregular ) {
		if( trace ) print("regridding irregular data")
		rr <- collapse3missing( x, y, z )
		x <- rr$a
		y <- rr$b
		z <- rr$c
		xirreg <- x
		yirreg <- y
		zirreg <- z
		if(verbose) print("contouring irregular data -- A")
		if( use_TPS ) {
			library("fields")
			np <- length(x)
			if(verbose) print(paste("np=",np))
			if( (length(y) != np) || (length(z) != np)) 
				stop("length of x, y, and z must all be same if irregular data!")
			bigx <- array(0.,dim=c(np,2))
			bigx[,1] <- x[1:np]
			bigx[,2] <- y[1:np]
			if(verbose) print("about to call Tps")
			Tps_result <- Tps(bigx,z,lambda=lambda,m=Tps_m)
	##Tps_result <- Krig(bigx,z,lambda=lambda)
			print(paste("lambda used:",Tps_result$lambda))

			if( verbose ) {
				fmt = '%9.3f'
				for( i in 1:np ) 
					print(paste('x:', mformatC(fmt,x[i]),
						    'y:', mformatC(fmt,y[i]),
						    'z:', mformatC(fmt,z[i]),
						    'fitted value:', mformatC(fmt,Tps_result$fitted.values[i]),
						    'resid:',mformatC(fmt,Tps_result$residuals[i])))
				}
			rmse = sqrt(mean( Tps_result$residuals**2 ))
			print(paste("RMS of residuals:", rmse ))
			p <- predict.surface(Tps_result,nx=regridnx,ny=regridny)
			x <- p$x
			y <- p$y
			z <- p$z
			}
		else
			{
			rv = gmt_surface( x, y, z, xinc=gmt_xinc, yinc=gmt_yinc, t=gmt_t )
			x = rv$x
			y = rv$y
			z = rv$z
			}
		retval <- list()
		retval$x <- x
		retval$y <- y
		retval$z <- z
		if(verbose) {
			print(x)
			print(y)
			print("done manip irreg data")
			}

		if( irreg_land_sea_mask ) {
			lsm = land_sea_mask( x, y )
			z[ lsm == 0 ] = NA
			}

		if( ! showplot )
			return(retval)
		}

	nx <- dim(z)[1]
	ny <- dim(z)[2]
	if( length(x) != nx )
		stop(paste("error, data array is size",nx,"by",ny,"but passed",length(x),"X values!"))
	if( length(y) != ny )
		stop(paste("error, data array is size",nx,"by",ny,"but passed",length(y),"Y values!"))

	if( trace ) print("contour: ==a")
	if( (length(xlim)==1) && (is.na(xlim)))
		xlim <- range(x)
	if( (length(ylim)==1) && (is.na(ylim)))
		ylim <- c( y[1], y[ny] )

	#-----------------------
	# if inverted, flip here
	#-----------------------
	if( x[2] < x[1] ) {
		if( trace ) print("mcontour: flipping data in X")
		tz <- z
		tx <- x
		nx <- length(x)
		for( i in 1:nx ) {
			tx[i] <- x[nx-i+1]
			tz[i,] <- z[nx-i+1,]
			}
		x <- tx
		z <- tz
		}
	if( y[2] < y[1] ) {
		print(paste("mcontour: flipping data in Y, here are y vals:", y[1], y[2], y[3], y[4], y[5], '...'))
		y <- y[length(y):1]
		z <- flip(z)
		ylim = c( ylim[2], ylim[1] )
print(paste("Reversed Y, now set ylim to ", pv(ylim) ))
		}

	if( trace ) print("contour: ==b")
	zmean <- mean(z,na.rm=TRUE)
	z2 <- ifelse( is.na(z), zmean, z )

	rg <- range(z2)
	if( (length(levels)==1) && is.na(levels) )
		levels <- pretty(range(z2), nlevels)

	nl <- length(levels)
	if( nl > 1 )
		delta <- levels[2]-levels[1]
	else
		delta = 0.1
	eps <- abs(delta*1.e-4)

	if( trace ) print("contour: ==c")
	for( i in 1:nl ) {
		if( abs(levels[i]) < eps )
			levels[i] <- 0.0
		}

	# Separate levels out into pos and neg values
	if( trace ) print("contour: ==d")
	nNeg <- 0
	nPos <- 0
	for( i in 1:length(levels) ) {
		if( levels[i] < -eps ) {
			if( nNeg == 0 ) 
				negLevs <- levels[i]
			else
				negLevs <- c(negLevs, levels[i])
			nNeg <- nNeg + 1
			}
		else if( levels[i] > eps ) {
			if( nPos == 0 ) 
				posLevs <- levels[i]
			else
				posLevs <- c(posLevs, levels[i])
			nPos <- nPos + 1
			}
		}

	if( verbose ) {
		print("xlim:")
		print(xlim)
		print("ylim:")
		print(ylim)
		}

	if( use_proj ) {
		rv = mapproject( x, y, '' )	# must run with proj='' to keep existing projection
		x = rv$x
		y = rv$y
		}

	if( nchar(proj4) > 2 ) {
		rv = project( cbind(x,y), proj4, degrees=TRUE )
		}

	if( TRUE ) print(paste("contour: ==e, docolor=", docolor ))
	locadd <- add
	if( docolor ) {
		colorNeg  <- negposcolors[1]
		colorZero <- negposcolors[2]
		colorPos  <- negposcolors[3]
		ltyNeg <- ifelse(is.na(lty),1,lty[1])
		ltyPos <- ifelse(is.na(lty),1,lty[2])
		}
	else
		{
		colorNeg  <- col
		colorPos  <- col
		colorZero <- col
		ltyNeg <- ifelse(is.na(lty),3,lty[1])
		ltyPos <- ifelse(is.na(lty),1,lty[2])
		}
	if( trace ) print("contour: ==f")
	if( nNeg > 0 ) {
print(paste("contour neg with color", colorNeg ))
		contour(x,y,z,levels=negLevs,col=colorNeg,add=locadd,lty=ltyNeg,
			xlim=xlim,ylim=ylim,lwd=lwd,...)
		locadd <- T
		}
	if( trace ) print("contour: ==g")
	if( nPos > 0 ) {
print(paste("contour pos with color", colorPos ))
		contour(x,y,z,levels=posLevs,...,col=colorPos,add=locadd,
			lwd=lwd,xlim=xlim,ylim=ylim,lty=ltyPos)
		locadd <- T
		}
	if( trace ) print("contour: ==h")
	if( (rg[1] < 0.) && (rg[2] > 0.) && (! nozero)) {
		contour(x,y,z,levels=c(0.),lwd=lwd*zerothick,...,col=colorZero,add=locadd)
		locadd <- T
		}

	if( trace ) print("contour: ==i")
	if( irregular ) {
		if( trace ) print("contour: ==j")
		if( ! is.na(showpointspch) ) {
			print("adding points")
			points(xirreg, yirreg, pch=showpointspch,cex=showpointscex)
			}
		if( ! is.na(showvalscex) ) {
			text(xirreg,yirreg,formatC(zirreg,width=3),cex=showvalscex, adj=c(0.5,0.5), col=showvalscol)
			}
		if( return_Tps ) 
			return(Tps_result)
		else
			return(retval)
		}
	if( trace ) print("contour: ==z")

}

#=============================================================================
# eps_size_cm: c(width,height) in cm
#
mdev.off <- function( thisfile, date=TRUE, gv=TRUE, gs=FALSE, eps_size_cm=NA, add_bb=TRUE, bb_pad=NA, xv=NA, convert_to_gif=FALSE ) {

	# Is the 

	if( !is.na(xv)) {
		cmd = paste( "xv ", xv );
		system( cmd );
		return()
		}

	if( date ) 
		timestamp( thisfile )
	dev.off()

	#---------------------------------------
	# Put a valid bounding box into the file
	#---------------------------------------
	if( add_bb && (postscriptoutputfilename != "") ) {
		print('mdev.off: adding bounding box')
print(paste("bb pad: ", pv(bb_pad) ))
		if( length(bb_pad) > 1 ) {
			if( length(bb_pad) == 4 ) 
				mydo(paste("add_bb -pllx",bb_pad[1], "-plly", bb_pad[2], 
					"-purx",bb_pad[3], "-pury", bb_pad[4], postscriptoutputfilename))
			else
				mydo(paste("add_bb -padx",bb_pad[1], "-pady", bb_pad[2], postscriptoutputfilename))
			}
		else
			system(paste("add_bb", postscriptoutputfilename))
		}

	if( length(eps_size_cm) > 1) {
		llx = 1
		lly = 1
		cm2inch  = 0.393701 
		cm2point = cm2inch * 72
		urx = floor(eps_size_cm[1]*cm2point + 0.99)
		ury = floor(eps_size_cm[2]*cm2point + 0.99)
		tmpfname = paste(postscriptoutputfilename, ".tmp_eps", sep='' )
		cmd = paste( "epsffit ", llx, lly, urx, ury, 
			postscriptoutputfilename, tmpfname )
		mydo( cmd )
		system( paste( 'mv ', tmpfname, postscriptoutputfilename ))
		if( gs )
			system(paste("gs ",postscriptoutputfilename))
		else if( gv )
			system(paste("gv -antialias",postscriptoutputfilename))
		return()
		}

	if( globaloutputfileiseps ) {
		print("......... about to run epstopdf ")
		eps_fname = paste(myname, '.eps', sep='' )
		pdf_fname = paste(myname, '.pdf', sep='' )
		mydo( paste('epstopdf -outfile=', pdf_fname, ' ', eps_fname, sep='' ))
		if( globaloutputfileembedfonts ) {
			print(paste('..... about to embed Fonts in output eps file ', eps_fname))
			embedFonts( eps_fname, options='-dEPSCrop' )	# -dEPSCrop added 2021-03-08
			}
		}

	if( postscriptoutputfilename != "" ) {
		if( gs ) {
			#print('mdev.off: starting GS')
			system(paste("gs ",postscriptoutputfilename))
			}
		else if( gv ) {
			#print('mdev.off: starting GV')
			system(paste("gv -antialias",postscriptoutputfilename))
			}
		}

	if( convert_to_gif ) {
		cmd = paste( 'pstogif ', postscriptoutputfilename )
		mydo( cmd )
		}
}

#=============================================================================
timestamp <- function( thisfile ) {

	dir     <- Sys.getenv("PWD")
	dateStr <- date()
	str <- paste( dir, "/", thisfile, " ", dateStr, sep="" )

	cex0 = par()$cex

	mtext(side=1,str,adj=1,cex=.7*cex0,outer=TRUE)
}

#=============================================================================
mpostscript <- function( myname, onefile=TRUE, horizontal=TRUE, height=NA, width=NA, paper="letter",
	mgp=c(1.2,0.2,0),mar=c(3,4,2,1), eps=FALSE, pdf=FALSE, embedfonts=FALSE, family='Helvetica', ... ) {

	debug = TRUE

	if( debug ) print(paste('mpostscript: entering with myname=', myname ))

	if( pdf )
		suffix <- '.pdf'
	else
		suffix <- '.ps'

	if( debug ) print(paste('mpostscript: suffix=', suffix ))

	globaloutputfileiseps <<- FALSE
	globaloutputfileembedfonts <<- FALSE

	if( eps ) {
		if( debug ) print(paste('mpostscript: working on EPS'))

		horizontal <- FALSE
		onefile    <- FALSE
		paper      <- 'special'
		suffix     <- '.eps'
		if( is.na(width))
			width <- 8.0
		if( is.na(height))
			height <- 11.
		globaloutputfileiseps <<- TRUE
		use_my_family = FALSE
		if( embedfonts ) {
			print('>>> NOTE: mpostscript is substituting URWHelvetica for Helvetica since embedfonts was specifed <<<')
			family = 'URWHelvetica'
			globaloutputfileembedfonts <<- TRUE
			}
		}
	else if( embedfonts ) {
		stop('Error, can only specify embedfonts when eps==TRUE')
		}

	outf <- paste(myname,suffix,sep="")

	if( debug ) print(paste('mpostscript: outf:', outf ))

	if( pdf ) {
		if( debug ) print(paste('mpostscript: about to run pdf command'))
		pdf(outf,onefile=onefile,height=height,width=width,paper=paper,...)
		}
	else
		{
		if( debug ) print(paste('mpostscript: about to run postscript command'))
		postscript(outf,onefile=onefile,horizontal=horizontal, 
			height=height,width=width,paper=paper, 
			family=family, ...)
		}

	if( debug ) print(paste('mpostscript: ...done'))

	postscriptoutputfilename <<- outf

	par(mar=mar,oma=c(1,0,6,1),mgp=mgp,tck=0.01,...)

	globalframenumber <<- 0
}

#=============================================================================
# offsetx,y are between 0 and 1.  Set to .5 to have stipplings overlap without
# overwriting.
#
polystipple <- function( polyX, polyY, density=40, offsetx=0.1, offsety=0.1, border=NA, cex=.2, 
		logx=NA, logy=NA, debug=FALSE, ... ) {
	
	wh <- par()$pin
	width  <- wh[1]
	height <- wh[2]
	aspect <- width/height

	xr <- range(polyX)
	yr <- range(polyY)
	if( debug ) {
		print(paste('polystipple: xrange=', xr[1], xr[2] ))
		print(paste('polystipple: yrange=', yr[1], yr[2] ))
		}

	u <- par()$usr
	llx <- u[1]
	urx <- u[2]
	lly <- u[3]
	ury <- u[4]

	dx <- abs((urx-llx)/density)
	dy <- abs((ury-lly)/density*aspect)

	maxx <- as.integer(abs(ceiling(xr[2]-xr[1])/dx))
	maxy <- as.integer(abs(ceiling(yr[2]-yr[1])/dy))
	if( debug ) print(paste('polystipple: dx=',dx,' dy=',dy, ' maxx=', maxx, ' maxy=', maxy))

	flipflop <- as.integer(1)

	if( !is.na(logx) ) 
		x0 <- min(polyX[polyX>0])
	else
		x0 <- xr[1] + offsetx*dx

	if( !is.na(logy) ) 
		y0 <- min(polyY[polyY>0])
	else
		y0 <- yr[1] + offsety*dy

	if( (!is.na(logx) && (logx < 1.00001)))
		stop(paste("bad value for logx, should be ratio between pts (about 1.2 to 2):",logx))
	if( (!is.na(logy) && (logy < 1.00001)))
		stop(paste("bad value for logy, should be ratio between pts (about 1.05 to 2):",logy))

	tmp <- array(0.,1000)
	if( !is.na(logx) ) {
		nx <- 1
		tmp[1] <- x0
		while( tmp[nx] < xr[2] ) {
			nx <- nx + 1
			tmp[nx] <- tmp[nx-1] * logx
			}
		xvals <- tmp[1:nx]
		}
	else
		{
		dx <- abs(dx)
		if( x0 > xr[2] )
			xvals <- seq(xr[2],x0,by=dx)
		else
			xvals <- seq(x0,xr[2],by=dx)
		}

	if( !is.na(logy) ) {
		if( y0 <= 0 ) 
			stop(paste("error, can't have min y <= 0 with logy!  y0=",y0))
		ny <- 1
		tmp[1] <- y0
		while( tmp[ny] < yr[2] ) {
			ny <- ny + 1
			tmp[ny] <- tmp[ny-1] * logy
			}
		yvals <- tmp[1:ny]
		}
	else
		{
		dy <- abs(dy)
		if( y0 < yr[2] )
			yvals <- seq(y0,yr[2],by=dy)
		else
			yvals <- seq(yr[2],y0,by=dy)
		}


	nx <- length(xvals)
	ny <- length(yvals)
	ptx <- array(0.,nx*ny)
	pty <- array(0.,nx*ny)

	for( iy in 1:ny ) {
		flipflop <- as.integer(1 - flipflop)
		for( ix in 1:nx ) {
			x <- xvals[ix]
			y <- yvals[iy]
			if( !is.na(logx)) {
				if( flipflop ) {
					xp1 <- x * logx 
					mid <- (x + xp1)*.5
					xx <- mid
					}
				else
					xx <- x
				}
			else
				{
				if( flipflop )
					xx <- x + dx/2
				else
					xx <- x
				}
			idx <- ix + (iy-1)*nx
			ptx[idx] <- xx
			pty[idx] <- y
			}
		}

	inpoly <- pointInPolygonC( ptx, pty, polyX, polyY )
	ptx <- ptx[inpoly]
	pty <- pty[inpoly]

	if( length(ptx) > 0 ) 
		points(ptx,pty,pch=19,cex=cex,...)

	if( !is.na(border))
		polygon(polyX, polyY, border=border)
}

#=======================================================================
# Remember, 'col' will give a filled circle, set 'border' for the outline
# of a circle with no fill
#
# To make an actual circular mark on a geographical plot, set
# expand_x = 1 / cos(lat)
# 
# where lat is the latitude of the circle center in radians
#
circle <- function( x, y, r, start_deg=0, end_deg=360, expand_x=1.0, ... ) {
	
	pi <- 3.1415926536

	if( (start_deg==0) && (end_deg == 360)) {
		npts <- 360.
		xv <- array(0.,npts)
		yv <- array(0.,npts)
		dtheta <- 2*pi/npts
		theta <- 0.
		for( i in 1:npts ) {
			xv[i] <- cos(theta)*r*expand_x + x
			yv[i] <- sin(theta)*r + y
			theta <- theta + dtheta
			}
		polygon(xv,yv,...)
		}
	else
		{
		ddeg = end_deg - start_deg
		npts = floor( ddeg ) + 1
		xv = array(x, npts)
		yv = array(y, npts)
		dtheta_deg <- 0.1
		theta_deg <- start_deg
		i = 2
		while( theta_deg < end_deg ) {
			xv[i] <- cos(theta_deg/180*pi)*r*expand_x + x
			yv[i] <- sin(theta_deg/180*pi)*r + y
			theta_deg <- theta_deg + dtheta_deg
			i = i + 1
			}
		polygon(xv,yv,...)
		}
}

#==============================================================================================
# This is used to draw wind speed vectors on a geographical plot that is assumed to be in 
# non-projected form, but scaled properly so that the canvas is square (i.e., canvas( is_geographical=TRUE)
# confirms that the plot is correctly scaled for width/height.
#
# xc, yc		: geographical origin of vector
# uu, vv		: wind speed
# scale			: controls how long the drawn vector is for a given wind speed
# length		: controls length of arrow head
#
dp_arrows = function( xc, yc, uu, vv, scale=1.0, length=0.07, ... ) {

	inner_scale = 1.0

	lat_rad = yc / 180. * pi
	ex = 1/cos(lat_rad)

	dx = uu*scale*inner_scale * ex
	dy = vv*scale*inner_scale

	arrows( xc, yc, xc+dx, yc+dy, length=length, ... )
}

#==============================================================================================
# Note: can be irregular data with X,Y vals per point in x[i], y[i], OR
# can be gridded data with u and v on the X and Y grid locaitons.
# Plot vectors, vector fields
#
# Decimate is the substep to take. So 1 means no decimation, 2 means
# half the vectors, 3 means 1/3rd, etc.
#
# For a good illustration of how to use this, see:
#	/cir1/loca/version_30/card10_winds_crossval/plot_particular_days.R
# or
#	/home/pierce/projects/CEC_hourly_wind_vector_dscaling/daily_ds/MERRA2/try_4_2020-03-13/compare_vector_days.R
#
# To draw a single speed vector on a geopgraphical plot, use dp_arrows()
#
vecplot <- function( x, y, u, v, add=FALSE, xlim=NA, ylim=NA, scale=0.05, 
	main=NA, gmtborder=NA, col='black', headscale=2.0, decimate=1, 
	scale_x=NA, scale_y=NA, scale_boxcol=NA, scale_size_x=0.1, 
	scale_size_y=0.1, scale_units='',
	draw_vert_scale=FALSE, scale_box_llx_lly_urx_ury=NA, is_geographical=TRUE,
	drawscale=NA, lwd=par("lwd"), ... ) {

	if( ! add ) 
		globalframenumber <<- globalframenumber + 1

	tGridded <- 1
	tIrreg   <- 2

	if( length(u) == length(x) )
		type <- tIrreg
	else if( length(u) == length(x)*length(y))
		type <- tGridded
	else
		{
		print(paste("length x:", length(x) ))
		print(paste("length y:", length(y) ))
		print(paste("length u:", length(u) ))
		print(paste("length v:", length(v) ))
		stop("Can't figure out data, seems to be neither gridded nor irregular!")
		}

	if( type == tIrreg ) {
		np <- length(x)
		if( length(y) != np ) 
			stop(paste("Given irregular data, but lenght of X and Y are different!",
				'length x:', length(x), 
				'length y:', length(y),
				'dim  u:', pv(dim(u)),
				'dim  v:', pv(dim(u)) ))
		}

	fudgefact <- 0.05


	if( add ) {
		pp = par()
		xlim = pp$usr[1:2]
		ylim = pp$usr[3:4]
		}
	else
		{
		if( (length(xlim)==1) && (is.na(xlim))) {
			dxr  <- max(x) - min(x)
			xmin <- min(x) - dxr*fudgefact
			xmax <- max(x) + dxr*fudgefact
			xlim <- c(xmin,xmax)
			}
		if( (length(ylim)==1) && (is.na(ylim))) {
			dyr  <- max(y) - min(y)
			ymin <- min(y) - dyr*fudgefact
			ymax <- max(y) + dyr*fudgefact
			ylim <- c(ymin,ymax)
			}
		}

	#---------------------------------------------------------
	# Check the geometry of the plot and make sure it's square
	# We assume X and Y are in degrees
	# Note: aspect ratio is width/height
	#---------------------------------------------------------
	if( is_geographical ) check_plot_aspect_ratio( 'vecplot', xlim, ylim )

	if( ! add ) 
		mplot( c(xlim[0],xlim[1]), c(ylim[0],ylim[1]), type='p', pch='.', col='white', 
			incr_global_frameno=FALSE,
			xlim=xlim, ylim=ylim, main=main, gmtborder=gmtborder, ...)

	spd  <- sqrt(u*u + v*v)

	if( type == tIrreg ) {

		for( i in 1:np ) {
			x1 <- x[i] + u[i]*scale
			y1 <- y[i] + v[i]*scale
			len <- spd[i]*.1*scale*headscale
			#len <- 0.25
			lenOK <- (len>0.) && (!is.na(len))
			if( ((x[i]!=x1)||(y[i]!=y1)) && lenOK )
				arrows(x[i],y[i],x1,y1,length=len,col=col, lwd=lwd)
			}
		}

	else
		{
		nx <- length(x)
		ny <- length(y)
		for( j in seq(1,ny,by=decimate) )
		for( i in seq(1,nx,by=decimate) ) {
			x1 <- x[i] + u[i,j]*scale
			y1 <- y[j] + v[i,j]*scale
			len <- spd[i,j]*.1*scale*headscale
			#len <- 0.25
			lenOK <- (len>0.) && (!is.na(len))
			if( ((x[i]!=x1)||(y[j]!=y1)) && lenOK )
				arrows(x[i],y[j],x1,y1,length=len,col=col, lwd=lwd)
			}
		}

	#----------------------------------------------------------
	# Value of drawscale is length of vector to draw as a scale
	#----------------------------------------------------------
	if( !is.na(drawscale) ) {

		dx = max(xlim) - min(xlim)
		dy = max(ylim) - min(ylim)

		if( is.na(scale_x))
			scale_x = min(xlim)
		if( is.na(scale_y))
			scale_y = min(ylim)

		user_spec_box = (length( scale_box_llx_lly_urx_ury ) > 1 )
		if( user_spec_box ) {
			box_llx = scale_box_llx_lly_urx_ury[1]
			box_lly = scale_box_llx_lly_urx_ury[2]
			box_urx = scale_box_llx_lly_urx_ury[3]
			box_ury = scale_box_llx_lly_urx_ury[4]
			polygon( c(box_llx,box_llx,box_urx,box_urx),
				c(box_lly,box_ury,box_ury,box_lly),
				col='white', border='black' )
			}

		box_llx = scale_x
		box_lly = scale_y
		box_urx = box_llx + dx*scale_size_x
		box_ury = box_lly + dy*scale_size_y

		print(paste("drawing scale, box=", box_llx, box_lly, box_urx, box_ury ))

		if( (!user_spec_box) && (!is.na(scale_boxcol))) {
			polygon( c(box_llx,box_llx,box_urx,box_urx),
				c(box_lly,box_ury,box_ury,box_lly),
				col='white', border='white' )
print(paste("vecplot: drawing scale box ll, lly, urx, ury=", box_llx, box_lly, box_urx, box_ury ))
			}

		x0 = box_llx + dx*0.01
		y0 = box_lly + 0.5*(box_ury - box_lly)
		x1 <- x0 + drawscale*scale
		y1 <- y0 
		spd = drawscale
		len <- spd*.1*scale*headscale
		arrows(x0,y0,x1,y1,length=len,col='black', lwd=lwd)
		text( x1+0.1*(x1-x0), y1, paste( drawscale, scale_units), adj=c(0,0.5), col='blue' )

		if( draw_vert_scale ) {
			x1v = x0
			y1v = y0 + drawscale*scale
			arrows(x0,y0,x1v,y1v,length=len,col='red', lwd=lwd)
			}
		}
}

#================================================================================
# Takes 4 control points (each of which is a 3-element array, for X, Y, and Z
# locations) and computes the Bezier curve starting at p1 and ending at p4.
# The distance along the curve is given my mu, with 0<=mu<=1.  At mu=0, you are
# at p1; at mu=1., you are at p4.
#
bezier3d <- function( p1, p2, p3, p4, mu ) {

	mum1  <- 1 - mu
	mum13 <- mum1 * mum1 * mum1;
	mu3   <- mu * mu * mu;

	x <- 1
	y <- 2
	z <- 3

	p <- array(0.,3)

	p[x] <- mum13*p1[x] + 3*mu*mum1*mum1*p2[x] + 3*mu*mu*mum1*p3[x] + mu3*p4[x]
	p[y] <- mum13*p1[y] + 3*mu*mum1*mum1*p2[y] + 3*mu*mu*mum1*p3[y] + mu3*p4[y]
	p[z] <- mum13*p1[z] + 3*mu*mum1*mum1*p2[z] + 3*mu*mu*mum1*p3[z] + mu3*p4[z]

	return(p)
}

#================================================================================
cross3 <- function( u, v ) {

	x <- 1
	y <- 2
	z <- 3

	r <- array(0,3)
	r[x] <- u[y]*v[z] - u[z]*v[y]
	r[y] <- u[z]*v[x] - u[x]*v[z]
	r[z] <- u[x]*v[y] - u[y]*v[x]

	return(r)
}

#================================================================================
dot3 <- function( u, v ) {

	x <- 1
	y <- 2
	z <- 3

	r <- u[x]*v[x] + u[y]*v[y] + u[z]*v[z]

	return(r)
}

#================================================================================
# Given as inputs:
#	1) an abitrary point "p" (array of 3); and
#	2) a plane described by a point on that plane (planepoint[3]), and
#	the plane's normal at that point (normal[3]), this calculates a
#	point on the plane "b" such that the line connecting "b" to "p" is
# 	normal to the plane.
#
perpendicularpoint <- function( p, planepoint, normal ) {

	sn <- -dot3( normal, (p-planepoint) )
	sd <- dot3(normal, normal)
	sb <- sn/sd

	b <- p + sb*normal
	return(b)
}

#=================================================================================
# Draw a "raw" whisker given all the parameters. See also whiskerplot and
# whiskerplot_old, below
#
whisker_raw = function( plot_pos, width, min_dot, min_bar, min_box, mid, max_box, max_bar, max_dot, horiz=FALSE, 
	col='black', ... ) {

	if( horiz ) {
		# plot_pos is plotting Y value, width is dy
		llx = min_box
		urx = max_box
		lly = plot_pos - width/2
		ury = plot_pos + width/2

		# box
		polygon( c( llx,llx,urx,urx), c(lly,ury,ury,lly), border=col, ... )

		# middle bar
		lines( c(mid,mid), c(lly,ury), col=col, ... )

		# lines: left and then right
		lines( c(min_bar, min_box), c(plot_pos,plot_pos), col=col, ... )
		lines( c(max_box, max_bar), c(plot_pos,plot_pos), col=col, ... )

		# left then right bar
		lines( c(min_bar,min_bar), c(lly,ury), col=col, ... )
		lines( c(max_bar,max_bar), c(lly,ury), col=col, ... )

		# Dots
		points( min_dot, plot_pos, pch=19, col=col, ... )
		points( max_dot, plot_pos, pch=19, col=col, ... )
		}
	else
		{
		# plot_pos is X value, width is dx
		stop('case not done yet')
		}
}

#=================================================================================
whiskerplot <- function( xloc, data, boxwidth, pch=19, cex=1.0, col='black', col_dots=col, 
		crossbar=FALSE,
		show_dots=TRUE,
		use_mean=FALSE, lwd=1, lwd_box=lwd, 
		conf_interval=90,
		key_labels=NA,
		... ) {

	data = sort( data )

	nn = length( data )

	#-----------------------------------------------
	# Identify all points in the confidence interval
	#-----------------------------------------------
	pp = dat2percentiles( data )
	excl = 100 - conf_interval
	iblo = excl/2
	ibhi = 100 - excl/2
	inbox = (pp >= iblo) & (pp <= ibhi)

	if( use_mean )
		med <- mean(data, na.rm=TRUE)
	else
		med <- median(data, na.rm=TRUE)
	qq <- quantile( data, probs=c(.25,.75), na.rm=TRUE)

	# Draw the 'box'
	polygon( c(xloc-boxwidth/2,xloc+boxwidth/2,xloc+boxwidth/2,xloc-boxwidth/2),
		 c(qq[1],qq[1],qq[2],qq[2]),border=col, lwd=lwd_box, ...)

	#  Draw median line
	lines(c(xloc-boxwidth/2,xloc+boxwidth/2), c(med,med), lwd=lwd_box, col=col, ...)

	#---------------------------------------------------
	# Draw extra dots that are not in the box or whisker
	#---------------------------------------------------
	data_outbox = data[ ! inbox ]
	xx_outbox = array( xloc, length(data_outbox) )
	if( show_dots ) {
		points( xx_outbox, data_outbox, pch=pch, col=col_dots, cex=cex, ... )
		}

	# Draw the 90% CI 'whisker'
	idxlo = min( (1:nn)[inbox] )
	v0lo = data[idxlo]
	v1lo = qq[1]
	lines(c(xloc,xloc), c(v0lo,v1lo), col=col, lwd=lwd, ...)

	idxhi = max( (1:nn)[inbox] )
	v0hi = data[idxhi]
	v1hi = qq[2]
	lines(c(xloc,xloc), c(v0hi,v1hi), col=col, lwd=lwd, ...)

	if( crossbar ) {
		x0 = xloc - boxwidth*0.3
		x1 = xloc + boxwidth*0.3
		lines( c(x0,x1), c(v0lo,v0lo), col=col, lwd=lwd, ... )
		lines( c(x0,x1), c(v0hi,v0hi), col=col, lwd=lwd, ... )
		}
}


#=================================================================================
# If style==1 (default), draws a box with the median (or, if use_mean, the mean), 
# interquartile range, and min/max vals.  
# If style==2, draws a box with the median (or, if use_mean, the mean), interquartile
# range, 90 percentile range, and dots at the min/max values
# If style==3, draw box with median (or, if use_mean, the mean), interquartile
# range, and 95th percentile range
#
# Arg 'labels' is a list with the following elements:
#	$labels	: can be an array of strings, in which case each element of the
# 		box/whiskerplot is labeled accoringly. This is used to draw a key indicating
# 		what each component of the plot means. Example: 
#
# 		labels$labels = c( 'min', '5-ptile', '25-ptile', 'Median', '75-ptile', '95-ptile', 'max' )
#
# 		OR it can be the single string 'AUTO', in which case the datea VALUES are used as labels
#
#	$cex 	: label cex value
#	$col	: label col value
#	$format : C style format to use
#
whiskerplot_old <- function( xloc, data, boxwidth, pch=19, cex=1.0, col='black', 
		crossbar=FALSE,
		use_mean=FALSE, style=1, n_extrema_dots=1, lwd=1, lwd_box=2, 
		labels=NULL,
		... ) {

	if( use_mean )
		med <- mean(data, na.rm=TRUE)
	else
		med <- median(data, na.rm=TRUE)
	qq <- quantile( data, probs=c(.25,.75), na.rm=TRUE)

	val_min = NA
	val_max = NA
	val_whisk0 = NA
	val_whisk1 = NA
	val_box0 = NA
	val_box1 = NA
	val_center = NA

	if( style == 1 ) {
		# Draw the 'whisker'
		lines(c(xloc,xloc), c(qq[2],max(data,na.rm=TRUE)), col=col, lwd=lwd, ...)
		lines(c(xloc,xloc), c(qq[1],min(data,na.rm=TRUE)), col=col, lwd=lwd, ...)
		
		if( crossbar ) {
			x0 = xloc - boxwidth*0.3
			x1 = xloc + boxwidth*0.3
			lines( c(x0,x1), c(max(data),max(data,na.rm=TRUE)), col=col, lwd=lwd, ... )
			lines( c(x0,x1), c(min(data),min(data,na.rm=TRUE)), col=col, lwd=lwd, ... )
			}
		}
	else if( style == 2 ) {
		if( n_extrema_dots > 0 ) {
			# Draw the extrema dots
			ttdat = data
			for( idot in 1:n_extrema_dots ) {
				ixmax = which.max( ttdat )
				ixmin = which.min( ttdat )
				points( c(xloc,xloc), c(data[ixmax], data[ixmin]), col=col, pch=pch, cex=cex, ... )
				val_min =  data[ixmin]
				val_max =  data[ixmax]
				ttdat[ ixmax ] = NA
				ttdat[ ixmin ] = NA
				}
			}

		# Draw the 90% CI 'whisker'
		pp <- percentiles( data, c(5,95))
		lines(c(xloc,xloc), c(qq[2],pp[2]), col=col, lwd=lwd, ...)
		lines(c(xloc,xloc), c(qq[1],pp[1]), col=col, lwd=lwd, ...)

		val_whisk0 = pp[1]
		val_whisk1 = pp[2]

		val_box0 = qq[1]
		val_box1 = qq[2]

		if( crossbar ) {
			x0 = xloc - boxwidth*0.3
			x1 = xloc + boxwidth*0.3
			lines( c(x0,x1), c(pp[2],pp[2]), col=col, lwd=lwd, ... )
			lines( c(x0,x1), c(pp[1],pp[1]), col=col, lwd=lwd, ... )
			}
		}

	else if( style == 3 ) {
		# Draw the 95% CI 'whisker'
		pp <- percentiles( data, c(2.5,97.5))
		lines(c(xloc,xloc), c(qq[2],pp[2]), col=col, lwd=lwd, ...)
		lines(c(xloc,xloc), c(qq[1],pp[1]), col=col, lwd=lwd, ...)

		if( crossbar ) {
			x0 = xloc - boxwidth*0.3
			x1 = xloc + boxwidth*0.3
			lines( c(x0,x1), c(pp[2],pp[2]), col=col, lwd=lwd, ... )
			lines( c(x0,x1), c(pp[1],pp[1]), col=col, lwd=lwd, ... )
			}

		if( n_extrema_dots > 0 ) {
			# Draw the extrema dots
			ttdat = data
			for( idot in 1:n_extrema_dots ) {
				ixmax = which.max( ttdat )
				ixmin = which.min( ttdat )
				points( c(xloc,xloc), c(data[ixmax], data[ixmin]), col=col, pch=pch, cex=cex, ... )
				ttdat[ ixmax ] = NA
				ttdat[ ixmin ] = NA
				}
			}
		}
	else
		stop("bad style passed!")

	# Draw the 'box'
	polygon( c(xloc-boxwidth/2,xloc+boxwidth/2,xloc+boxwidth/2,xloc-boxwidth/2),
		 c(qq[1],qq[1],qq[2],qq[2]),border=col, lwd=lwd_box, ...)

	#  Draw median line
	lines(c(xloc-boxwidth/2,xloc+boxwidth/2), c(med,med), lwd=lwd_box, col=col, ...)

	val_center = med

	if( !is.null(labels) ) {

		# Check for required elements
		req_els = c( 'labels', 'cex', 'col' )
		nre = length( req_els )
		kk = names( labels )
		for( ire in 1:nre ) {
			if( ! (req_els[ire] %in% kk ))
				stop(paste("error, supplied labels list must have these elements: ", pv(req_els), 
					"MISSING REQUIRED ELEMENT:", req_els[ire] ))
			}

		n_req = 5	# whisker ends, IQR, mean/median
		if( n_extrema_dots > 0 ) {
			n_req = n_req + 2	# extrema
			}

		if( storage.mode( labels ) != 'list'  ) 
			stop( paste( 'Optional arg labels must be a list -- see code for doc' ))

		#----------------------------------------------
		# Generate auto-labels if not specified by user
		#----------------------------------------------
		if( length( labels$labels ) == 1 ) {
			if( (labels$labels[1] == 'AUTO' )) {
				if( is.null( labels$format ))
					stop(paste("error, supplied labels list must have element format!"))
				labels2use = array( '', n_req )
				if( n_extrema_dots > 0 ) {
					labels2use[1] = mformatC( labels$format, val_min )
					labels2use[n_req] = mformatC( labels$format, val_max )

					labels2use[2] = mformatC( labels$format, val_whisk0 )
					labels2use[n_req-1] = mformatC( labels$format, val_whisk1 )

					labels2use[3] = mformatC( labels$format, val_box0 )
					labels2use[n_req-2] = mformatC( labels$format, val_box1 )
					 
					labels2use[4] = mformatC( labels$format, val_center )
					}
				}
			else
				stop(paste("sorry, specified arg labels$labels with one element that is NOT AUTO -- see doc in code"))
			}
		else
			labels2use = labels$labels

		if( length(labels2use) != n_req ) 
			stop(paste("Error, if labels2use is specified, it must be a list OR an array of character strings with length", n_req, ' but got length', length(labels2use) ))
		adj = list()
		for( ilab in 1:n_req )
			adj[[ilab]] = c(1.0,0.5)

		if( n_extrema_dots > 0 ) {
			ylab = c( min(data), pp[1], qq[1], med, qq[2], pp[2], max(data) )
			adj[[3]] = c(1.0,0.2)
			adj[[5]] = c(1.0,0.8)
			}
		else
			{
			ylab = c( pp[2], qq[2], med, qq[1], pp[1]            )
			adj[[2]] = c(1.0,0.2)
			adj[[4]] = c(1.0,0.8)
			}

		x0 = xloc - boxwidth/2
		ddx = boxwidth * 0.10
		x0 = x0 - ddx 

		for( ilab in 1:n_req ) {
			print(paste('drawing key label:', x0, ylab[ilab], labels2use[ilab] ))
			text( x0, ylab[ilab], labels2use[ilab], adj=adj[[ilab]], col=labels$col, cex=labels$cex )
			}
		}
}

#=============================================================================
# This is made to shade a particular region on a graph.  Typically
# you would use it to shade the +/- 1-sigma region about a central
# curve.  To use it, you just give the upper and lower lines;
# shading will be applied between them.
#
shadeRegion <- function( xvals, yvalsUp, yvalsLo, ... ) {

	x <- c(xvals, rev(xvals))
	y <- c(yvalsUp,rev(yvalsLo))
	polygon(x,y,...)
}

#=======================================================================
# The idea here is you are comparing a bunch of model results to one
# observed result.  The observed standard deviation goes in 'signal_stddev'.
# The correlation between each model and the obs is given by 'correlations'.
# The standard devs of each model run are given in 'stddevs'.
#
taylor_diagram <- function( signal_stddev, correlations, stddevs, main='', 
	pch=19, pch_col='black', pch_lwd=1, pch_bg=pch_col, pch_cex=1,
	pch_signal=19, theta_max_deg=90, nseg=100, col_signal='red', 
	col_signal_bg=col_signal, cex_signal=1.5,
	xmax=NA, corr_labels=c(seq(0.,.9,by=.1),.95,.99), add=FALSE,
	point_label_vals=NA, point_label_adj=NA, point_label_cex=1.0,
	connect_with_lines=FALSE,
	draw_corr_val=NA, draw_corr_lty=1, draw_corr_col='black',
	shade_corr_below=NA, shade_corr_col=rgb(.8,.8,.8),
	xlab='Standard Deviation', draw_stddev_arc=TRUE) {

	if( ! add ) 
		globalframenumber <<- globalframenumber + 1

	theta_max <- theta_max_deg/180.*pi

	if( ! add ) {
		sdmax <- max(c(signal_stddev, stddevs ))
		if( is.na(xmax))
			xmax <- sdmax

		if( theta_max_deg >= 90 ) {
			xmin <- xmax*cos(theta_max)
			xlab2use <- ''
			ylab <- ''
			}
		else
			{
			xmin <- 0.
			xlab2use <- xlab
			ylab <- xlab
			}

		#------------------------------
		# A few constants we need below
		#------------------------------
		ticklen <- xmax*0.02
		r_arc   <- xmax*1.05  # Radius of drawn 'correlation arc'
		yext <- xmax*sin(theta_max)  # Y extent of 'Y' axis

		plot( c(0,1), c(0,1), type='p', col='white',
			xlim=c(xmin,xmax), ylim=c(0,xmax),
			xlab=xlab2use, grid=FALSE, ylab=ylab,
			xaxs='i', yaxs='i', bty='l', axes=FALSE,
			main=main )


		if( !is.na(shade_corr_below)) {
			#-------------------------------------------------
			# Shade all correlations below the requested value
			#-------------------------------------------------
			px <- c(0,xmin)
			py <- c(0,yext) 
			#-------------------------------------------------
			# Shaded region that is bounded by correlation arc
			#-------------------------------------------------
			theta_0 <- theta_max_deg/180.*pi
			theta_1 <- acos( shade_corr_below )
			dt <- (theta_0-theta_1)/nseg
			for( i in 1:nseg ) {
				x0 <- r_arc*cos(theta_0 - dt*i)
				y0 <- r_arc*sin(theta_0 - dt*i)
				px <- c(px,x0)
				py <- c(py,y0)
				}
			px <- c(px,0)
			py <- c(py,0)
			polygon(px,py,col=shade_corr_col,xpd=NA,border=shade_corr_col)
			}

		if( theta_max_deg <= 90 ) {
			axis(1)
			axis(2)
			}
		else
			{
			#-----------------------
			# Put on positive X axis
			#-----------------------
			lines(c(0,xmax),c(0,0))
			xticks <- pretty(c(0,xmax))
			nt <- length(xticks)
			for( i in 1:nt ) {
				x <- xticks[i]
				text(x, 0., x, adj=c(0.5,1.5), xpd=NA)
				lines(c(x,x), c(0.,ticklen)) 
				}
			#----------------------
			# Positive X axis label
			#----------------------
			cx <- xmax/2.
			text(cx,0.,xlab,adj=c(0.5,3.0),xpd=NA)

			#----------------------------------------------------
			# Put on 'Y' axis (could almost be a negative X axis)
			#----------------------------------------------------
			#------------
			# Main Y axis
			#------------
			lines( c(0,xmin), c(0,yext), xpd=NA ) 
			#------------------
			# Y axis tic labels
			#------------------
			theta_perp <- theta_max + pi
			par(srt=theta_perp*180/pi)
			for( i in 1:nt ) {
				x <- xticks[i]
				if( x != 0 ) {
					xp <- x*cos(theta_max)
					yp <- x*sin(theta_max)
					text(xp, yp, x, adj=c(0.5,1.5), xpd=NA)
					# Y axis ticks
					xpp <- ticklen*cos(theta_max-pi/2)
					ypp <- ticklen*sin(theta_max-pi/2)
					lines(c(xp,xp+xpp), c(yp,yp+ypp))
					}
				}
			par(srt=0.)
			#----------------------
			# Positive Y axis label
			#----------------------
			cx <- xmin/2.
			cy <- yext/2.
			par(srt=(theta_max+pi)/pi*180)
			text(cx,cy,xlab,adj=c(0.5,3.0),xpd=NA)
			par(srt=0.)
			}

		#----------------------------------
		# Put on correlation arc & label it
		#----------------------------------
		guard_deg <- 2	# guardband 
		theta_0_deg <- guard_deg
		theta_1_deg <- theta_max_deg - guard_deg
		dt_deg <- (theta_1_deg-theta_0_deg)/nseg
		dt <- dt_deg/180.*pi
		theta_0 <- theta_0_deg/180*pi
		#----------------
		# Correlation Arc
		#----------------
		for( i in 1:nseg ) {
			x0 <- r_arc*cos(dt*(i-1) + theta_0)
			x1 <- r_arc*cos(dt*i + theta_0)
			y0 <- r_arc*sin(dt*(i-1) + theta_0)
			y1 <- r_arc*sin(dt*i + theta_0)
			lines(c(x0,x1), c(y0,y1), xpd=NA)
			}
		#------------------------------
		# Correlation Arc tics & labels
		#------------------------------
		ncl <- length(corr_labels)
		for( icl in 1:ncl ) {
			theta <- acos( corr_labels[icl] )
			if( theta <= theta_max ) {
				xt <- r_arc*cos(theta)
				yt <- r_arc*sin(theta)
				if( corr_labels[icl] >= 0 ) {
					srt <- theta/pi*180.
					xadj <- -0.09
					}
				else
					{
					srt <- (theta+pi)/pi*180.
					xadj <- 1.09
					}
				par(srt=srt)
				text(xt,yt,corr_labels[icl],xpd=NA,adj=c(xadj,0.5))

				# Ticks
				xp <- ticklen*cos(theta+0.*pi/2.)
				yp <- ticklen*sin(theta+0.*pi/2.)
				lines(c(xt,xt-xp),c(yt,yt-yp),xpd=NA)
				}
			}
		#----------------------
		# Correlation Arc title
		#----------------------
		theta <- pi/4.
		xt <- r_arc*cos(theta)
		yt <- r_arc*sin(theta)
		srt <- theta/pi*180. - 90.
		par(srt=srt)
		text(xt,yt,'Correlation',xpd=NA,adj=c(0.5,-3.0))
		par(srt=0)

		#-------------------------------------------
		# Add constant correlation line if requested
		#-------------------------------------------
		if( (length(draw_corr_val)>1) || ( (length(draw_corr_val)==1) && (!is.na(draw_corr_val[1])))) {
			ncv <- length(draw_corr_val)
			for( idc in 1:ncv ) {
				theta <- acos(draw_corr_val[idc])
				x1 <- r_arc*cos(theta)
				y1 <- r_arc*sin(theta)
				lines( c(0,x1), c(0,y1), col=draw_corr_col, lty=draw_corr_lty, xpd=NA )
				}
			}

		#------------------
		# Put on stddev arc 
		#------------------
		if( draw_stddev_arc ) {
			guard_deg <- 0	# guardband 
			theta_0_deg <- guard_deg
			theta_1_deg <- theta_max_deg - guard_deg
			r_arc <- signal_stddev
			dt_deg <- (theta_1_deg-theta_0_deg)/nseg
			dt <- dt_deg/180.*pi
			theta_0 <- theta_0_deg/180*pi
			for( i in 1:nseg ) {
				x0 <- r_arc*cos(dt*(i-1) + theta_0)
				x1 <- r_arc*cos(dt*i + theta_0)
				y0 <- r_arc*sin(dt*(i-1) + theta_0)
				y1 <- r_arc*sin(dt*i + theta_0)
				lines(c(x0,x1), c(y0,y1), xpd=NA, lty=3)
				}
			}

		#------------------------------
		# Add on the one 'signal' point
		#------------------------------
		points( signal_stddev, 0., pch=pch_signal, cex=cex_signal, 
			col=col_signal, bg=col_signal_bg, xpd=NA, lwd=pch_lwd )
		}

	#-----------------------------------------------
	# Calculate X and Y locations of the MAIN points
	#-----------------------------------------------
	theta <- acos(correlations)
	r     <- stddevs
	xvals <- r*cos(theta)
	yvals <- r*sin(theta)

	#---------------------------------------
	# Connect points with lines if requested
	#---------------------------------------
	if( connect_with_lines ) {
		npl <- length(xvals)
		if( npl > 1 ) {
			for( i in 2:npl ) {
				lines( c(xvals[i-1],xvals[i]), c(yvals[i-1],yvals[i]))
				}
			}
		}

	#---------------------
	# Draw the MAIN points
	#---------------------
	points( xvals, yvals, pch=pch, col=pch_col, bg=pch_bg, lwd=pch_lwd, cex=pch_cex )

	#--------------------------------------
	# Add labels to the points if requested
	#--------------------------------------
	if( (length(point_label_vals)>1) || ( (length(point_label_vals)==1) && (!is.na(point_label_vals[1])))) {
		npl <- length(point_label_vals)
		if( npl != length(xvals))
			stop("need to have same number of point LABELS as you have POINTS!")
		if( length(point_label_adj) == 1 )
			point_label_adj <- array(-0.2,dim=c(npl,2))
		for( ip in 1:npl ) {
			text(xvals[ip], yvals[ip], point_label_vals[ip], adj=point_label_adj[ip,], cex=point_label_cex)
			}
		}
}

#=========================================================================================
# This makes a blank 'canvas' to later draw points and lines on.
#
canvas = function( xlim=NA, ylim=NA, type='b', pch='.', bg=NA, box=TRUE, xlab='', is_geographical=FALSE, ... ) {

	globalframenumber <<- globalframenumber + 1
	
	if( is.na(xlim) && is.na(ylim)) {
		#---------------------------------------------------
		# A TRULY blank canvas, for skipping a plotting spot
		#---------------------------------------------------
		mplot( c(100,101), c(100,101), xlim=c(0,1), ylim=c(0,1), pch='.', col='white', axes=FALSE, box=FALSE, grid=FALSE, xlab='' )
		return()
		}

	mplot( c(-10000.1,-10000), c(-10000.1,-10000), type=type, xlim=xlim, ylim=ylim, box=box, xlab=xlab, ... )

	if( ! is.na(bg)) {
		usr = par()$usr   # vector of the extrems of the user coordinates of the PLOTTING region
		x0 = usr[1]
		x1 = usr[2]
		y0 = usr[3]
		y1 = usr[4]
		polygon( c(x0,x1,x1,x0), c(y0,y0,y1,y1), col=bg, border=bg )
		}

	#---------------------------------------------------------
	# Check the geometry of the plot and make sure it's square
	# We assume X and Y are in degrees
	# Note: aspect ratio is width/height
	#---------------------------------------------------------
	if( is_geographical ) check_plot_aspect_ratio( 'canvas', xlim, ylim )
}

#====================================================================================================
check_plot_aspect_ratio = function( ftn_name, xlim, ylim ) {
	
	#--------------------------------------------------------
	# Try to skip this if we are NOT doing a geophysical plot
	#--------------------------------------------------------
	if( (xlim[2] > 360) || (xlim[1] < -180) || (ylim[2] > 90) || (ylim[1] < -90))
		return()

	if( (xlim[1] == 0) && (xlim[2] == 0))
		return()

	xr <- xlim[2] - xlim[1]
	yr <- ylim[2] - ylim[1]
	km_per_degree = 110.
	midy_deg = mean( ylim )
	midy_rad = midy_deg/180*pi
	dx_km = xr * cos( midy_rad ) * km_per_degree
	dy_km = yr * km_per_degree
	fin = par()$fin
	plt = par()$plt
	dx_plot = fin[1] * (plt[2] - plt[1])
	dy_plot = fin[2] * (plt[4] - plt[3])
	aspect_dat  = dx_km/dy_km
	aspect_plot = dx_plot/dy_plot

	aerr = abs( aspect_dat - aspect_plot)
	aerr_pct = aerr / aspect_plot * 100.
	err_too_big = aerr_pct > 1.0

	if( err_too_big ) {
		if(aspect_dat > aspect_plot) 
			tag = "make plot wider (or shorter)"
		else if( aspect_dat < aspect_plot )
			tag = "make the plot thinner (or taller)"
		}
	else
		tag = paste( "plot OK, error in pct:", aerr_pct, '%' )

	#print(paste(ftn_name, ':', "Aspect ratio data:", aspect_dat, 'fig:', aspect_plot, tag ))
}

#=========================================================================================
# This plots lines with overplotted symbols, BUT the overplotted symbols
# are not plotted at every point (as is standard with R).  Instead they
# are plotted fractionally along the length of the (windy) line segments.
# They are plotted 'subsample' distance apart with an offset of 'offset',
# where both those values are normalized to 1, representing the length
# of the total line segment.  So, for example, you might have subsample=.2
# and offset=.1 to have 5 points plotted along the line.
#
# Note: pch can be set to a string, such as "Test 1".
#
lines_subsample_pts <- function( x, y, subsample, offset=0, whiteout=TRUE, col='black', lwd=1, pch=NA, 
			lwd_hershey=1,
			hershey_whiteout_factor=7,
			hershey_hover_inch=NA,
			hershey_str_len_fudge=1,
			hershey_angle_fudge=0.,
			str_height_inch=0.1, 
			debug=FALSE,
			... ) {

	lines( x, y, ..., col=col, lwd=lwd, type='l' )

	nsegs <- length(x)-1
	del   <- 1/nsegs
	seg_frac_start <- (1:nsegs-1)*del
	seg_frac_end   <- (1:nsegs)*del

	#------------------------------------------------------------------
	# Convert x and y to absolute inches; needed for angle calculations
	#------------------------------------------------------------------
	pin <- par('pin')
	usr <- par('usr')
	nn <- length(x)
	x_inch <- (x-usr[1])/(usr[2]-usr[1])*pin[1]
	y_inch <- (y-usr[3])/(usr[4]-usr[3])*pin[2]

	npts2plot <- floor( 1/subsample + .001 )

	par(family='Helvetica')
	#par(family='HersheySans')
	pin <- par('pin')
	plot_width_in  <- pin[1]
	plot_height_in <- pin[2]
	usr <- par('usr')
	dx_user <- usr[2] - usr[1]
	dy_user <- usr[4] - usr[3]

	for( i in 1:npts2plot ) {
		frac <- (i-1)*subsample + offset
		if( (frac>=0) && (frac <= 1)) {
			#--------------------------------------
			# Find what segment this fraction is on
			#--------------------------------------
			myseg <- -1
			for( iseg in 1:nsegs ) {
				if( (frac >= seg_frac_start[iseg]) && (frac <= seg_frac_end[iseg])) {
					myseg <- iseg
					break
					}
				}
			if( myseg == -1 ) 
				stop("coding error, did not find seg for this point!")
			frac_along_seg <- (frac-seg_frac_start[myseg])/(seg_frac_end[myseg]-seg_frac_start[myseg])
			px0 <- x[myseg]
			py0 <- y[myseg]
			px1 <- x[myseg+1]
			py1 <- y[myseg+1]
			px  <- px0 + frac_along_seg*(px1-px0)	# 'px' is the location of the symbol we will be plotting, or center of char string
			py  <- py0 + frac_along_seg*(py1-py0)	# 'py' is the location of the symbol we will be plotting, or center of char string
			if( storage.mode(pch) == 'character' ) {
				if( debug ) print(paste("lines_subsample_pts: Working on char label:", pch))
				col_time_thru <- c('white', col)
				lwd_time_thru <- c(lwd_hershey*hershey_whiteout_factor,lwd_hershey)
				for( itimethru in 1:length(col_time_thru)) {

					fontinfo <- dphershey_fontinfo(1)

					sw_user <- dphershey_strwidth(pch, str_height_inch, fontinfo)  # Width of string in USER units
					sw_inch <- dphershey_strwidth(pch, str_height_inch, fontinfo, units='inches')  # Width of string in INCHES

					#----------------------------------------------------
					# If we are "hovering" over the line, then move it in
					# the indicated direction before calculating the 
					# requested line segment
					#----------------------------------------------------
					if( debug ) print(paste("lines_subsample_pts: calculating hover: ", hershey_hover_inch))
					px_inch <- (px-usr[1])/dx_user*plot_width_in	# SCALAR X loc, in inches, of center of the string to be plotted
					py_inch <- (py-usr[3])/dy_user*plot_height_in	# SCALAR Y loc, in inches, of center of the string to be plotted
					if( is.na(hershey_hover_inch)) {
						x_inch_2use <- x_inch		# VECTOR: 'x_inch' is the requested X values to plot in inches instead of user units
						y_inch_2use <- y_inch		# VECTOR: 'y_inch' is the requested Y values to plot in inches instead of user units
						px_inch_2use <- px_inch		
						py_inch_2use <- py_inch
						}
					else
						{
						#-----------------------
						# Move line as requested
						#-----------------------
						x_inch_2use <- array(0., nn)
						y_inch_2use <- array(0., nn)
						for( ii  in 1:nn ) {
							idx_l <- ii-1
							if(ii==1) idx_l <- 1
							idx_r <- ii+1
							if(ii==nn) idx_r <- nn
							#------------------
							# Get angle of line
							#------------------
							ddx <- x_inch[idx_r] - x_inch[idx_l]
							ddy <- y_inch[idx_r] - y_inch[idx_l]
							theta <- atan2( ddy, ddx )
							#-----------------------------------------------------
							# Project location along line perpindicular to tangent
							#-----------------------------------------------------
							projang <- theta+pi/2
							x_inch_2use[ii] <- x_inch[ii] + hershey_hover_inch*cos(projang)
							y_inch_2use[ii] <- y_inch[ii] + hershey_hover_inch*sin(projang)
							}
						#-------------------------------
						# Move center point as requested
						#-------------------------------
						ddx <- px1 - px0
						ddy <- py1 - py0
						theta <- atan2( ddy, ddx )
						#-----------------------------------------------------
						# Project location along line perpindicular to tangent
						#-----------------------------------------------------
						projang <- theta+pi/2
						px_inch_2use <- px_inch + hershey_hover_inch*sin(projang)
						py_inch_2use <- py_inch + hershey_hover_inch*cos(projang)

						if(FALSE) {
							#------------------
							# Draw 'hover' line
							#------------------
							x_user_2use <- x_inch_2use/plot_width_in*dx_user + usr[1]
							y_user_2use <- y_inch_2use/plot_height_in*dy_user + usr[3]
							lines( x_user_2use, y_user_2use, col='purple' )
							}
						}

					#--------------------------------------------------------
					# Find points along the curve, centered at (px,py), where
					# the length along this line segment is the requested
					# length.
					#------------------------------------------------------
					if( debug ) print(paste("lines_subsample_pts: getting base curve"))
					requested_length <- sw_inch*hershey_str_len_fudge
					if( debug ) print(paste("width of string to draw, in inches: ", sw_inch ))
					rv <- find_disk_entry_exit( x_inch_2use, y_inch_2use, px_inch_2use, py_inch_2use, requested_length )
					curve_x_inch <- rv$pts_along_curve_x
					curve_y_inch <- rv$pts_along_curve_y

					if(debug) {
						ncurve <- length(curve_x_inch)
						for( kk in 1:ncurve ) {
							xx_user <- curve_x_inch[kk]/plot_width_in*dx_user + usr[1]
							yy_user <- curve_y_inch[kk]/plot_height_in*dy_user + usr[3]
							points( xx_user, yy_user, pch=19, col='orange', cex=.2 )
							}
						entry_x_user = rv$entry[1]/plot_width_in*dx_user + usr[1]
						entry_y_user = rv$entry[2]/plot_height_in*dy_user + usr[3]
						exit_x_user  = rv$exit[1]/plot_width_in*dx_user + usr[1]
						exit_y_user  = rv$exit[2]/plot_height_in*dy_user + usr[3]
						points( entry_x_user, entry_y_user, pch=11, cex=2.0, col='red' )
						points( exit_x_user, exit_y_user,  pch=11, cex=2.0, col='darkgreen' )
						print(paste("Curve along which to draw string: (", rv$entry[1],",",rv$entry[2],") to (",
							rv$exit[1],",",rv$exit[2],")" ))
						dxi = rv$entry[1] - rv$exit[1]
						dyi = rv$entry[2] - rv$exit[2]
						print(paste("Length of curve in inches: ", sqrt(dxi*dxi + dyi*dyi)))
						print('curve X, Y values in inches:')
						print(curve_x_inch)
						print(curve_y_inch)
						}

					#----------------------------------------------------
					# Now go along our text string, and place each letter
					# at the proper spot and with the proper rotation
					#----------------------------------------------------
					z0 <- array(0., length(curve_x_inch))
					nc <- nchar( pch )
					#-----------------------------------------
					# Get location of BEGINNING each character
					#-----------------------------------------
					charloc_x_inch <- array(curve_x_inch[1], nc)	# This sets the beginning for character #1, do rest of chars in loop below
					charloc_y_inch <- array(curve_y_inch[1], nc)
					if( debug ) print(paste("lines_subsample_pts: getting loc of beginning of chars"))
					if( nc > 1 ) {
						for( ic  in 2:nc ) {
							#-----------------------------------------------------------------
							# For the BEGINNING of the character, we need the substring up to,
							# but not including, the current character
							#-----------------------------------------------------------------
							if(debug) print(paste("Calculating beginning of char #",ic,"(",substr(pch,ic,1),")"))
							thissubstr = substr(pch,1,ic-1)
							sw_this_ss = dphershey_strwidth( thissubstr, str_height_inch, fontinfo, units='inches'  )
							swfrac     = sw_this_ss / sw_inch
							if( debug )
								print(paste("Stringwidth of substring ",thissubstr,
									", which is UP TO BUT NOT INCLUDING our char, in inches, is: ", 
									sw_this_ss, " stringwidth whole string=", sw_inch,
									" stringwidth fraction for this substring is ", swfrac))
							t <- terpolate3line( curve_x_inch, curve_y_inch, z0, swfrac )
							charloc_x_inch[ic] <- t[,1]
							charloc_y_inch[ic] <- t[,2]
							if( debug ) {	
								print(paste(swfrac,"along the following X,Y vaules is X=", charloc_x_inch[ic], "Y=", charloc_y_inch[ic] ))
								print(curve_x_inch)
								print(curve_y_inch)
								}	
							}
						}
					if( debug ) {
						for( ic in 1:nc )
							print(paste("lines_subsample_pts: beginning of char", ic, "is at:", charloc_x_inch[ic], charloc_y_inch[ic], "(inches)"))
						}
					#-----------------------------------------
					# Get location of ENDING of each character
					#-----------------------------------------
					ncurve <- length(curve_x_inch)
					if( nc > 1 ) {
						if( debug ) print(paste("lines_subsample_pts: getting loc of ending of char", ic))
						charloc_x_inch_end <- c(charloc_x_inch[2:nc], curve_x_inch[ncurve] )
						charloc_y_inch_end <- c(charloc_y_inch[2:nc], curve_y_inch[ncurve] )
						}
					else
						{
						charloc_x_inch_end <- curve_x_inch[ncurve]
						charloc_y_inch_end <- curve_y_inch[ncurve]
						}
					if( debug ) {
						for( ic in 1:nc ) 
							print(paste("lines_subsample_pts: ending of char", ic, "is at:", charloc_x_inch_end[ic], charloc_y_inch_end[ic], "(inches)"))
						}

					#-------------------------------------------
					# Now actually print the chars on the output
					#-------------------------------------------
					for( ic  in 1:nc ) {
						#-------------
						# Char to plot
						#-------------
						cc <- substr(pch,ic,ic)
						if( debug ) print(paste("lines_subsample_pts: emitting char number", ic, "which is: ", cc))
						#-----------------------
						# Get angle of character
						#-----------------------
						ddx <- charloc_x_inch_end[ic]-charloc_x_inch[ic] 
						ddy <- charloc_y_inch_end[ic]-charloc_y_inch[ic]
						theta <- atan2( ddy, ddx )
						theta_deg <- theta/pi*180 + hershey_angle_fudge
						if( debug ) print(paste("lines_subsample_pts: angle (degrees) of char number", ic, "is: ", theta_deg))
						#----------------------
						# Convert to user units
						#----------------------
						charloc_x_user <- charloc_x_inch[ic]/plot_width_in*dx_user + usr[1]
						charloc_y_user <- charloc_y_inch[ic]/plot_height_in*dy_user + usr[3]

						if( debug ) {
							points( charloc_x_user, charloc_y_user, pch=4, col='darkgreen', cex=0.2 )
							xx_user <- charloc_x_inch_end[ic]/plot_width_in*dx_user + usr[1]
							yy_user <- charloc_y_inch_end[ic]/plot_height_in*dy_user + usr[3]
							points( xx_user, yy_user, pch=4, col='orange', cex=0.2 )
							}

						dphershey_emit_string( charloc_x_user, charloc_y_user, cc,
							str_height_inch, fontinfo, srt=theta_deg, col=col_time_thru[itimethru],
							lwd=lwd_time_thru[itimethru])
						}
					}

				}
			else
				{
				if( whiteout ) 
					points( px, py, pch=pch, ..., type='p', col='white', lwd=2*lwd )
				points( px, py, pch=pch, ..., type='p', col=col, lwd=lwd )
				}
			}
		}
}

#=========================================================================================
# Given some set of X,Y points that draw a line, this supersamples the vaules and
# returns a high-resolution (but not smoothed or interpolated!) version of the line
#
supersample_line = function( xorig, yorig, nsamp ) {

	nn = length(xorig)
	if( nn < 3 )
		stop(paste("error, must be supplied with at least 3 points!"))
	xnew = array(xorig[nn], nn*nsamp + 1)	# This sets the LAST point in the 'xnew', 'ynew' arrays to the last point in the xorig,yorig arrays
	ynew = array(yorig[nn], nn*nsamp + 1)
	for( i in 1:(nn-1) ) {
		for( ss in 1:nsamp ) {
			idx = (i-1)*nsamp + ss
			dx = xorig[i+1] - xorig[i]
			dy = yorig[i+1] - yorig[i]
			xnew[idx] = xorig[i] + dx*(ss-1)/nsamp
			ynew[idx] = yorig[i] + dy*(ss-1)/nsamp
			}
		}

	return( list( x=xnew, y=ynew ))
}

#=========================================================================================
# Given xx and yy values that define a curve, find the point where the curve first
# enters (and then exits) a disk of radius "rad" around center point (px,py)
#
find_disk_entry_exit <- function( xx_orig, yy_orig, px, py, rad, fine_res_npts=1000, debug=FALSE ) {

	rv = supersample_line( xx_orig, yy_orig, 50 )
	xx = rv$x
	yy = rv$y

	if( debug ) {
		print("++++++++++++++++++++++++++++++++++++++++++++++++++")
		print(paste("find_disk_entry_exit: entering with center point=", px, py, "and radius=", rad ))
		print("here are X values on entry:")
		print(xx)
		print("here are Y values on entry:")
		print(yy)
		}

	retval <- array(NA, 4)	# format: [1] is entry_x, [2] is entry_y, [3] is exit_x, [4] is exit_y
				# 'NA' means that the curve does not enter or exit at that point

	halfrad <- rad/2

	np <- length(xx)
	if( length(yy) != np )
		stop(paste("error, length xx does not match length of yy! xx=", np, 'yy=', length(yy)))

	INDISK  <- 1
	OUTDISK <- 2

	idx_px <- wherenearest( px, xx )

	nsubsample <- 100

	pts_along_curve_x_left <- NA

	#------------------------------------------------------------
	# Get entry point ... start at disk center, move "left" along
	# the curve until we have gotten 'halfrad' away from pt
	#------------------------------------------------------------
	if(debug) print(paste("looking for entry point, index of point closest to center is", 
		idx_px," which is pt",xx[idx_px],yy[idx_px]))
	if( idx_px != 1 ) {
		dist_so_far <- 0.
		if( debug ) print(paste("looking for entry point, counting down from ",idx_px,"to 2"))
		for( ip in idx_px:2 ) {
			dx <- xx[ip-1] - xx[ip]
			dy <- yy[ip-1] - yy[ip]
			dd <- sqrt( dx*dx + dy*dy )
			dist_so_far <- dist_so_far + dd
			if( dist_so_far > halfrad ) {
				#-----------------------------------
				# Point to "left" is outside of disk
				#-----------------------------------
				if(debug) print(paste("point to left is outside of disk"))
				xsub <- (0:nsubsample)/nsubsample*(xx[ip]-xx[ip-1]) + xx[ip-1]
				ysub <- (0:nsubsample)/nsubsample*(yy[ip]-yy[ip-1]) + yy[ip-1]
				distinc <- sqrt( (xsub-xx[ip])**2 + (ysub-yy[ip])**2 )
				dd2 <- (dist_so_far - dd) + distinc
				idx <- wherenearest( halfrad, dd2 )
				retval[1] <- xsub[idx]
				retval[2] <- ysub[idx]
				if(debug) print(paste("Entry point to disk is", retval[1], retval[2]))
				pts_along_curve_x_left <- c(retval[1], xx[ip:idx_px] )  # Note: include center pt
				pts_along_curve_y_left <- c(retval[2], yy[ip:idx_px] )  # Note: include center pt
				if(debug) {
					print("points along curve to the left, X:")
					print(pts_along_curve_x_left)
					print("points along curve to the left, Y:")
					print(pts_along_curve_y_left)
					}
				break
				}
			}
		}
	#----------------------------------------------------
	# We exited the disk prematurely, BUT the disk is not 
	# actually at the end of the curve
	#----------------------------------------------------
	if( (idx_px != 1) && is.na(retval[1])) {
		if(debug) print("We exited the disk prematurely, BUT the disk is not actually at the end of the curve")
		retval[1] <- xx[1]
		retval[2] <- yy[2]
		pts_along_curve_x_left <- xx[1:idx_px]
		pts_along_curve_y_left <- yy[1:idx_px]
		}

	pts_along_curve_x_right <- NA

	#------------------------------------------------------------
	# Get exit point ... start at disk center, move "right" along
	# the curve until we exit the disk
	#------------------------------------------------------------
	if(debug) print('Getting exit point')
	if( idx_px != np ) {
		dist_so_far <- 0.
		if(debug) print(paste('Getting exit point: counding up from',idx_px,"to",np-1))
		for( ip in idx_px:(np-1) ) {
			dx <- xx[ip+1] - xx[ip]
			dy <- yy[ip+1] - yy[ip]
			dd <- sqrt( dx*dx + dy*dy )
			dist_so_far <- dist_so_far + dd
			
			if( dist_so_far > halfrad ) {
				#------------------------------------
				# Point to "right" is outside of disk
				#------------------------------------
				if(debug) print(paste("point to the right is outside of disk"))
				xsub <- (0:nsubsample)/nsubsample*(xx[ip+1]-xx[ip]) + xx[ip]
				ysub <- (0:nsubsample)/nsubsample*(yy[ip+1]-yy[ip]) + yy[ip]
				distinc <- sqrt( (xsub-xx[ip])**2 + (ysub-yy[ip])**2 )
				dd2 <- (dist_so_far - dd) + distinc
				idx <- wherenearest( halfrad, dd2 )
				retval[3] <- xsub[idx]
				retval[4] <- ysub[idx]
				if(debug) print(paste("Exit point of disk is", retval[3], retval[4]))
				pts_along_curve_x_right <- c(xx[(idx_px+1):ip],retval[3])  
				pts_along_curve_y_right <- c(yy[(idx_px+1):ip],retval[4])
				if(debug) {
					print("points along curve to the right, X:")
					print(pts_along_curve_x_right)
					print("points along curve to the right, Y:")
					print(pts_along_curve_y_right)
					}
				break;
				}
			}
		}
	if( (idx_px != np) && is.na(retval[3])) {
		retval[3] <- xx[np]
		retval[4] <- yy[np]
		pts_along_curve_x_right <- xx[(idx_px+1):np]
		pts_along_curve_y_right <- yy[(idx_px+1):np]
		}

	#---------------------------------------------------------
	# If we have an entry point but no exit point, make a fake
	# exit point that just extends the slope in that direction
	#---------------------------------------------------------
	if( !is.na(retval[1]) && is.na(retval[3])) {
		retval[3] <- px + (px-retval[1])
		retval[4] <- py + (py-retval[2])
		xx_pad <- c(xx,retval[3])
		yy_pad <- c(yy,retval[4])
		pts_along_curve_x_right <- retval[3]
		pts_along_curve_y_right <- retval[4]
		}
	if( !is.na(retval[3]) && is.na(retval[1])) {
		retval[1] <- px + (px-retval[3])
		retval[2] <- py + (py-retval[4])
		xx_pad <- c(retval[1],xx)
		yy_pad <- c(retval[2],yy)
		pts_along_curve_x_left <- retval[1]
		pts_along_curve_y_left <- retval[2]
		}

	#---------------------------------------------------------
	# Collect the points along the line from the entry to exit
	#---------------------------------------------------------
	if( dpisna(pts_along_curve_x_left) && dpisna(pts_along_curve_x_right)) {
		pts_along_curve_x <- NA
		pts_along_curve_y <- NA
		}
	else
		{
		if( dpisna(pts_along_curve_x_left) ) {
			pts_along_curve_x <- pts_along_curve_x_right
			pts_along_curve_y <- pts_along_curve_y_right
			}
		else if( dpisna(pts_along_curve_x_right) ) {
			pts_along_curve_x <- pts_along_curve_x_left
			pts_along_curve_y <- pts_along_curve_y_left
			}
		else
			{
			pts_along_curve_x <- c(pts_along_curve_x_left, pts_along_curve_x_right)
			pts_along_curve_y <- c(pts_along_curve_y_left, pts_along_curve_y_right)
			}
		}
		
	retval_list <- list( entry=c(retval[1],retval[2]), exit=c(retval[3],retval[4]), 
			pts_along_curve_x=pts_along_curve_x, pts_along_curve_y=pts_along_curve_y )

	if(debug) { 
		print("-------------------------------------------------------------------")
		print("disk_entry_exit: exiting")
		}
	return( retval_list )
}

#=========================================================================================
# This plots lines with overplotted symbols, BUT the overplotted symbols
# are not plotted at every point (as is standard with R).  Instead they
# are only plotted at every subsample'd point
#
lines_subsample_pts_old <- function( x, y, subsample, ioffset=0, ... ) {

	lines( x, y, ..., type='l' )

	n <- length(x)
	if( length(y) != n ) 
		stop("must have same number of X and Y values!")

	nsub <- floor(n/subsample)
	idx  <- (0:(nsub-1))*subsample + 1 + ioffset
	xsub <- x[idx]
	ysub <- y[idx]

	points( xsub, ysub, ..., type='p' )
}

#=================================================================================
# this takes a cumultative distrubution function (CDF), which is an s-shaped
# curve.  It plots the CDF using an inverse erf translation on the Y axis,
# so that gaussians end up as straight lines in the plot.  'xvals' are the
# values on the x axis of the 's' curve, and typically indicate the physical 
# values associated with the CDF. 'prob' are the values on the Y axis of the
# S-curve, and must be between 0.0 and 1.0, inclusive.  So, for example, 
# the probability of getting a measurement of xvals[N] or less is prob[N].
#
# This is specifically for CDF plots.  If you just want a general plot on a
# erf-type probability axis, use plot_erf_axis, below.
#
probplot <- function( xvals, prob, col='black', col_grid=rgb(.6,.9,.6), lty_grid=1, 
		add=FALSE, xlim=range(xvals,na.rm=TRUE), ylim=NA, 
		plotit=TRUE,
		subsample=FALSE,
		x_xform=NA, put_x_labels_at=NA,
		show_act_y_vals=FALSE,
		yvals=c(.0001, .001,.01,.05,.1,.25,.5,.75,.9,.95,.99,.999, .9999),
		... ) {

print(paste("probplot entering ======================================="))
print(paste("probplot: entering with ylim=", pv(ylim) ))
print(paste("probplot: entering with yvals=", pv(yvals) ))

	if( !add )
		globalframenumber <<- globalframenumber + 1

	prob = prob[ !is.na(prob) ]
	if( length(prob) < 10 ) {
		stop(paste("PROBPLOT: Error, after taking out NA's, length of prob is only ", length(prob)))
		}

	if( (min(prob)<0.0) || (max(prob)>1.0))
		stop("PROBPLOT: bad range in prob: must be between 0.0 and 1.0, inclusive")

	xrange <- seq(-5.6,5.6,by=.05)	# in standard deviations; do not extend this past 5.6, R runs out of resolution
	yerf   <- myerf(xrange)/2 + .5	# S-shaped, range of 0 to 1

	xvals_orig = xvals

	if( (! is.finite(xlim[1])) || (! is.finite(xlim[2])))
		xlim_orig = range( xvals, na.rm=TRUE )
	else
		xlim_orig = xlim

	if( !is.na(x_xform)) {
		xvals = x_xform(xvals)
		xlim = x_xform(xlim)
		}

	ymax = max( c( prob, yvals, ylim ), na.rm=TRUE )

	#-----------------------------------------------------
	# Now reverse-transform prob values using erf function
	#-----------------------------------------------------
	print(paste("orig range of prob vals:", pv(range(prob)) ))
	prob_xform <- tableTranslate( prob, yerf, xrange )  # note how x and y switched since inverting!
	print(paste("transformed range of prob vals:", pv(range(prob_xform)) ))

	if( add ) {
		if( plotit ) {
			if( subsample ) {
				rvss = subsample_erf( xvals, prob_xform )
				xvals = rvss$x
				prob_xform = rvss$y
				}
			lines( xvals, prob_xform, col=col, xlim=xlim, ... )
			}
		return()
		}
	else
		{
		if( length(ylim) == 2) {
			mask_keep = ((prob >= ylim[1]) & (prob <= ylim[2]))	# note we mask on prob, but then apply it to prob_xform
			prob_xform = prob_xform[ mask_keep ]
			xvals      = xvals[ mask_keep ]
			}
		else
			{
			#ylim = range( prob_xform )
			ylim = c( min( prob ), ymax )
print(paste("probplot: ylim not supplied, so I'm setting it to:", pv(ylim) ))
			}
			
		ylim_xform = tableTranslate( ylim, yerf, xrange ) 

if( sum(is.na(ylim_xform)) > 0 ) {
	print(paste('Error in probplot, ylim_form in NA!'))
	print(paste('applied yerf:'))
	print(yerf)
	print(paste('ylim:', pv(ylim)))
	print(paste('xrange:', pv(xrange)))
	stop('bad value in probplot')
	}
print(paste("probplot: final units-ful ylim being applied: ", pv(ylim), " final xformed ylim:", pv(ylim_xform) ))

		#---------------------------------------------
		# This is the main call that produces the plot
		#---------------------------------------------
print('=== mlib_plot probplot ==a')
		if( plotit ) {
			if( subsample ) {
				rvss = subsample_erf( xvals, prob_xform )
				xvals = rvss$x
				prob_xform = rvss$y
				}
			if( (! is.finite(xlim[1])) || (! is.finite(xlim[2])))
				xlim = range( xvals, na.rm=TRUE )
			mplot( xvals, prob_xform, axes=FALSE, grid=FALSE, xaxs='i', yaxs='i', col=col, xlim=xlim, ylim=ylim_xform, ... )
			}
print('=== mlib_plot probplot ==b')
		}

	if( show_act_y_vals ) {
		xoff = mean(xvals)
		for( ty in seq(-3,3,by=0.25)) {
			if( plotit )
				text( xoff, ty, ty, adj=c(0,0), col='orange')
			}
		}

	tt <- par('usr')
	xlim <- tt[1:2]

	#---------------------
	# Put on X axis labels
	#---------------------
	if( length(put_x_labels_at) == 1 ) {
		if( is.na(xlim))
			pp <- pretty(xvals_orig)
		else	
			pp <- pretty(xlim_orig)
		}
	else
		pp <- put_x_labels_at

	for( ix in 1:length(pp)) {
		if( !is.na(x_xform)) 
			x0 = x_xform( pp[ix] )
		else
			x0 = pp[ix]

		y_val_for_x_labels = ylim[1]

		if( (pp[ix] <= xlim_orig[2]) && (pp[ix] >= xlim_orig[1]) ) {
			if( is.na(col_grid)) {
				uu = par()$usr
				ddy = uu[4] - uu[3]
				ytic0 = uu[3]
				ytic1 = ytic0 + ddy*0.03
				if( plotit ) 
					lines( c(x0,x0), c(ytic0,ytic1) )
				}
			else
				{
				if( plotit ) 
					abline(v=x0, lty=lty_grid, col=col_grid )
				}

			ylab <- tableTranslate( y_val_for_x_labels, yerf, xrange )

			if( plotit )
				text( x0, ylab, pp[ix], adj=c(0.5,1.4), xpd=NA )
			}
		}

	#--------------
	# Y axis labels
	#--------------
#print('probplot putting on Y axis labels')
	tty   <- tableTranslate( yvals, yerf, xrange )
	ny <- length(yvals)
	for( iy in 1:ny ) {
		yoff <- tty[iy]
#print(paste('probplot y label', iy, 'orig value: ', yvals[iy], ' transormed y:', tty[iy] ))
		#if( (yoff < max(prob_xform)) && (yoff > prob_xform[1])) {
		if( (yvals[iy] <= ylim[2]) && (yvals[iy] >= ylim[1])) {
			#print(paste("...NOT skipping this label, yvals[iy]=", yvals[iy], 'ylim=', pv(ylim) )) 
			if( plotit )
				text( xlim[1], yoff, yvals[iy], adj=c(1.1,0.5), xpd=NA)
			if( is.na(col_grid)) {
				uu = par()$usr
				dx = uu[2] - uu[1]
				xtic0 = uu[1]
				xtic1 = uu[1] + dx*0.02
				if( plotit )
					lines( c(xtic0,xtic1), c(yoff,yoff) )
				}
			else
				{
				if( plotit ) 
					abline(h=yoff, lty=lty_grid, col=col_grid)
				}
			}
		else
			print(paste('...skipping this y label; yoff=', yoff, ' max(prob_xform)=', max(prob_xform), 
				'prob_xform[1]=', prob_xform[1] ))
		}

	if( plotit ) 
		lines( xvals, prob_xform, col=col, ... )

	#---------------
	# Example useage
	#---------------
	# Make CDF
	#---------
	# x<-seq(-3,3,by=.1)
	# y <- (myerf(x)/2) + .5
	# xvals <- x * 100.0
	# probplot( xvals, y )

	#------------------------------------------------------------------------------
	# You can plot more points on the plot using the return value as follows.  Say
	# the point you want to plot should fall on the graph at X=0.4, Y=0.99.
	# Then: 
	#	x <- 0.4
	#	y <- tableTranslate( 0.99, rv$yerf, rv$xrange )
	#   	points( x, y )
	#------------------------------------------------------------------------------
	rv <- list( yerf=yerf, xrange=xrange, x_xform=x_xform, xvals=xvals, yvals=prob_xform )	

print(paste("probplot exiting ======================================="))

	return( rv )
}

#=============================================================================================
plot_erf_axis <- function( xvals, prob, add=FALSE, grid_col=rgb(.6,.9,.6), 
		x_xform=NA,
		yvals_axis=c(.0001, .0005, .001,.005,.01,.05,.1,.25,.5,.75,.9,.95,.99,.995,.999, .9995, .9999),
		grid_show=TRUE,
		grid_lty=1, grid_lwd=1, 
		xlim=range(xvals), 
		max_pt_pch=4,
		max_pt_cex=1.0,
		ylim=range(prob), 
		... ) {

	if( !add )
		globalframenumber <<- globalframenumber + 1

	prob = prob[ !is.na(prob) ]
	if( length(prob) < 10 ) {
		stop(paste("Error, after taking out NA's, length of prob is only ", length(prob)))
		}

	if( (min(prob)<0.0) || (max(prob)>1.0))
		stop(paste("bad range in prob: must be between 0.0 and 1.0, inclusive; min=", min(prob), 'max=', max(prob)))

	if( !is.na(x_xform)) {
		xvals_orig = xvals
		xlim_orig = xlim
		xvals = x_xform(xvals)
		xlim = x_xform(xlim)
		}

	xrange <- seq(-8,8,by=.05)	# in standard deviations
	yerf   <- myerf(xrange)/2 + .5	# S-shaped, range of 0 to 1

	#-----------------------------------------------------
	# Now reverse-transform prob values using erf function
	#-----------------------------------------------------
	prob_xform <- tableTranslate( prob, yerf, xrange )  # note how x and y switched since inverting!

	#---------------
	# Tranform Y lim
	#---------------
	ymin_xform <- tableTranslate( ylim[1], yerf, xrange )
	ymax_xform <- tableTranslate( ylim[2], yerf, xrange )

	if( add ) {
		lines( xvals, prob_xform, ... )

		#-----------------------------------
		# Highlight max Y point if requested
		#-----------------------------------
		if( !is.na( max_pt_pch )) {
			idx_max = which.max( prob_xform )
			points( xvals[idx_max], prob_xform[idx_max], pch=max_pt_pch, cex=max_pt_cex, ... )
			}
		return()
		}
	else
		{
		ylim_xform = c(ymin_xform,ymax_xform)
		mplot( xvals, prob_xform, grid=FALSE, xaxs='i', yaxs='i',
			axes=FALSE,
			ylim=ylim_xform, xlim=xlim, ... )

		#----------------------------------------------------
		# Add vertical grid lines ... based on ORIG values if
		# there is an X transform in effect
		#----------------------------------------------------
		if( !is.na(x_xform)) {
			pp_orig <- pretty(xlim_orig, n=10)
			pp = x_xform( pp_orig )
			}
		else
			{
			pp <- pretty(xlim)
			pp_orig = pp
			}

		ylab = ymin_xform
		for( ix in 1:length(pp)) {
			if( pp[ix] < xlim[2] ) {
				if( grid_show ) 
					abline(v=pp[ix], lty=grid_lty, col=grid_col)
				text( pp[ix], ylab, pp_orig[ix], adj=c(0.5,1.2), xpd=NA)
				}
			}

		#-----------------------------------
		# Highlight max Y point if requested
		#-----------------------------------
		if( !is.na( max_pt_pch )) {
			idx_max = which.max( prob_xform )
			points( xvals[idx_max], prob_xform[idx_max], pch=max_pt_pch, cex=max_pt_cex, ... )
			}
		}

	#--------------------------------------
	# Put on labels that are in Y lim range
	#--------------------------------------
	yvals_axis <- yvals_axis[ yvals_axis >= ylim[1] ]
	yvals_axis <- yvals_axis[ yvals_axis <= ylim[2] ]

	tt <- par('usr')
	xlim <- tt[1:2]

	tty   <- tableTranslate( yvals_axis, yerf, xrange )
	ny <- length(yvals_axis)
	for( iy in 1:ny ) {
		if( (yvals_axis[iy] >= ylim[1]) && (yvals_axis[iy] <= ylim[2])) {
			yoff <- tty[iy]
			text( xlim[1], yoff, yvals_axis[iy], adj=c(1.1,0.5), xpd=NA)
			#--------------------------
			# Add horizontal grid lines
			#--------------------------
			if( grid_show ) 
				abline(h=yoff, lty=grid_lty, col=grid_col, lwd=grid_lwd)
			}
		}

	#-----------------------------------------
	# Replot since plotted grid on top of data
	#-----------------------------------------
	lines( xvals, prob_xform, ... )

	#------------------------------------------------------------------------------
	# You can plot more points on the plot using the return value as follows.  Say
	# the point you want to plot should fall on the graph at X=0.4, Y=0.99.
	# Then: 
	#	x <- 0.4
	#	y <- tableTranslate( 0.99, rv$yerf, rv$xrange )
	#   	points( x, y )
	#------------------------------------------------------------------------------
	rv <- list( yerf=yerf, xrange=xrange, xlim=xlim, x_xform=x_xform, ylim=c(ymin_xform,ymax_xform) )	

	return(rv)
}

#=============================================================================================
# This plots data over time in a recognizable time/day format.  First input argument
# is a list of NT utDate elements.  Second argument is the Y data to plot.
#
day_time_plot <- function( dates, y, ylim=range(y), bg_night=rgb(.7,.7,.7), 
	label_hours=TRUE, ... ) {

	nt <- length(dates)
	if( nt != length(y))
		stop(paste("error, length of dates is",nt,"but I was supplied",
			length(y),"Y values to plot!  The must be the same"))

	
	canvas( xlim=c(1,nt), ylim=ylim, axes=FALSE, xlab='', yaxs='i', ... )
	axis(2)

	yr <- abs(ylim[2] - ylim[1])
	
	#--------------------------------------------------------
	# Draw the grey boxes for nighttime values (6 pm to 6 am)
	#--------------------------------------------------------
	for( it in 1:nt ) {
		night <- FALSE
		if( (dates[[it]]$hour < 6) || (dates[[it]]$hour > 18))
			night <- TRUE
		if( night ) {
			x0 <- it-.5
			x1 <- it+.1
			y0 <- ylim[1] - yr/10
			y1 <- ylim[2] + yr/10
			polygon( c(x0,x1,x1,x0), c(y0,y0,y1,y1), col=bg_night, border=bg_night )
			}
		}

	if( label_hours ) {
		}

}

#==========================================================================================
#==========================================================================================
#==========================================================================================


#======================================================================================
plot_3d_line <- function( scat, x0, y0, z0, x1, y1, z1, col='grey', verbose=FALSE, ... ) {

	p0 <- scat$xyz.convert( x0, y0, z0 )
	p1 <- scat$xyz.convert( x1, y1, z1 )

	if(verbose) print(paste("line from",x0,y0,z0," to ", x1,y1,z1))
	lines( c(p0$x, p1$x), c(p0$y, p1$y), col=col, ... )
}

#======================================================================================
# extend has a '1' where the glyph has extent in that axis, and 0 otherwise
#
plot_3d_glyphs <- function( scat, xx, yy, zz, xr, yr, zr, extend, col='purple' ) {

	scale <- .015

	dx <- (xr[2] - xr[1])*scale
	dy <- (yr[2] - yr[1])*scale
	dz <- (zr[2] - zr[1])*scale
	
	np <- length(xx)

	if( length(col) == 1 )
		col <- array(col, np)
	else if( length(col) != np )
		stop("col array must have either 1 entry or 1 entry per point")

	for( ip in 1:np ) {
		llx <- xx[ip] - dx*extend[1]
		lly <- yy[ip] - dy*extend[2]
		llz <- zz[ip] - dz*extend[3]

		urx <- xx[ip] + dx*extend[1]
		ury <- yy[ip] + dy*extend[2]
		urz <- zz[ip] + dz*extend[3]

		ll <- scat$xyz.convert( llx, lly, llz )
		ur <- scat$xyz.convert( urx, ury, urz )

		lwd <- 1
		v <- FALSE
		if( extend[3] == 0 ) {
			plot_3d_line( scat, llx, lly, llz, urx, lly, llz, col=col[ip], lwd=lwd, verbose=v )
			plot_3d_line( scat, urx, lly, llz, urx, ury, llz, col=col[ip], lwd=lwd, verbose=v )
			plot_3d_line( scat, urx, ury, llz, llx, ury, llz, col=col[ip], lwd=lwd, verbose=v )
			plot_3d_line( scat, llx, ury, llz, llx, lly, llz, col=col[ip], lwd=lwd, verbose=v )
			}
		if( extend[2] == 0 ) {
			plot_3d_line( scat, llx, lly, llz, urx, lly, llz, col=col[ip], lwd=lwd, verbose=v )
			plot_3d_line( scat, urx, lly, llz, urx, lly, urz, col=col[ip], lwd=lwd, verbose=v )
			plot_3d_line( scat, urx, lly, urz, llx, lly, urz, col=col[ip], lwd=lwd, verbose=v )
			plot_3d_line( scat, llx, lly, urz, llx, lly, llz, col=col[ip], lwd=lwd, verbose=v )
			}
		if( extend[1] == 0 ) {
			plot_3d_line( scat, llx, lly, llz, llx, ury, llz, col=col[ip], lwd=lwd, verbose=v )
			plot_3d_line( scat, llx, ury, llz, llx, ury, urz, col=col[ip], lwd=lwd, verbose=v )
			plot_3d_line( scat, llx, ury, urz, llx, lly, urz, col=col[ip], lwd=lwd, verbose=v )
			plot_3d_line( scat, llx, lly, urz, llx, lly, llz, col=col[ip], lwd=lwd, verbose=v )
			}
		}
}

#======================================================================================
plot_3d_swarm <- function( xx, yy, zz, field_to_color_by=zz, 
		dropline_to=c('xx'),
		zrange_field_to_color_by=range(field_to_color_by),
		grid=FALSE, box=FALSE, cex.symbols=1.2, 
		dropline_col='black', point_col='orange', pch=19, cex=1.0, distance_cex_fact=2.0,
		blackpoint_cex=1.5,
		xlim=range(xx,na.rm=TRUE), ylim=range(yy,na.rm=TRUE), 
		zlim=range(zz,na.rm=TRUE), ... ) {

	cols3d = mcolormap( 'bluegreenred', field_to_color_by, zrange=zrange_field_to_color_by )

	np <- length(xx)
	if( np != length(yy))
		stop("error, x and y lengths differ!")
	if( np != length(zz))
		stop("error, x and z lengths differ!")
	print(paste('npts=', np))

	library(scatterplot3d)
	print('about to call scatterplot3d')
	scat <- scatterplot3d( xx, yy, zz, pch=pch, cex.symbols=cex.symbols*.1, grid=grid, box=box, 
		xlim=xlim, ylim=ylim, zlim=zlim, xaxs='i', yaxs='i', zaxs='i', color=point_col, ... )
	print('back from call to scatterplot3d')

	xlim <- c( scat$dpvals$x.min, scat$dpvals$x.max )
	ylim <- c( scat$dpvals$y.min, scat$dpvals$y.max )
	zlim <- c( scat$dpvals$z.min, scat$dpvals$z.max )

	orig <- scat$xyz.convert( xx, yy, zz )
	base <- scat$xyz.convert( xx, yy, array(zlim[1],length(xx)))
	back <- scat$xyz.convert( xx, array(ylim[2],length(xx)), zz)
	left <- scat$xyz.convert( array(xlim[1],length(xx)), yy, zz)

	print(paste("xlim=", xlim[1], xlim[2] ))
	print(paste("ylim=", ylim[1], ylim[2] ))
	print(paste("zlim=", zlim[1], zlim[2] ))

	#-----------------
	# Draw grid planes
	#-----------------
	xl <- pretty(xlim)
	yl <- pretty(ylim)
	zl <- pretty(zlim)
	lwd_axis <- 2
	for( ix in 1:length(xl)) {
		x0 <- xl[ix]
		y0 <- ylim[1]
		x1 <- xl[ix]
		y1 <- ylim[2]
		z0 <- zlim[1]
		z1 <- zlim[1]
		lwd <- 1
		col <- 'grey'
		if( (ix==1) || (ix==length(xl))) {
			lwd <- lwd_axis
			col <- 'black'
			}
		plot_3d_line( scat, x0, y0, z0, x1, y1, z1, col=col, lwd=lwd )
		}
	for( iy in 1:length(yl)) {
		x0 <- xlim[1]
		y0 <- yl[iy]
		z0 <- zlim[1]
		x1 <- xlim[2]
		y1 <- yl[iy]
		z1 <- zlim[1]
		lwd <- 1
		col <- 'grey'
		if( (iy==1) || (iy==length(yl))) {
			lwd <- lwd_axis
			col <- 'black'
			}
		plot_3d_line( scat, x0, y0, z0, x1, y1, z1, col=col, lwd=lwd )
		}

	for( iz in 1:length(zl)) {
		x0 <- xlim[1]
		x1 <- xlim[1]
		y0 <- ylim[1]
		y1 <- ylim[2]
		z0 <- zl[iz]
		z1 <- z0
		lwd <- 1
		col <- 'grey'
		if( (iz==1) || (iz==length(zl))) {
			lwd <- lwd_axis
			col <- 'black'
			}
		plot_3d_line( scat, x0, y0, z0, x1, y1, z1, col=col, lwd=lwd )
		}
	for( iy in 1:length(yl)) {
		x0 <- xlim[1]
		x1 <- xlim[1]
		y0 <- yl[iy]
		y1 <- y0
		z0 <- zlim[1]
		z1 <- zlim[2]
		lwd <- 1
		col <- 'grey'
		if( (iy==1) || (iy==length(yl))) {
			lwd <- lwd_axis
			col <- 'black'
			}
		plot_3d_line( scat, x0, y0, z0, x1, y1, z1, col=col, lwd=lwd )
		}

	for( ix in 1:length(xl)) {
		x0 <- xl[ix]
		x1 <- x0
		y0 <- ylim[2]
		y1 <- y0
		z0 <- zlim[1]
		z1 <- zlim[2]
		lwd <- 1
		col <- 'grey'
		if( (ix==1) || (ix==length(xl))) {
			lwd <- lwd_axis
			col <- 'black'
			}
		plot_3d_line( scat, x0, y0, z0, x1, y1, z1, col=col, lwd=lwd )
		}

	for( iz in 1:length(zl)) {
		x0 <- xlim[1]
		x1 <- xlim[2]
		y0 <- ylim[2]
		y1 <- y0
		z0 <- zl[iz]
		z1 <- z0
		lwd <- 1
		col <- 'grey'
		if( (iz==1) || (iz==length(zl))) {
			lwd <- lwd_axis
			col <- 'black'
			}
		plot_3d_line( scat, x0, y0, z0, x1, y1, z1, col=col, lwd=lwd )
		}

	nn <- length(orig$x)

	#-------------------------------
	# Now plot the dots on each axis
	#-------------------------------
	plot_3d_glyphs( scat, xx, yy, array(zlim[1],length(xx)), xlim, ylim, zlim, c(1,1,0), col=cols3d )

	plot_3d_glyphs( scat, xx, array(ylim[2],length(xx)), zz, xlim, ylim, zlim, c(1,0,1), col=cols3d )

	plot_3d_glyphs( scat, array(xlim[1],length(xx)), yy, zz, xlim, ylim, zlim, c(0,1,1), col=cols3d )

	if( 'xx' %in% dropline_to )
		segments( orig$x, orig$y, base$x, base$y, col=dropline_col )

	if( 'yy' %in% dropline_to )
		segments( orig$x, orig$y, back$x, back$y, col=dropline_col )

	if( 'zz' %in% dropline_to )
		segments( orig$x, orig$y, left$x, left$y, col=dropline_col )

	#---------------------------------------------------------------
	# This is the actual (re-)plotting of the dots, we do it last so
	# they are not covered up by other things, such as the droplines
	#---------------------------------------------------------------

	#-----------------------------------------------------------
	# For distance, simplify by just using Y coordinate.  More
	# negative values are CLOSER, more positive vals are FARTHER
	#-----------------------------------------------------------
	ymin <- min(yy,na.rm=TRUE)
	ymax <- max(yy,na.rm=TRUE)
	dist <- (yy-ymin)/(ymax-ymin)	# in range 0->1
	dist <- 1 - dist		# since neg vals are closer, not farther
	cex <- array(cex, nn)
	tt <- 1:nn
	lo <- 2/(1+distance_cex_fact)
	hi <- lo*distance_cex_fact
	del <- hi-lo
	fact <- dist*del + lo
	cex <- cex*fact

	for( ipt in 1:nn ) {
		col = cols3d[ipt]
		#points(orig$x[ipt],orig$y[ipt],col='black',pch=pch,cex=cex[ipt]*blackpoint_cex,...)
		points(orig$x[ipt],orig$y[ipt],col='black',bg=col,pch=21,cex=cex[ipt],...)
		}
}

#===============================================================================================
# Similar to a color contour plot, but optimized for a few blocks and labels on the
# X and Y axes.  The main difference between this and a color contour plot is that
# this assumes values are at the CENTER of color blocks, not the edges.
#
# 'gutterx' is the space to the RIGHT of each block. So if there are N blocks in the X
# directoin, gutterx should either be a constant or an array length(N-1).
#
# Since the X values of the centers of the blocks can be complicated if gutterx is not
# a constant, center_x is returned.
#
# Same statements as above apply to 'guttery', which is the space ABOVE each block.
#
# See discussion under 'mcolorbar' for useage of 'discretevals'.
#
portrait_plot <- function( xlab, ylab, data, cmap='bluegreenred', 
		ylab_col=array('black', length(ylab)),
		discretevals=NA,
		blockdx=0.8,
		blockdy=0.8,
		gutterx=array(0.1,length(xlab)-1),
		guttery=array(0.1,length(ylab)-1),
		right_hand_row_labels=NA,
		emphasize_best_model=FALSE,
		exp=1.0,				# passed to mcolormap
		zrange=range(data), ... ) {

	ndims <- length(dim(data))
	if( ndims != 2 ) 
		stop(paste("Error, data must be 2-D; passed data had",ndims,"dimensions"))

	nx <- dim(data)[1]
	ny <- dim(data)[2]
	if( length(xlab) != nx )
		stop(paste("Error, must have same number of X labels as X entries in data; # x labels=", length(xlabels), " X entries in data=", nx))
	if( length(ylab) != ny )
		stop(paste("Error, must have same number of Y labels as Y entries in data; # y labels=", length(ylabels), " Y entries in data=", ny))

	#---------------------------------------------------------------------------
	# Get the overall X location of the left and right hand edge, and same for Y
	#---------------------------------------------------------------------------
	x00 <- 1
	xbl <- array(0., nx)	# 'x block left'
	xbr <- array(0., nx)	# 'x block right'
	xc  <- array(0., nx)	# 'x center'
	xbl[1] <- x00
	xbr[1] <- xbl[1] + blockdx
	for( i in 2:nx ) {
		xbl[i] <- xbr[i-1] + gutterx[i-1]
		xbr[i] <- xbl[i] + blockdx
		}
	xc  <- 0.5*(xbl + xbr)
	x11 <- xbr[nx]

	y00 <- 1
	ybl <- array(0., ny)	# 'y block lower'
	ybu <- array(0., ny)	# 'y block upper'
	yc  <- array(0., ny)	# 'y center'
	ybl[1] <- y00
	ybu[1] <- ybl[1] + blockdy
	for( i in 2:ny ) {
		ybl[i] <- ybu[i-1] + guttery[i-1]
		ybu[i] <- ybl[i] + blockdy
		}
	yc  <- 0.5*(ybl + ybu)
	y11 <- ybu[ny]

	canvas( xlim=c(x00,x11), ylim=c(y00,y11), box=FALSE, grid=FALSE,
		xaxs='i', yaxs='i', xlab='',
		axes=FALSE, ... )
	for( jj in 1:ny )
	for( ii in 1:nx ) {

		dat_missing = FALSE

		if( is.na(discretevals)) 
			col <- mcolormap( cmap, data[ii,jj], zrange=zrange, exp=exp, ... )
		else
			{
			ndv <- length( discretevals$values )
			ndc <- length( discretevals$colors )
			if( ndv == ndc ) {
				iidx <- wherenearest( data[ii,jj], discretevals$values )
				col  <- discretevals$colors[iidx]
				}
			else
				{
				if( is.na( data[ii,jj] )) {
					dat_missing = TRUE
					}
				else if( data[ii,jj] < discretevals$values[1] ) {
					col <- discretevals$colors[1]
					if( is.null( discretevals$left_arrow_col ))
						col <- discretevals$colors[1]
					else
						col <- discretevals$left_arrow_col
					}
				else if( data[ii,jj] > discretevals$values[ndv] )
					col <- discretevals$colors[ndc]
				else {
					iidx <- -1
					for( interval in 1:ndc )
						if( ( data[ii,jj] >= discretevals$values[interval] ) && ( data[ii,jj] < discretevals$values[interval+1] ))
							iidx <- interval
					if( iidx < 0 ) 
						stop(paste("error, didn't find interval in color discrete values! data=", data[ii,jj]))
					col <- discretevals$colors[iidx]
					}
				}
			}

		if( dat_missing ) {
			lines( c( xbl[ii], xbr[ii]), c( ybl[jj], ybu[jj] ))
			lines( c( xbr[ii], xbl[ii]), c( ybl[jj], ybu[jj] ))
			}
		else
			polygon( c(xbl[ii],xbr[ii],xbr[ii],xbl[ii]),
				 c(ybl[jj],ybl[jj],ybu[jj],ybu[jj]),
				 col=col, border=col )
		}

	#---------------------
	# Left hand row labels
	#---------------------
	for( jj in 1:ny ) {
		if( jj == ny && emphasize_best_model ) par(font=2)
		text( x00-0.1, yc[jj], ylab[jj], adj=c(1.0,0.5), xpd=NA, col=ylab_col[jj], ... )
		if( jj == ny && emphasize_best_model ) par(font=1)
		}

	#----------------------
	# Right hand row labels
	#----------------------
	if( length(right_hand_row_labels) == ny ) {
		for( jj in 1:ny ) {
			if( jj == ny && emphasize_best_model ) par(font=2)
			text( x11, yc[jj], right_hand_row_labels[jj], adj=c(-0.02,0.5), xpd=NA, ... )
			if( jj == ny && emphasize_best_model ) par(font=1)
			}
		}

	#---------------------------
	# Lables along bottom X axis
	#---------------------------
	par(srt=60)
	for( ii in 1:nx )
		text( xc[ii], y00-0.5, xlab[ii], adj=c(1.0,0.5), xpd=NA, ... )
	par(srt=0)

	retval <- list( xc=xc, yc=yc )

	return( retval )
}

#===========================================================================================
big_vert_paren <- function( llx, lly, urx, ury, dx=0.05, abs=FALSE, ... ) {

	if( abs ) {
		p = fractional_to_user_units( llx, lly )
		llx = p[1]
		lly = p[2]
		p = fractional_to_user_units( urx, ury )
		urx = p[1]
		ury = p[2]
		}

	#-------------------------------------------------
	# First build it in box (0,0) to (1,1), then scale
	#-------------------------------------------------
	cc <- 0.35	# curve factor; 0.5 is no curve at all
	yy <- c( 0, dx*cc, dx, .5-dx, .5-dx*cc, .5, .5+dx*cc, .5+dx, 1-dx, 1-dx*cc, 1 )
	xx <- c( 1, .75, .5, .5, .25, 0, .25, .5, .5, .75, 1 )

	width  <- urx - llx
	height <- ury - lly

	xvals <- llx + xx*width
	yvals <- lly + yy*height

	lines( xvals, yvals, xpd=NA, ... )
}

#===========================================================================================
big_horiz_bracket <- function( llx, lly, urx, ury, center_split_width=0, square=FALSE, ... ) {

	if( square ) {
		lines( c(llx,llx), c(lly,ury), xpd=NA, ... )
		lines( c(llx,urx), c(ury,ury), xpd=NA, ... )
		lines( c(urx,urx), c(ury,lly), xpd=NA, ... )
		return
		}

	#-------------------------------------------------
	# First build it in box (0,0) to (1,1), then scale
	#-------------------------------------------------
	xx <- c( 0, 0, 0.5-center_split_width/2, 0.5+center_split_width/2, 1, 1 )
	yy <- c( 0, 1, 1, 1, 1, 0 )

	width  <- urx - llx
	height <- ury - lly

	xvals <- llx + xx*width
	yvals <- lly + yy*height

	lines( xvals[1:3], yvals[1:3], xpd=NA, ... )
	lines( xvals[4:6], yvals[4:6], xpd=NA, ... )
}

#===========================================================================================
big_horiz_paren <- function( llx, lly, urx, ury, dx=0.05, fixed_dx=NA, abs=FALSE, ... ) {

	if( abs ) {
		p = fractional_to_user_units( llx, lly )
		llx = p[1]
		lly = p[2]
		p = fractional_to_user_units( urx, ury )
		urx = p[1]
		ury = p[2]
		}

	width  <- urx - llx
	height <- ury - lly

	if( !is.na(fixed_dx)) {
		width = urx - llx
		if( fixed_dx > 0.3*width )
			stop(paste("Error, fixed_dx is too large; max is", 0.3*width, "but you supplied", fixed_dx))
		dx = fixed_dx / width
		}

	print(paste("using dx=", dx))
		
	#-------------------------------------------------
	# First build it in box (0,0) to (1,1), then scale
	#-------------------------------------------------
	cc <- 0.35	# curve factor; 0.5 is no curve at all
	xx <- c( 0, dx*cc, dx, .5-dx, .5-dx*cc, .5, .5+dx*cc, .5+dx, 1-dx, 1-dx*cc, 1 )
	yy <- c( 1, .75, .5, .5, .25, 0, .25, .5, .5, .75, 1 )

	xvals <- llx + xx*width
	yvals <- lly + yy*height

	lines( xvals, yvals, xpd=NA, ... )
}

#==========================================================================================
# Given X, Y, and a horizontal line, this colors the area above the line but
# below the curve one color, and the area below the line but above the curve
# another color.  Typically, this would be used to do wet and dry spells, or
# hot and cold, etc.
#
fill_curve_color_above_below <- function( x, y, above_col='red', above_val=mean(y),
		below_col='blue', below_val=mean(y) ) {

	cl_below_low = 1
	cl_on_low    = 2
	cl_middle    = 3
	cl_on_hi     = 4
	cl_above_hi  = 5

	above_below_same = ( above_val == below_val ) 

	#------------------------------------------------------------------
	# Life is easier if we transform the data to always have an X entry
	# where it crosses either the below or above line ... do that now
	#------------------------------------------------------------------
	np <- length(x)
	nmax = 3*np
	xx = array(0., nmax)
	yy = array(0., nmax)
	cl = array(as.integer(0), nmax)

	np2 = 1
	if( y[1] < below_val ) {
		xx[1] = x[1] - (x[2]-x[1])*1.e-6
		yy[1] = below_val
		cl[1] = cl_on_low
		xx[2] = x[1]
		yy[2] = y[1]
		np2 = 2
		}
	else if( y[1] == below_val ) {
		cl[1] = cl_on_low
		xx[1] = x[1]
		yy[1] = y[1]
		np2 = 1
		}
	else if( (!above_below_same) && ( y[1] == above_val )) {
		cl[1] = cl_on_hi
		xx[1] = x[1]
		yy[1] = y[1]
		}
	else if( y[1] > above_val ) {
		xx[1] = x[1] - (x[2]-x[1])*1.e-6
		yy[1] = above_val
		cl[1] = cl_on_hi
		xx[2] = x[1]
		yy[2] = y[1]
		np2 = 2
		}

	for( ip in 2:np ) {
		crossed_low = FALSE
		crossed_hi  = FALSE
		#---------------------------
		# Did we cross the low line?
		#---------------------------
		if( ((y[ip-1] > below_val) && (y[ip] < below_val)) || 
		    ((y[ip-1] < below_val) && (y[ip] > below_val)) ) {
			crossed_low = TRUE
			dy_have = y[ip-1] - y[ip]
			dy_want = y[ip-1] - below_val	
			fracx = dy_want / dy_have		# fractional, 0 to 1
			np2 = np2 + 1
			xx[np2] = x[ip-1] + fracx*( x[ip+1] - x[ip] )
			yy[np2] = below_val
			cl[np2] = cl_on_low
		    	}

		#---------------------------------------------------------------
		# Did we cross the high line? Only applies if high and low lines 
		# are distinct
		#---------------------------------------------------------------
		if( above_val != below_val ) {
			if( ((y[ip-1] > above_val) && (y[ip] < above_val)) || 
			    ((y[ip-1] < above_val) && (y[ip] > above_val)) ) {
				crossed_hi = TRUE
				dy_have = y[ip-1] - y[ip]
				dy_want = y[ip-1] - above_val	
				fracx = dy_want / dy_have		# fractional, 0 to 1
				np2 = np2 + 1
				xx[np2] = x[ip-1] + fracx*( x[ip+1] - x[ip] )
				yy[np2] = above_val
				cl[np2] = cl_on_hi
				}
			}

		#--------------------------------------------
		# If we crossed BOTH, order might be switched
		#--------------------------------------------
		if( crossed_low && crossed_hi ) {
			if( xx[np2] < xx[np2-1] ) {
				# Switch
				tx = xx[np2]
				xx[np2] = xx[np2-1]
				xx[np2-1] = tx
				ty = yy[np2]
				yy[np2] = yy[np2-1]
				yy[np2-1] = ty
				tc = cl[np2]
				cl[np2] = cl[np2-1]
				cl[np2-1] = tc
				}
			}
		np2 = np2 + 1
		xx[np2] = x[ip]
		yy[np2] = y[ip]
		if( yy[np2] < below_val )
			cl[np2] = cl_below_low
		else if( yy[np2] == below_val ) 
			cl[np2] = cl_on_low
		else if( (!above_below_same) && ( yy[np2] == above_val ))
			cl[np2] = cl_on_hi
		else if( yy[np2] > above_val )
			cl[np2] = cl_above_hi
		}

	if( y[np] < below_val ) {
		np2 = np2 + 1
		xx[np2] = x[np] + (x[np]-x[np-1])*1.e-6
		yy[np2] = below_val
		cl[np2] = cl_on_low
		}
	else if( y[np] > above_val ) {
		np2 = np2 + 1
		xx[np2] = x[np] + (x[np]-x[np-1])*1.e-6
		yy[np2] = above_val
		cl[np2] = cl_on_hi
		}

	xx = xx[1:np2]
	yy = yy[1:np2]
	cl = cl[1:np2]

	#---------------------------------------------------------
	# Now go through and color blocks of above or below points
	#---------------------------------------------------------
	
	ip = 1
	while( ip < np2 ) {
		#----------------------------------------------------------
		# Advance to first point that is on the above or below line
		#----------------------------------------------------------
		online = ((cl[ip] == cl_on_low) || (cl[ip] == cl_on_hi))
		while( (! online) && (ip < np2) ) {
			ip = ip + 1
			online = ((cl[ip] == cl_on_low) || (cl[ip] == cl_on_hi))
			}
		if( ip == np2 ) 
			return()

		#------------------------------------------------------------
		# Advance to next point that is of the same class as this one
		#------------------------------------------------------------
		ip2 = ip + 1
		sameclass = (cl[ip2] == cl[ip])
		while( (! sameclass) && (ip2 <= np2)) {
			ip2 = ip2 + 1
			sameclass = (cl[ip2] == cl[ip])
			}

		#----------------
		# Fill this block
		#----------------
		tx = xx[ip:ip2]
		ty = yy[ip:ip2]
		tn = length(tx)
		if( cl[ip2-1] == cl_below_low )
			col2use = below_col
		else if( cl[ip2-1] == cl_above_hi )
			col2use = above_col
		else
			col2use = 'green'
		polygon( tx, ty, col=col2use, border=col2use )

		ip = ip2 
		}
}

#===================================================================================
char_is_ascender <- function( cc ) {

	if( (cc=='b') || (cc=='d') || (cc=='f') || (cc=='h') || (cc=='i') ||
	    (cc=='k') || (cc=='l') || (cc=='t'))
		return( TRUE )
	return(FALSE)
}

#===================================================================================
char_is_descender <- function( cc ) {

	if( (cc=='g') || (cc=='j') || (cc=='p') || (cc=='y'))
		return( TRUE )
	return(FALSE)
}

#===================================================================================
char_is_upper <- function( cc ) {

	if( (cc=='A') || (cc=='B') || (cc=='C') || (cc=='D') ||
	    (cc=='E') || (cc=='F') || (cc=='G') || (cc=='H') ||
	    (cc=='I') || (cc=='J') || (cc=='K') || (cc=='L') ||
	    (cc=='M') || (cc=='N') || (cc=='O') || (cc=='P') ||
	    (cc=='Q') || (cc=='R') || (cc=='S') || (cc=='T') ||
	    (cc=='U') || (cc=='V') || (cc=='W') || (cc=='X') ||
	    (cc=='Y') || (cc=='Z') )
		return( TRUE )
	return(FALSE)
}

#=====================================================================================
draw_inch_reticule <- function( factor, n2draw=10, ... ) {

	pin <- par('pin')	# 'pin' The current plot dimensions, '(width, height)', in inches
	usr <- par('usr')	#  A vector 'c(x1, x2, y1, y2)' giving extremes of the user coordis of the plot region

	plot_width_in  <- pin[1]
	plot_height_in <- pin[2]

	dx_user <- usr[2] - usr[1]
	dy_user <- usr[4] - usr[3]

	for( i in 1:(n2draw*10) ) {
		x_in <- .5 + (i-1)*factor/10
		y0 <- usr[3]
		y1 <- usr[4]
		x_usr <- x_in/plot_width_in * dx_user + usr[1]

		lines( c(x_usr,x_usr), c(y0,y1), col='grey' )
		}

	for( i in 1:n2draw ) {
		x_in <- .5 + (i-1)*factor
		y0 <- usr[3]
		y1 <- usr[4]
		x_usr <- x_in/plot_width_in * dx_user + usr[1]

		lines( c(x_usr,x_usr), c(y0,y1), ... )
		}

	for( i in 1:(n2draw*10) ) {
		y_in <- .5 + (i-1)*factor/10
		x0 <- usr[1]
		x1 <- usr[2]
		y_usr <- y_in/plot_height_in * dy_user + usr[3]

		lines( c(x0,x1), c(y_usr,y_usr), col='grey' )
		}

	for( i in 1:n2draw ) {
		y_in <- .5 + (i-1)*factor
		x0 <- usr[1]
		x1 <- usr[2]
		y_usr <- y_in/plot_height_in * dy_user + usr[3]

print(paste(x0,x1,y_usr))
		lines( c(x0,x1), c(y_usr,y_usr), ... )
		}
}

#===================================================================================
# Give the coordinates of the lower left and upper right points in a box
# in USER UNITS of an EXISTING PLOT, and the next plot to be drawn will appear
# in that location.
#
# Useage:
#	rv = pre_inset_plot( ... )
#	inset_plot()
#	post_inset_plot( rv )
#
# NOTE: you might have to do a "par(new=FALSE)" to start a clean sheet
# of paper after filling the page.
#
# NOTE: if you have MULTIPLE INSET PLOTS ON ONE CANVAS, you probably want
# to set post_inset_plot( rv, advance=FALSE )
# 
# Example:
# /home/pierce/src/mine/bcca2/version_01/cross_validate/cccma_cgcm3_1_grid/compare_dist_mod_to_coarse_obs_tasmax.R
#
# NOTE: if you are doing LOG AXES, you pass the log10() of the coords to plot.
#	I guess R is plotting in log space but labeling in linear, it does
#	not do the conversion for you.
#
pre_inset_plot = function( inset_llx, inset_lly, inset_urx, inset_ury, logx=FALSE, logy=FALSE ) {

print(paste("orig inset (llx,lly,urx,ury): ", inset_llx, inset_lly, inset_urx, inset_ury ))
	if( logx ) {
		inset_llx = log10( inset_llx )
		inset_urx = log10( inset_urx )
		}
	if( logy ) {
		inset_lly = log10( inset_lly )
		inset_ury = log10( inset_ury )
		}

	z = par()

	#-------------------------------------------------------------------
	# Non-dimensional location of the current plot, range 0-1.  Remember
	# that if a plot is taking up less than the entire page, this fact
	# will be reflected in these numbers 
	#-------------------------------------------------------------------
	llx = z$fig[1]
	urx = z$fig[2]
	lly = z$fig[3]
	ury = z$fig[4]

	xr = urx - llx
	yr = ury - lly

	#-------------------------------------------------
	# The dimensional area of the plot in user units
	# On any axis that is log, this is the LOG of what
	# value appears on the plot's axis
	#-------------------------------------------------
	usrx0 = z$usr[1]
	usrx1 = z$usr[2]
	usry0 = z$usr[3]
	usry1 = z$usr[4]

	usr_xr = usrx1 - usrx0	# if x axis is log, this is in logs
	usr_yr = usry1 - usry0	# if y axis is log, this is in logs

	#-----------------------------------------------
	# Convert inset to normalized units in 0-1 range
	#-----------------------------------------------
	figx0 = (inset_llx-usrx0)/usr_xr
	figx1 = (inset_urx-usrx0)/usr_xr
	figy0 = (inset_lly-usry0)/usr_yr
	figy1 = (inset_ury-usry0)/usr_yr

	#-------------------------------------------------------------------------
	# Take into account the fact that the plot might just cover part of a page
	#-------------------------------------------------------------------------
	figx0 = figx0*xr + llx
	figx1 = figx1*xr + llx
	figy0 = figy0*yr + lly
	figy1 = figy1*yr + lly

	if( figx1 > 1 ) figx1 = 1
	if( figy1 > 1 ) figy1 = 1
	if( figx0 < 0 ) figx0 = 0
	if( figy0 < 0 ) figy0 = 0

	par( fig=c(figx0, figx1, figy0, figy1), new=TRUE)

	return(z)
}

#==========================================================================
post_inset_plot = function( par_orig, advance=TRUE ) {

	par( par_orig )
	par( new=TRUE )

	# set par back to what it should be
	curr_x   = par_orig$mfg[2]
	curr_y   = par_orig$mfg[1]
	nplots_x = par_orig$mfg[4]
	nplots_y = par_orig$mfg[3]

	if( advance )
		plotinc = 1
	else
		plotinc = 0

	curr_x = curr_x + plotinc
	if( curr_x > nplots_x ) {
		curr_x = 1
		curr_y = curr_y + 1
		if( curr_y > nplots_y ) {
			curr_x = nplots_x
			curr_y = nplots_y
			}
		}

	par( mfg=c(curr_y, curr_x, nplots_y, nplots_x ))

	print(paste("current par mfg: ", pv(par()$mfg) ))
}

#===============================================================================================
# Given coordinates in user units, this converts to coordinates in inches relative to the
# bottom left hand corner of the plotting area
#
# This routine assumes linear scaling at the moment
#
# It is safe to pass ux, uy as arrays
#
user_units_to_inches = function( ux, uy, verbose=FALSE ) {

verbose=FALSE

	if( storage.mode(ux) != 'double' )
		stop(paste("Error, on input ux must be a number (storage mode double), but has storage mode:", 
			storage.mode(ux) ))
	if( storage.mode(uy) != 'double' )
		stop(paste("Error, on input uy must be a number (storage mode double), but has storage mode:", 
			storage.mode(uy) ))

	if( verbose ) print(paste("user_units_to_inches: entering with ux, uy=", ux, uy ))

	par = par()
	usr = par$usr	# vector of the extremes of the user coordinates of the PLOTTING region

	if( verbose ) print(paste("user_units_to_inches: usr=", pv(usr) ))
	x0_user_plot = usr[1]
	x1_user_plot = usr[2]
	y0_user_plot = usr[3]
	y1_user_plot = usr[4]

	fracx = (ux - x0_user_plot) / (x1_user_plot - x0_user_plot)
	fracy = (uy - y0_user_plot) / (y1_user_plot - y0_user_plot)

	pin = par$pin	# The current plot dimensions, '(width, height)', in inches
	plot_width_in  = pin[1]
	plot_height_in = pin[2]

	ux_inch = fracx * plot_width_in
	uy_inch = fracy * plot_height_in

	if( verbose ) {
		print(paste("user_units_to_inches: plot width height in inches:", plot_width_in, plot_height_in ))
		print(paste("user_units_to_inches: user coords of", ux, uy,
			"are equal to inches (from lower left corner) of", ux_inch, uy_inch ))
		}

	return( list( x=ux_inch, y=uy_inch ))
}

#===============================================================================================
# ux_inch, uy_inch are relative to lower left corner of the plot
#
inches_to_user_units = function( ux_inch, uy_inch ) {

	par = par()

	usr = par$usr	# vector of the extremes of the user coordinates of the PLOTTING region

	x0_user_plot = usr[1]
	x1_user_plot = usr[2]
	y0_user_plot = usr[3]
	y1_user_plot = usr[4]

	pin = par$pin	# The current plot dimensions, '(width, height)', in inches
	plot_width_in  = pin[1]
	plot_height_in = pin[2]

	fracx = ux_inch / plot_width_in
	fracy = uy_inch / plot_height_in

	ux = fracx*(x1_user_plot - x0_user_plot) + x0_user_plot
	uy = fracy*(y1_user_plot - y0_user_plot) + y0_user_plot

	return( c( ux, uy ))
}

#===============================================================================================
# Let's say you want to plot something at an absolution location, but you have just done
# a user-space plot.  How do you know what cordinates to plot at?  This routine takes
# the absolute location (normalized, 0->1 on each axis) and returns the location of the 
# point in user units. 
#
# See also "text_abs", below
#
fractional_to_user_units = function( xx, yy ) {

	par = par()

	fig = par$fig	# NDC coords of figure region in display region of device
	fin = par$fin	# figure region dimnesions, width,height, in inches
	plt = par$plt	# coords of the PLOT region as fractions of the current FIGURE region
	usr = par$usr	# vector of the extremes of the user coordinates of the PLOTTING region
	din = par$din	# device size in inches
	omi = par$omi	# the size of the outer margins in inches
	omd = par$omd	# the region _inside_ outer margins in NDC (= normalized device coordinates), i.e., as fraction (in [0, 1]) of the device region.

	#--------------------------------------------------------------
	# Calcuate extent of FIGURE (not plotting) region in user units
	#--------------------------------------------------------------
	x0_user_plot = usr[1]
	x1_user_plot = usr[2]
	y0_user_plot = usr[3]
	y1_user_plot = usr[4]
	dx_user_plot = x1_user_plot - x0_user_plot
	dy_user_plot = y1_user_plot - y0_user_plot

	x0_frac_plot = plt[1]
	x1_frac_plot = plt[2]
	y0_frac_plot = plt[3]
	y1_frac_plot = plt[4]
	dx_frac_plot = x1_frac_plot - x0_frac_plot
	dy_frac_plot = y1_frac_plot - y0_frac_plot

	x0_user_fig = x0_user_plot - dx_user_plot/dx_frac_plot*x0_frac_plot
	x1_user_fig = x1_user_plot + dx_user_plot/dx_frac_plot*(1-x1_frac_plot)
	y0_user_fig = y0_user_plot - dy_user_plot/dy_frac_plot*y0_frac_plot
	y1_user_fig = y1_user_plot + dy_user_plot/dy_frac_plot*(1-y1_frac_plot)

	#abline(v=x0_user_fig, xpd=NA, col='blue' )
	#abline(v=x1_user_fig, xpd=NA, col='red' )
	#abline(h=y0_user_fig, xpd=NA, col='pink')
	#abline(h=y1_user_fig, xpd=NA, col='orange' )

	#-----------------------------------------------------------
	# Calculate extent of DISPLAY region of device in user units
	#-----------------------------------------------------------
	dx_user_fig = x1_user_fig - x0_user_fig
	dy_user_fig = y1_user_fig - y0_user_fig
	x0_ncd_fig = fig[1]
	x1_ncd_fig = fig[2]
	y0_ncd_fig = fig[3]
	y1_ncd_fig = fig[4]
	dx_ncd_fig = x1_ncd_fig - x0_ncd_fig
	dy_ncd_fig = y1_ncd_fig - y0_ncd_fig
	x0_user_disp = x0_user_fig - dx_user_fig/dx_ncd_fig*x0_ncd_fig
	x1_user_disp = x1_user_fig + dx_user_fig/dx_ncd_fig*(1 - x1_ncd_fig)
	y0_user_disp = y0_user_fig - dy_user_fig/dy_ncd_fig*y0_ncd_fig
	y1_user_disp = y1_user_fig + dy_user_fig/dy_ncd_fig*(1 - y1_ncd_fig)
	#abline( v=x0_user_disp, xpd=NA, col='lawngreen' )
	#abline( v=x1_user_disp, xpd=NA, col='brown' )
	#abline( h=y0_user_disp, xpd=NA, col='green' )
	#abline( h=y1_user_disp, xpd=NA, col='blue', lty=6 )

	#----------------------------------------------------------------
	# Actual plottable area is larger than "display" due to the outer
	# margins.
	#----------------------------------------------------------------
	display_width  = din[1] - (omi[2] + omi[4])
	display_height = din[2] - (omi[1] + omi[3])
	frac_incr_left  = omi[2]/display_width
	frac_incr_right = omi[4]/display_width
	frac_incr_bot   = omi[1]/display_height
	frac_incr_top   = omi[3]/display_height

	x0_user_plot = x0_user_disp - (frac_incr_left )*(x1_user_disp - x0_user_disp) 
	x1_user_plot = x1_user_disp + (frac_incr_right)*(x1_user_disp - x0_user_disp) 
	y0_user_plot = y0_user_disp - (frac_incr_bot  )*(y1_user_disp - y0_user_disp)
	y1_user_plot = y1_user_disp + (frac_incr_top  )*(y1_user_disp - y0_user_disp)

	#abline( h=y0_user_plot, xpd=NA, col='black', lty=6 )
	#abline( h=y1_user_plot, xpd=NA, col='black', lty=6 )
	#abline( v=x0_user_plot, xpd=NA, col='black', lty=6 )
	#abline( v=x1_user_plot, xpd=NA, col='black', lty=6 )

	dx = x1_user_plot - x0_user_plot
	x_uu = dx*xx + x0_user_plot
	dy = y1_user_plot - y0_user_plot
	y_uu = dy*yy + y0_user_plot

	return( c(x_uu, y_uu))
}

#======================================================================================
text_abs = function( xx, yy, str, ... ) {

	p = fractional_to_user_units( xx, yy )
	text( p[1], p[2], str, xpd=NA, ... )
}

#======================================================================================
# Input: h2d[nx,ny]
#
hist2d_bounding_box = function( h2d ) {

	nx = dim(h2d)[1]
	ny = dim(h2d)[2]

	x0 = -1
	for( i in 1:nx ) {
		nz = sum( h2d[i,] == 0)
		if( (nz < ny) && (x0 == -1))
			x0 = i
		}

	x1 = -1
	for( i in nx:1 ) {
		nz = sum( h2d[i,] == 0)
		if( (nz < ny) && (x1 == -1))
			x1 = i
		}

	y0 = -1
	for( j in 1:ny ) {
		nz = sum( h2d[,j] == 0 )
		if( (nz < nx) && (y0 == -1))
			y0 = j
		}

	y1 = -1
	for( j in ny:1 ) {
		nz = sum( h2d[,j] == 0 )
		if( (nz < nx) && (y1 == -1))
			y1 = j
		}

	return( c( x0, x1, y0, y1 ))
}

#======================================================================================
# Inputs: d1[nt], d2[nt]
#
# NOTE: breaks MUST encompass the data
# Breaks must be monotonically increasing
#
hist2d = function( d1, breaks1, d2, breaks2, style='new', ... ) {

	dim(d1) = NULL
	dim(d2) = NULL

	if( length(d1) != length(d2))
		stop(paste("Error, d1 is not same length as d2"))

	nx = length(breaks1) - 1
	ny = length(breaks2) - 1

	retval = array(as.integer(0), dim=c(nx,ny))

	nb1 = length( breaks1 )
	for( ii in 2:nb1 )
		if( breaks1[ii] <= breaks1[ii-1] )
			stop(paste("breaks1 must be monotonic and increasing"))

	nb2 = length( breaks2 )
	for( ii in 2:nb2 )
		if( breaks2[ii] <= breaks2[ii-1] )
			stop(paste("breaks2 must be monotonic and increasing"))

	ndat = length(d1)

	for( id in 1:ndat ) {
		if( (!is.na(d1[id])) && (!is.na(d2[id]))) {

			if( (!is.na(d1[id])) && (d1[id] < breaks1[1]))
				stop(paste('d1: data off left end',d1[id]))
			if( (!is.na(d1[id])) && (d1[id] > breaks1[nb1] ))
				stop(paste('d1: data off right end', d1[id]))
			if( (!is.na(d2[id])) && (d2[id] < breaks2[1]))
				stop(paste('d2: data off left end',d2[id]))
			if( (!is.na(d2[id])) && (d2[id] > breaks2[nb2] ))
				stop(paste('d2: data off right end', d2[id]))

			wn = wherenearest( d1[id], breaks1 )
			if( d1[id] > breaks1[wn] )
				ix = wn
			else
				ix = wn - 1

			wn = wherenearest( d2[id], breaks2 )
			if( d2[id] > breaks2[wn] )
				iy = wn 
			else
				iy = wn - 1

			retval[ix,iy] = retval[ix,iy] + 1
			}
		}

	if( style != 'new' ) 
		return( retval )

	xvals = array(0., nx)
	for( ix in 1:nx ) 
		xvals[ix] = mean( c( breaks1[ix] , breaks1[ix+1] ))

	yvals = array(0., ny)
	for( iy in 1:ny ) 
		yvals[iy] = mean( c( breaks2[iy] , breaks2[iy+1] ))

	return( list(
		xvals	= xvals,
		yvals	= yvals,
		hist2d	= retval ))
}

#=================================================================================================
#	fontinfo <- dphershey_fontinfo(1)
#
dbl_dphershey_emit_string = function(  x, y, str, height_inch, fontinfo, bg_lwd=5, bg_col='white', 
		text_lwd=1.0, col='black', ... ) {

	orig_col = col

	dphershey_emit_string( x, y, str, height_inch, fontinfo, col=bg_col, lwd=bg_lwd, ... )
	dphershey_emit_string( x, y, str, height_inch, fontinfo, lwd=text_lwd, col=orig_col, ... )
}

#=================================================================================================
dph_legend = function( x, y, labels, line_dx, label_dy, height_inch, fontinfo,
	col=array('black',length(labels)), lwd=array(1,length(labels)),
	label_gap_fact=0.15, bg_lwd=5, bg_col='white', text_lwd=0.1 ) {

	nn = length(labels)
	for( ii in 1:nn ) {
		yoff = y - (ii-1)*label_dy
		lines( c(x,x+line_dx), c(yoff,yoff), lwd=bg_lwd, col=bg_col )
		lines( c(x,x+line_dx), c(yoff,yoff), lwd=lwd[ii], col=col[ii] )

		lx = x+line_dx*(1+label_gap_fact)
		dbl_dphershey_emit_string( lx, yoff, labels[ii], height_inch, fontinfo,
			bg_lwd=bg_lwd, bg_col=bg_col, text_lwd=text_lwd )
		}
}

#=================================================================================================
rgb255 = function( r, g, b ) {

	return( rgb(r/255,g/255,b/255))
}

#======================================================================================
# Given some array 'zorig' that has x values 'xorig' and y values 'yorig',
# this limits all of x, y, and z to the range x range xlim and the y range ylim
#
clipout = function( xlim, ylim, xorig, yorig, zorig ) {

	nx = length( xorig )
	ny = length( yorig )
	if( nx != dim(zorig)[1] ) 
		stop(paste("internal error: passed xvals of length", nx, "but dim of z is", pv(dim(zorig)) ))
	if( ny != dim(zorig)[2] ) 
		stop(paste("internal error: passed yvals of length", ny, "but dim of z is", pv(dim(zorig)) ))

	i0 = wherenearest( xlim[1], xorig )
	i1 = wherenearest( xlim[2], xorig )
	j0 = wherenearest( ylim[1], yorig )
	j1 = wherenearest( ylim[2], yorig )

	if( i0 > 1 ) i0 = i0 - 1
	if( i1 < nx ) i1 = i1 + 1

	y_rev = ylim[1] > ylim[2]
	if( y_rev ) {
		if( j0 < ny ) j0 = j0 + 1
		if( j1 > 1  ) j1 = j1 - 1
		}
	else
		{
		if( j0 > 1 ) j0 = j0 - 1
		if( j1 < ny ) j1 = j1 + 1
		}

	x = xorig[i0:i1]
	y = yorig[j0:j1]
	z = zorig[ i0:i1, j0:j1 ]

	return( list( x=x, y=y, z=z ))
}

#===========================================================================================================
error_bar = function( x0, y0, dx, dy, ... ) {

	lines( c(x0,x0), c(y0-dy, y0+dy), ... )
	lines( c(x0-dx,x0+dx), c(y0-dy,y0-dy), ... )
	lines( c(x0-dx,x0+dx), c(y0+dy,y0+dy), ... )
}

#===========================================================================================================
# This is for adding geographical axes that are labeld in "degrees"
# Call ccontour with axes=FALSE, then call this
#
# xory = 1 for X axis, =2 for Y axis
# ticvals should be negative for W or S, positive for E or N.
#
label_axes_in_degrees = function( xory, ticvals ) {

	nn = length( ticvals ) 
	ne_or_sw = array(' ', nn)
	deg = array(' ', nn )

	if( xory == 1 ) {
		char_neg = 'W'
		char_pos = 'E'
		char_zero = ' '
		}
	else
		{
		char_neg = 'S'
		char_pos = 'N'
		char_zero = ' '
		}

	for( i in 1:nn ) {
		if( xory == 1 ) {
			if( ticvals[i] < 180 ) {
				ne_or_sw[i] = char_pos
				deg[i] = '*degree*'
				}
			else if( ticvals[i] > 180 ) {
				ne_or_sw[i] = char_neg
				deg[i] = '*degree*'
				}
			else
				{
				ne_or_sw[i] = char_zero
				deg[i] = '*degree'
				}
			}
		else if( xory == 2 ) {
			if( ticvals[i] < 0 ) {
				ne_or_sw[i] = char_neg
				deg[i] = '*degree*'
				}
			else if( ticvals[i] > 0 ) {
				ne_or_sw[i] = char_pos
				deg[i] = '*degree*'
				}
			else
				{
				ne_or_sw[i] = char_zero
				deg[i] = '*degree'
				}
			}
		}

	str = array( ' ', nn )
	for( i in 1:nn ) {
		vv = abs(ticvals[i])
		if( (ne_or_sw[i] == 'W') && (vv > 180.001))
			vv = 360 - vv
		pp = paste( vv, deg[i], ne_or_sw[i] )
		str[i] = parse(text=pp)
		}

	if( xory == 2 )
		par(las=1)
	axis( xory, ticvals, str )
}

#==============================================================================================================
#==============================================================================================================
# There are two routines that make it easy to produce X-Y lon-lat plots that are not geographically
# distorted (at their center).
#
# "mar_adjust" adjusts the margins of a plot to make it geographically undistorted. Useage:
#
#	xlim = c()
#	ylim = c()
#	ps = mar_adjust( xlim=xlim, ylim=ylim )
#	mplot( xlim=xlim, ylim=ylim ) or ccontour( xlim=xlim, ylim=ylim )
#	mar_adjust( prevstate=ps )	# if you forget to call after mplot, margins will not be rest to orig vals
#
# "get_geo_lim", by contrast, figures out the data range (xlim or ylim) needed to make the plot fill
# the entire currently-allocated space in an undistorted way. You have to already have in mind either
# an xlimit and ycenter, OR a ylimit and xcenter. The routine will then return either the ylimit or
# xlimit, respectively.
# Useage:
#
#	xlim = c()
#	yc = y_center
#	ylim = get_geo_lim( xlim=xlim, yc=yc )
# or
#	ylim = c()
#	xc = x_center
#	xlim = get_geo_lim( ylim=ylim, xc=xc )
#
#	mplot or ccontour( xlim=xlim, ylim=ylim, etc ) 
#
#=====================================================================================================
get_asp_ratio = function( xlim, ylim ) {

	xr <- xlim[2] - xlim[1]
	yr <- ylim[2] - ylim[1]
	km_per_degree = 110.
	midy_deg = mean( ylim )
	midy_rad = midy_deg/180*pi
	dx_km = xr * cos( midy_rad ) * km_per_degree
	dy_km = yr * km_per_degree
	fin = par()$fin
	plt = par()$plt
	dx_plot = fin[1] * (plt[2] - plt[1])
	dy_plot = fin[2] * (plt[4] - plt[3])
	aspect_dat  = dx_km/dy_km
	aspect_plot = dx_plot/dy_plot

	return( c( aspect_dat, aspect_plot ))	# aspect ratio is width / height
}

#=====================================================================================================
mar_adjust = function( xlim=NA, ylim=NA, prevstate=NA ) {

	ps_not_set = (length(prevstate) == 1) && (is.na(prevstate))
	print(paste('mar_adjust entering: prevstate set: ', !ps_not_set ))
	if( ! ps_not_set ) {
		#-----------------------------------------------------------------
		# if it is NOT true that prevstate is NOT set, then it must BE set
		#-----------------------------------------------------------------
		mar0 = prevstate$mar0
		par(mar=mar0)
		return()
		}

	arr = get_asp_ratio( xlim, ylim )

	mar0 = par()$mar
	prevstate = list( mar0=mar0 )

	print(paste('mar0:', pv(mar0)))
	print(paste('aspect ratio 0:', pv(arr) ))

	DATA = 1
	PLOT = 2

	mar = mar0
	if( arr[PLOT] == arr[DATA] ) 
		return

	delta = 0.01
	if( arr[PLOT] > arr[DATA] ) {
		#-----------------
		# Plot is too wide
		#-----------------
		done = FALSE
		dx = delta
		while( ! done ) {
			par(mar=c(mar0[1], mar0[2]+dx, mar0[3], mar0[4]+dx))
			arr = get_asp_ratio( xlim, ylim )
			if( arr[PLOT] <= arr[DATA] ) {
				done = TRUE
				dx = dx - delta/2
				par(mar=c(mar0[1], mar0[2]+dx, mar0[3], mar0[4]+dx))
				}
			dx = dx + delta
			}
		}
	else
		{
		#-----------------
		# Plot is too tall
		#-----------------
		done = FALSE
		dy = delta
		while( ! done ) {
			par(mar=c(mar0[1]+dy, mar0[2], mar0[3]+dy, mar0[4]))
			arr = get_asp_ratio( xlim, ylim )
			if( arr[PLOT] >= arr[DATA] ) {
				done = TRUE
				dy = dy - delta/2
				par(mar=c(mar0[1]+dy, mar0[2], mar0[3]+dy, mar0[4]))
				}
			dy = dy + delta
			}
		}

	print(paste('aspect ratio 1:', pv(arr) ))

	return( prevstate )
}

#======================================================================================
# Given an X -or- Y lim, this calculates the other one to make a correct aspect ratio
# geographic plot
#
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# THIS ONLY REALLY WORKS IF xaxs='i', yaxs='i'
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#
# Useage: Either one of these:
#
#	ylim = get_geo_lim( xlim=c(x0,x1), yc=y_center )
# or
#	xlim = get_geo_lim( ylim=c(x0,x1), xc=x_center )
#
get_geo_lim = function( xlim=NA, ylim=NA, xc=NA, yc=NA ) {
	
	if( is.na(xlim) && is.na(yc) && (!is.na(ylim)) && (length(ylim)==2) && (!is.na(xc))) {
		calc_xlim = TRUE
		midy_deg = mean( ylim )
		}
	else if( is.na(ylim) && is.na(xc) && (!is.na(xlim)) && (length(xlim)==2) && (!is.na(yc))) {
		calc_xlim = FALSE
		midy_deg = yc
		}
	else
		stop(paste("Error, call either with xlim and yc set (returns ylim), or ylim and xc set (returns xlim)"))

	km_per_degree = 110.
	midy_rad = midy_deg/180*pi

	fin = par()$fin
	plt = par()$plt
	dx_plot = fin[1] * (plt[2] - plt[1])
	dy_plot = fin[2] * (plt[4] - plt[3])
	aspect_plot = dx_plot/dy_plot

	if( calc_xlim ) {
		yr = ylim[2] - ylim[1]
		dy_km = yr * km_per_degree
		dx_km_needed = dy_km  * aspect_plot
		xr_deg = dx_km_needed / ( cos(midy_rad) * km_per_degree) 
		xlim = c( xc - xr_deg/2, xc + xr_deg/2 )

		ar = get_asp_ratio( xlim, ylim ) 
		print(paste("calculated xlim, returned aspect ratios of data, plot=", ar[1], ar[2] ))

		return( xlim )
		}
	else
		{
		#----------
		# Calc ylim
		#----------
		xr = xlim[2] - xlim[1]
		dx_km = xr * km_per_degree
		dy_km_needed = dx_km * aspect_plot
		yr_deg = dy_km_needed / (cos(midy_rad) * km_per_degree )
		ylim = c( yc - yr_deg/2, yc + yr_deg/2 )

		ar = get_asp_ratio( xlim, ylim ) 
		print(paste("calculated ylim, returned aspect ratios of data, plot=", ar[1], ar[2] ))

		return( ylim )
		}
}

#===============================================================================================
two_lines_to_polygon = function( x, y1, y2, ... ) {

	bad = (is.na(x) | is.na(y1) | is.na(y2))

	x 	= x [ ! bad ]
	y1 	= y1[ ! bad ]
	y2 	= y2[ ! bad ]

	xr = reverse( x )
	y2r = reverse( y2 )

	polygon( c(x, xr), c( y1, y2r), ... )
}

#=================================================================================================
# Given a set of model results over time, this computes the time series of min and max values
# at each time step.
#
# Input:
#	zz[nmodels,nt]
#
# Outputs:
#	three time series, one of the lowest values each time step, one of the mean, one of the highest values
#
model_set_to_range = function( zz ) {

	nt = dim( zz )[2]
	tser_low  = array(0., nt)
	tser_hi   = array(0., nt)
	tser_mmea = array(0., nt)

	for( it in 1:nt ) {
		tser_low [it] = min( zz[,it], na.rm=TRUE )
		tser_hi  [it] = max( zz[,it], na.rm=TRUE )
		tser_mmea[it] = mean( zz[,it], na.rm=TRUE )
		}

	return( list( 
		tser_low	= tser_low,	# [nt]
		tser_mmea	= tser_mmea,	# [nt]
		tser_hi	 	= tser_hi	# [nt]
		))
}

#=================================================================================================
# Given two time sequences of high and low ranges, this plots the ranges as colored polygons,
# plus their intersection in a different color, plus the MMEA lines 
#
# Inputs: 
#	tvals[nt]	: common time vals of the time series
#	zlow1[nt]	: low values for time series 1
#	zhi1[nt]	: high values for time series 1
#	zlow2[nt]	: low values for time series 2
#	zhi2[nt]	: high values for time series 2
#	col1		: color to plot range 1
#	col2		: color to plot range 2
#	coli		: color to plot range intersection of 1 and 2
#
# See also utility function model_set_to_range that returns the low and hi value
# time series for a set of model results
#
plot_overlapping_ranges = function( tvals, zlow1, zhi1, zlow2, zhi2, col1, col2, coli ) { 

	library( 'sp' )
	library( 'rgeos' )

	proj4str = CRS('+proj=tmerc')

	px = c( tvals, reverse(tvals), tvals[1] )

	#--------------
	# Make Polygon1
	#--------------
	py = c( zlow1, reverse(zhi1), zlow1[1] )
	xy = cbind( px, py )
	p1 = Polygon( xy, hole=FALSE )
	p2 = Polygons( list(p1), 'p1' )
	poly1 = SpatialPolygons( list(p2), proj4string=proj4str )

	#--------------
	# Make Polygon2
	#--------------
	py = c( zlow2, reverse(zhi2), zlow2[1] )
	xy = cbind( px, py )
	p1 = Polygon( xy, hole=FALSE )
	p2 = Polygons( list(p1), 'p1' )
	poly2 = SpatialPolygons( list(p2), proj4string=proj4str )

	#-------------------------------------
	# Plot intersection of the two regions
	#-------------------------------------
	pboth = gIntersection( poly1, poly2 )
	plot( pboth, add=TRUE, col=coli, border=coli )

	#---------------------------------------------
	# Plot region 1 that does not overlap region 2
	#---------------------------------------------
	pdiff = gDifference( poly1, poly2 )
	plot( pdiff, add=TRUE, col=col1, border=col1 )

	#---------------------------------------------
	# Plot region 2 that does not overlap region 1
	#---------------------------------------------
	pdiff = gDifference( poly2, poly1 )
	plot( pdiff, add=TRUE, col=col2, border=col2 )
}

#============================================================================================
# Given a set of model results over time, this plots the range as a polygon, and the MMEA
# as a central line.
#
# This is intended for historical periods. See also plot_overlapping_ranges for future periods
#
# Inputs:
#	tvals[nt]	: time values
#	zz[nmods,nt]	: Data from every model over time
#	col_range	: color to plot the range in 
#	col_mmea	: color to plot the MMEA in 
#	lwd_mmea	: line width to plot the MMEA in 
#
# Set extend_right to 1 to extend the time series and ranges by 1 to eliminate a gap between drawn polygons
#
plot_model_range_and_mmea = function( tvals, zz, col_range='grey', col_mmea='black', lwd_mmea=1, extend_right=NA ) {

	if( ! is.na( extend_right ) ) {
		nt = length( tvals )
		nt2 = nt + extend_right
		tvals2 = array(0., nt2)
		nmods = dim( zz )[1]
		zz2 = array( 0., dim=c(nmods,nt2))

		dt = tvals[2] - tvals[1]

		tvals2[1:nt] = tvals
		zz2[,1:nt] = zz

		for( it in 1:extend_right ) {
			tvals2[nt+it] = tvals[nt] + dt*it
			zz2[,nt+it] = zz[,nt]
			}

		tvals = tvals2
		zz = zz2
		}

	rv = model_set_to_range( zz )

	px = c( tvals, reverse( tvals ))
	py = c( rv$tser_low, reverse( rv$tser_hi ))

	nt = length( tvals )
	mmea = array(0, nt )
	for( it in 1:nt ) 
		mmea[it] = mean( zz[,it], na.rm=TRUE )

	polygon( px, py, col=col_range, border=col_range )
	lines( tvals, mmea, col=col_mmea, lwd=lwd_mmea )

	return( list(
		tvals	= tvals,
		mmea	= mmea ))
}

#=================================================================================
add_righthand_y_axis = function( left_hand_equiv_vals, tic_label_vals, axis_lab, labline=3, tic_len=0.03 ) {

	uu = par()$usr

	xx = uu[2]
	y0 = uu[3]
	y1 = uu[4]

	dx = uu[2] - uu[1]

	nv = length( left_hand_equiv_vals )

	for( iv in 1:nv ) {
		tx = xx + dx * 0.01
		ty = left_hand_equiv_vals[iv]
		if( (ty >= y0) && (ty <= y1)) {
			text( tx, ty, tic_label_vals[iv], adj=c(0.0, 0.5), xpd=NA )
			lines( c(xx-dx*tic_len, xx), c(ty,ty))
print(paste("label at", tx, ty, tic_label_vals[iv] ))
			}
		}

	par(srt=90)
	cy = 0.5*(y0+y1)
	text( xx + dx*labline*0.05, cy, axis_lab, xpd=NA )
	par( srt=0)
}

#========================================================================================
outline_figure_region = function( col=rgb(0.9,0.09,0.9), guardband=0.01, ...) {
	
	#------------------------------------
	# Get NDC coords of the figure region
	#------------------------------------
	ff = par()$fig	# x1, x2, y1, y2
	llx_ndc = ff[1]
	urx_ndc = ff[2]
	lly_ndc = ff[3]
	ury_ndc = ff[4]

	#-----------------------
	# Convert to user coords
	#-----------------------
	from = "ndc"
	to   = "user"

	llx  = grconvertX( llx_ndc, from, to )
	urx  = grconvertX( urx_ndc, from, to )

	lly  = grconvertY( lly_ndc, from, to )
	ury  = grconvertY( ury_ndc, from, to )

	dx = urx - llx
	dy = ury - lly
	delx = dx * guardband
	dely = dy * guardband

	llx = llx + delx
	lly = lly + dely
	urx = urx - delx
	ury = ury - dely

	lines( c( llx, llx), c(lly, ury), col=col, xpd=NA, ... )
	lines( c( llx, urx), c(ury, ury), col=col, xpd=NA, ... )
	lines( c( urx, urx), c(ury, lly), col=col, xpd=NA, ... )
	lines( c( urx, llx), c(lly, lly), col=col, xpd=NA, ... )
}

#====================================================================================
# Start a new page of plots
#
newpage = function() {

	while( ! par('page'))
		plot.new()
}
	
#====================================================================================
subsample_erf = function( xx, yy ) {

	nn = length( xx )
	if( nn < 1000 ) 
		return( list( x=xx, y=yy ))

	keep = array( FALSE, nn )

	#---------------------
	# Always keep the ends
	#---------------------
	n_keep_ends = 30
	keep[1:n_keep_ends] = TRUE
	keep[(nn-n_keep_ends):nn] = TRUE

	#-----------------------------
	# Keep half pts in this region
	#-----------------------------
	iz = floor( nn * 0.10 )
	it0 = n_keep_ends + 1
	it1 = iz
	keep[seq(it0,it1,by=2)] = TRUE
	it0 = nn - iz
	it1 = nn
	keep[seq(it0,it1,by=2)] = TRUE

	#-------------------------------
	# Keep 1 out of 10 pts in middle
	#-------------------------------
	it0 = iz + 1
	it1 = nn - iz - 1
	keep[seq(it0,it1,by=20)] = TRUE

	nkeep = sum( keep )
	print(paste('subsample_erf: ntot=', nn, 'nkeep=', nkeep ))

	xx = xx[ keep ]
	yy = yy[ keep ]

	return( list( x=xx, y=yy ))
}

#=========================================================================================
# Inputs:
#	xvals[nx]
# 	yvals[nx,n_elements}
#	col[n_elements]
#
# each element is plotted atop the sum of the previous elements
#
# 
stackplot = function( xvals, yvals, col, xlim=NULL, ylim=NULL, xtics=NULL, ... ) {

	nx = length( xvals )
	ne = dim( yvals )[2]

	if( dim(yvals)[1] != nx ) stop(paste("Error, yvals must be shape [nx,nelements]"))

	if( length(col) != ne ) stop(paste("Error, col must be length n_elements"))

	if( is.null(xlim))
		xlim = range( xvals )

	if( is.null(ylim)) {
		yy = array(0, nx)
		for( ie in 1:ne )
			yy = yy + yvals[,ie]
		ylim = c(0, max( yy ))
		}

	canvas( xlim=xlim, ylim=ylim, ... )

	sum_so_far = array(0, nx)

	for( ie in 1:ne ) {
		new_sum = sum_so_far + yvals[,ie]
		two_lines_to_polygon( xvals, sum_so_far, new_sum, col=col[ie], ... )

		sum_so_far = new_sum
		}

print('==a')
	uu = par()$usr	# llx, urx, lly, ury
print('==b')
	lly = uu[3]
	ury = uu[4]
	dy = ury - lly
	ticlen = dy * 0.025
print('==c')

	if( ! is.null( xtics )) {
		nt = length( xtics )

print('==d')
		for( itic in 1:nt ) {

			xx = xtics[itic]
			y0 = lly
			y1 = lly+ticlen

			lines( c(xx,xx), c(y0,y1) )
			}
		}
}

#==========================================================================================
# Given a time series of x, y values, this labels the peaks in the time series.
# The top n2lab can be labeled
# by default, the X value is shown on the label, but if supplied,
# lab[nt] can be used instead
# If set, exclusion radius means that no other point within exclusion_radius
# timesteps will be labeled. It is given in POINTS, not X UNITS.
#
label_peaks = function( xx, yy_orig, n2lab=10, lab=NULL, exclusion_radius=0, yoffset_fact=0.01, ... ) {

	nt = length( yy_orig )

	notta_max = min( yy_orig, na.rm=TRUE )
	yy = yy_orig
	yy[ is.na( yy )] = notta_max

	if( is.null(lab))
		lab2use = xx
	else
		lab2use = lab

	yy2 = yy

	uu = par()$usr
	plotdy = uu[4] - uu[3]
	dy = plotdy * yoffset_fact

	for( kk in 1:n2lab ) {

		six = sort( yy2, index.return=TRUE, decreasing=TRUE )

		ix0 = six$ix[1]

		xmax = xx[ix0]
		ymax = yy2[ix0]
		labmax = lab2use[ix0]

print(paste("label", kk, "x, y max:", xmax, ymax, "lab:", labmax ))

		par( srt=90 )
		text( xmax, ymax+dy, labmax, adj=c(0.0,0.5), ... )
		par( srt=0 )

		for( kk in -exclusion_radius:exclusion_radius ) {
			kk2 = kk + ix0
			if( kk2 < 1 ) kk2 = 1
			if( kk2 > nt ) kk2 = nt
			yy2[kk2] = notta_max
			}
		}
}

#========================================================================
# This is only needed if you want to put string tic labels on.
#
#  xticlvas[nxlab] : values for the x tics that will be labeled
#  xticlabs[nxlab] : labels for the x tic vals that will be labeled
#  dx_subtic	   : delta-X (user units) if you want subticks besides the labeled tics
#
draw_x_axis = function( xticvals, xticlabs, dx_subtic=NA, ticlen=0.02, lab_y_offset=0.01 ) {

	nx = length( xticvals )
	if( length( xticlabs ) != nx ) stop(paste("Error, supplied xticvals is length", nx, "but length of xticlabs is", 
		length( xticlabs ), 'they must be identical!' ))

	uu = par()$usr		# (x1, x2, y1, y2)

	dy = (uu[4] - uu[3]) 

	ty0 = uu[3] - dy*ticlen
	ty1 = uu[3]

	ddx = abs(xticvals[2] - xticvals[1])

	if( ! is.na( dx_subtic )) {
		xleft = xticvals[1] - 10*ddx
		xright = xticvals[nx] + 10*ddx
		xminortic = seq( xleft, xright, by=dx_subtic )
		nn = length( xminortic )
		for( kk in 1:nn ) {
			xmt = xminortic[kk]
			if( (xmt >= uu[1]) && (xmt <= uu[2])) {
				lines( c( xmt, xmt ), c(ty0,ty1), xpd=NA )
				}
			}
		}

	for( kk in 1:nx ) {
		xtv = xticvals[kk]
		lines( c( xtv, xtv ), c(ty0,ty1), xpd=NA )
		text( xtv, ty0-(lab_y_offset*dy), xticlabs[kk], adj=c(0.5, 1.0), xpd=NA )
		}
}

#========================================================================
# Let's say that you have some X and Y and want to make a line plot,
# BUT you want to rearrange the X order. To do this:
#
# 1. Make a permutation array that is INDEXED by the ORIGINAL X value,
#    and holds where you want it to be plotted.
#
# Ex: to reverse a plot that goes from 1:10:
#
# Standard version:
#    xx = 1:10
#    yy = 15:30
#    xticvals = 1:10
#    xticlabs = paste(1:10)
#    mplot( xx, yy, axes=FALSE )
#    par( las=1 )
#    axis(2)
#    draw_x_axis( xticvals, xticlabs )
#
# REVERSED VERSION:
#
#    perm = 10:1	# i.e., what was ONCE plotted at loc 1 will now be plotted at loc perm[1] = 10
#    mplot( xx, perm_x_axis(yy,perm), axes=FALSE )
#    par( las=1 )
#    axis(2)
#    draw_x_axis( xticvals, perm_x_axis(xticlabs,perm) )
#
# NOTE HOW YOU CALL PERM_X_AXIS ON THE Y VALUES, NOT THE X VALUES!!!
#
perm_x_axis = function( yvals_orig, perm ) {

	if( length(yvals_orig) != length(perm)) 
		stop(paste("Error, length of yvals_orig and perm mus be identical! Length of yvals_orig:",
			 length(yvals_orig), 'length of perm:', length(perm) ))

	return( yvals_orig[perm] )
}

